#pragma once

#include "defs.h"

#include <vector>
#include <string>

#if DEBUG
#define CHECK_RESULT(cond) assert(cond)
#else
#define CHECK_RESULT(cond) cond
#endif
#define MAX_ASYNC_CONTEXTS 15

namespace gfx
{
    typedef void* Shader;
    typedef void* Buffer;
    typedef void* Texture;
    typedef void* Sampler;
    typedef void* RenderPass;
    typedef void* Framebuffer;
    typedef void* RenderPipeline;
    typedef void* ComputePipeline;
    typedef void* Semaphore;

    namespace e_featureFlags
    {
        enum featureFlags
        {
            none = 0,
            shaderFloat64 = 1 << 0
        };
    }
    typedef u32 FeatureFlags;

    namespace e_bindFlags
    {
        enum bindFlags
        {
            vertexBuffer = 1 << 0,
            indexBuffer = 1 << 1,
            storage = 1 << 2,
            uniform = 1 << 3
        };
    }
    typedef u32 BindFlags;

    namespace e_indexType
    {
        enum indexType
        {
            uint16,
            uint32
        };
    }
    typedef e_indexType::indexType IndexType;

    // semantically resources that are device local can still be read/written using intermediate
    // resources - this sort of update is suitable for resources that have infrequent cpu access
    // worth investiagating how explictly requiring flush/invalidate after cpu data update affects
    // the performance and handling of resources - some useful stuff there?
    namespace e_accessFlags
    {
        enum accessFlags
        {
            none = 0,
            write = 1 << 0,
            read = 1 << 1
        };
    }
    typedef u32 AccessFlags;

    // looks a bit mental but certain channels are reliant on there being a previous channel
    // e.g. you can't have a format with only 'g' channel, it has to also have an 'r' channel
    namespace e_formatChannelFlags
    {
        enum formatChannelFlags
        {
            r = (1 << 0),
            rg = (1 << 1 | r),
            rgb = (1 << 2 | rg),
            rgba = (1 << 3 | rgb),
            d = (1 << 4),
            ds = (1 << 5 | d)
        };
    }
    typedef u32 FormatChannelFlags;

    namespace e_formatType
    {
        enum formatType
        {
            unsignedNorm,
            signedNorm,
            unsignedScaled,
            signedScaled,
            unsignedInt,
            signedInt,
            signedFloat
        };
    }
    typedef e_formatType::formatType FormatType;

    namespace e_textureUsageFlags
    {
        enum textureUsageFlags
        {
            transferSrc = 1 << 0,
            transferDst = 1 << 1,
            shaderRead = 1 << 2,
            shaderWrite = 1 << 3,
            storage = 1 << 4,
            depthStencil = 1 << 5,
            inputAttachment = 1 << 6 // in vulkan input attachment allows the output of one subpass to be sampled as an input to a subsequent subpass
        };
    }
    typedef u32 TextureUsageFlags;

    namespace e_bufferUsageFlags
    {
        enum bufferUsageFlags
        {
            transferSrc = 1 << 0,
            transferDst = 1 << 1,
            uniform = 1 << 2,
            storage = 1 << 3,
            vertex = 1 << 4,
            index = 1 << 5,
            indirect = 1 << 6
        };
    }
    typedef u32 BufferUsageFlags;

    namespace e_loadOpType
    {
        enum loadOpType
        {
            clear,
            load,
            undefined
        };
    }
    typedef e_loadOpType::loadOpType LoadOp;

    namespace e_storeOpType
    {
        enum storeOpType
        {
            store,
            undefined
        };
    }
    typedef e_storeOpType::storeOpType StoreOp;

    namespace e_pipelineStageType
    {
        enum pipelineStageType
        {
            top,            // pre command execution
            vertexInput,    // post vertex input
            vertexOutput,   // post vertex shader execution
            fragmentInput,  // post tessellation
            fragmentOutput, // post fragment shader execution
            colorOutput,    // post colour output (blending, resolve, etc.)
            compute,
            transfer,
            bottom
        };
    }
    typedef e_pipelineStageType::pipelineStageType PipelineStage;

    namespace e_memoryAccessFlags
    {
        enum memoryAccessFlags
        {
            none = 0,
            indirect = 1 << 0,
            index = 1 << 1,
            vertex = 1 << 2,
            uniform = 1 << 3,
            inputAttachment = 1 << 4,
            colorRead = 1 << 5,
            colorWrite = 1 << 6,
            depthStencilRead = 1 << 7,
            depthStencilWrite = 1 << 8
        };
    }
    typedef u32 MemoryAccessFlags;

    namespace e_shaderStageFlags
    {
        enum shaderStageFlags
        {
            vertex = 1 << 0,
            tessellationControl = 1 << 1,
            tessellationEvaluation = 1 << 2,
            geometry = 1 << 3,
            fragment = 1 << 4,
            compute = 1 << 5
        };
    }
    typedef u32 ShaderStageFlags;

    namespace e_polygonModeType
    {
        enum polygonModeType
        {
            fill,
            line,
            point
        };
    }
    typedef e_polygonModeType::polygonModeType PolygonMode;

    namespace e_frontFaceType
    {
        enum frontFaceType
        {
            counterClockwise,
            clockwise
        };
    }
    typedef e_frontFaceType::frontFaceType FrontFace;

    namespace e_cullModeFlags
    {
        enum cullModeFlags
        {
            none = 0,
            front = 1 << 0,
            back = 1 << 1
        };
    }
    typedef u32 CullMode;

    namespace e_compareOpType
    {
        enum compareOpType
        {
            never,
            less,
            equal,
            lessOrEqual,
            greater,
            notEqual,
            greatorOrEqual,
            always
        };
    }
    typedef e_compareOpType::compareOpType CompareOp;

    namespace e_stencilOpType
    {
        enum stencilOpType
        {
            keep,
            zero,
            replace,
            incrementClamp,
            decrementClamp,
            invert,
            incrementWrap,
            decrementWrap
        };
    }
    typedef e_stencilOpType::stencilOpType StencilOp;

    namespace e_dynamicStateFlags
    {
        enum dynamicStateFlags
        {
            viewport = 1 << 0,
            scissor = 1 << 1,
            lineWidth = 1 << 2,
            depthBias = 1 << 3,
            blendConstants = 1 << 4,
            depthBounds = 1 << 5,
            stencilCompareMask = 1 << 6,
            stencilWriteMask = 1 << 7,
            stencilReference = 1 << 8
        };
    }
    typedef u32 DynamicState;

    namespace e_blendFactorType
    {
        enum blendFactorType
        {
            zero,
            one,
            srcColor,
            oneMinusSrcColor,
            dstColor,
            oneMinusDstColor,
            srcAlpha,
            oneMinusSrcAlpha,
            dstAlpha,
            oneMinusDstAlpha,
            constantColor,
            oneMinusConstantColor,
            constantAlpha,
            oneMinusConstantAlpha,
            src1Color,
            oneMinusSrc1Color,
            src1Alpha,
            oneMinusSrc1Alpha
        };
    }
    typedef e_blendFactorType::blendFactorType BlendFactor;

    namespace e_blendOpType
    {
        enum blendOpType
        {
            add,
            subtract,
            reverseSubtract,
            min,
            max
        };
    }
    typedef e_blendOpType::blendOpType BlendOp;

    namespace e_primitiveTopology
    {
        enum primitiveTopology
        {
            pointList,
            lineList,
            lineStrip,
            triangleList,
            triangleStrip,
            triangleFan,
            lineListAdjacency,
            lineStripAdjacency,
            triangleListAdjacency,
            triangleStripAdjacency,
            patchList
        };
    }
    typedef e_primitiveTopology::primitiveTopology PrimitiveTopology;

    namespace e_logicOpType
    {
        enum logicOpType
        {
            clear,
            and,
            andReverse,
            copy,
            andInverted,
            noOp,
            xor,
            or ,
            nor,
            equivalent,
            invert,
            orReverse,
            copyInverted,
            orInverted,
            nand,
            set
        };
    }
    typedef e_logicOpType::logicOpType LogicOp;

    namespace e_descriptorType
    {
        enum descriptorType
        {
            combinedImageSampler,
            sampler,
            sampledImage,
            storageImage,
            uniformBuffer,
            storageBuffer,
            uniformBufferDynamic,
            storageBufferDynamic,
            inputAttachment
        };
    }
    typedef e_descriptorType::descriptorType DescriptorType;

    namespace e_inputRateType
    {
        enum inputRateType
        {
            vertex,
            instance
        };
    }
    typedef e_inputRateType::inputRateType InputRate;

    namespace e_filterType
    {
        enum filterType
        {
            nearest,
            linear
        };
    }
    typedef e_filterType::filterType Filter;

    namespace e_mipmapMode
    {
        enum mipmapMode
        {
            nearest,
            linear
        };
    }
    typedef e_mipmapMode::mipmapMode MipmapMode;

    namespace e_addressMode
    {
        enum addressMode
        {
            repeat,
            mirroredRepeat,
            clampToEdge,
            clampToBorder,
            mirrorClampToEdge
        };
    }
    typedef e_addressMode::addressMode AddressMode;

    namespace e_borderColor
    {
        enum borderColor
        {
            transparentBlack,
            opaqueBlack,
            opaqueWhite
        };
    }
    typedef e_borderColor::borderColor BorderColor;

    namespace e_barrierFlags
    {
        enum barrierFlags
        {
            buffer = 1 << 0,
            texture = 1 << 1,
            global = buffer | texture
        };
    }
    typedef e_barrierFlags::barrierFlags BarrierFlags;

    namespace e_descriptorFlags
    {
        enum descriptorFlags
        {
            unbounded = 1 << 0
        };
    }
    typedef u32 DescriptorFlags;

    struct Viewport
    {
        float x;
        float y;
        float m_width;
        float m_height;
        float m_minDepth;
        float m_maxDepth;
    };

    struct Rect
    {
        int x;
        int y;
        u32 w;
        u32 h;
    };

    struct Clear
    {
        union
        {
            struct
            {
                float r, g, b, a;
            };
            struct
            {
                float depth;
                u8 stencil;
            };
        };
    };

    // supplying format as a combination of elements rather than a single enum might allow
    // streamlined code/more flexibility depending on format enumeration in the graphics api
    struct FormatDesc
    {
        FormatChannelFlags  m_channels = e_formatChannelFlags::r;
        u8                  m_bitDepth = 8;
        FormatType          m_type = e_formatType::unsignedNorm;
        bool                m_standard = false;
        bool                m_reverse = false;
    };

    struct DescriptorDesc
    {
        std::string     m_name = "";
        DescriptorType  m_type = e_descriptorType::uniformBuffer;
        u32             m_count = 0;
        u32             m_binding = 0;
        u32             m_mapping = -1;
        DescriptorFlags m_flags = 0;
    };

    struct PushConstantDesc
    {
        u32 m_offset;
        u32 m_size;
        u32 m_binding;
    };

    struct ShaderDesc
    {
        Mem                             m_data;
        ShaderStageFlags                m_stage;
        std::string                     m_entry;
        std::vector<DescriptorDesc>     m_descriptors;
        std::vector<PushConstantDesc>   m_pushConstants;
    };

    struct BufferDesc
    {
        Mem                 m_data;
        AccessFlags         m_access = e_accessFlags::none;
        BufferUsageFlags    m_usage = 0;
    };

    struct TextureDesc
    {
        u32                 m_width = 1;
        u32                 m_height = 1;
        u32                 m_depth = 1;
        u32                 m_mips = 1;
        u32                 m_layers = 1;
        u32                 m_samples = 1;
        AccessFlags         m_access = e_accessFlags::none;
        FormatDesc          m_format;
        TextureUsageFlags   m_usage = 0;
        bool                m_buffered = false;
        Mem                 m_data;
    };

    struct AttachmentDesc
    {
        TextureDesc*    m_texture; // uses pointers to share texture attachments
        LoadOp          m_loadOp;
        StoreOp         m_storeOp;
        LoadOp          m_stencilLoadOp;
        StoreOp         m_stencilStoreOp;
    };

    struct SamplerDesc
    {
        Filter      m_magFilter = gfx::e_filterType::nearest;
        Filter      m_minFilter = gfx::e_filterType::nearest;
        MipmapMode  m_mipmapMode = gfx::e_mipmapMode::nearest;
        AddressMode m_addressModeU = gfx::e_addressMode::clampToEdge;
        AddressMode m_addressModeV = gfx::e_addressMode::clampToEdge;
        AddressMode m_addressModeW = gfx::e_addressMode::clampToEdge;
        float       m_mipLodBias = 0.0f;
        bool        m_anisotropyEnable = false;
        float       m_maxAnisotropy = 1.0f;
        bool        m_compareEnable = false;
        CompareOp   m_compareOp = gfx::e_compareOpType::always;
        float       m_minLod = -1000;
        float       m_maxLod = 1000;
        BorderColor m_borderColor = e_borderColor::transparentBlack;
        bool        m_unnormalizedCoordinates = false;
    };

    struct SubpassDesc
    {
        std::vector<AttachmentDesc> m_inputAttachments;
        std::vector<AttachmentDesc> m_colorAttachments;
        std::vector<AttachmentDesc> m_depthAttachments;
        std::vector<AttachmentDesc> m_resolveAttachments;
        PipelineStage               m_prevSubpassStageDependency = e_pipelineStageType::top; // the stage in the previous subpass which must be completed first
        PipelineStage               m_subpassWaitStage = e_pipelineStageType::top;           // the stage in this subpass that cannot be executed before the previous subpass has reached its required stage
        MemoryAccessFlags           m_prevSubpassAccessFlags = 0;                           // mask for which memory accesses in previous subpass trigger a dependency 
        MemoryAccessFlags           m_subpassAccessFlags = 0;                               // mask for which memory accesses in current subpass trigger a dependency
        bool                        m_computeSubpass = false;
    };

    struct RenderPassDesc
    {
        std::vector<SubpassDesc> m_subpasses;
    };

    struct FramebufferDesc
    {
        RenderPass                  m_pass;
        std::vector<Texture>        m_textures;            // assume a null pointer passed in desc means the current back buffer image
        std::vector<gfx::LoadOp>    m_loadOps;
    };

    struct InputAssemblyDesc
    {
        PrimitiveTopology   m_topology = e_primitiveTopology::triangleList;
        bool                m_primitiveRestartEnable = false;
    };

    struct DynamicDesc
    {
        DynamicState m_dynamicState = 0;
    };

    struct BlendDesc
    {
        bool        m_blendEnable;
        BlendFactor m_srcColorBlendFactor;
        BlendFactor m_dstColorBlendFactor;
        BlendOp     m_colorBlendOp;
        BlendFactor m_srcAlphaBlendFactor;
        BlendFactor m_dstAlphaBlendFactor;
        BlendOp     m_alphaBlendOp;
    };

    struct MrtBlendDesc
    {
        bool                    m_logicOpEnable = false;
        LogicOp                 m_logicOp = e_logicOpType::equivalent;
        std::vector<BlendDesc>  m_blends;
        float                   m_blendConstants[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
    };

    struct MultisampleDesc
    {
        u32     m_samples = 1;
        bool    m_sampleShading = false;
        float   m_minSampleShading = false;
        bool    m_alphaToCoverage = false;
        bool    m_alphaToOne = false;
    };

    struct RasterDesc
    {
        bool        m_depthClamp = false;
        bool        m_rasterizerDiscard = false;
        PolygonMode m_polygonMode = PolygonMode::fill;
        CullMode    m_cullMode = e_cullModeFlags::none;
        FrontFace   m_frontFace = FrontFace::counterClockwise;
        bool        m_depthBias = false;
        float       m_depthBiasConstant = 0.0f;
        float       m_depthBiasClamp = 0.0f;
        float       m_depthBiasSlope = 0.0f;
        float       m_lineWidth = 1.0f;
    };

    struct TessellationDesc
    {
        u32 m_controlPoints;
    };

    struct StencilDesc
    {
        StencilOp   m_failOp;
        StencilOp   m_passOp;
        StencilOp   m_depthFailOp;
        CompareOp   m_compareOP;
        u32         m_compareMask;
        u32         m_writeMask;
        u32         m_reference;
    };

    struct DepthStencilDesc
    {
        bool        m_depthTest = false;
        bool        m_depthWrite = false;
        CompareOp   m_depthCompareOp = CompareOp::never;
        bool        m_depthBounds = false;
        bool        m_stencilTest = false;
        StencilDesc m_front = {};
        StencilDesc m_back = {};
        float       m_minDepthBounds = 0.0f;
        float       m_maxDepthBounds = 1.0f;
    };

    struct VertexBindingDesc
    {
        u32         m_binding;
        size_t      m_stride;
        InputRate   m_rate;
    };

    struct VertexAttributeDesc
    {
        u32         m_location;
        u32         m_binding;
        FormatDesc  m_format;
        u32         m_offset;
    };

    struct VertexInputDesc
    {
        std::vector<VertexBindingDesc>      m_bindings;
        std::vector<VertexAttributeDesc>    m_attributes;
    };

    struct RenderPipelineDesc
    {
        std::vector<Shader>     m_shaders;
        VertexInputDesc         m_vertexInput;
        InputAssemblyDesc       m_inputAssembly;
        TessellationDesc        m_tessellation;
        std::vector<Viewport>   m_viewports;
        std::vector<Rect>       m_scissors;
        RasterDesc              m_rasterizer;
        MultisampleDesc         m_multisample;
        DepthStencilDesc        m_depthStencil;
        MrtBlendDesc            m_blendDesc;
        DynamicDesc             m_dynamic;
        RenderPass              m_renderPass;
        u32                     m_subpass;
    };

    struct ComputePipelineDesc
    {
        Shader m_shader;
    };

    struct TextureUpdateDesc
    {
        u32 x = 0;
        u32 y = 0;
        u32 z = 0;
        u32 m_width;
        u32 m_height;
        u32 m_depth = 0;
        u32 m_mip = 0;
        Mem m_data;
    };

    struct Geometry
    {
        Buffer m_vertexBuffer = nullptr;
        Buffer m_indexBuffer = nullptr;
        u32    m_numVertices = 0;
        u32    m_numIndices = 0;
        float  m_minAABB[4];
        float  m_maxAABB[4];
    };
}