#pragma once

#include "gfx_defs.h"

namespace gfx
{
    struct DynamicBufferDestroy
    {
        Buffer m_buffer;
        size_t m_destroyFrame = 0;
    };

    struct DynamicBuffer
    {
    public:
        void create(Mem initial, gfx::BufferUsageFlags usage);
        void update(Mem newMem);
        void destroy();

        Mem                                 m_data;
        gfx::BufferUsageFlags               m_usage = e_bufferUsageFlags::uniform;
        Buffer                              m_gpuBuffer = nullptr;
        size_t                              m_gpuSize = 0;
        std::vector<DynamicBufferDestroy>   m_destroyBuffers;
    };

    namespace e_materialFlags
    {
        enum materialFlags
        {
            diffuseMap  = 1 << 0,
            specularMap = 1 << 1,
            bumpMap     = 1 << 2,
            alphaMap    = 1 << 3,
        };
    };
    typedef u32 MaterialFlags;

    struct MaterialBuffer
    {
        float       m_ambient[4];
        float       m_diffuse[4];
        float       m_specular[4];
        float       m_shininess;
        float       m_ior;
        float       m_dissolve;
        float       m_illum;
    };

    struct SharedResource
    {
    public:
        void increment() { m_referenceCount++; }
        bool decrement()
        {
            m_referenceCount--;
            if (m_referenceCount == 0)
                return true;
            return false;
        }
    protected:
        u32 m_referenceCount = 0;
    };

    struct MaterialPipeline : public SharedResource
    {
    public:
        u32             m_flags = 0;
        RenderPipeline  m_pipeline;
    };
    
    struct Material : public SharedResource
    {
    public:
        std::string m_name = "";
        
        Texture*    m_textures = nullptr;
        size_t      m_numTextures = 0;
        Buffer      m_data = nullptr;

        MaterialPipeline* m_pipeline;
    protected:

    };

    struct Model : public SharedResource
    {
    public:
        Buffer  m_vertexBuffer = nullptr;
        Buffer  m_indexBuffer = nullptr;
        u32     m_indexCount = 0;     
        u32     m_vertexBufferIndexOffset = 0;
        size_t  m_vertexBufferByteOffset = 0;
        u32     m_indexBufferOffset = 0;

        float   m_aabbMin[4];
        float   m_aabbMax[4];

    protected:

    };

    struct ModelInstance
    {
    public:
        ModelInstance(Model* model, Material* material)
        {
            m_model = model;
            model->increment();
            m_material = material;
            if (material)
            {
                material->increment();
            }
        }

        Model*      m_model;
        Material*   m_material;
        Mem         m_instanceData;
    };

    struct ShaderLoadDesc
    {
        std::string         m_filepath;
        ShaderStageFlags    m_stage;
    };

    struct RenderPipelineLoadDesc
    {
        std::vector<ShaderLoadDesc> m_shaders;
        InputAssemblyDesc           m_input;
        RasterDesc                  m_raster;
        MultisampleDesc             m_multisample;
        std::vector<BlendDesc>      m_blends;
        DepthStencilDesc            m_depthStencil;
        DynamicDesc                 m_dynamic;
        RenderPass                  m_renderPass;
        u32                         m_subpass;
    };

    struct OBJLoadDesc
    {
        bool                    m_singleBuffer;
        std::string             m_objFilepath;
        std::string             m_texturePath;
        std::string             m_materialTemplate;
        RenderPipelineLoadDesc  m_pipeline;
    };

    RenderPipeline              loadRenderPipeline(const RenderPipelineLoadDesc& desc);
    ComputePipeline             loadComputePipeline(const std::string& filepath);

    std::vector<ModelInstance*> loadOBJ(const OBJLoadDesc& params);
    Texture                     loadTexture(const std::string& filepath);

    // mesh functions
    u32                         generateQuad(Buffer& vertexBuffer, Buffer& indexBuffer, u32 res = 1);
    u32                         generateSphere(Buffer& vertexBuffer, Buffer& indexBuffer, u32 res = 12);
}