#include "dev_ui.h"

#include "gfx.h"
#include "platform.h"
#include "gfx_utils.h"
#include <imgui.h>
#include <algorithm>
#include <unordered_map>

using namespace gfx;

namespace
{
    struct Context
    {
        RenderPass m_renderPass = nullptr;
        RenderPipeline m_pipeline = nullptr;
        Sampler m_sampler = nullptr;
        Texture m_fontTexture = nullptr;
        DynamicBuffer m_vertexBuffer;
        DynamicBuffer m_indexBuffer;
        int m_width = 0;
        int m_height = 0;
    };
    Context s_context;
}

namespace DevUI
{
    void setup()
    {
        IMGUI_CHECKVERSION();
        ImGui::CreateContext();
        ImGuiIO& io = ImGui::GetIO(); (void)io;
        io.BackendFlags |= ImGuiBackendFlags_HasMouseCursors;
        io.BackendFlags |= ImGuiBackendFlags_HasSetMousePos;

        ImGui::StyleColorsDark();
        
        auto ctx = getMainContext();

        SamplerDesc samplerInfo;
        samplerInfo.m_magFilter = e_filterType::linear;
        samplerInfo.m_minFilter = e_filterType::linear;
        samplerInfo.m_mipmapMode = e_mipmapMode::linear;
        samplerInfo.m_addressModeU = e_addressMode::repeat;
        samplerInfo.m_addressModeV = e_addressMode::repeat;
        samplerInfo.m_addressModeW = e_addressMode::repeat;
        samplerInfo.m_minLod = -1000;
        samplerInfo.m_maxLod = 1000;
        samplerInfo.m_maxAnisotropy = 1.0f;
        s_context.m_sampler = ctx->createSampler(samplerInfo);

        ShaderLoadDesc vertShader;
        vertShader.m_filepath = "data/shaders/dev_ui_vert.shader";
        vertShader.m_stage = gfx::e_shaderStageFlags::vertex;
        
        ShaderLoadDesc fragShader;
        fragShader.m_filepath = "data/shaders/dev_ui_frag.shader";
        fragShader.m_stage = gfx::e_shaderStageFlags::fragment;
        gfx::RenderPipelineLoadDesc pld;
        pld.m_shaders.push_back(vertShader);
        pld.m_shaders.push_back(fragShader);

        VertexInputDesc vertInput;
        vertInput.m_bindings.push_back({ 0, sizeof(ImDrawVert), e_inputRateType::vertex });
        vertInput.m_attributes.push_back({ 0, 0, { e_formatChannelFlags::rg, 32, e_formatType::signedFloat, false}, IM_OFFSETOF(ImDrawVert, pos) });
        vertInput.m_attributes.push_back({ 1, 0, { e_formatChannelFlags::rg, 32, e_formatType::signedFloat, false }, IM_OFFSETOF(ImDrawVert, uv) });
        vertInput.m_attributes.push_back({ 2, 0, { e_formatChannelFlags::r, 32, e_formatType::unsignedInt, false }, IM_OFFSETOF(ImDrawVert, col) });

        InputAssemblyDesc& inputAssembly = pld.m_input;
        inputAssembly.m_primitiveRestartEnable = false;
        inputAssembly.m_topology = e_primitiveTopology::triangleList;

        RasterDesc& raster = pld.m_raster;
        raster.m_polygonMode = e_polygonModeType::fill;
        raster.m_cullMode = e_cullModeFlags::none;
        raster.m_frontFace = e_frontFaceType::counterClockwise;
        raster.m_lineWidth = 1.0f;

        MultisampleDesc& multisample = pld.m_multisample;
        multisample.m_samples = 1;

        BlendDesc blend;
        blend.m_blendEnable = true;
        blend.m_srcColorBlendFactor = e_blendFactorType::srcAlpha;
        blend.m_dstColorBlendFactor = e_blendFactorType::oneMinusSrcAlpha;
        blend.m_colorBlendOp = e_blendOpType::add;
        blend.m_srcAlphaBlendFactor = e_blendFactorType::oneMinusSrcAlpha;
        blend.m_dstAlphaBlendFactor = e_blendFactorType::zero;
        blend.m_alphaBlendOp = e_blendOpType::add;
        pld.m_blends.push_back(blend);

        DepthStencilDesc& depthStencil = pld.m_depthStencil;

        DynamicDesc& dynamic = pld.m_dynamic;
        dynamic.m_dynamicState = e_dynamicStateFlags::viewport | e_dynamicStateFlags::scissor;

        TextureDesc textureDesc = ctx->getBackbufferTextureDesc();

        AttachmentDesc colorAttachment = {};
        colorAttachment.m_loadOp = e_loadOpType::load;
        colorAttachment.m_storeOp = e_storeOpType::store;
        colorAttachment.m_texture = &textureDesc;

        SubpassDesc subpass;
        subpass.m_colorAttachments = { colorAttachment };
        subpass.m_prevSubpassStageDependency = e_pipelineStageType::colorOutput;
        subpass.m_subpassWaitStage = e_pipelineStageType::colorOutput;

        RenderPassDesc renderPassDesc;
        renderPassDesc.m_subpasses = { subpass };
        s_context.m_renderPass = ctx->createRenderPass(renderPassDesc);
        
        pld.m_subpass = 0;
        pld.m_renderPass = s_context.m_renderPass;

        s_context.m_pipeline = gfx::loadRenderPipeline(pld);

        unsigned char* pixels;
        int w, h;
        io.Fonts->GetTexDataAsRGBA32(&pixels, &w, &h);
        size_t size = w * h * 4 * sizeof(char);

        TextureDesc textureInfo;
        textureInfo.m_format = { e_formatChannelFlags::rgba, 8, e_formatType::unsignedNorm };
        textureInfo.m_width = w;
        textureInfo.m_height = h;
        textureInfo.m_depth = 1;
        textureInfo.m_mips = 1;
        textureInfo.m_layers = 1;
        textureInfo.m_samples = 1;
        textureInfo.m_usage = e_textureUsageFlags::shaderRead | e_textureUsageFlags::transferDst;
        textureInfo.m_access = e_accessFlags::none;
        textureInfo.m_data.m_data = pixels;
        textureInfo.m_data.m_size = size;
        s_context.m_fontTexture = ctx->createTexture(textureInfo);
    }

    void shutdown()
    {
        auto ctx = gfx::getMainContext();
        s_context.m_vertexBuffer.destroy();
        s_context.m_indexBuffer.destroy();
        ctx->destroyRenderPass(s_context.m_renderPass);
        ctx->destroyRenderPipeline(s_context.m_pipeline);
        ctx->destroySampler(s_context.m_sampler);
        ctx->destroyTexture(s_context.m_fontTexture);
    }

    void update(float dt, int w, int h)
    {
        ImGuiIO& io = ImGui::GetIO();
        io.DisplaySize = ImVec2((float)w, (float)h);
        io.DisplayFramebufferScale = ImVec2(1.0f, 1.0f);
        io.DeltaTime = std::max(dt, 1.0f / 60.0f);
        
        int mouseX, mouseY;
        utils::getMousePos(mouseX, mouseY);
        io.MousePos = ImVec2((float)mouseX, (float)mouseY);

        bool mouseDown = utils::isMouseButtonDown(utils::e_mouseButton::left);
        if (!mouseDown && io.MouseDown[0])
        {
            io.MouseReleased[0] = true;
            io.MouseClicked[0] = true;
        }
        else
        {
            io.MouseReleased[0] = false;
            io.MouseClicked[0] = false;
        }
        io.MouseDown[0] = mouseDown;

        ImGui::NewFrame();
    }

    void render()
    {
        auto ctx = getMainContext();

        ImGui::Render();

        ImGuiIO& io = ImGui::GetIO();
        ImDrawData* drawData = ImGui::GetDrawData();

        // obtain back buffer for this frame
        FramebufferDesc fd = {};
        fd.m_textures = { nullptr };
        fd.m_pass = s_context.m_renderPass;
        Framebuffer framebuffer = ctx->obtainFramebuffer(fd);

        // update buffers
        size_t vertSize = drawData->TotalVtxCount * sizeof(ImDrawVert);
        size_t indexSize = drawData->TotalIdxCount * sizeof(ImDrawIdx);       

        if (!vertSize || !indexSize)
            return;

        ImVector<ImDrawVert> verts;
        ImVector<ImDrawIdx> indices;
        for (int i = 0; i < drawData->CmdListsCount; ++i)
        {
            ImDrawList* cmdList = drawData->CmdLists[i];
            for (const auto& v : cmdList->VtxBuffer)
                verts.push_back(v);
            for (const auto& in : cmdList->IdxBuffer)
                indices.push_back(in);
        }
        if (vertSize)
        {
            Mem mem;
            mem.m_size = vertSize;  
            mem.m_data = verts.Data;
            if (!s_context.m_vertexBuffer.m_gpuBuffer)
            {
                s_context.m_vertexBuffer.create(mem, e_bufferUsageFlags::vertex);
            }
            else
            {
                s_context.m_vertexBuffer.update(mem);
            }
        }

        if (indexSize)
        {
            Mem mem;
            mem.m_size = indexSize;
            mem.m_data = indices.Data;
            if (!s_context.m_indexBuffer.m_gpuBuffer)
            {
                BufferDesc desc;
                desc.m_usage = e_bufferUsageFlags::index;
                desc.m_access = e_accessFlags::write;
                desc.m_data = mem;
                s_context.m_indexBuffer.create(mem, e_bufferUsageFlags::index);
            }
            else
            {
                s_context.m_indexBuffer.update(mem);
            }
        }

        // begin render commands
        ctx->beginRenderPass(s_context.m_renderPass, framebuffer);
        ctx->bindRenderPipeline(s_context.m_pipeline);
        ctx->bindTexture(s_context.m_fontTexture, s_context.m_sampler, 0, 0);
        ctx->bindBuffer(s_context.m_vertexBuffer.m_gpuBuffer, 0, e_bindFlags::vertexBuffer);

        ImVec2 clipOff = drawData->DisplayPos;
        ImVec2 clipScale = drawData->FramebufferScale;

        Viewport viewport;
        viewport.x = 0;
        viewport.y = 0;
        viewport.m_width = io.DisplaySize.x;
        viewport.m_height = io.DisplaySize.y;
        viewport.m_minDepth = 0.0f;
        viewport.m_maxDepth = 1.0f;
        ctx->setViewport(viewport);

        float scale[2];
        scale[0] = 2.0f / drawData->DisplaySize.x;
        scale[1] = 2.0f / drawData->DisplaySize.y;
        float translate[2];
        translate[0] = -1.0f - drawData->DisplayPos.x * scale[0];
        translate[1] = -1.0f - drawData->DisplayPos.y * scale[1];
        ctx->updateConstants(e_shaderStageFlags::vertex, sizeof(float) * 2, sizeof(float) * 0, scale);
        ctx->updateConstants(e_shaderStageFlags::vertex, sizeof(float) * 2, sizeof(float) * 2, translate);
        
        int vertOffset = 0;
        int indexOffset = 0;
        for (int i = 0; i < drawData->CmdListsCount; ++i)
        {
            const ImDrawList* cmdList = drawData->CmdLists[i];
            for (int j = 0; j < cmdList->CmdBuffer.Size; ++j)
            {
                const ImDrawCmd* pcmd = &cmdList->CmdBuffer[j];
                Rect clipRect;
                clipRect.x = (int)((pcmd->ClipRect.x - clipOff.x) * clipScale.x);
                clipRect.y = (int)((pcmd->ClipRect.y - clipOff.y) * clipScale.y);
                clipRect.w = (u32)((pcmd->ClipRect.z - clipOff.x) * clipScale.x);
                clipRect.h = (u32)((pcmd->ClipRect.w - clipOff.y) * clipScale.y);

                if (clipRect.x < viewport.m_width && clipRect.y < viewport.m_height && clipRect.w >= 0.0f && clipRect.h >= 0.0f)
                {
                    // Negative offsets are illegal
                    if (clipRect.x < 0.0f)
                        clipRect.x = 0;
                    if (clipRect.y < 0.0f)
                        clipRect.y = 0;

                    // Apply scissor/clipping rectangle
                    Rect scissor;
                    scissor.x = (int32_t)(clipRect.x);
                    scissor.y = (int32_t)(clipRect.y);
                    scissor.w = (uint32_t)(clipRect.w - clipRect.x);
                    scissor.h = (uint32_t)(clipRect.h - clipRect.y);
                    ctx->setScissorRect(scissor);

                    // Draw
                    ctx->drawIndexed(pcmd->ElemCount, s_context.m_indexBuffer.m_gpuBuffer, e_indexType::uint16, indexOffset * sizeof(u16), vertOffset);
                }
                indexOffset += pcmd->ElemCount;
            }
            vertOffset += cmdList->VtxBuffer.Size;
        }
        ctx->endRenderEncoding();
        ctx->enqueue();
        ctx->commit();
    }
}