#include "font.h"

#include "gfx.h"
#include "platform.h"

#define STB_TRUETYPE_IMPLEMENTATION
#include <stb_truetype.h>

namespace gfx
{
    struct CharData
    {
        int x0;
        int y0;
        int x1;
        int y1;
        int xadvance;
        int xoffset;
        int yoffset;
    };

    struct TextVertex
    {
        float m_pos[4];
        float m_uv[2];
    };

    class FontInternal : public Font
    {
        stbtt_fontinfo  m_font = {};
        Texture         m_fontTexture = nullptr;
        void*           m_fontTextureMem = nullptr;
        int             m_fontTextureWidth = 1024;
        int             m_fontTextureHeight = 1024;
        CharData*       m_charData = nullptr;

        bool load(const FontDesc& desc) override
        {
            void* buf = nullptr;
            size_t bufSize = 0;
            utils::readFileToBuffer(desc.m_filename.c_str(), buf, bufSize);
            if (!buf)
            {
                return false;
            }

            int res = stbtt_InitFont(&m_font, (unsigned char*)buf, 0);

            if (!res)
                return false;

            m_fontTextureMem = memAlloc(m_fontTextureWidth * m_fontTextureHeight);
            unsigned char* bmp = (unsigned char*)m_fontTextureMem;
            memset(bmp, 0, m_fontTextureWidth * m_fontTextureHeight);

            float scale = stbtt_ScaleForPixelHeight(&m_font, desc.m_scale);
                        
            // init char data storage
            size_t charmax = 0;
            for (auto& codepoints : desc.m_codepoints)
                charmax = std::max(charmax, (size_t)codepoints.m_last);
            m_charData = memNewArray<CharData>(charmax + 1);

            // load codepoints into bitmap mem        
            int xpos = 1;
            int ypos = 1;
            int bottom = 1;
            for (auto& codepoints : desc.m_codepoints)
            {
                char c = codepoints.m_first;
                while (c <= codepoints.m_last)
                {
                    int advance, lsb, x0, y0, x1, y1, gw, gh;
                    int g = stbtt_FindGlyphIndex(&m_font, c);
                    stbtt_GetGlyphHMetrics(&m_font, g, &advance, &lsb);
                    stbtt_GetGlyphBitmapBox(&m_font, g, scale, scale, &x0, &y0, &x1, &y1);
                    gw = x1 - x0;
                    gh = y1 - y0;
                    // move to next line if not enough space for the character
                    if (xpos + x0 + (x1 - x0) > m_fontTextureWidth)
                    {
                        xpos = 1;
                        ypos = bottom;
                        if (ypos >= m_fontTextureHeight)
                            break;
                    }
                    stbtt_MakeGlyphBitmap(&m_font, bmp + xpos + ypos * m_fontTextureWidth, gw, gh, m_fontTextureWidth, scale, scale, g);

                    // set char lookup for this character
                    m_charData[c].x0 = xpos;
                    m_charData[c].y0 = ypos;
                    m_charData[c].x1 = xpos + gw;
                    m_charData[c].y1 = ypos + gh;
                    m_charData[c].xadvance = (int)(scale * advance);
                    m_charData[c].xoffset = x0;
                    m_charData[c].yoffset = y0;

                    // advance
                    xpos = xpos + gw + 1;

                    // update bottom to include largest char on this row
                    if (ypos + gh + 1 > bottom)
                        bottom = ypos + gh + 1;
                    ++c;
                }
            }
            utils::freeFileBuffer(buf);
            

            return true;
        }

        bool createTexture() override
        {
            auto ctx = gfx::getMainContext();

            gfx::TextureDesc td = {};
            td.m_width = m_fontTextureWidth;
            td.m_height = m_fontTextureHeight;
            td.m_depth = 1;
            td.m_format.m_channels = gfx::e_formatChannelFlags::r;
            td.m_format.m_type = gfx::e_formatType::unsignedNorm;
            td.m_format.m_bitDepth = 8;
            td.m_usage = gfx::e_textureUsageFlags::shaderRead;
            td.m_data.m_data = m_fontTextureMem;
            td.m_data.m_size = m_fontTextureWidth * m_fontTextureHeight;
            m_fontTexture = ctx->createTexture(td);

            // free the font texture memory
            memFree(m_fontTextureMem);
            m_fontTextureMem = nullptr;

            return true;
        }

        Geometry createTextMesh(const std::string& text, TextHAlign hAlign, TextVAlign vAlign) override
        {
            auto ctx = gfx::getMainContext();

            Geometry geo = {};

            // first do a loop to determine the text constraints
            float width = 0.0f;
            float height = 0.0f;
            for (auto& c : text)
            {
                CharData chardata = m_charData[c];
                width = width + chardata.xadvance;
                height = std::max(height, (float)(chardata.y1 - chardata.y0));
            }

            float pixToUvX = 1.0f / m_fontTextureWidth;
            float pixToUvY = 1.0f / m_fontTextureHeight;

            std::vector<TextVertex> vertices;
            std::vector<u32> indices;
            float xpos, ypos;
            
            switch (hAlign)
            {
            case e_textHAlign::left:
                xpos = 0.0f;
                break;
            case e_textHAlign::right:
                xpos = -width;
                break;
            case e_textHAlign::centre:
                xpos = -width * 0.5f;
            }

            ypos = 0.0f;

            geo.m_minAABB[0] = FLT_MAX;
            geo.m_minAABB[1] = FLT_MAX;
            geo.m_minAABB[2] = 0.0f;
            geo.m_minAABB[3] = 1.0f;
            geo.m_maxAABB[0] = -FLT_MAX;
            geo.m_maxAABB[1] = -FLT_MAX;
            geo.m_maxAABB[2] = 0.0f;
            geo.m_maxAABB[3] = 1.0f;

            for (auto& c : text)
            {
                CharData chardata = m_charData[c];

                xpos += chardata.xoffset;

                float ybase = ypos;

                float ybottom, ytop;

                float charheight = (float)(chardata.y1 - chardata.y0);

                switch (vAlign)
                {
                case e_textVAlign::bottom:
                {
                    ybottom = ybase;
                    ytop = ybase + charheight;
                }
                break;
                case e_textVAlign::top:
                {
                    ybottom = ybase - charheight;
                    ytop = ybase;
                }
                break;
                case e_textVAlign::centre:
                {
                    float halfheight = charheight * 0.5f;
                    ybottom = ybase - halfheight;
                    ytop = ybase + halfheight;
                }
                break;
                }
                TextVertex v0 = {};                
                v0.m_pos[0] = xpos;
                v0.m_pos[1] = ytop;
                v0.m_pos[2] = 0.0f;
                v0.m_pos[3] = 1.0f;
                v0.m_uv[0] = chardata.x0 * pixToUvX;
                v0.m_uv[1] = chardata.y0 * pixToUvY;

                TextVertex v1 = {};
                v1.m_pos[0] = xpos + (chardata.x1 - chardata.x0);
                v1.m_pos[1] = ybottom;
                v1.m_pos[2] = 0.0f;
                v1.m_pos[3] = 1.0f;
                v1.m_uv[0] = chardata.x1 * pixToUvX;
                v1.m_uv[1] = chardata.y1 * pixToUvY;

                TextVertex v2 = {};
                v2.m_pos[0] = xpos + (chardata.x1 - chardata.x0);
                v2.m_pos[1] = ytop;
                v2.m_pos[2] = 0.0f;
                v2.m_pos[3] = 1.0f;
                v2.m_uv[0] = chardata.x1 * pixToUvX;
                v2.m_uv[1] = chardata.y0 * pixToUvY;

                TextVertex v3 = {};
                v3.m_pos[0] = xpos;
                v3.m_pos[1] = ybottom;
                v3.m_pos[2] = 0.0f;
                v3.m_pos[3] = 1.0f;
                v3.m_uv[0] = chardata.x0 * pixToUvX;
                v3.m_uv[1] = chardata.y1 * pixToUvY;

                u32 v0i = (u32)vertices.size();

                vertices.push_back(v0);
                vertices.push_back(v1);
                vertices.push_back(v2);
                vertices.push_back(v3);

                geo.m_minAABB[0] = std::min(geo.m_minAABB[0], v0.m_pos[0]);
                geo.m_minAABB[1] = std::min(geo.m_minAABB[1], v0.m_pos[1]);
                geo.m_maxAABB[0] = std::max(geo.m_maxAABB[0], v1.m_pos[0]);
                geo.m_maxAABB[1] = std::max(geo.m_maxAABB[1], v1.m_pos[1]);

                indices.push_back(v0i);
                indices.push_back(v0i + 1);
                indices.push_back(v0i + 2);
                indices.push_back(v0i);
                indices.push_back(v0i + 3);
                indices.push_back(v0i + 1);

                xpos += chardata.xadvance;
            }

            geo.m_numVertices = (u32)vertices.size();
            geo.m_numIndices = (u32)indices.size();

            gfx::BufferDesc vbd = {};
            vbd.m_usage = e_bufferUsageFlags::vertex;
            vbd.m_data.m_size = sizeof(TextVertex) * vertices.size();
            vbd.m_data.m_data = vertices.data();
            geo.m_vertexBuffer = ctx->createBuffer(vbd);

            gfx::BufferDesc ibd = {};
            ibd.m_usage = e_bufferUsageFlags::index;
            ibd.m_data.m_size = sizeof(u32) * indices.size();
            ibd.m_data.m_data = indices.data();
            geo.m_indexBuffer = ctx->createBuffer(ibd);

            return geo;
        }

        void unload() override
        {
            auto ctx = gfx::getMainContext();

            if (m_fontTexture)
            {
                ctx->destroyTexture(m_fontTexture);
                m_fontTexture = nullptr;
            }

            if (m_fontTextureMem)
            {
                memFree(m_fontTextureMem);
                m_fontTextureMem = nullptr;
            }
        }

        gfx::Texture getTexture() override
        {
            return m_fontTexture;
        }
    };

    Font* createFont()
    {
        return memNew<FontInternal>();
    }

    void destroyFont(Font* font)
    {
        memDelete(font);
    }
}
