#include "gfx_utils.h"
#include "gfx.h"
#include "platform.h"

#define _USE_MATH_DEFINES
#include <math.h>
#include <unordered_set>
#include <unordered_map>
#include <rapidjson\document.h>
#include <glm\glm.hpp>
#define STB_IMAGE_IMPLEMENTATION
#include <stb_image.h>
#define TINYOBJLOADER_IMPLEMENTATION
#include <tiny_obj_loader.h>
#include <pystring.h>

namespace
{
    struct TexturedVertex
    {
        glm::vec3 m_pos;
        glm::vec2 m_tex;
    };

    struct Indices
    {
        u32     index;
        float   values[3];

        bool operator==(const Indices& other) const
        {
            return memcmp(this->values, other.values, sizeof(values)) == 0;
        }
    };

    std::vector<gfx::MaterialPipeline*> s_materialPipelines;

    const char* k_shaderStageNames[] =
    {
        "vert",
        "tesc",
        "tese",
        "geom",
        "frag",
        "comp"
    };
}

namespace std
{
    template<> struct hash<Indices>
    {
        size_t operator()(Indices const& indices) const
        {
            return (hash<float>()(indices.values[0]) ^ (hash<float>()(indices.values[1]) << 1) ^ (hash<float>()(indices.values[2]) << 2));
        }
    };
}

namespace gfx
{
    void DynamicBuffer::create(Mem initial, gfx::BufferUsageFlags usage)
    {
        auto ctx = gfx::getMainContext();

        m_usage = usage;
        m_data.m_size = initial.m_size;
        m_data.m_data = (void*)new u8[initial.m_size];
        if (initial.m_data)
        {
            memcpy(m_data.m_data, initial.m_data, initial.m_size);
        }

        gfx::BufferDesc bd = {};
        bd.m_data = m_data;
        bd.m_usage = m_usage;
        bd.m_access = gfx::e_accessFlags::write;
        m_gpuBuffer = ctx->createBuffer(bd);
    }

    void DynamicBuffer::update(Mem mem)
    {
        auto ctx = gfx::getMainContext();

        bool resize = mem.m_size != m_data.m_size;
        // take cpu copy of data
        if (resize)
        {
            delete[] m_data.m_data;
            m_data.m_size = mem.m_size;
            m_data.m_data = (void*)new u8[m_data.m_size];

        }
        memcpy(m_data.m_data, mem.m_data, mem.m_size);

        if (resize)
        {
            if (m_gpuBuffer)
            {
                // don't destroy immediatly in case buffer is in flight
                m_destroyBuffers.push_back({ m_gpuBuffer, utils::getFrame() });
            }
            gfx::BufferDesc bd = {};
            bd.m_data = m_data;
            bd.m_usage = m_usage;
            bd.m_access = gfx::e_accessFlags::write;
            m_gpuBuffer = ctx->createBuffer(bd);
        }
        else
        {
            ctx->updateBuffer(m_gpuBuffer, m_data);
        }

        // destroy any old buffers that are no longer in flight
        size_t current = utils::getFrame();
        for (intptr_t i = (intptr_t)m_destroyBuffers.size() - 1; i >= 0; --i)
        {
            if (current - m_destroyBuffers[i].m_destroyFrame > 3)
            {
                ctx->destroyBuffer(m_destroyBuffers[i].m_buffer);
                m_destroyBuffers.erase(m_destroyBuffers.begin() + i);
            }
        }
    }

    void DynamicBuffer::destroy()
    {
        auto ctx = gfx::getMainContext();

        // destroy any remaining buffers - they're probably not in flight any more
        for (size_t i = 0; i < m_destroyBuffers.size(); ++i)
        {
            ctx->destroyBuffer(m_destroyBuffers[i].m_buffer);
        }
        m_destroyBuffers.clear();

        // destroy current buffer
        if (m_gpuBuffer)
        {
            ctx->destroyBuffer(m_gpuBuffer);
            m_gpuBuffer = nullptr;
        }

        // clean cpu copy
        if (m_data.m_data)
        {
            delete[] m_data.m_data;
            m_data.m_data = nullptr;
            m_data.m_size = 0;
        }
    }

    Shader loadShader(const ShaderLoadDesc& desc, VertexInputDesc* vertexInput = nullptr)
    {
        auto ctx = gfx::getMainContext();

        void* buf = nullptr;
        size_t len = 0;
        utils::readFileToBuffer(desc.m_filepath.c_str(), buf, len);
        rapidjson::Document d;
        bool result = d.Parse<rapidjson::kParseStopWhenDoneFlag>((const char*)buf).HasParseError();
        if (result)
        {
            auto error = d.GetParseError();
            printf("Error!");
        }

        std::string stageName = "";
        for (u32 i = 0; i < 6; ++i)
        {
            if (desc.m_stage == (1 << i))
            {
                stageName = k_shaderStageNames[i];
            }
        }

        const rapidjson::Value& s = (d.HasMember("variations")) ? d["variations"][0][stageName.c_str()] : d[stageName.c_str()];

        ShaderDesc shaderDesc;
        size_t offset = 0;
        for (u32 i = 0, n = s["Push Constants"].GetArray().Size(); i < n; ++i)
        {
            const rapidjson::Value& pushConstant = s["Push Constants"][i];
            PushConstantDesc pushDesc;
            pushDesc.m_size = 0;
            for (u32 j = 0; j < pushConstant["Members"].GetArray().Size(); ++j)
            {
                const rapidjson::Value& member = pushConstant["Members"][j];
                size_t size = 0;
                std::string typeString = member["Type"].GetString();
                if (typeString == "uint")
                    size = sizeof(uint32_t);
                else if (typeString == "int")
                    size = sizeof(int);
                else if (typeString == "float")
                    size = sizeof(float);
                size = size * member["Count"].GetUint();
                pushDesc.m_size += (u32)size;
            }
            pushDesc.m_offset = (u32)offset;
            offset += pushDesc.m_size;
            pushDesc.m_binding = pushConstant["Binding"].GetUint();
            shaderDesc.m_pushConstants.push_back(pushDesc);
        }

        for (int i = 0, n = s["Uniforms"].GetArray().Size(); i < n; ++i)
        {
            const rapidjson::Value& uniform = s["Uniforms"][i];
            DescriptorDesc descriptor;
            std::string typeString = uniform["Type"].GetString();
            if (typeString == "Combined Sampler")
                descriptor.m_type = e_descriptorType::combinedImageSampler;
            else if (typeString == "Sampler")
                descriptor.m_type = e_descriptorType::sampler;
            else if (typeString == "Texture")
                descriptor.m_type = e_descriptorType::sampledImage;
            else if (typeString == "Storage Image")
                descriptor.m_type = e_descriptorType::storageImage;
            else if (typeString == "Uniform Buffer")
                descriptor.m_type = e_descriptorType::uniformBuffer;
            else if (typeString == "Storage Buffer")
                descriptor.m_type = e_descriptorType::storageBuffer;
            descriptor.m_count = uniform["Count"].GetUint();
            if (descriptor.m_count == 0)
                descriptor.m_count = 128;
            descriptor.m_binding = uniform["Binding"].GetUint();
            if (uniform.FindMember("Mapping") != uniform.MemberEnd())
                descriptor.m_mapping = uniform["Mapping"].GetUint();
            descriptor.m_name = uniform["Name"].GetString();
            shaderDesc.m_descriptors.push_back(descriptor);
        }

        delete[] buf;
        std::string api = gfx::getAPI();
        assert(s["Shader Files"].HasMember(api.c_str()));
        auto& shaderJson = s["Shader Files"][api.c_str()];
        std::string shaderData = std::string(shaderJson.GetString(), shaderJson.GetStringLength());
        shaderDesc.m_data.m_data = (void*)shaderData.c_str();
        shaderDesc.m_data.m_size = s["Shader Files"][api.c_str()].GetStringLength();
        shaderDesc.m_entry = "main";
        shaderDesc.m_stage = desc.m_stage;
        Shader shader = ctx->createShader(shaderDesc);

        if (desc.m_stage & e_shaderStageFlags::vertex && vertexInput)
        {
            // read vertex layout from vertex shader input
            size_t perVertexOffset = 0;
            size_t perInstanceOffset = 0;
            for (int i = 0, n = s["Inputs"].GetArray().Size(); i < n; ++i)
            {
                const rapidjson::Value& input = s["Inputs"][i];
                bool instanced = input["Instanced"].GetBool();
                u32 count = input["Count"].GetUint();
                u32 numIndices = count > 4 ? count / 4 : 1;
                count = count > 4 ? 4 : count;
                VertexAttributeDesc desc;
                desc.m_location = i;
                desc.m_binding = instanced ? 1 : 0;
                desc.m_format.m_channels = (1 << count) - 1;
                desc.m_format.m_bitDepth = 32;
                const char* typeString = input["Type"].GetString();
                if (strcmp(typeString, "float") == 0)
                    desc.m_format.m_type = e_formatType::signedFloat;
                else if (strcmp(typeString, "uint") == 0)
                    desc.m_format.m_type = e_formatType::unsignedInt;
                else if (strcmp(typeString, "int") == 0)
                    desc.m_format.m_type = e_formatType::signedInt;
                else
                    desc.m_format.m_type = e_formatType::unsignedNorm;

                for (u32 j = 0; j < numIndices; ++j)
                {
                    desc.m_offset = instanced ? (u32)perInstanceOffset : (u32)perVertexOffset;
                    if (instanced)
                        perInstanceOffset += (size_t)((desc.m_format.m_bitDepth / 8) * input["Count"].GetUint() * (1.0f / numIndices));
                    else
                        perVertexOffset += (size_t)((desc.m_format.m_bitDepth / 8) * input["Count"].GetUint() * (1.0f / numIndices));
                    desc.m_location = i + j;
                    vertexInput->m_attributes.push_back(desc);
                }
            }
            vertexInput->m_bindings.push_back({ 0, perVertexOffset, e_inputRateType::vertex });
            if (perInstanceOffset > 0)
                vertexInput->m_bindings.push_back({ 1, perInstanceOffset, e_inputRateType::instance });
        }
        else if (desc.m_stage & e_shaderStageFlags::fragment)
        {
            // read render targets from fragment shader output

        }

        return shader;
    }

    RenderPipeline loadRenderPipeline(const RenderPipelineLoadDesc& desc)
    {
        auto ctx = gfx::getMainContext();

        // read shader files
        std::vector<Shader> s;
        VertexInputDesc vertInput;
        for (auto& shaderDesc : desc.m_shaders)
            s.push_back(loadShader(shaderDesc, &vertInput));

        MrtBlendDesc mrtBlend;
        mrtBlend.m_blends = desc.m_blends;

        RenderPipelineDesc pipelineInfo;
        pipelineInfo.m_shaders = s;
        pipelineInfo.m_vertexInput = vertInput;
        pipelineInfo.m_inputAssembly = desc.m_input;
        pipelineInfo.m_viewports = { { 0, 0, 1, 1, 0, 1 } };
        pipelineInfo.m_scissors = { {} };
        pipelineInfo.m_rasterizer = desc.m_raster;
        pipelineInfo.m_multisample = desc.m_multisample;
        pipelineInfo.m_blendDesc = mrtBlend;
        pipelineInfo.m_depthStencil = desc.m_depthStencil;
        pipelineInfo.m_dynamic = desc.m_dynamic;
        pipelineInfo.m_renderPass = desc.m_renderPass;
        pipelineInfo.m_subpass = 0;
        return ctx->createRenderPipeline(pipelineInfo);
    }

    ComputePipeline loadComputePipeline(const std::string& filepath)
    {
        auto ctx = gfx::getMainContext();
        ShaderLoadDesc loadDesc;
        loadDesc.m_filepath = filepath;
        loadDesc.m_stage = e_shaderStageFlags::compute;
        Shader shader = loadShader(loadDesc);

        ComputePipelineDesc desc;
        desc.m_shader = shader;
        return ctx->createComputePipeline(desc);
    }

    Texture loadTexture(const std::string& filepath)
    {
        auto ctx = gfx::getMainContext();

        int width, height, channels;
        stbi_uc* pixels = stbi_load(filepath.c_str(), &width, &height, &channels, STBI_rgb_alpha);
        if (!pixels)
            return nullptr;
        
        TextureDesc desc;
        desc.m_width = width;
        desc.m_height = height;
        desc.m_format.m_channels = e_formatChannelFlags::rgba;
        desc.m_format.m_bitDepth = 8;
        desc.m_format.m_type = e_formatType::unsignedNorm;
        desc.m_usage = e_textureUsageFlags::shaderRead;
        desc.m_data.m_data = pixels;
        desc.m_data.m_size = width * height * 4;
        return ctx->createTexture(desc);
    }

    Material* loadMaterial(const tinyobj::material_t& params, const std::string& texturePath, const std::string& materialTemplate, const RenderPipelineLoadDesc& pipelineLoadDesc)
    {
        auto ctx = gfx::getMainContext();

        Material* material = memNew<Material>();
        material->m_name = params.name;
        
        MaterialBuffer data;
        memcpy(data.m_ambient, params.ambient, sizeof(float) * 3);
        data.m_ambient[3] = 1.0f;
        memcpy(data.m_diffuse, params.diffuse, sizeof(float) * 3);
        data.m_diffuse[3] = 1.0f;
        memcpy(data.m_specular, params.specular, sizeof(float) * 3);
        data.m_specular[3] = 1.0f;
        data.m_shininess = params.shininess;
        data.m_ior = params.ior;
        data.m_dissolve = params.dissolve;
        data.m_illum = (float)params.illum;

        u32 materialFlags = 0;
        u32 binding = 2;
        std::string header = "";
        std::vector<Texture> textures;
        if (!params.diffuse_texname.empty())
        {
            std::string texPath = pystring::os::path::join(texturePath, pystring::os::path::basename(params.diffuse_texname));
            textures.push_back(loadTexture(texPath));
            materialFlags |= e_materialFlags::diffuseMap;
            header += "#define DIFFUSE_MAP 1\n";
            header += "#define DIFFUSE_BINDING " + std::to_string(binding++) + "\n";
        }

        if (!params.specular_texname.empty())
        {
            std::string texPath = pystring::os::path::join(texturePath, pystring::os::path::basename(params.specular_texname));
            textures.push_back(loadTexture(texPath));
            materialFlags |= e_materialFlags::specularMap;
            header += "#define SPECULAR_MAP 1\n";
            header += "#define SPECULAR_BINDING " + std::to_string(binding++) + "\n";
        }

        if (!params.bump_texname.empty())
        {
            std::string texPath = pystring::os::path::join(texturePath, pystring::os::path::basename(params.bump_texname));
            textures.push_back(loadTexture(texPath));
            materialFlags |= e_materialFlags::bumpMap;
            header += "#define BUMP_MAP 1\n";
            header += "#define BUMP_BINDING " + std::to_string(binding++) + "\n";
        }

        if (!params.alpha_texname.empty())
        {
            std::string texPath = pystring::os::path::join(texturePath, pystring::os::path::basename(params.alpha_texname));
            textures.push_back(loadTexture(texPath));
            materialFlags |= e_materialFlags::alphaMap;
            header += "#define ALPHA_MAP 1\n";
            header += "#define ALPHA_BINDING " + std::to_string(binding++) + "\n";
        } 

        BufferDesc desc;
        desc.m_usage = e_bufferUsageFlags::uniform;
        desc.m_access = e_accessFlags::none;
        desc.m_data.m_size = sizeof(MaterialBuffer);
        desc.m_data.m_data = &data;
        material->m_data = ctx->createBuffer(desc);
        material->m_numTextures = textures.size();
        material->m_textures = new Texture[textures.size()];
        for (size_t i = 0; i < textures.size(); ++i)
        {
            material->m_textures[i] = textures[i];
        }

        MaterialPipeline* pipeline = nullptr;
        for (auto& p : s_materialPipelines)
        {
            if (p->m_flags == materialFlags)
            {
                pipeline = p;
                break;
            }
        }
        if(!pipeline)
        {
            pipeline = memNew<MaterialPipeline>();
            pipeline->m_flags = materialFlags;

            // create the required shader from template file
            void* buf = nullptr;
            size_t len = 0;
            utils::readFileToBuffer(materialTemplate.c_str(), buf, len);
            std::string templateString = std::string((char*)buf);
            templateString.resize(len);
            int templatePos = pystring::find(templateString, "# TEMPLATE_BEGIN");
            templateString = templateString.insert(templatePos, header);
            templateString = pystring::replace(templateString, "# TEMPLATE_BEGIN", "");

            std::string base, dir;
            pystring::os::path::split(dir, base, materialTemplate);
            base = pystring::replace(base, ".template", ".frag");
            std::string completedTemplate = pystring::os::path::join(dir, std::to_string(materialFlags) + "_" + base);
            utils::writeFileFromBuffer(completedTemplate.c_str(), (void*)templateString.c_str(), templateString.size());
            
            std::string compiled = pystring::os::path::join(dir, std::to_string(materialFlags) + "_" + pystring::replace(base, ".", "_") + ".shader");
            std::string command = "ShaderTool -i ";
            command += completedTemplate;
            command += " -o ";
            command += compiled;
            utils::runCommand(command.c_str());

            RenderPipelineLoadDesc pipelineDesc = pipelineLoadDesc;
            pipelineDesc.m_shaders.push_back({ compiled, e_shaderStageFlags::fragment });
            pipeline->m_pipeline = loadRenderPipeline(pipelineDesc);

            std::remove(completedTemplate.c_str());
            std::remove(compiled.c_str());

            s_materialPipelines.push_back(pipeline);
        }
        material->m_pipeline = pipeline;
        pipeline->increment();

        return material;
    }

    void createModelBuffers(std::vector<float>& verts, std::vector<u32>& indices, Buffer& vertexBuffer, Buffer& indexBuffer)
    {
        auto ctx = gfx::getMainContext();

        BufferDesc vertDesc;
        vertDesc.m_access = e_accessFlags::none;
        vertDesc.m_usage = e_bufferUsageFlags::vertex;
        vertDesc.m_data.m_data = verts.data();
        vertDesc.m_data.m_size = sizeof(float) * verts.size();
        vertexBuffer = ctx->createBuffer(vertDesc);

        BufferDesc indexDesc;
        indexDesc.m_access = e_accessFlags::none;
        indexDesc.m_usage = e_bufferUsageFlags::index;
        indexDesc.m_data.m_data = indices.data();
        indexDesc.m_data.m_size = sizeof(u32) * indices.size();
        indexBuffer = ctx->createBuffer(indexDesc);
    }

    ModelInstance* createModelInstanceSingleBuffer( const std::vector<float>& verts, 
                                                    const std::vector<u32>& indices, 
                                                    const size_t& numVertFloats, 
                                                    Material* material, 
                                                    float minVert[4],
                                                    float maxVert[4],
                                                    std::vector<float>& bufferVerts, 
                                                    std::vector<u32>& bufferInds)
    {
        Model* model = memNew<Model>();

        model->m_indexCount = (u32)indices.size();
        model->m_vertexBufferIndexOffset = (u32)(bufferVerts.size() / numVertFloats);
        model->m_vertexBufferByteOffset = bufferVerts.size() * sizeof(float);
        model->m_indexBufferOffset = (u32)(bufferInds.size() * sizeof(u32));
        memcpy(model->m_aabbMin, minVert, sizeof(float) * 4);
        memcpy(model->m_aabbMax, maxVert, sizeof(float) * 4);

        for (auto& v : verts)
        {
            bufferVerts.push_back(v);
        }

        for (auto& i : indices)
        {
            bufferInds.push_back(i);
        }

        return new ModelInstance(model, material);
    }
    
    ModelInstance* createModelInstance(std::vector<float>& verts, std::vector<u32>& indices, Material* material, float minVert[4], float maxVert[4])
    {
        auto ctx = gfx::getMainContext();

        Model* model = memNew<Model>();
        createModelBuffers(verts, indices, model->m_vertexBuffer, model->m_indexBuffer);
        model->m_indexCount = (u32)indices.size();
        memcpy(model->m_aabbMin, minVert, sizeof(float) * 4);
        memcpy(model->m_aabbMax, maxVert, sizeof(float) * 4);

        return new ModelInstance(model, material);
    }

    std::vector<ModelInstance*> loadOBJ(const OBJLoadDesc& params)
    {
        auto ctx = getMainContext();

        tinyobj::attrib_t attrib;
        std::vector<tinyobj::shape_t> shapes;
        std::vector<tinyobj::material_t> mats;
        std::string err, warn;
        std::string materialPath = pystring::os::path::dirname(params.m_objFilepath) + "/";
        if (!tinyobj::LoadObj(&attrib, &shapes, &mats, &err, &warn, params.m_objFilepath.c_str(), materialPath.c_str()))
        {
            printf("Failed to load obj: %s\n", params.m_objFilepath.c_str());
            return std::vector<ModelInstance*>();
        }

        // create materials
        std::vector<Material*> materials(mats.size());
        for (size_t i = 0; i < mats.size(); ++i)
        {
            auto& mat = mats[i];

            materials[i] = loadMaterial(mat, params.m_texturePath, params.m_materialTemplate, params.m_pipeline);
        }
        
        const size_t numVertFloats = 9;
        auto& verts = attrib.vertices;
        auto& norms = attrib.normals;
        auto& texs = attrib.texcoords;
        auto& colors = attrib.colors;

        std::vector<ModelInstance*> instances;
        
        std::vector<float> singleBufferVertData;
        std::vector<u32> singleBufferIndexData;

        for (auto& shape : shapes)
        {
            std::vector<float> vertData;
            std::vector<u32> indexData;
            std::unordered_set<Indices> uniqueIndices;
            int prevMat = -1;
            float minVert[4] = { FLT_MAX, FLT_MAX, FLT_MAX, 0.0f };
            float maxVert[4] = { -FLT_MAX, -FLT_MAX, -FLT_MAX, 0.0f };
            
            auto& inds = shape.mesh.indices;
            for (size_t i = 0, n = inds.size(); i < n; ++i)
            {
                size_t faceIndex = i / 3;

                Indices indices;
                indices.values[0] = inds[i].vertex_index;
                indices.values[1] = inds[i].texcoord_index;
                indices.values[2] = inds[i].normal_index;

                // split shapes additionally per material
                int mat = shape.mesh.material_ids[faceIndex];
                if (prevMat >= 0 && mat != prevMat)
                {
                    ModelInstance* instance = nullptr;
                    if (params.m_singleBuffer)
                    {
                        instance = createModelInstanceSingleBuffer( vertData,
                                                                    indexData,
                                                                    numVertFloats,
                                                                    materials[prevMat],
                                                                    minVert,
                                                                    maxVert,
                                                                    singleBufferVertData,
                                                                    singleBufferIndexData);
                    }
                    else
                    {
                        instance = createModelInstance(vertData, indexData, materials[prevMat], minVert, maxVert);
                    }
                    instances.push_back(instance);

                    vertData.clear();
                    indexData.clear();
                    uniqueIndices.clear();
                    for (int n = 0; i < 3; ++i)
                    {
                        minVert[n] = FLT_MAX;
                        maxVert[n] = -FLT_MAX;
                    }
                }
                prevMat = mat;

                auto unique = uniqueIndices.find(indices);
                if (unique == uniqueIndices.end())
                {
                    indices.index = (u32)(vertData.size() / numVertFloats);

                    // add a new vertex
                    // position
                    glm::vec4 pos = glm::vec4(0.0f, 0.0f, 0.0f, 1.0f);
                    if (indices.values[0] >= 0)
                    {
                        pos = glm::vec4(
                            verts[3 * indices.values[0] + 0],
                            verts[3 * indices.values[0] + 1],
                            verts[3 * indices.values[0] + 2],
                            1.0f
                        );
                    }
                    vertData.push_back(pos.x);
                    vertData.push_back(pos.y);
                    vertData.push_back(pos.z);
                    vertData.push_back(pos.w);
                    
                    // calculate aabb
                    for (int n = 0; n < 3; ++n)
                    {
                        if (pos[n] < minVert[n])
                            minVert[n] = pos[n];

                        if (pos[n] > maxVert[n])
                            maxVert[n] = pos[n];
                    }

                    // texcoord
                    glm::vec2 tex = glm::vec2(0.0f);
                    if (indices.values[1] >= 0)
                    {
                        tex = glm::vec2(
                            texs[2 * indices.values[1] + 0],
                            1.0f - texs[2 * indices.values[1] + 1]
                        );
                    }
                    vertData.push_back(tex.x);
                    vertData.push_back(tex.y);

                    // normal
                    glm::vec3 normal = glm::vec3(0.0f);
                    if (indices.values[2] >= 0)
                    {
                        normal = glm::vec3(
                            norms[3 * indices.values[2] + 0],
                            norms[3 * indices.values[2] + 1],
                            norms[3 * indices.values[2] + 2]
                        );
                    }
                    vertData.push_back(normal.x);
                    vertData.push_back(normal.y);
                    vertData.push_back(normal.z);
                    
                    uniqueIndices.insert(indices);
                    indexData.push_back(indices.index);
                }
                else
                {
                    indexData.push_back(unique->index);
                }
            }

            ModelInstance* instance = nullptr;
            if (params.m_singleBuffer)
            {
                instance = createModelInstanceSingleBuffer( vertData, 
                                                            indexData, 
                                                            numVertFloats, 
                                                            materials[prevMat],
                                                            minVert,
                                                            maxVert,
                                                            singleBufferVertData, 
                                                            singleBufferIndexData);
            }
            else
            {
                instance = createModelInstance(vertData, indexData, materials[prevMat], minVert, maxVert);
            }
            instances.push_back(instance);
        }

        if (params.m_singleBuffer)
        {
            Buffer vertexBuffer, indexBuffer;
            createModelBuffers(singleBufferVertData, singleBufferIndexData, vertexBuffer, indexBuffer);

            for (auto& i : instances)
            {
                i->m_model->m_vertexBuffer = vertexBuffer;
                i->m_model->m_indexBuffer = indexBuffer;
            }
        }
           
        return instances;
    }

    u32 generateSphere(Buffer& vertexBuffer, Buffer& indexBuffer, u32 res)
    {
        float longitude = (float)res;
        float latitude = (float)res;
        std::vector<TexturedVertex> verts;
        std::vector<u32> inds;
        // generate vertices
        for (float y = 0.0f; y < longitude; y += 1.0f)
        {
            if (y > 0.0f && y < longitude - 1.0f)
            {
                for (float x = 0.0f; x < latitude; x += 1.0f)
                {
                    TexturedVertex vert;
                    vert.m_pos.x = sinf(M_PI * (x / latitude)) * cosf(2.0f * M_PI * (y / longitude));
                    vert.m_pos.y = sinf(M_PI * (x / latitude)) * sinf(2.0f * M_PI * (y / longitude));
                    vert.m_pos.z = cosf(M_PI * (x / latitude));
                    verts.push_back(vert);
                }
            }
            else
            {
                // use a single point at the poles
                TexturedVertex vert;
                vert.m_pos.x = 0.0f;
                vert.m_pos.y = (y > 0.0f) ? 1.0f : -1.0f;
                vert.m_pos.z = 0.0f;
                verts.push_back(vert);
            }
        }

        // generate indices
        for (int y = 0; y < longitude - 1; ++y)
        {
            for (int x = 0; x < latitude; ++x)
            {
                if (y > 0 && y < longitude - 2)
                {
                    u32 indOffsetA = (u32)(1 + latitude * (y - 1));
                    u32 indOffsetB = (u32)(1 + latitude * y);
                    inds.push_back(indOffsetA + x);
                    inds.push_back(indOffsetA + ((x + 1) % (int)latitude));
                    inds.push_back(indOffsetB + ((x + 1) % (int)latitude));
                    inds.push_back(indOffsetA + x);
                    inds.push_back(indOffsetB + ((x + 1) % (int)latitude));
                    inds.push_back(indOffsetB + x);

                }
                else if (y > 0)
                {
                    u32 indOffsetA = (u32)(1 + latitude * (y - 1));
                    inds.push_back(indOffsetA + x);
                    inds.push_back(indOffsetA + ((x + 1) % (int)latitude));
                    inds.push_back((u32)verts.size() - 1);
                }
                else
                {
                    inds.push_back(0);
                    u32 indOffset = (u32)(1 + latitude * (y - 1));
                    inds.push_back(indOffset + x);
                    inds.push_back(indOffset + ((x + 1) % (u32)latitude));
                }
            }
        }

        // create buffers
        auto ctx = getMainContext();
        BufferDesc vertDesc;
        vertDesc.m_access = e_accessFlags::none;
        vertDesc.m_usage = e_bufferUsageFlags::vertex;
        vertDesc.m_data.m_data = verts.data();
        vertDesc.m_data.m_size = sizeof(TexturedVertex) * verts.size();
        vertexBuffer = ctx->createBuffer(vertDesc);

        BufferDesc indexDesc;
        indexDesc.m_access = e_accessFlags::none;
        indexDesc.m_usage = e_bufferUsageFlags::index;
        indexDesc.m_data.m_data = inds.data();
        indexDesc.m_data.m_size = sizeof(u32) * inds.size();
        indexBuffer = ctx->createBuffer(indexDesc);

        return (u32)inds.size();
    }

    u32 generateQuad(Buffer& vertexBuffer, Buffer& indexBuffer, u32 res)
    {
        float verts[] = {
            -1.0f,  1.0f, 0.0f, 1.0f,
             0.0f,  1.0f,
             1.0f,  1.0f, 0.0f, 1.0f,
             1.0f,  1.0f,
             1.0f, -1.0f, 0.0f, 1.0f,
             1.0f,  0.0f,
            -1.0f, -1.0f, 0.0f, 1.0f,
             0.0f,  0.0f
        };

        u32 inds[] = {
            0, 1, 2,
            0, 2, 3
        };

        auto ctx = getMainContext();
        BufferDesc vertDesc;
        vertDesc.m_access = e_accessFlags::none;
        vertDesc.m_usage = e_bufferUsageFlags::vertex;
        vertDesc.m_data.m_data = verts;
        vertDesc.m_data.m_size = sizeof(float) * 24;
        vertexBuffer = ctx->createBuffer(vertDesc);

        BufferDesc indexDesc;
        indexDesc.m_access = e_accessFlags::none;
        indexDesc.m_usage = e_bufferUsageFlags::index;
        indexDesc.m_data.m_data = inds;
        indexDesc.m_data.m_size = sizeof(u32) * 6;
        indexBuffer = ctx->createBuffer(indexDesc);

        return 6;
    }
}
