#include "maths_utils.h"

#include <glm/gtx/transform.hpp>
#include <glm/gtx/norm.hpp>

namespace maths
{
	float min(float a, float b)
	{
		return a < b ? a : b;
	}

	float max(float a, float b)
	{
		return a > b ? a : b;
	}

	bool rayAABBIntersection(const Ray& ray, const AABB& aabb, float& t)
	{		
		float tx1 = (aabb.m_min.x - ray.m_origin.x) * ray.m_invRay.x;
		float tx2 = (aabb.m_max.x - ray.m_origin.x) * ray.m_invRay.x;

		float tmin = min(tx1, tx2);
		float tmax = max(tx1, tx2);

		float ty1 = (aabb.m_min.y - ray.m_origin.y) * ray.m_invRay.y;
		float ty2 = (aabb.m_max.y - ray.m_origin.y) * ray.m_invRay.y;
		
		tmin = max(tmin, min(ty1, ty2));
		tmax = min(tmax, max(ty1, ty2));

		float tz1 = (aabb.m_min.z - ray.m_origin.z) * ray.m_invRay.z;
		float tz2 = (aabb.m_max.z - ray.m_origin.z) * ray.m_invRay.z;

		tmin = max(tmin, min(tz1, tz2));
		tmax = min(tmax, max(tz1, tz2));

		t = tmin;

		return tmax >= max(0.0f, tmin);
	}

	bool frustumOBBIntersection(const Frustum& frustum, const glm::mat4& view, const AABB& aabb)
	{
		float near = frustum.m_nearPlane;
		float far = frustum.m_farPlane;
		float right = frustum.m_nearRight;
		float top = frustum.m_nearTop;

		glm::vec3 corners[4] = 
		{
			{aabb.m_min.x, aabb.m_min.y, aabb.m_min.z},
			{aabb.m_max.x, aabb.m_min.y, aabb.m_min.z},
			{aabb.m_min.x, aabb.m_max.y, aabb.m_min.z},
			{aabb.m_min.x, aabb.m_min.y, aabb.m_max.z}
		};

		for (size_t i = 0; i < 4; ++i)
			corners[i] = (view * glm::vec4(corners[i], 1.0f));

		glm::vec3 axes[3] =
		{
			corners[1] - corners[0],
			corners[2] - corners[0],
			corners[3] - corners[0]
		};

		glm::vec3 centre = corners[0] + 0.5f * (axes[0] + axes[1] + axes[2]);
		glm::vec3 extents = glm::vec3(glm::length(axes[0]), glm::length(axes[1]), glm::length(axes[2]));
		axes[0] = axes[0] / extents.x;
		axes[1] = axes[1] / extents.y;
		axes[2] = axes[2] / extents.z;
		extents *= 0.5f;

		// near/far test
		{
			float mc = centre.z;
			float radius = 0.0f;
			for (size_t i = 0; i < 3; ++i)
				radius += fabsf(axes[i].z) * extents[i];

			float obbMin = mc - radius;
			float obbMax = mc + radius;

			if (obbMin > near || obbMax < far)
				return false;
		}

		// frustum planes
		{
			glm::vec3 m[4] =
			{
				{ near, 0.0f, right },
				{ -near, 0.0f, right },
				{ 0.0f, -near, top },
				{ 0.0f, near, top }
			};

			for (size_t i = 0; i < 4; ++i)
			{
				float mx = fabsf(m[i].x);
				float my = fabsf(m[i].y);
				float mz = m[i].z;
				float mc = glm::dot(m[i], centre);

				float radius = 0.0f;
				for (size_t j = 0; j < 3; ++j)
					radius += fabsf(glm::dot(m[i], axes[j])) * extents[j];

				float obbMin = mc - radius;
				float obbMax = mc + radius;

				float p = right * mx + top * my;

				float tau0 = near * mz - p;
				float tau1 = near * mz + p;

				if (tau0 < 0.0f)
					tau0 *= far / near;

				if (tau1 > 0.0f)
					tau1 *= far / near;

				if (obbMin > tau1 || obbMax < tau0)
					return false;
			}
		}

		// OBB axes
		{
			for (size_t i = 0; i < 3; ++i)
			{
				float mx = fabsf(axes[i].x);
				float my = fabsf(axes[i].y);
				float mz = axes[i].z;
				float mc = glm::dot(axes[i], centre);

				float radius = extents[i];

				float obbMin = mc - radius;
				float obbMax = mc + radius;

				float p = right * mx + top * my;
				float tau0 = near * mz - p;
				float tau1 = near * mz + p;

				if (tau0 < 0.0f)
					tau0 *= far / near;
				if (tau1 > 0.0f)
					tau1 *= far / near;

				if (obbMin > tau0 || obbMax < tau1)
					return false;
			}
		}

		{
			for (size_t i = 0; i < 3; ++i)
			{
				glm::vec3 m = { 0.0f, -axes[i].z, axes[i].y };
				float mx = 0.0f;
				float my = fabsf(m.y);
				float mz = m.z;
				float mc = my * centre.y + m.z * centre.z;

				float radius = 0.0f;
				for (size_t j = 0; j < 3; ++j)
					radius += fabsf(glm::dot(m, axes[j])) * extents[j];

				float obbMin = mc - radius;
				float obbMax = mc + radius;

				float p = top * my;
				float tau0 = near * mz - p;
				float tau1 = near * mz + p;
				if (tau0 < 0.0f)
					tau0 *= far / near;
				if (tau1 > 0.0f)
					tau1 *= far / near;

				if (obbMin > tau1 || obbMax < tau0)
					return false;
			}
		}

		{
			for (size_t i = 0; i < 3; ++i)
			{
				glm::vec3 m = { axes[i].z, 0.0f, -axes[i].x };
				float mx = fabsf(m.x);
				float my = 0.0f;
				float mz = m.z;
				float mc = m.x * centre.x + m.z * centre.z;

				float radius = 0.0f;
				for (size_t j = 0; j < 3; ++j)
					radius += fabsf(glm::dot(m, axes[j])) * extents[j];

				float obbMin = mc - radius;
				float obbMax = mc + radius;

				float p = right * mx;
				float tau0 = near * mz - p;
				float tau1 = near * mz + p;
				if (tau0 < 0.0f)
					tau0 *= far / near;
				if (tau1 > 0.0f)
					tau1 *= far / near;

				if (obbMin > tau1 || obbMax < tau0)
					return false;
			}
		}

		{
			for (size_t i = 0; i < 3; ++i)
			{
				glm::vec3 m[4] =
				{
					glm::cross({-right, 0.0f, near}, axes[i]),
					glm::cross({ right, 0.0f, near}, axes[i]),
					glm::cross({ 0.0f,  top,  near}, axes[i]),
					glm::cross({ 0.0f, -top,  near}, axes[i])
				};

				for (size_t j = 0; j < 4; ++j)
				{
					float mx = fabsf(m[j].x);
					float my = fabsf(m[j].y);
					float mz = m[j].z;

					constexpr float epsilon = 1e-4f;
					if (mx < epsilon && my < epsilon && fabsf(mz) < epsilon)
						continue;

					float mc = glm::dot(m[j], centre);

					float radius = 0.0f;
					for (size_t k = 0; k < 3; ++k)
						radius += fabsf(glm::dot(m[j], axes[k])) * extents[k];

					float obbMin = mc - radius;
					float obbMax = mc + radius;

					float p = right * mx + top * my;
					float tau0 = near * mz - p;
					float tau1 = near * mz + p;
					if (tau0 < 0.0f)
						tau0 *= far / near;
					if (tau1 > 0.0f)
						tau1 *= far / near;
					
					if (obbMin > tau1 || obbMax < tau0)
						return false;

				}
			}
		}

		return true;
	}

	bool within(float edge0, float edge1, float t)
	{
		return t >= edge0 && t <= edge1;
	}

	bool frustumOBBIntersectionSimple(const glm::mat4& vp, const OBB& obb)
	{
		// define corners
		glm::vec4 corners[8] =
		{
			{ obb.m_min.x, obb.m_min.y, obb.m_min.z, 1.0 },
			{ obb.m_max.x, obb.m_min.y, obb.m_min.z, 1.0 },
			{ obb.m_min.x, obb.m_max.y, obb.m_min.z, 1.0 },
			{ obb.m_max.x, obb.m_max.y, obb.m_min.z, 1.0 },
			{ obb.m_min.x, obb.m_min.y, obb.m_max.z, 1.0 },
			{ obb.m_max.x, obb.m_min.y, obb.m_max.z, 1.0 },
			{ obb.m_min.x, obb.m_max.y, obb.m_max.z, 1.0 },
			{ obb.m_max.x, obb.m_max.y, obb.m_max.z, 1.0 }
		};

		// check if any corners fall within projection space
		for (size_t i = 0; i < 8; ++i)
		{
			corners[i] = vp * corners[i];
			
			bool x = within(-corners[i].w, corners[i].w, corners[i].x);
			bool y = within(-corners[i].w, corners[i].w, corners[i].y);
			bool z = within(0.0f, corners[i].w, corners[i].z);
			if ( x && y && z)
			{
				return true;
			}
		}

		// check if the projected aabb falls intersects with any of the frustum edges
		AABB projected = {};
		projected.m_min = glm::vec3(FLT_MAX);
		projected.m_max = glm::vec3(-FLT_MAX);
		for (size_t i = 0; i < 8; ++i)
		{
			corners[i] /= corners[i].w;
			projected.m_min = glm::min(projected.m_min, glm::vec3(corners[i]));
			projected.m_max = glm::max(projected.m_max, glm::vec3(corners[i]));
		}

		glm::vec3 dirs[4] =
		{
			{ -1.0f, -1.0f, 1.0f },
			{  1.0f, -1.0f, 1.0f },
			{ -1.0f,  1.0f, 1.0f },
			{  1.0f,  1.0f, 1.0f }
		};

		Ray r = {};
		r.m_origin = glm::vec3(0.0f);
		for (size_t i = 0; i < 4; ++i)
		{
			r.m_ray = dirs[i];
			r.compute();
			float t = 0.0f;
			if (rayAABBIntersection(r, projected, t) && within(0.0f, 1.0f, t))
			{
				return true;
			}
		}

		return false;
	}


	bool sphereOBBIntersection(const glm::vec3& centre, float radius, const OBB& obb)
	{
		glm::vec3 d = centre - obb.m_centre;
		glm::vec3 closest = obb.m_centre;
		
		for (size_t i = 0; i < 3; ++i)
		{
			float dist = glm::dot(d, obb.m_axes[i]);
			dist = glm::clamp(dist, -obb.m_extents[i], obb.m_extents[i]);
			closest += dist * obb.m_axes[i];
		}

		float sqDist = glm::dot(closest - centre, closest - centre);
		return sqDist <= (radius * radius);
	}


	glm::mat4 createVectorAlignMatrix(const glm::vec3& a, const glm::vec3& b)
	{
		glm::mat4 r = glm::mat4(1.0f);
		if (a != b)
		{
			if (a == -b)
			{
				// special case for if a is inverse of b
				r = glm::rotate(glm::pi<float>(), glm::vec3(0.0f, 1.0f, 0.0f));
			}
			else
			{
				glm::vec3 v = -glm::cross(b, a);
				float c = acos(glm::dot(b, a));
				r = glm::rotate(c, v);
			}
		}

		return r;
	}

	glm::mat4 createLookAtMatrix(const glm::vec3& eye, const glm::vec3& centre, const glm::vec3& up)
	{
		glm::vec3 a = glm::normalize(centre - eye);

		glm::vec3 b = up;
		if (a == b || a == -b)
		{
			glm::mat4 r = createVectorAlignMatrix(glm::vec3(0.0f, 0.0f, 1.0f), a);
			b = glm::mat3x3(r) * up;
		}
		
		return glm::lookAt(eye, eye + a, b);
	}

	glm::mat4 createRotationMatrix(float yaw, float pitch, float roll)
	{
		glm::mat4 r = glm::mat4(1.0f);
		r = glm::rotate(r, glm::radians(yaw), glm::vec3(0.0f, 1.0f, 0.0f));
		r = glm::rotate(r, glm::radians(pitch), glm::vec3(1.0f, 0.0f, 0.0f));
		r = glm::rotate(r, glm::radians(roll), glm::vec3(0.0f, 0.0f, 1.0f));
		return r;
	}

	Frustum createFrustum(float near, float far, float fov, float aspect)
	{
		float tanFov = tan(0.5f * fov);
		Frustum out = {};
		out.m_nearRight = near * tanFov * aspect;
		out.m_nearTop = near * tanFov;
		out.m_nearPlane = -near;
		out.m_farPlane = -far;
		return out;
	}

	float random(glm::vec2 st)
	{
		return glm::fract(sin(glm::dot(st, glm::vec2(12.9898f, 78.233f))) * 43758.5453123f);
	}

	float noise(glm::vec2 st)
	{
		glm::vec2 i = glm::floor(st);
		glm::vec2 f = glm::fract(st);

		// Four corners in 2D of a tile
		float a = random(i);
		float b = random(i + glm::vec2(1.0f, 0.0f));
		float c = random(i + glm::vec2(0.0f, 1.0f));
		float d = random(i + glm::vec2(1.0f, 1.0f));

		glm::vec2 u = f * f * (3.0f - 2.0f * f);

		return glm::mix(a, b, u.x) + (c - a) * u.y * (1.0f - u.x) + (d - b) * u.x * u.y;
	}

	float generateFbmNoise(glm::vec2 st, float seed)
	{
		float value = 0.0f;
		float amplitude = 0.5f;
		
		st = st + seed;

		// Loop of octaves
		for (int i = 0; i < 6; i++) 
		{
			value += amplitude * noise(st);
			st *= 2.0f;
			amplitude *= 0.5f;
		}
		return value;
	}
}