#include "gfx.h"
#include "platform.h"

#include <algorithm>
#include <thread>
#include <GL/glew.h>

// --- windows specific begin ---
#define NOMINMAX
#include "gfx_win32.h"
#include <windows.h>
#include <GL/wglew.h>

namespace utils
{
    extern HWND createWindow();
}

// --- windows specific end ---

#define CHECK_ERROR(func) func; { int err = glGetError(); if(err != GL_NO_ERROR && !s_debugToolAttached) printf("GL Error: %i\n", err); assert(err == GL_NO_ERROR || s_debugToolAttached); }

#define NO_SWIZZLE		0
#define SWIZZLE_RGB_BGR 1
#define SWIZZLE_R_A		2

// defines for debug tool extension
#define GL_DEBUG_TOOL_EXT                 0x6789
#define GL_DEBUG_TOOL_NAME_EXT            0x678A
#define GL_DEBUG_TOOL_PURPOSE_EXT         0x678B

namespace
{
    bool s_debugToolAttached = false;


    void APIENTRY openGLDebugCallback(GLenum source, GLenum type, GLuint id, GLenum severity, GLsizei length, const GLchar* message, const void* userParam)
    {
        if (type == GL_DEBUG_TYPE_ERROR || type == GL_DEBUG_TYPE_UNDEFINED_BEHAVIOR || type == GL_DEBUG_TYPE_DEPRECATED_BEHAVIOR)
            printf("GL Error: %s \n", message);
    }
}

namespace gfx
{
    namespace e_internalBindFlags
    {
        enum internalBindFlags
        {
            vertexBuffer = 1 << 0,
            indexBuffer = 1 << 1,
            storage = 1 << 2,
            uniform = 1 << 3,
            dispatchIndirect = 1 << 4
        };
    }
    typedef u32 InternalBindFlags;

    struct GLShader
    {
        GLuint                          m_shader;
        std::vector<DescriptorDesc>     m_uniforms;
        std::vector<PushConstantDesc>   m_pushConstants;
    };

    struct GLPixelFormat
    {
        GLint   m_sizedFormat;
        GLenum  m_format;
        GLenum  m_type;
        u32		m_swizzle;
    };

    struct GLTexture
    {
        GLuint          m_texture;
        GLPixelFormat   m_format;
        GLint           m_width;
        GLint           m_height;
        GLenum			m_target;
        GLint			m_maxLevel;
        TextureDesc*	m_desc;
    };

    struct GLFramebuffer
    {
        std::vector<GLTexture*> m_textures;
        GLuint                  m_fbo;
    };

    struct GLRenderPass
    {
        RenderPassDesc      m_desc;
        size_t              m_currentSubpass;
    };

    struct GLVertexFormat
    {
        GLint   m_size;
        GLenum  m_type;
    };

    struct GLVertexAttrib
    {
        GLuint          m_location;
        GLVertexFormat  m_format;
        GLuint          m_bufferIndex;
        size_t          m_offset;
    };

    struct GLVertexBufferLayout
    {
        GLuint m_stride;
        GLuint m_divisor;
    };

    struct GLVertexLayout
    {
        std::vector<GLVertexAttrib>         m_vertexAttribs;
        std::vector<GLVertexBufferLayout>   m_vertexBuffers;
    };

    struct GLBlendState
    {
        struct
        {
            bool r;
            bool g;
            bool b;
            bool a;
        }           m_writeMask;
        bool        m_enabled;
        GLenum      m_srcColorBlend;
        GLenum      m_dstColorBlend;
        GLenum      m_colorBlendOp;
        GLenum      m_srcAlphaBlend;
        GLenum      m_dstAlphaBlend;
        GLenum      m_alphaBlendOp;
    };

    struct GLMrtBlendState
    {
        GLBlendState*   m_blendStates;
        u32             m_numBlendStates;
    };
    
    struct GLRasterState
    {
        GLenum  m_cullMode = GL_NONE;
        GLenum  m_frontFace = GL_CCW;
        bool    m_depthClipEnable = false;
        bool    m_scissorEnable = false;
    };

    struct GLDepthStencilState
    {
        bool    m_depthEnable = false;
        bool    m_depthWrite = false;
        GLint   m_depthFunc;
        bool    m_stencilEnable = false;
        u8      m_stencilReadMask;
        u8      m_stencilWriteMask;
        GLenum  m_frontStencilFailOp;
        GLenum  m_frontStencilDepthFailOp;
        GLenum  m_frontStencilPassOp;
        GLint   m_frontStencilFunc;
        GLenum  m_backStencilFailOp;
        GLenum  m_backStencilDepthFailOp;
        GLenum  m_backStencilPassOp;
        GLint   m_backStencilFunc;
    };
    
    struct GLUniformDescriptor
    {
        GLint           m_location;
        GLint           m_binding;
        DescriptorType  m_type;
    };
        
    struct GLSampler
    {
        GLuint  m_sampler;
        float   m_border[4];
        float   m_mipLodBias;
        float   m_maxAnisotropy;
    };

    struct GLBuffer
    {
        std::vector<GLuint> m_buffers;
        Mem                 m_data;
        bool                m_dynamic;
        GLbitfield          m_accessFlags;
        int                 m_write = 0;
        int                 m_read = -1;
        bool                m_dirty = true;
        GLenum              m_usage;

        void destroy()
        {
            delete[] m_data.m_data;
            if (m_buffers.empty())
                return;
            CHECK_ERROR(glDeleteBuffers((GLsizei)m_buffers.size(), m_buffers.data()));
            m_buffers.clear();
        }

        void reset()
        {
            m_write = 0;
            m_read = -1;
            m_dirty = true; // make sure at least one buffer is written before trying to bind
        }

        GLuint bind(InternalBindFlags bindFlags)
        {
            // cpu data has changed, new buffer needed
            if (m_dirty)
            {
                if (m_write >= m_buffers.size())
                {
                    // create a new buffer
                    GLuint buffer;
                    CHECK_ERROR(glGenBuffers(1, &buffer));
                    if (bindFlags & e_internalBindFlags::uniform)
                    {
                        CHECK_ERROR(glBindBuffer(GL_UNIFORM_BUFFER, buffer));
                        CHECK_ERROR(glBufferData(GL_UNIFORM_BUFFER, m_data.m_size, m_data.m_data, m_usage));
                    }
                    else if (bindFlags & e_internalBindFlags::vertexBuffer)
                    {
                        CHECK_ERROR(glBindBuffer(GL_ARRAY_BUFFER, buffer));
                        CHECK_ERROR(glBufferData(GL_ARRAY_BUFFER, m_data.m_size, m_data.m_data, m_usage));
                    }
                    else if (bindFlags & e_internalBindFlags::indexBuffer)
                    {
                        CHECK_ERROR(glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer));
                        CHECK_ERROR(glBufferData(GL_ELEMENT_ARRAY_BUFFER, m_data.m_size, m_data.m_data, m_usage));
                    }
                    else if (bindFlags & e_internalBindFlags::dispatchIndirect)
                    {
                        CHECK_ERROR(glBindBuffer(GL_DISPATCH_INDIRECT_BUFFER, buffer));
                        CHECK_ERROR(glBufferData(GL_DISPATCH_INDIRECT_BUFFER, m_data.m_size, m_data.m_data, m_usage));
                    }
                    m_buffers.push_back(buffer);
                }
                else
                {
                    // reuse an existing buffer
                    CHECK_ERROR(glBindBuffer(GL_UNIFORM_BUFFER, m_buffers[m_write]));
                    void* mappedMem = CHECK_ERROR(glMapBuffer(GL_UNIFORM_BUFFER, GL_WRITE_ONLY));
                    if (mappedMem)
                    {
                        memcpy(mappedMem, m_data.m_data, m_data.m_size);
                    }
                    CHECK_ERROR(glUnmapBuffer(GL_UNIFORM_BUFFER));
                }
                ++m_write;
                ++m_read;
                m_dirty = false;
            }
            return m_buffers[m_read];
        }

        void dirty()
        {
            m_dirty = true;
        }
    };

    struct GLPipeline
    {
        GLuint                              m_program;
        std::vector<GLUniformDescriptor>    m_uniformDescriptors;
        GLBuffer*                           m_constantBuffer = nullptr;
        u32                                 m_constantBufferBinding;
    };

    struct GLRenderPipeline : GLPipeline
    {
        GLint                               m_vFlipLoc;
        GLenum                              m_primitiveType;
        GLVertexLayout*                     m_vertexLayout;
        GLMrtBlendState                     m_blendState;
        GLRasterState                       m_rasterState;
        GLDepthStencilState                 m_depthStencilState;
    };

    struct GLComputePipeline : public GLPipeline
    {
    };

    inline InternalBindFlags translateBindFlags(BindFlags flags)
    {
        InternalBindFlags intern = 0;
        if (flags & e_bindFlags::vertexBuffer)
        {
            intern |= e_internalBindFlags::vertexBuffer;
        }
        if (flags & e_bindFlags::indexBuffer)
        {
            intern |= e_internalBindFlags::indexBuffer;
        }
        if (flags & e_bindFlags::uniform)
        {
            intern |= e_internalBindFlags::uniform;
        }
        if (flags & e_bindFlags::storage)
        {
            intern |= e_internalBindFlags::storage;
        }

        return intern;
    }

    inline GLPixelFormat translateFormat(const FormatDesc& desc)
    {
        GLPixelFormat format;

        if (desc.m_type == e_formatType::unsignedNorm)
            format.m_type = GL_UNSIGNED_BYTE;
        else if (desc.m_type == e_formatType::signedNorm)
            format.m_type = GL_BYTE;
        else if (desc.m_type == e_formatType::unsignedScaled)
            format.m_type = GL_UNSIGNED_BYTE;
        else if (desc.m_type == e_formatType::signedScaled)
            format.m_type = GL_BYTE;
        else if (desc.m_type == e_formatType::unsignedInt)
            format.m_type = GL_UNSIGNED_INT;
        else if (desc.m_type == e_formatType::signedInt)
            format.m_type = GL_INT;
        else if (desc.m_type == e_formatType::signedFloat)
            format.m_type = GL_FLOAT;

        if ((desc.m_channels & e_formatChannelFlags::rgba) == e_formatChannelFlags::rgba)
        {
            format.m_format = GL_RGBA;
            if (desc.m_reverse)
                format.m_swizzle = SWIZZLE_RGB_BGR;
            else
                format.m_swizzle = NO_SWIZZLE;
            if (desc.m_bitDepth == 8)
            {
                format.m_sizedFormat = GL_RGBA8;
            }
            else if (desc.m_bitDepth == 16)
            {
                format.m_sizedFormat = GL_RGBA16;
            }
            else if (desc.m_bitDepth == 32)
            {
                format.m_sizedFormat = GL_RGBA32F;
            }
            else if (desc.m_bitDepth == 64)
            {
                format.m_sizedFormat = GL_RGBA32F;
            }
        }
        else if ((desc.m_channels & e_formatChannelFlags::rgb) == e_formatChannelFlags::rgb)
        {
            format.m_format = GL_RGB;
            if (desc.m_reverse)
                format.m_swizzle = SWIZZLE_RGB_BGR;
            else
                format.m_swizzle = NO_SWIZZLE;

            if (desc.m_bitDepth == 8)
            {
                format.m_sizedFormat = GL_RGB8;
            }
            else if (desc.m_bitDepth == 16)
            {
                format.m_sizedFormat = GL_RGB16;
            }
            else if (desc.m_bitDepth == 32)
            {
                format.m_sizedFormat = GL_RGB32F;
            }
            else if (desc.m_bitDepth == 64)
            {
                format.m_sizedFormat = GL_RGB32F;
            }
        }
        else if ((desc.m_channels & e_formatChannelFlags::rg) == e_formatChannelFlags::rg)
        {
            format.m_format = GL_RG;
            if (desc.m_reverse)
                format.m_swizzle = SWIZZLE_RGB_BGR;
            else
                format.m_swizzle = NO_SWIZZLE;

            if (desc.m_bitDepth == 8)
            {
                format.m_sizedFormat = GL_RG8;
            }
            else if (desc.m_bitDepth == 16)
            {
                format.m_sizedFormat = GL_RG16;
            }
            else if (desc.m_bitDepth == 32)
            {
                format.m_sizedFormat = GL_RG32F;
            }
            else if (desc.m_bitDepth == 64)
            {
                format.m_sizedFormat = GL_RG32F;
            }
        }
        else if ((desc.m_channels & e_formatChannelFlags::r) == e_formatChannelFlags::r)
        {
            format.m_format = GL_R;
            if (desc.m_reverse)
                format.m_swizzle = SWIZZLE_RGB_BGR;
            else
                format.m_swizzle = NO_SWIZZLE;

            if (desc.m_bitDepth == 8)
            {
                format.m_sizedFormat = GL_R8;
            }
            else if (desc.m_bitDepth == 16)
            {
                format.m_sizedFormat = GL_R16;
            }
            else if (desc.m_bitDepth == 32)
            {
                format.m_sizedFormat = GL_R32F;
            }
            else if (desc.m_bitDepth == 64)
            {
                format.m_sizedFormat = GL_R32F;
            }
        }
        else if ((desc.m_channels & e_formatChannelFlags::d) == e_formatChannelFlags::d)
        {
            format.m_format = GL_DEPTH_COMPONENT;
            if (desc.m_bitDepth == 16)
                format.m_sizedFormat = GL_DEPTH_COMPONENT16;
            else if (desc.m_bitDepth == 32)
                format.m_sizedFormat = GL_DEPTH_COMPONENT32;
        }
        else if ((desc.m_channels & e_formatChannelFlags::ds) == e_formatChannelFlags::ds)
        {
            format.m_format = GL_DEPTH_STENCIL;
            if (desc.m_bitDepth == 24)
                format.m_sizedFormat = GL_DEPTH24_STENCIL8;
            else if (desc.m_bitDepth == 32)
                format.m_sizedFormat = GL_DEPTH32F_STENCIL8;
        }
        return format;
    }

    inline GLenum translateBlendFactor(BlendFactor factor)
    {
        switch (factor)
        {
        case BlendFactor::zero:
            return GL_ZERO;
        case BlendFactor::one:
            return GL_ONE;
        case BlendFactor::srcColor:
            return GL_SRC_COLOR;
        case BlendFactor::oneMinusSrcColor:
            return GL_ONE_MINUS_SRC_COLOR;
        case BlendFactor::srcAlpha:
            return GL_SRC_ALPHA;
        case BlendFactor::oneMinusSrcAlpha:
            return GL_ONE_MINUS_SRC_ALPHA;
        case BlendFactor::dstAlpha:
            return GL_DST_ALPHA;
        case BlendFactor::oneMinusDstAlpha:
            return GL_ONE_MINUS_DST_ALPHA;
        case BlendFactor::dstColor:
            return GL_DST_COLOR;
        case BlendFactor::oneMinusDstColor:
            return GL_ONE_MINUS_DST_COLOR;
        case BlendFactor::src1Color:
            return GL_SRC_COLOR;
        case BlendFactor::oneMinusSrc1Color:
            return GL_ONE_MINUS_SRC_COLOR;
        case BlendFactor::src1Alpha:
            return GL_SRC_ALPHA;
        case BlendFactor::oneMinusSrc1Alpha:
            return GL_ONE_MINUS_SRC_ALPHA;
        }
        return GL_ONE;
    }

    inline GLenum translateBlendOp(BlendOp op)
    {
        switch (op)
        {
        case BlendOp::add:
            return GL_FUNC_ADD;
        case BlendOp::subtract:
            return GL_FUNC_SUBTRACT;
        case BlendOp::reverseSubtract:
            return GL_FUNC_REVERSE_SUBTRACT;
        case BlendOp::min:
            return GL_MIN;
        case BlendOp::max:
            return GL_MAX;
        }
        return GL_FUNC_ADD;
    }

    inline GLenum translatePrimitiveTopology(PrimitiveTopology topology)
    {
        switch (topology)
        {
        case e_primitiveTopology::pointList:
            return GL_POINTS;
            break;
        case e_primitiveTopology::lineList:
            return GL_LINES;
            break;
        case e_primitiveTopology::lineStrip:
            return GL_LINE_STRIP;
            break;
        case e_primitiveTopology::triangleList:
            return GL_TRIANGLES;
            break;
        case e_primitiveTopology::triangleStrip:
            return GL_TRIANGLE_STRIP;
            break;
        case e_primitiveTopology::triangleFan:
            return GL_TRIANGLE_FAN;
            break;
        case e_primitiveTopology::lineListAdjacency:
            return GL_LINES_ADJACENCY;
            break;
        case e_primitiveTopology::lineStripAdjacency:
            return GL_LINE_STRIP_ADJACENCY;
            break;
        case e_primitiveTopology::triangleListAdjacency:
            return GL_TRIANGLES_ADJACENCY;
            break;
        case e_primitiveTopology::triangleStripAdjacency:
            return GL_TRIANGLE_STRIP_ADJACENCY;
            break;
        case e_primitiveTopology::patchList:
            return GL_PATCHES;
            break;
        default:
            return GL_TRIANGLES;
            break;
        }
    }

    inline GLint translateCompareOp(CompareOp op)
    {
        switch (op)
        {
        case e_compareOpType::never:
            return GL_NEVER;
            break;
        case e_compareOpType::less:
            return GL_LESS;
            break;
        case e_compareOpType::equal:
            return GL_EQUAL;
            break;
        case e_compareOpType::lessOrEqual:
            return GL_LEQUAL;
            break;
        case e_compareOpType::greater:
            return GL_GREATER;
            break;
        case e_compareOpType::notEqual:
            return GL_NOTEQUAL;
            break;
        case e_compareOpType::greatorOrEqual:
            return GL_GEQUAL;
            break;
        case e_compareOpType::always:
            return GL_ALWAYS;
            break;
        default:
            return GL_NEVER;
            break;
        }
    }

    inline GLenum translateStencilOp(StencilOp op)
    {
        switch (op)
        {
        case e_stencilOpType::keep:
            return GL_KEEP;
            break;
        case e_stencilOpType::zero:
            return GL_ZERO;
            break;
        case e_stencilOpType::replace:
            return GL_REPLACE;
            break;
        case e_stencilOpType::incrementClamp:
            return GL_INCR;
            break;
        case e_stencilOpType::decrementClamp:
            return GL_DECR;
            break;
        case e_stencilOpType::invert:
            return GL_INVERT;
            break;
        case e_stencilOpType::incrementWrap:
            return GL_INCR_WRAP;
            break;
        case e_stencilOpType::decrementWrap:
            return GL_DECR_WRAP;
            break;
        default:
            return GL_KEEP;
            break;
        }
    }

    inline GLenum translatePolygonMode(PolygonMode mode)
    {
        switch (mode)
        {
        case e_polygonModeType::fill:
            return GL_FILL;
            break;
        case e_polygonModeType::line:
            return GL_LINE;
            break;
        case e_polygonModeType::point:
            return GL_POINT;
            break;
        default:
            return GL_FILL;
            break;
        }
    }

    inline GLenum translateCullMode(CullMode mode)
    {
        if (mode & e_cullModeFlags::front)
            return GL_FRONT;
        else if (mode & e_cullModeFlags::back)
            return GL_BACK;
        return GL_NONE;
    }

    inline GLenum translateFrontFace(FrontFace face)
    {
        switch (face)
        {
        case e_frontFaceType::counterClockwise:
            return GL_CCW;
            break;
        case e_frontFaceType::clockwise:
            return GL_CW;
            break;
        default:
            return GL_CCW;
            break;
        }
    }

    inline GLenum translateLogicOp(LogicOp op)
    {
        switch (op)
        {
        case e_logicOpType::clear:
            return GL_CLEAR;
            break;
        case e_logicOpType::and:
            return GL_AND;
            break;
        case e_logicOpType::andReverse:
            return GL_AND_REVERSE;
            break;
        case e_logicOpType::copy:
            return GL_COPY;
            break;
        case e_logicOpType::andInverted:
            return GL_AND_INVERTED;
            break;
        case e_logicOpType::noOp:
            return GL_NOOP;
            break;
        case e_logicOpType::xor:
            return GL_XOR;
            break;
        case e_logicOpType:: or :
            return GL_OR;
            break;
        case e_logicOpType::nor:
            return GL_NOR;
            break;
        case e_logicOpType::equivalent:
            return GL_EQUIV;
            break;
        case e_logicOpType::invert:
            return GL_INVERT;
            break;
        case e_logicOpType::orReverse:
            return GL_OR_REVERSE;
            break;
        case e_logicOpType::copyInverted:
            return GL_COPY_INVERTED;
            break;
        case e_logicOpType::orInverted:
            return GL_OR_INVERTED;
            break;
        case e_logicOpType::nand:
            return GL_NAND;
            break;
        case e_logicOpType::set:
            return GL_SET;
            break;
        default:
            return GL_NOOP;
            break;
        }
    }

    inline GLenum translateShaderStage(ShaderStageFlags stages)
    {
        if (stages & e_shaderStageFlags::vertex)
            return GL_VERTEX_SHADER;
        else if (stages & e_shaderStageFlags::fragment)
            return GL_FRAGMENT_SHADER;
        else if (stages & e_shaderStageFlags::compute)
            return GL_COMPUTE_SHADER;
        else if (stages & e_shaderStageFlags::tessellationControl)
            return GL_TESS_CONTROL_SHADER;
        else if (stages & e_shaderStageFlags::tessellationEvaluation)
            return GL_TESS_EVALUATION_SHADER;
        else if (stages & e_shaderStageFlags::geometry)
            return GL_GEOMETRY_SHADER;
        else
            return 0;
    }

    inline GLint translateFilter(Filter filter)
    {
        switch (filter)
        {
        case Filter::nearest:
            return GL_NEAREST;
        case Filter::linear:
            return GL_LINEAR;
        }
        return GL_NEAREST;
    }

    inline GLint translateAddressMode(AddressMode mode)
    {
        switch (mode)
        {
        case e_addressMode::repeat:
            return GL_REPEAT;
            break;
        case e_addressMode::mirroredRepeat:
            return GL_MIRRORED_REPEAT;
            break;
        case e_addressMode::clampToEdge:
            return GL_CLAMP_TO_EDGE;
            break;
        case e_addressMode::clampToBorder:
            return GL_CLAMP_TO_BORDER;
            break;
        case e_addressMode::mirrorClampToEdge:
            return GL_MIRROR_CLAMP_TO_EDGE;
            break;
        default:
            return GL_REPEAT;
            break;
        }
    }

    inline GLbitfield translateAccessFlags(AccessFlags flags)
    {
        GLbitfield accessFlags = 0;
        if (flags & e_accessFlags::write)
        {
            accessFlags |= GL_MAP_WRITE_BIT;
        }
        if (flags & e_accessFlags::read)
        {
            accessFlags |= GL_MAP_READ_BIT;
        }
        return accessFlags;
    }
    
    inline GLenum translateIndexType(IndexType index)
    {
        switch (index)
        {
        case IndexType::uint16:
            return GL_UNSIGNED_SHORT;
        case IndexType::uint32:
            return GL_UNSIGNED_INT;
        default:
            return GL_UNSIGNED_SHORT;
        }
    }

    u32 s_numBufferedResources = 1;
    u32 s_bb = 0;
    u32 s_fb = 0;

    void incrementWrapBufferedResources()
    {
        s_bb = (s_bb + 1) % s_numBufferedResources;
        s_fb = (s_fb + 1) % s_numBufferedResources;
    }

    class GLContext : public Context
    {
    protected:

    public:
        void setup(void* platform);
        void setupAsync(GLContext* main, const std::thread::id& id);
        void shutdown();
        void shutdownAsync();

        Shader createShader(const ShaderDesc& desc) override;
        Buffer createBuffer(const BufferDesc& desc) override;
        Texture createTexture(const TextureDesc& desc) override;
        Sampler createSampler(const SamplerDesc& desc) override;
        RenderPass createRenderPass(const RenderPassDesc& desc) override;
        RenderPipeline createRenderPipeline(const RenderPipelineDesc& desc) override;
        ComputePipeline createComputePipeline(const ComputePipelineDesc& desc) override;
        Framebuffer obtainFramebuffer(const FramebufferDesc& desc) override;

        void destroyShader(Shader shader) override;
        void destroyBuffer(Buffer buffer) override;
        void destroyTexture(Texture texture) override;
        void destroySampler(Sampler sampler) override;
        void destroyRenderPass(RenderPass pass) override;
        void destroyRenderPipeline(RenderPipeline pipeline) override;
        void destroyComputePipeline(ComputePipeline pipeline) override;

        void begin() override;
        void setViewport(const Viewport& vp) override;
        void setScissorRect(const Rect& sr) override;
        void beginComputeEncoding() override;
        void beginRenderPass(RenderPass pass, Framebuffer framebuffer, Clear* clears = nullptr, size_t numClears = 0) override;
        void beginRenderPassInternal();
        void beginNextSubpass() override;
        void beginSubpass(const SubpassDesc& subpass, Clear* clears, size_t numClears);
        void bindRenderPipeline(RenderPipeline pipeline) override;
        void bindComputePipeline(ComputePipeline pipeline) override;
        void bindBuffer(Buffer buf, u32 slot, BindFlags flags, size_t offset = 0) override;
        void bindBufferInternal(Buffer buf, u32 slot, InternalBindFlags flags, size_t offset);
        void bindTexture(Texture tex, Sampler samp, u32 slot, BindFlags flags) override;
        void updateBuffer(Buffer buf, const Mem& data, size_t offset = 0) override;
        void updateTexture(Texture tex, const TextureUpdateDesc& data) override;
        void updateConstants(ShaderStageFlags stage, size_t size, size_t offset, void* data) override;
        void insertBarrier(PipelineStage srcStage, PipelineStage dstStage, BarrierFlags barrierType) override;
        void copyTexture(Texture src, Texture dst, u32 srcLayer = 0, u32 dstLayer = 0) override;

        void draw(u32 vertexCount, u32 startVertex = 0) override;
        void drawIndexed(u32 indexCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset = 0, u32 vertexOffset = 0) override;
        void drawIndexedInstanced(u32 indexCount, u32 instanceCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset = 0, u32 vertexOffset = 0) override;
        void drawIndirect(u32 drawCount, Buffer drawBuffer, u32 drawBufferOffset = 0) override;
        void drawIndexedIndirect(u32 drawCount, Buffer indexBuffer, IndexType indexType, Buffer drawBuffer, u32 indexBufferOffset = 0, u32 drawBufferOffset = 0) override;
        void dispatchCompute(u32 grid[3], u32 numThreads[3]) override;
        void dispatchComputeIndirect(Buffer drawBuffer, u32 offset = 0) override;

        void        endRenderEncoding() override;
        void        endComputeEncoding() override;
        void        enqueue() override;
        Semaphore   commit(Semaphore* waits = nullptr, size_t count = 0, PipelineStage waitStage = e_pipelineStageType::top) override;
        void        present() override;
        void        flushCommandBuffers() override;

        bool        isSignalled(Semaphore semaphore) override;

        HDC             m_hdc;
        GLFramebuffer*  m_backBufferFramebuffer = nullptr;
        GLVertexLayout* m_currentVertexLayout = nullptr;
        Viewport        m_currentViewport;
        GLenum          m_currentPrimitiveType = GL_TRIANGLES;
        GLuint          m_vao = 0;
        bool            m_doFramebufferFlip = false;
        GLPipeline*     m_currentPipeline = nullptr;
        GLRenderPass*   m_currentRenderPass = nullptr;

        std::thread::id m_threadID = std::thread::id();

        std::vector<GLBuffer*>      m_buffers;
        std::vector<GLFramebuffer*> m_framebuffers;
    };

    void GLContext::setup(void* platform)
    {
        PlatformGFX* plat = (PlatformGFX*)platform;

        // --- windows specific begins ---

        // create temp device
        PIXELFORMATDESCRIPTOR pfd;
        memset(&pfd, 0, sizeof(pfd)); 
        pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
        pfd.nVersion = 1;
        pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
        pfd.iLayerType = PFD_MAIN_PLANE;
        pfd.iPixelType = PFD_TYPE_RGBA;
        pfd.cColorBits = 24;
        pfd.cAlphaBits = 8;

        HDC hdc = GetDC(plat->m_hwnd);
        int pixelFormat = ChoosePixelFormat(hdc, &pfd);
        DescribePixelFormat(hdc, pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
        int result = SetPixelFormat(hdc, pixelFormat, &pfd);

        HGLRC temp = wglCreateContext(hdc);
        wglMakeCurrent(hdc, temp);
        m_hdc = hdc;

        // attempt to create opengl device
        int error = glewInit();
        if (wglewIsSupported("WGL_ARB_create_context") == 1)
        {
            int ctxAttribs[] =
            {
                WGL_CONTEXT_MAJOR_VERSION_ARB, 4,
                WGL_CONTEXT_MINOR_VERSION_ARB, 3,
                WGL_CONTEXT_FLAGS_ARB, 0,
                0
            };

            pixelFormat = 0;
            u32 numFormats = 0;
            int sampleCount = 1;
            int pfAttribs[] = {
                WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
                WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
                WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
                WGL_PIXEL_TYPE_ARB, WGL_TYPE_RGBA_ARB,
                WGL_COLOR_BITS_ARB, 24,
                WGL_ALPHA_BITS_ARB, 8,
                WGL_DEPTH_BITS_ARB, pfd.cDepthBits,
                WGL_STENCIL_BITS_ARB, pfd.cStencilBits,
                WGL_SAMPLE_BUFFERS_ARB, 1,
                WGL_SAMPLES_ARB, sampleCount,
                0
            };

            // close and re-create window
            HWND prev = plat->m_hwnd;
            HWND wnd = utils::createWindow();
            hdc = GetDC(wnd);

            wglChoosePixelFormatARB(hdc, pfAttribs, nullptr, 1, &pixelFormat, &numFormats);
            memset(&pfd, 0, sizeof(pfd));
            DescribePixelFormat(hdc, pixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd);
            result = SetPixelFormat(hdc, pixelFormat, &pfd);
            DWORD err = GetLastError();

            HGLRC hglrc = wglCreateContextAttribsARB(hdc, 0, ctxAttribs);
            wglMakeCurrent(NULL, NULL);
            wglDeleteContext(temp);

            result = wglMakeCurrent(hdc, hglrc);

            m_hdc = hdc;
        }
        else
        {
            HGLRC hglrc = temp;
        }
        
        // --- windows specific ends ---
        
        s_debugToolAttached = glIsEnabled(GL_DEBUG_TOOL_EXT);
        glGetError();
        if (s_debugToolAttached)
        {
            printf("Debug tool \"%s\" is attached\n", glGetString(GL_DEBUG_TOOL_NAME_EXT));
        }

        glGenVertexArrays(1, &m_vao);
        glBindVertexArray(m_vao);

        glEnable(GL_DEBUG_OUTPUT);
        glDebugMessageCallback(openGLDebugCallback, nullptr);

        // set up default framebuffer info
        m_backBufferFramebuffer = new GLFramebuffer();
        GLint bb = 0;
        glGetIntegerv(GL_DRAW_FRAMEBUFFER_BINDING, &bb);
        m_backBufferFramebuffer->m_fbo = bb;
    }

    void GLContext::setupAsync(GLContext* mainContext, const std::thread::id& id)
    {
        
    }

    void GLContext::shutdown()
    {

    }

    void GLContext::shutdownAsync()
    {

    }
    
    Texture GLContext::createTexture(const TextureDesc& desc)
    {
        GLTexture* tex = new GLTexture;
        tex->m_format = translateFormat(desc.m_format);
        tex->m_width = (GLint)desc.m_width;
        tex->m_height = (GLint)desc.m_height;
        tex->m_maxLevel = desc.m_mips - 1;
        CHECK_ERROR(glGenTextures(1, &tex->m_texture));

        tex->m_desc = new TextureDesc;
        memcpy(tex->m_desc, &desc, sizeof(TextureDesc));

        // determine image and view type from params
        GLenum target = GL_TEXTURE_1D;
        size_t depth = desc.m_depth;
        size_t layers = desc.m_layers;
        if (desc.m_height > 1)
        {
            target = (desc.m_layers > 1) ? GL_TEXTURE_2D_ARRAY : GL_TEXTURE_2D;
        }
        if (desc.m_depth > 1)
        {
            target = GL_TEXTURE_3D;
        }
        else if (desc.m_width == desc.m_height && desc.m_layers % 6 == 0)
        {
            target = GL_TEXTURE_CUBE_MAP;
        }

        if (target == GL_TEXTURE_2D_ARRAY)
        {
            depth = desc.m_layers;
            layers = 1;
        }
        tex->m_target = target;

        CHECK_ERROR(glBindTexture(target, tex->m_texture));
        char* data = (char*)desc.m_data.m_data;
        for (u32 l = 0; l < layers; ++l)
        {
            GLint w = (GLint)desc.m_width;
            GLint h = (GLint)desc.m_height;
            GLint d = (GLint)depth;

            for (u32 mip = 0; mip < desc.m_mips; ++mip)
            {
                // need to implement the function to get size
                GLint mipSize = 0;

                if (target == GL_TEXTURE_2D)
                {
                    CHECK_ERROR(glTexImage2D(target + l, mip, tex->m_format.m_sizedFormat, w, h, 0, tex->m_format.m_format, tex->m_format.m_type, data));
                }
                else
                {
                    CHECK_ERROR(glTexImage3D(target, mip, tex->m_format.m_sizedFormat, w, h, d, 0, tex->m_format.m_format, tex->m_format.m_type, data));
                }

                if (data)
                    data += mipSize;

                w /= 2;
                h /= 2;
                d /= 2;
                
                w = std::max<u32>(1, w);
                h = std::max<u32>(1, h);
                d = std::max<u32>(1, d);
            }
        }

        return (Texture)tex;
    }

    Shader GLContext::createShader(const ShaderDesc& desc)
    {
        GLShader* shader = new GLShader;
        shader->m_shader = CHECK_ERROR(glCreateShader(translateShaderStage(desc.m_stage)));
        const char* src = (const char*)desc.m_data.m_data;
        GLint size = (GLint)desc.m_data.m_size;
        CHECK_ERROR(glShaderSource(shader->m_shader, 1, &src, &size));
        CHECK_ERROR(glCompileShader(shader->m_shader));
        GLint compiled = GL_FALSE;
        CHECK_ERROR(glGetShaderiv(shader->m_shader, GL_COMPILE_STATUS, &compiled));
        if (!compiled)
        {
            GLint infoLogLen = 0;
            CHECK_ERROR(glGetShaderiv(shader->m_shader, GL_INFO_LOG_LENGTH, &infoLogLen));
            if (infoLogLen > 0)
            {
                GLchar infoLog[256];
                glGetShaderInfoLog(shader->m_shader, std::min(infoLogLen, 255), NULL, infoLog);
                infoLog[std::min(infoLogLen + 1, 254)] = '\0';
                printf("Could not compile shader:\n%s\n", infoLog);
            }
            glDeleteShader(shader->m_shader);
            shader->m_shader = 0;
        }
        shader->m_uniforms = desc.m_descriptors;
        return (Shader)shader;
    }

    Buffer GLContext::createBuffer(const BufferDesc& desc)
    {
        GLBuffer* buf = new GLBuffer;
        buf->m_accessFlags = translateAccessFlags(desc.m_access);
        buf->m_dynamic = (desc.m_access & e_accessFlags::write);
        GLenum usage = (buf->m_dynamic) ? GL_DYNAMIC_DRAW : GL_STATIC_DRAW;
        buf->m_usage = usage;
        buf->m_data.m_size = desc.m_data.m_size;
        buf->m_data.m_data = new char[buf->m_data.m_size];
        if (desc.m_data.m_data)
            memcpy(buf->m_data.m_data, desc.m_data.m_data, desc.m_data.m_size);
        m_buffers.push_back(buf);
        return (Buffer)buf;
    }

    Sampler GLContext::createSampler(const SamplerDesc& desc)
    {
        GLSampler* sampler = new GLSampler;
        CHECK_ERROR(glGenSamplers(1, &sampler->m_sampler));
        CHECK_ERROR(glSamplerParameteri(sampler->m_sampler, GL_TEXTURE_MIN_FILTER, translateFilter(desc.m_minFilter)));
        CHECK_ERROR(glSamplerParameteri(sampler->m_sampler, GL_TEXTURE_MAG_FILTER, translateFilter(desc.m_magFilter)));
        CHECK_ERROR(glSamplerParameteri(sampler->m_sampler, GL_TEXTURE_WRAP_S, translateAddressMode(desc.m_addressModeU)));
        CHECK_ERROR(glSamplerParameteri(sampler->m_sampler, GL_TEXTURE_WRAP_T, translateAddressMode(desc.m_addressModeV)));
        CHECK_ERROR(glSamplerParameteri(sampler->m_sampler, GL_TEXTURE_WRAP_R, translateAddressMode(desc.m_addressModeW)));
        CHECK_ERROR(glSamplerParameteri(sampler->m_sampler, GL_TEXTURE_COMPARE_FUNC, translateCompareOp(desc.m_compareOp)));
        CHECK_ERROR(glSamplerParameterf(sampler->m_sampler, GL_TEXTURE_MIN_LOD, desc.m_minLod));
        CHECK_ERROR(glSamplerParameterf(sampler->m_sampler, GL_TEXTURE_MAX_LOD, desc.m_maxLod));
        sampler->m_mipLodBias = desc.m_mipLodBias;
        sampler->m_maxAnisotropy = desc.m_maxAnisotropy;
        switch (desc.m_borderColor)
        {
        case e_borderColor::transparentBlack:
            sampler->m_border[0] = 0.0f;
            sampler->m_border[1] = 0.0f;
            sampler->m_border[2] = 0.0f;
            sampler->m_border[3] = 0.0f;
            break;
        case e_borderColor::opaqueBlack:
            sampler->m_border[0] = 0.0f;
            sampler->m_border[1] = 0.0f;
            sampler->m_border[2] = 0.0f;
            sampler->m_border[3] = 1.0f;            
            break;
        case e_borderColor::opaqueWhite:
            sampler->m_border[0] = 1.0f;
            sampler->m_border[1] = 1.0f;
            sampler->m_border[2] = 1.0f;
            sampler->m_border[3] = 1.0f;
            break;
        }
        return (Sampler)sampler;
    }

    RenderPass GLContext::createRenderPass(const RenderPassDesc& desc)
    {
        GLRenderPass* rp = new GLRenderPass;
        rp->m_desc = desc;
        rp->m_currentSubpass = 0;
        return rp;
    }

    Framebuffer GLContext::obtainFramebuffer(const FramebufferDesc& desc)
    {
        GLFramebuffer* framebuffer = nullptr;

        if (desc.m_textures.size() == 1 && !desc.m_textures[0])
        {
            // backbuffer framebuffer
            return m_backBufferFramebuffer;
        }

        for (auto& f : m_framebuffers)
        {
            // check if framebuffers match
            if (desc.m_textures.size() <= f->m_textures.size())
            {
                bool found = true;
                for (size_t i = 0; desc.m_textures.size(); ++i)
                {
                    if (desc.m_textures[i] != f->m_textures[i])
                    {
                        found = false;
                        break;
                    }
                }
                if (found)
                {
                    framebuffer = f;
                    break;
                }
            }
        }

        if (!framebuffer)
        {
            GLRenderPass* rp = (GLRenderPass*)desc.m_pass;

            // create a new framebuffer
            framebuffer = new GLFramebuffer();

            CHECK_ERROR(glGenFramebuffers(1, &framebuffer->m_fbo));
            for (size_t i = 0; i < desc.m_textures.size(); ++i)
            {
                framebuffer->m_textures.push_back((GLTexture*)desc.m_textures[i]);
            }
        }

        return framebuffer;
    }

    RenderPipeline GLContext::createRenderPipeline(const RenderPipelineDesc& desc)
    {
        GLRenderPipeline* pipeline = new GLRenderPipeline;
        std::vector<GLShader*> shaders;
        for (auto& s : desc.m_shaders)
        {
            shaders.push_back((GLShader*)s);
        }

        // we must take a copy of vertex layout
        pipeline->m_vertexLayout = new GLVertexLayout;
        for (auto& l : desc.m_vertexInput.m_attributes)
        {
            GLVertexAttrib attrib;
            attrib.m_location = l.m_location;
            attrib.m_bufferIndex = l.m_binding;
            attrib.m_offset = l.m_offset;
            attrib.m_format.m_type = translateFormat(l.m_format).m_type;
            attrib.m_format.m_size = log2(l.m_format.m_channels + 1);
            pipeline->m_vertexLayout->m_vertexAttribs.push_back(attrib);
        }

        for (auto& l : desc.m_vertexInput.m_bindings)
        {
            GLVertexBufferLayout layout;
            layout.m_stride = l.m_stride;
            layout.m_divisor = l.m_rate;
            pipeline->m_vertexLayout->m_vertexBuffers.push_back(layout);
        }

        pipeline->m_primitiveType = translatePrimitiveTopology(desc.m_inputAssembly.m_topology);
        pipeline->m_program = CHECK_ERROR(glCreateProgram());
        for (auto& s : shaders)
        {
            CHECK_ERROR(glAttachShader(pipeline->m_program, s->m_shader));
        }
        CHECK_ERROR(glLinkProgram(pipeline->m_program));
        GLint linked = GL_FALSE;
        CHECK_ERROR(glGetProgramiv(pipeline->m_program, GL_LINK_STATUS, &linked));
        if (!linked)
        {
            GLint infoLogLen = 0;
            CHECK_ERROR(glGetProgramiv(pipeline->m_program, GL_INFO_LOG_LENGTH, &infoLogLen));
            if (infoLogLen > 0)
            {
                GLchar infoLog[256];
                glGetProgramInfoLog(pipeline->m_program, std::min(infoLogLen, 255), NULL, infoLog);
                infoLog[std::min(infoLogLen + 1, 255)] = '\0';
                printf("Could not link shader program:\n%s\n", infoLog);
            }
            glDeleteProgram(pipeline->m_program);
            pipeline->m_program = 0;
        }

        // get uniform shader locations in the linked program
        std::vector<std::string> usedNames;
        size_t pushConstantsSize = 0;
        u32 pushConstantBinding = 0;
        for (int i = 0; i < shaders.size(); ++i)
        {
            for (const auto& uDesc : shaders[i]->m_uniforms)
            {
                std::string name = uDesc.m_name;
                bool found = false;
                for (auto& used : usedNames)
                {
                    if (used == name)
                    {
                        found = true;
                        break;
                    }
                }
                if (found)
                    continue;
                GLUniformDescriptor uniform;
                uniform.m_type = uDesc.m_type;
                if (uDesc.m_type == DescriptorType::sampledImage)
                {
                    uniform.m_location = CHECK_ERROR(glGetUniformLocation(pipeline->m_program, name.c_str()));
                    uniform.m_binding = uDesc.m_binding;
                }
                else if (uDesc.m_type == DescriptorType::uniformBuffer)
                {
                    uniform.m_binding = uDesc.m_binding;
                    uniform.m_location = CHECK_ERROR(glGetUniformBlockIndex(pipeline->m_program, name.c_str()));
                    if (uniform.m_location >= 0)
                        CHECK_ERROR(glUniformBlockBinding(pipeline->m_program, uniform.m_location, uniform.m_binding));
                }
                if (uniform.m_location >= 0)
                {
                    pipeline->m_uniformDescriptors.push_back(uniform);
                    usedNames.push_back(name);
                }
            }

            for (const auto& cDesc : shaders[i]->m_pushConstants)
            {
                pushConstantsSize += cDesc.m_size;
                pushConstantBinding = cDesc.m_binding;
            }
        }
        pipeline->m_vFlipLoc = CHECK_ERROR(glGetUniformLocation(pipeline->m_program, "vFlip"));
        
        if (pushConstantsSize > 0)
        {
            BufferDesc bufferDesc;
            bufferDesc.m_usage = e_bufferUsageFlags::uniform;
            bufferDesc.m_access = e_accessFlags::write;
            bufferDesc.m_data.m_data = nullptr;
            bufferDesc.m_data.m_size = pushConstantsSize;
            pipeline->m_constantBuffer = (GLBuffer*)createBuffer(bufferDesc);
            pipeline->m_constantBufferBinding = pushConstantBinding;
        }

        pipeline->m_blendState.m_blendStates = new GLBlendState[desc.m_blendDesc.m_blends.size()];
        for (u32 i = 0; i < desc.m_blendDesc.m_blends.size(); ++i)
        {
            const BlendDesc& blend = desc.m_blendDesc.m_blends[i];
            GLBlendState& glBlend = pipeline->m_blendState.m_blendStates[i];
            glBlend.m_enabled = blend.m_blendEnable;
            if (glBlend.m_enabled)
            {
                glBlend.m_srcColorBlend = translateBlendFactor(blend.m_srcColorBlendFactor);
                glBlend.m_dstColorBlend = translateBlendFactor(blend.m_dstColorBlendFactor);
                glBlend.m_colorBlendOp = translateBlendOp(blend.m_colorBlendOp);
                glBlend.m_srcAlphaBlend = translateBlendFactor(blend.m_srcAlphaBlendFactor);
                glBlend.m_dstAlphaBlend = translateBlendFactor(blend.m_dstAlphaBlendFactor);
                glBlend.m_alphaBlendOp = translateBlendOp(blend.m_alphaBlendOp);
            }
            glBlend.m_writeMask.r = 1;
            glBlend.m_writeMask.g = 1;
            glBlend.m_writeMask.b = 1;
            glBlend.m_writeMask.a = 1;
        }

        GLRasterState& rs = pipeline->m_rasterState;
        rs.m_cullMode = translateCullMode(desc.m_rasterizer.m_cullMode);
        rs.m_frontFace = translateFrontFace(desc.m_rasterizer.m_frontFace);
        rs.m_depthClipEnable = false;
        rs.m_scissorEnable = false;

        GLDepthStencilState& dss = pipeline->m_depthStencilState;
        const DepthStencilDesc& dsDesc = desc.m_depthStencil;
        dss.m_depthEnable = dsDesc.m_depthTest;
        dss.m_depthWrite = dsDesc.m_depthWrite;
        dss.m_depthFunc = translateCompareOp(dsDesc.m_depthCompareOp);
        dss.m_stencilEnable = dsDesc.m_stencilTest;
        dss.m_stencilReadMask = 0;
        dss.m_stencilWriteMask = 0;
        dss.m_frontStencilFailOp = translateStencilOp(dsDesc.m_front.m_failOp);
        dss.m_frontStencilDepthFailOp = translateStencilOp(dsDesc.m_front.m_depthFailOp);
        dss.m_frontStencilPassOp = translateStencilOp(dsDesc.m_front.m_passOp);
        dss.m_frontStencilFunc = translateCompareOp(dsDesc.m_front.m_compareOP);
        dss.m_backStencilFailOp = translateStencilOp(dsDesc.m_back.m_failOp);
        dss.m_backStencilDepthFailOp = translateStencilOp(dsDesc.m_back.m_depthFailOp);
        dss.m_backStencilPassOp = translateStencilOp(dsDesc.m_back.m_failOp);
        dss.m_backStencilFunc = translateCompareOp(dsDesc.m_back.m_compareOP);

        return (RenderPipeline)pipeline;
    }

    ComputePipeline GLContext::createComputePipeline(const ComputePipelineDesc& desc)
    {
        GLComputePipeline* pipeline = new GLComputePipeline;
        GLShader* shader = (GLShader*)desc.m_shader;

        pipeline->m_program = CHECK_ERROR(glCreateProgram());
        CHECK_ERROR(glAttachShader(pipeline->m_program, shader->m_shader));
        CHECK_ERROR(glLinkProgram(pipeline->m_program));
        GLint linked = GL_FALSE;
        CHECK_ERROR(glGetProgramiv(pipeline->m_program, GL_LINK_STATUS, &linked));
        if (!linked)
        {
            GLint infoLogLen = 0;
            CHECK_ERROR(glGetProgramiv(pipeline->m_program, GL_INFO_LOG_LENGTH, &infoLogLen));
            if (infoLogLen > 0)
            {
                GLchar infoLog[256];
                glGetProgramInfoLog(pipeline->m_program, std::min(infoLogLen, 255), NULL, infoLog);
                infoLog[std::min(infoLogLen + 1, 255)] = '\0';
                printf("Could not link shader program:\n%s\n", infoLog);
            }
            glDeleteProgram(pipeline->m_program);
        }

        // get uniform shader locations in the linked program
        std::vector<std::string> usedNames;
        for (const auto& uDesc : shader->m_uniforms)
        {
            std::string name = uDesc.m_name;
            bool found = false;
            for (auto& used : usedNames)
            {
                if (used == name)
                {
                    found = true;
                    break;
                }
            }
            if (found)
                continue;
            GLUniformDescriptor uniform;
            uniform.m_type = uDesc.m_type;
            if (uDesc.m_type == DescriptorType::sampledImage)
            {
                uniform.m_location = CHECK_ERROR(glGetUniformLocation(pipeline->m_program, name.c_str()));
                uniform.m_binding = uDesc.m_binding;
            }
            else if (uDesc.m_type == DescriptorType::uniformBuffer)
            {
                uniform.m_binding = uDesc.m_binding;
                uniform.m_location = CHECK_ERROR(glGetUniformBlockIndex(pipeline->m_program, name.c_str()));
                if (uniform.m_location >= 0)
                    CHECK_ERROR(glUniformBlockBinding(pipeline->m_program, uniform.m_location, uniform.m_binding));
            }
            if (uniform.m_location >= 0)
            {
                pipeline->m_uniformDescriptors.push_back(uniform);
                usedNames.push_back(name);
            }
        }

        size_t pushConstantsSize = 0;
        u32 pushConstantBinding = 0;
        for (const auto& cDesc : shader->m_pushConstants)
        {
            pushConstantsSize += cDesc.m_size;
            pushConstantBinding = cDesc.m_binding;
        }

        if (pushConstantsSize > 0)
        {
            BufferDesc bufferDesc;
            bufferDesc.m_usage = e_bufferUsageFlags::uniform;
            bufferDesc.m_access = e_accessFlags::write;
            bufferDesc.m_data.m_data = nullptr;
            bufferDesc.m_data.m_size = pushConstantsSize;
            pipeline->m_constantBuffer = (GLBuffer*)createBuffer(bufferDesc);
            pipeline->m_constantBufferBinding = pushConstantBinding;
        }

        return (ComputePipeline)pipeline;
    }

    void GLContext::destroyShader(Shader shader)
    {
        GLShader* glShdr = (GLShader*)shader;
        CHECK_ERROR(glDeleteShader(glShdr->m_shader));
        delete glShdr;
    }

    void GLContext::destroyBuffer(Buffer buffer)
    {
        GLBuffer* glBuf = (GLBuffer*)buffer;

        // remove from tracked buffers
        for (size_t i = 0; i < m_buffers.size(); ++i)
        {
            if (m_buffers[i] == glBuf)
            {
                m_buffers.erase(m_buffers.begin() + i);
                break;
            }
        }

        glBuf->destroy();
        delete glBuf;
    }

    void GLContext::destroyTexture(Texture texture)
    {
        GLTexture* glTex = (GLTexture*)texture;
        CHECK_ERROR(glDeleteTextures(1, &glTex->m_texture));
        if (glTex->m_desc)
        {
            delete glTex->m_desc;
            glTex->m_desc = nullptr;
        }
        delete glTex;
    }

    void GLContext::destroySampler(Sampler sampler)
    {
        GLSampler* glSampler = (GLSampler*)sampler;
        CHECK_ERROR(glDeleteSamplers(1, &glSampler->m_sampler));
        delete glSampler;
    }

    void GLContext::destroyRenderPass(RenderPass pass)
    {
        GLRenderPass* rp = (GLRenderPass*)pass;
        delete rp;
    }
    
    void GLContext::destroyRenderPipeline(RenderPipeline pipeline)
    {
        GLRenderPipeline* glPipeline = (GLRenderPipeline*)pipeline;
        CHECK_ERROR(glDeleteProgram(glPipeline->m_program));
        if (glPipeline->m_constantBuffer)
        {
            destroyBuffer(glPipeline->m_constantBuffer);
            glPipeline->m_constantBuffer = nullptr;
        }
        delete glPipeline;
    }

    void GLContext::destroyComputePipeline(ComputePipeline pipeline)
    {
        GLComputePipeline* glPipeline = (GLComputePipeline*)pipeline;
        CHECK_ERROR(glDeleteProgram(glPipeline->m_program));
        if (glPipeline->m_constantBuffer)
        {
            destroyBuffer(glPipeline->m_constantBuffer);
            glPipeline->m_constantBuffer = nullptr;
        }
        delete glPipeline;
    }

    void GLContext::setViewport(const Viewport& vp)
    {
        m_currentViewport = vp;
        glViewport((GLint)vp.x, (GLint)vp.y, (GLsizei)vp.m_width, (GLsizei)vp.m_height);
        glDepthRangef(vp.m_minDepth, vp.m_maxDepth);
    }

    void GLContext::setScissorRect(const Rect& sr)
    {
        glEnable(GL_SCISSOR_TEST);
        GLint top = (GLint)m_currentViewport.m_height - (GLint)sr.y;
        CHECK_ERROR(glScissor(sr.x, sr.y, sr.w, sr.h));
    }

    void GLContext::begin()
    {
        
    }
        
    void GLContext::beginComputeEncoding()
    {
        
    }

    void GLContext::beginSubpass(const SubpassDesc& subpass, Clear* clears, size_t numClears)
    {
        int currentClear = 0;
        for (u32 i = 0; i < subpass.m_colorAttachments.size(); ++i)
        {
            if (subpass.m_colorAttachments[i].m_loadOp == e_loadOpType::clear)
            {
                const Clear& clear = clears[currentClear];
                float color[] = { clear.r, clear.g, clear.b, clear.a };
                CHECK_ERROR(glClearBufferfv(GL_COLOR, i, color));
            }
            currentClear++;
        }

        for (u32 i = 0; i < subpass.m_depthAttachments.size(); ++i)
        {
            if (subpass.m_depthAttachments[i].m_loadOp == e_loadOpType::clear)
            {
                const Clear& clear = clears[currentClear];
                CHECK_ERROR(glClearBufferfi(GL_DEPTH_STENCIL, i, clear.depth, clear.stencil));
            }
            currentClear++;
        }
    }

    void GLContext::beginRenderPass(RenderPass pass, Framebuffer framebuffer, Clear* clears, size_t numClears)
    {
        if (!pass)
            return;

        GLRenderPass* glPass = (GLRenderPass*)pass;
        GLFramebuffer* glFbo = (GLFramebuffer*)framebuffer;
        CHECK_ERROR(glBindFramebuffer(GL_FRAMEBUFFER, glFbo->m_fbo));

        // bind textures to fbo
        size_t colourBindings = 0;
        std::vector<GLenum> attachments;
        for(auto& t : glFbo->m_textures)
        {
            GLenum attachment;
            if (t->m_format.m_format == GL_DEPTH_COMPONENT)
            {
                attachment = GL_DEPTH_ATTACHMENT;
            }
            else if (t->m_format.m_format == GL_DEPTH_STENCIL)
            {
                attachment = GL_DEPTH_STENCIL_ATTACHMENT;
            }
            else
            {
                attachment = GL_COLOR_ATTACHMENT0 + colourBindings;
                colourBindings++;
                attachments.push_back(attachment);
            }
            CHECK_ERROR(glFramebufferTexture2D(GL_FRAMEBUFFER, attachment, GL_TEXTURE_2D, t->m_texture, 0));
        }
        if (framebuffer == m_backBufferFramebuffer)
        {
            attachments = { GL_FRONT_LEFT };
        }
        if (!attachments.empty())
        {
            CHECK_ERROR(glDrawBuffers(attachments.size(), attachments.data()));
        }

        // load action
        glPass->m_currentSubpass = 0;
        beginSubpass(glPass->m_desc.m_subpasses[0], clears, numClears);
        
        m_currentRenderPass = glPass;
        m_doFramebufferFlip = false; glFbo == m_backBufferFramebuffer;
    }


    void GLContext::beginNextSubpass()
    {
        m_currentRenderPass->m_currentSubpass = std::min(m_currentRenderPass->m_currentSubpass + 1, m_currentRenderPass->m_desc.m_subpasses.size() - 1);
        beginSubpass(m_currentRenderPass->m_desc.m_subpasses[m_currentRenderPass->m_currentSubpass], nullptr, 0);
    }

    void GLContext::bindRenderPipeline(RenderPipeline pipeline)
    {
        GLRenderPipeline* glPipeline = (GLRenderPipeline*)pipeline;
        GLVertexLayout* vertexLayout = glPipeline->m_vertexLayout;
        if (glPipeline->m_rasterState.m_cullMode != GL_NONE)
        {
            CHECK_ERROR(glEnable(GL_CULL_FACE));
            CHECK_ERROR(glCullFace(glPipeline->m_rasterState.m_cullMode));
            CHECK_ERROR(glFrontFace(glPipeline->m_rasterState.m_frontFace));
        }
        else
        {
            CHECK_ERROR(glDisable(GL_CULL_FACE));
        }

        CHECK_ERROR(glUseProgram(glPipeline->m_program));

        if (glPipeline->m_blendState.m_numBlendStates > 0)
        {
            const GLBlendState& blend = glPipeline->m_blendState.m_blendStates[0];
            CHECK_ERROR(glColorMask(blend.m_writeMask.r, blend.m_writeMask.g,
                blend.m_writeMask.b, blend.m_writeMask.a));
            if (blend.m_enabled)
            {
                glEnable(GL_BLEND);
                CHECK_ERROR(glBlendFuncSeparate(blend.m_srcColorBlend, blend.m_dstColorBlend,
                    blend.m_srcAlphaBlend, blend.m_dstAlphaBlend));
                CHECK_ERROR(glBlendEquationSeparate(blend.m_colorBlendOp, blend.m_alphaBlendOp));
            }
            else
            {
                glDisable(GL_BLEND);
            }
        }
        else
        {
            CHECK_ERROR(glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE));
            glDisable(GL_BLEND);
        }

        auto& dss = glPipeline->m_depthStencilState;
        if (dss.m_depthEnable)
        {
            CHECK_ERROR(glEnable(GL_DEPTH_TEST));
            CHECK_ERROR(glDepthMask(dss.m_depthWrite));
            CHECK_ERROR(glDepthFunc(dss.m_depthFunc));            
            CHECK_ERROR(glStencilFuncSeparate(GL_FRONT, dss.m_frontStencilFunc, dss.m_stencilWriteMask, dss.m_stencilReadMask));
            CHECK_ERROR(glStencilFuncSeparate(GL_BACK, dss.m_backStencilFunc, dss.m_stencilWriteMask, dss.m_stencilReadMask));
            CHECK_ERROR(glStencilOpSeparate(GL_FRONT, dss.m_frontStencilFailOp, dss.m_frontStencilDepthFailOp, dss.m_frontStencilPassOp));
            CHECK_ERROR(glStencilOpSeparate(GL_BACK, dss.m_backStencilFailOp, dss.m_backStencilDepthFailOp, dss.m_backStencilPassOp));
        }
        else
        {
            CHECK_ERROR(glDisable(GL_DEPTH_TEST));
        }

        if (dss.m_stencilEnable)
        {
            CHECK_ERROR(glEnable(GL_STENCIL_TEST));
        }
        else
        {
            CHECK_ERROR(glDisable(GL_STENCIL_TEST));
        }

        for (auto& desc : glPipeline->m_uniformDescriptors)
        {
            if (desc.m_type == e_descriptorType::sampledImage)
                CHECK_ERROR(glUniform1i(desc.m_location, desc.m_binding));
        }

        if (m_doFramebufferFlip && glPipeline->m_vFlipLoc >= 0)
        {
            CHECK_ERROR(glUniform1f(glPipeline->m_vFlipLoc, -1.0f));
            CHECK_ERROR(glFrontFace(GL_CW));
        }
        else if (glPipeline->m_vFlipLoc >= 0)
        {
            CHECK_ERROR(glUniform1f(glPipeline->m_vFlipLoc, 1.0f));
            CHECK_ERROR(glFrontFace(GL_CCW));
        }
        else
        {
            CHECK_ERROR(glFrontFace(GL_CCW));
        }

        if (glPipeline->m_constantBuffer)
        {
            bindBuffer(glPipeline->m_constantBuffer, glPipeline->m_constantBufferBinding, e_bindFlags::uniform, 0);
        }

        m_currentPipeline = glPipeline;
        m_currentVertexLayout = vertexLayout;
    }

    void GLContext::bindComputePipeline(ComputePipeline pipeline)
    {
        GLComputePipeline* glPipeline = (GLComputePipeline*)pipeline;

        CHECK_ERROR(glUseProgram(glPipeline->m_program));
        for (auto& desc : glPipeline->m_uniformDescriptors)
        {
            if (desc.m_type == e_descriptorType::sampledImage)
                CHECK_ERROR(glUniform1i(desc.m_location, desc.m_binding));
        }

        if (glPipeline->m_constantBuffer)
        {
            bindBuffer(glPipeline->m_constantBuffer, glPipeline->m_constantBufferBinding, e_bindFlags::uniform, 0);
        }

        m_currentPipeline = glPipeline;
    }

    void GLContext::bindBufferInternal(Buffer buf, u32 slot, InternalBindFlags flags, size_t offset)
    {
        GLBuffer* buffer = (GLBuffer*)buf;
        if (flags & e_internalBindFlags::vertexBuffer)
        {
            CHECK_ERROR(glBindBuffer(GL_ARRAY_BUFFER, buffer->bind(flags)));
            if (m_currentVertexLayout)
            {
                for (size_t i = 0; i < m_currentVertexLayout->m_vertexAttribs.size(); ++i)
                {
                    GLVertexAttrib& attrib = m_currentVertexLayout->m_vertexAttribs[i];
                    if (slot != attrib.m_bufferIndex)
                        continue;
                    GLVertexBufferLayout& bufferLayout = m_currentVertexLayout->m_vertexBuffers[attrib.m_bufferIndex];
                    CHECK_ERROR(glEnableVertexAttribArray(attrib.m_location));
                    CHECK_ERROR(glVertexAttribPointer(attrib.m_location, attrib.m_format.m_size, attrib.m_format.m_type, attrib.m_format.m_type == GL_UNSIGNED_BYTE, bufferLayout.m_stride, (void*)(attrib.m_offset + offset)));
                    CHECK_ERROR(glVertexAttribDivisor(attrib.m_location, bufferLayout.m_divisor));
                }
            }
        }
        else if (flags & e_internalBindFlags::indexBuffer)
        {
            CHECK_ERROR(glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, buffer->bind(flags)));
        }
        else if (flags & e_internalBindFlags::uniform)
        {
            if (offset == 0)
            {
                CHECK_ERROR(glBindBufferBase(GL_UNIFORM_BUFFER, slot, buffer->bind(flags)));
            }
            else
            {
                CHECK_ERROR(glBindBufferRange(GL_UNIFORM_BUFFER, slot, buffer->bind(flags), offset, buffer->m_data.m_size));
            }
        }
        else if (flags & e_internalBindFlags::dispatchIndirect)
        {
            CHECK_ERROR(glBindBuffer(GL_DISPATCH_INDIRECT_BUFFER, buffer->bind(flags)));
        }
    }

    void GLContext::bindBuffer(Buffer buf, u32 slot, BindFlags flags, size_t offset)
    {
        bindBufferInternal(buf, slot, translateBindFlags(flags), offset);
    }
    
    void GLContext::bindTexture(Texture tex, Sampler samp, u32 slot, BindFlags flags)
    {
        GLTexture* texture = (GLTexture*)tex;
        GLSampler* sampler = (GLSampler*)samp;

        CHECK_ERROR(glActiveTexture(GL_TEXTURE0 + slot));
        CHECK_ERROR(glBindTexture(texture->m_target, texture->m_texture));

        switch (texture->m_format.m_swizzle)
        {
        case NO_SWIZZLE:
            break;
        case SWIZZLE_RGB_BGR:
            CHECK_ERROR(glTexParameteri(texture->m_target, GL_TEXTURE_SWIZZLE_R, GL_BLUE));
            CHECK_ERROR(glTexParameteri(texture->m_target, GL_TEXTURE_SWIZZLE_G, GL_GREEN));
            CHECK_ERROR(glTexParameteri(texture->m_target, GL_TEXTURE_SWIZZLE_B, GL_RED));
            break;
        case SWIZZLE_R_A:
            CHECK_ERROR(glTexParameteri(texture->m_target, GL_TEXTURE_SWIZZLE_A, GL_RED));
            break;
        }

        CHECK_ERROR(glTexParameteri(texture->m_target, GL_TEXTURE_MAX_LEVEL, texture->m_maxLevel));

        if (samp)
        {
            CHECK_ERROR(glBindSampler(slot, sampler->m_sampler));
        }
    }

    void GLContext::updateBuffer(Buffer buf, const Mem& data, size_t offset)
    {
        GLBuffer* buffer = (GLBuffer*)buf;
        char* mappedOffset = ((char*)buffer->m_data.m_data) + offset;
        memcpy(mappedOffset, data.m_data, data.m_size);
        buffer->dirty();
    }

    void GLContext::updateTexture(Texture tex, const TextureUpdateDesc& desc)
    {
        GLTexture* texture = (GLTexture*)tex;
        CHECK_ERROR(glActiveTexture(GL_TEXTURE0));
        CHECK_ERROR(glBindTexture(GL_TEXTURE_2D, texture->m_texture));
        CHECK_ERROR(glTexSubImage2D(GL_TEXTURE_2D, desc.m_mip, desc.x, desc.y, (GLint)desc.m_width, (GLint)desc.m_height, texture->m_format.m_format, texture->m_format.m_type, desc.m_data.m_data));
    }

    void GLContext::updateConstants(ShaderStageFlags stage, size_t size, size_t offset, void* data)
    {
        if (!m_currentPipeline)
            return;

        Mem mem;
        mem.m_data = data;
        mem.m_size = size;
        updateBuffer(m_currentPipeline->m_constantBuffer, mem, offset);
    }

    void GLContext::insertBarrier(PipelineStage srcStage, PipelineStage dstStage, BarrierFlags barrierType)
    {
        
    }

    void GLContext::copyTexture(Texture src, Texture dst, u32 srcLayer, u32 dstLayer)
    {
        GLTexture* srcTex = (GLTexture*)src;
        GLTexture* dstTex = (GLTexture*)dst;

        GLuint fbo;
        CHECK_ERROR(glGenFramebuffers(1, &fbo));
        CHECK_ERROR(glBindFramebuffer(GL_READ_FRAMEBUFFER, fbo));
        CHECK_ERROR(glFramebufferTexture2D(GL_READ_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, srcTex->m_texture, 0));
        CHECK_ERROR(glReadBuffer(GL_COLOR_ATTACHMENT0));
        CHECK_ERROR(glCopyTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, 0, 0, srcTex->m_width, srcTex->m_height));
        CHECK_ERROR(glFramebufferTexture2D(GL_READ_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, 0, 0));
        CHECK_ERROR(glDeleteFramebuffers(1, &fbo));
    }
    
    void GLContext::draw(u32 vertexCount, u32 startVertex)
    {
        CHECK_ERROR(glDrawArrays(m_currentPrimitiveType, startVertex, vertexCount));
    }

    void GLContext::drawIndexed(u32 indexCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset, u32 vertexOffset)
    {
        bindBuffer(indexBuffer, 0, e_bindFlags::indexBuffer, 0);
        CHECK_ERROR(glDrawElements(m_currentPrimitiveType, indexCount, translateIndexType(indexType), (void*)(intptr_t)indexBufferOffset));
    }

    void GLContext::drawIndexedInstanced(u32 indexCount, u32 instanceCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset, u32 vertexOffset)
    {
        bindBuffer(indexBuffer, 0, e_bindFlags::indexBuffer, 0);
        CHECK_ERROR(glDrawElementsInstanced(m_currentPrimitiveType, indexCount, translateIndexType(indexType), (void*)(intptr_t)indexBufferOffset, instanceCount));        
    }

    void GLContext::drawIndirect(u32 drawCount, Buffer drawBuffer, u32 drawBufferOffset)
    {
        GLBuffer* buffer = (GLBuffer*)drawBuffer;
        char* data = (char*)buffer->m_data.m_data + drawBufferOffset;
        CHECK_ERROR(glDrawArraysIndirect(m_currentPrimitiveType, data));
    }

    void GLContext::drawIndexedIndirect(u32 drawCount, Buffer indexBuffer, IndexType indexType, Buffer drawBuffer, u32 indexBufferOffset, u32 drawBufferOffset)
    {
        GLBuffer* buffer = (GLBuffer*)drawBuffer;
        char* data = (char*)buffer->m_data.m_data + drawBufferOffset;
        bindBuffer(indexBuffer, 0, e_bindFlags::indexBuffer, 0);
        CHECK_ERROR(glDrawElementsIndirect(m_currentPrimitiveType, translateIndexType(indexType), data));
    }

    void GLContext::dispatchCompute(u32 grid[3], u32 numThreads[3])
    {
        CHECK_ERROR(glDispatchCompute(grid[0], grid[1], grid[2]));
    }

    void GLContext::dispatchComputeIndirect(Buffer drawBuffer, u32 offset)
    {
        bindBufferInternal(drawBuffer, 0, e_internalBindFlags::dispatchIndirect, offset);
        CHECK_ERROR(glDispatchComputeIndirect(offset));
    }

    void GLContext::endRenderEncoding()
    {
        glDisable(GL_SCISSOR_TEST);
    }

    void GLContext::endComputeEncoding()
    {

    }

    void GLContext::enqueue()
    {

    }

    Semaphore GLContext::commit(Semaphore* waits = nullptr, size_t count = 0, PipelineStage stage)
    {
        return nullptr;
    }

    void GLContext::present()
    {
        for (auto& buffer : m_buffers)
            buffer->reset();
        incrementWrapBufferedResources();

        SwapBuffers(m_hdc);
    }

    // async context clean up used command buffers without calling begin()
    void GLContext::flushCommandBuffers()
    {
        
    }

    bool GLContext::isSignalled(Semaphore semaphore)
    {
        return true;
    }

    static gfx::GLContext* s_mainContext;
    static gfx::GLContext* s_asyncContexts[MAX_ASYNC_CONTEXTS];
    void createMainContext(void* platform)
    {
        s_mainContext = new GLContext();
        s_mainContext->setup(platform);

        // make sure all the async contexts are null before creating any
        memset(s_asyncContexts, 0, sizeof(s_asyncContexts[0]) * MAX_ASYNC_CONTEXTS);
    }

    Context* getMainContext()
    {
        return s_mainContext;
    }

    Context* getAsyncContext()
    {
        // search for existing async context for this thread
        std::thread::id threadID = std::this_thread::get_id();
        int first = INT_MAX;
        for (int i = 0; i < MAX_ASYNC_CONTEXTS; ++i)
        {
            if (s_asyncContexts[i] && s_asyncContexts[i]->m_threadID == threadID)
                return s_asyncContexts[i];

            if (!s_asyncContexts[i] && i < first)
                first = i;
        }

        ASSERT(first < INT_MAX); // no free async contexts remaining!!!

                                 // create a new async context for this thread id
        s_asyncContexts[first] = new GLContext();
        s_asyncContexts[first]->setupAsync(s_mainContext, threadID);
        return s_asyncContexts[first];
    }

    void destroyContext(Context* context)
    {
        if (context == s_mainContext)
        {
            // shutdown main context
            ((GLContext*)context)->shutdown();
            delete context;
        }
        else
        {
            // shutdown async context
            for (int i = 0; i < MAX_ASYNC_CONTEXTS; ++i)
            {
                if (((GLContext*)context)->m_threadID == s_asyncContexts[i]->m_threadID)
                {
                    s_asyncContexts[i]->shutdownAsync();
                    delete s_asyncContexts[i];
                    s_asyncContexts[i] = nullptr;
                    break;
                }
            }
        }
    }

    std::string getAPI()
    {
        return "opengl";
    }
}
