#include "gfx.h"

#define VK_USE_PLATFORM_WIN32_KHR
#include <vulkan/vulkan.h>
#include <string>
#include <vector>
#include <limits>
#include <thread>
#include <stdint.h>
#include <map>

#include "platform.h"
#include "gfx_win32.h"

#define MAX_QUEUED_COMMAND_BUFFERS 16
#define INITIAL_NUM_COMMAND_BUFFERS 16
#define MAX_DESCRIPTOR_WRITES 32

namespace gfx
{
    VKAPI_ATTR VkBool32 VKAPI_CALL debugCallback(
        VkDebugReportFlagsEXT flags,
        VkDebugReportObjectTypeEXT objType,
        uint64_t obj,
        size_t location,
        int32_t code,
        const char* layerPrefix,
        const char* msg,
        void* userData) {
        printf("validation layer: %s", msg);
        ASSERT(false);
        return VK_FALSE;
    }

    VkFormat translateFormat(const FormatDesc& desc)
    {
        VkFormat format = VK_FORMAT_R8G8B8A8_UNORM;
        if ((desc.m_channels & e_formatChannelFlags::rgba) == e_formatChannelFlags::rgba)
        {
            if (desc.m_bitDepth == 8)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R8G8B8A8_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R8G8B8A8_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R8G8B8A8_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R8G8B8A8_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R8G8B8A8_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R8G8B8A8_SINT;

                if (desc.m_reverse)
                    format = (VkFormat)(format + (VK_FORMAT_B8G8R8A8_UNORM - VK_FORMAT_R8G8B8A8_UNORM));

                if (desc.m_standard)
                    format = (VkFormat)(format + (VK_FORMAT_R8G8B8A8_SRGB - VK_FORMAT_R8G8B8A8_UNORM));
            }
            else if (desc.m_bitDepth == 16)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R16G16B16A16_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R16G16B16A16_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R16G16B16A16_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R16G16B16A16_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R16G16B16A16_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R16G16B16A16_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R16G16B16A16_SFLOAT;
            }
            else if (desc.m_bitDepth == 32)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R32G32B32A32_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R32G32B32A32_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R32G32B32A32_SFLOAT;
            }
            else if (desc.m_bitDepth == 64)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R64G64B64A64_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R64G64B64A64_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R64G64B64A64_SFLOAT;
            }
        }
        else if ((desc.m_channels & e_formatChannelFlags::rgb) == e_formatChannelFlags::rgb)
        {
            if (desc.m_bitDepth == 8)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R8G8B8_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R8G8B8_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R8G8B8_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R8G8B8_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R8G8B8_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R8G8B8_SINT;
            }
            else if (desc.m_bitDepth == 16)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R16G16B16_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R16G16B16_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R16G16B16_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R16G16B16_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R16G16B16_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R16G16B16_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R16G16B16_SFLOAT;
            }
            else if (desc.m_bitDepth == 32)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R32G32B32_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R32G32B32_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R32G32B32_SFLOAT;
            }
            else if (desc.m_bitDepth == 64)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R64G64B64_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R64G64B64_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R64G64B64_SFLOAT;
            }

        }
        else if ((desc.m_channels & e_formatChannelFlags::rg) == e_formatChannelFlags::rg)
        {
            if (desc.m_bitDepth == 8)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R8G8_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R8G8_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R8G8_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R8G8_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R8G8_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R8G8_SINT;
            }
            else if (desc.m_bitDepth == 16)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R16G16_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R16G16_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R16G16_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R16G16_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R16G16_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R16G16_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R16G16_SFLOAT;
            }
            else if (desc.m_bitDepth == 32)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R32G32_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R32G32_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R32G32_SFLOAT;
            }
            else if (desc.m_bitDepth == 64)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R64G64_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R64G64_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R64G64_SFLOAT;
            }
        }
        else if ((desc.m_channels & e_formatChannelFlags::r) == e_formatChannelFlags::r)
        {
            if (desc.m_bitDepth == 8)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R8_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R8_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R8_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R8_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R8_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R8_SINT;
            }
            else if (desc.m_bitDepth == 16)
            {
                if (desc.m_type == e_formatType::unsignedNorm)
                    format = VK_FORMAT_R16_UNORM;
                else if (desc.m_type == e_formatType::signedNorm)
                    format = VK_FORMAT_R16_SNORM;
                else if (desc.m_type == e_formatType::unsignedScaled)
                    format = VK_FORMAT_R16_USCALED;
                else if (desc.m_type == e_formatType::signedScaled)
                    format = VK_FORMAT_R16_SSCALED;
                else if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R16_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R16_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R16_SFLOAT;
            }
            else if (desc.m_bitDepth == 32)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R32_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R32_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R32_SFLOAT;
            }
            else if (desc.m_bitDepth == 64)
            {
                if (desc.m_type == e_formatType::unsignedInt)
                    format = VK_FORMAT_R64_UINT;
                else if (desc.m_type == e_formatType::signedInt)
                    format = VK_FORMAT_R64_SINT;
                else if (desc.m_type == e_formatType::signedFloat)
                    format = VK_FORMAT_R64_SFLOAT;
            }
        }
        else if (desc.m_channels == e_formatChannelFlags::d)
        {
            if (desc.m_bitDepth == 16)
                format = VK_FORMAT_D16_UNORM;
            else if (desc.m_bitDepth == 24)
                format = VK_FORMAT_D24_UNORM_S8_UINT;
            else if (desc.m_bitDepth == 32)
                format = VK_FORMAT_D32_SFLOAT;
        }
        else if (desc.m_channels == e_formatChannelFlags::ds)
        {
            if (desc.m_bitDepth == 16)
                format = VK_FORMAT_D16_UNORM_S8_UINT;
            else if (desc.m_bitDepth == 24)
                format = VK_FORMAT_D24_UNORM_S8_UINT;
            else if (desc.m_bitDepth == 32)
                format = VK_FORMAT_D32_SFLOAT_S8_UINT;
        }
        return format;
    }

    VkImageUsageFlags translateImageUsage(TextureUsageFlags flags)
    {
        VkImageUsageFlags f = 0;
        if (flags & e_textureUsageFlags::transferSrc)
            f |= VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
        if (flags & e_textureUsageFlags::transferDst)
            f |= VK_IMAGE_USAGE_TRANSFER_DST_BIT;
        if (flags & e_textureUsageFlags::shaderRead)
            f |= VK_IMAGE_USAGE_SAMPLED_BIT;
        if (flags & e_textureUsageFlags::shaderWrite)
            f |= VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
        if (flags & e_textureUsageFlags::storage)
            f |= VK_IMAGE_USAGE_STORAGE_BIT;
        if (flags & e_textureUsageFlags::depthStencil)
            f |= VK_IMAGE_USAGE_DEPTH_STENCIL_ATTACHMENT_BIT;
        if (flags & e_textureUsageFlags::inputAttachment)
            f |= VK_IMAGE_USAGE_INPUT_ATTACHMENT_BIT;
        return f;
    }

    VkBufferUsageFlags translateBufferUsage(BufferUsageFlags flags)
    {
        VkBufferUsageFlags f = 0;
        if (flags & e_bufferUsageFlags::transferSrc)
            f |= VK_BUFFER_USAGE_TRANSFER_SRC_BIT;
        if (flags & e_bufferUsageFlags::transferDst)
            f |= VK_BUFFER_USAGE_TRANSFER_DST_BIT;
        if (flags & e_bufferUsageFlags::uniform)
            f |= VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT;
        if (flags & e_bufferUsageFlags::storage)
            f |= VK_BUFFER_USAGE_STORAGE_BUFFER_BIT;
        if (flags & e_bufferUsageFlags::vertex)
            f |= VK_BUFFER_USAGE_VERTEX_BUFFER_BIT;
        if (flags & e_bufferUsageFlags::index)
            f |= VK_BUFFER_USAGE_INDEX_BUFFER_BIT;
        if (flags & e_bufferUsageFlags::indirect)
            f |= VK_BUFFER_USAGE_INDIRECT_BUFFER_BIT;
        return f;
    }

    VkAttachmentLoadOp translateLoadOp(LoadOp op)
    {
        switch (op)
        {
        case LoadOp::clear:
            return VK_ATTACHMENT_LOAD_OP_CLEAR;
        case LoadOp::load:
            return VK_ATTACHMENT_LOAD_OP_LOAD;
        default:
            return VK_ATTACHMENT_LOAD_OP_DONT_CARE;
        }
    }

    VkAttachmentStoreOp translateStoreOp(StoreOp op)
    {
        switch (op)
        {
        case StoreOp::store:
            return VK_ATTACHMENT_STORE_OP_STORE;
        default:
            return VK_ATTACHMENT_STORE_OP_DONT_CARE;
        }
    }

    VkPipelineStageFlags translatePipelineStage(PipelineStage stage)
    {
        switch (stage)
        {
        case PipelineStage::top:
            return VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
        case PipelineStage::vertexInput:
            return VK_PIPELINE_STAGE_VERTEX_INPUT_BIT;
        case PipelineStage::vertexOutput:
            return VK_PIPELINE_STAGE_VERTEX_SHADER_BIT;
        case PipelineStage::fragmentInput:
            return VK_PIPELINE_STAGE_GEOMETRY_SHADER_BIT;
        case PipelineStage::fragmentOutput:
            return VK_PIPELINE_STAGE_FRAGMENT_SHADER_BIT;
        case PipelineStage::colorOutput:
            return VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
        case PipelineStage::compute:
            return VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT;
        case PipelineStage::transfer:
            return VK_PIPELINE_STAGE_TRANSFER_BIT;
        case PipelineStage::bottom:
            return VK_PIPELINE_STAGE_BOTTOM_OF_PIPE_BIT;
        default:
            return VK_PIPELINE_STAGE_ALL_COMMANDS_BIT;
        }
    }

    VkAccessFlags translateMemoryAccess(MemoryAccessFlags flags)
    {
        VkAccessFlags f = 0;
        if (flags & e_memoryAccessFlags::indirect)
            f |= VK_ACCESS_INDIRECT_COMMAND_READ_BIT;
        if (flags & e_memoryAccessFlags::index)
            f |= VK_ACCESS_INDEX_READ_BIT;
        if (flags & e_memoryAccessFlags::vertex)
            f |= VK_ACCESS_VERTEX_ATTRIBUTE_READ_BIT;
        if (flags & e_memoryAccessFlags::uniform)
            f |= VK_ACCESS_UNIFORM_READ_BIT;
        if (flags & e_memoryAccessFlags::inputAttachment)
            f |= VK_ACCESS_INPUT_ATTACHMENT_READ_BIT;
        if (flags & e_memoryAccessFlags::colorRead)
            f |= VK_ACCESS_COLOR_ATTACHMENT_READ_BIT;
        if (flags & e_memoryAccessFlags::colorWrite)
            f |= VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
        if (flags & e_memoryAccessFlags::depthStencilRead)
            f |= VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT;
        if (flags & e_memoryAccessFlags::depthStencilWrite)
            f |= VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT;
        return f;
    }

    std::vector<VkDynamicState> translateDynamicState(DynamicState flags)
    {
        std::vector<VkDynamicState> states;
        if (flags & e_dynamicStateFlags::viewport)
            states.push_back(VK_DYNAMIC_STATE_VIEWPORT);
        if (flags & e_dynamicStateFlags::scissor)
            states.push_back(VK_DYNAMIC_STATE_SCISSOR);
        if (flags & e_dynamicStateFlags::lineWidth)
            states.push_back(VK_DYNAMIC_STATE_LINE_WIDTH);
        if (flags & e_dynamicStateFlags::depthBias)
            states.push_back(VK_DYNAMIC_STATE_DEPTH_BIAS);
        if (flags & e_dynamicStateFlags::blendConstants)
            states.push_back(VK_DYNAMIC_STATE_BLEND_CONSTANTS);
        if (flags & e_dynamicStateFlags::depthBounds)
            states.push_back(VK_DYNAMIC_STATE_DEPTH_BOUNDS);
        if (flags & e_dynamicStateFlags::stencilCompareMask)
            states.push_back(VK_DYNAMIC_STATE_STENCIL_COMPARE_MASK);
        if (flags & e_dynamicStateFlags::stencilWriteMask)
            states.push_back(VK_DYNAMIC_STATE_STENCIL_WRITE_MASK);
        if (flags & e_dynamicStateFlags::stencilReference)
            states.push_back(VK_DYNAMIC_STATE_STENCIL_REFERENCE);
        return states;
    }

    VkBlendFactor translateBlendFactor(BlendFactor factor)
    {
        // vulkan blend factor enums are packed (factor, 1-factor)
        VkBlendFactor blend;
        bool invertFactor = (factor & (1 << 0));
        if (invertFactor)
            factor = (BlendFactor)(factor - 1);
        switch (factor)
        {
        case e_blendFactorType::zero:
            blend = VK_BLEND_FACTOR_ZERO;
            break;
        case e_blendFactorType::srcColor:
            blend = VK_BLEND_FACTOR_SRC_COLOR;
            break;
        case e_blendFactorType::dstColor:
            blend = VK_BLEND_FACTOR_DST_COLOR;
            break;
        case e_blendFactorType::srcAlpha:
            blend = VK_BLEND_FACTOR_SRC_ALPHA;
            break;
        case e_blendFactorType::dstAlpha:
            blend = VK_BLEND_FACTOR_DST_ALPHA;
            break;
        case e_blendFactorType::constantColor:
            blend = VK_BLEND_FACTOR_CONSTANT_COLOR;
            break;
        case e_blendFactorType::constantAlpha:
            blend = VK_BLEND_FACTOR_CONSTANT_ALPHA;
            break;
        case e_blendFactorType::src1Color:
            blend = VK_BLEND_FACTOR_SRC1_COLOR;
            break;
        case e_blendFactorType::src1Alpha:
            blend = VK_BLEND_FACTOR_SRC1_ALPHA;
            break;
        default:
            blend = VK_BLEND_FACTOR_ZERO;
            break;
        }
        if (invertFactor)
            blend = (VkBlendFactor)(blend + 1);
        return blend;
    }

    VkBlendOp translateBlendOp(BlendOp op)
    {
        switch (op)
        {
        case e_blendOpType::add:
            return VK_BLEND_OP_ADD;
            break;
        case e_blendOpType::subtract:
            return VK_BLEND_OP_SUBTRACT;
            break;
        case e_blendOpType::reverseSubtract:
            return VK_BLEND_OP_REVERSE_SUBTRACT;
            break;
        case e_blendOpType::min:
            return VK_BLEND_OP_MIN;
            break;
        case e_blendOpType::max:
            return VK_BLEND_OP_MAX;
            break;
        default:
            return VK_BLEND_OP_ADD;
            break;
        }
    }

    VkPrimitiveTopology translatePrimitiveTopology(PrimitiveTopology topology)
    {
        switch (topology)
        {
        case e_primitiveTopology::pointList:
            return VK_PRIMITIVE_TOPOLOGY_POINT_LIST;
            break;
        case e_primitiveTopology::lineList:
            return VK_PRIMITIVE_TOPOLOGY_LINE_LIST;
            break;
        case e_primitiveTopology::lineStrip:
            return VK_PRIMITIVE_TOPOLOGY_LINE_STRIP;
            break;
        case e_primitiveTopology::triangleList:
            return VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
            break;
        case e_primitiveTopology::triangleStrip:
            return VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP;
            break;
        case e_primitiveTopology::triangleFan:
            return VK_PRIMITIVE_TOPOLOGY_TRIANGLE_FAN;
            break;
        case e_primitiveTopology::lineListAdjacency:
            return VK_PRIMITIVE_TOPOLOGY_LINE_LIST_WITH_ADJACENCY;
            break;
        case e_primitiveTopology::lineStripAdjacency:
            return VK_PRIMITIVE_TOPOLOGY_LINE_STRIP_WITH_ADJACENCY;
            break;
        case e_primitiveTopology::triangleListAdjacency:
            return VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST_WITH_ADJACENCY;
            break;
        case e_primitiveTopology::triangleStripAdjacency:
            return VK_PRIMITIVE_TOPOLOGY_TRIANGLE_STRIP_WITH_ADJACENCY;
            break;
        case e_primitiveTopology::patchList:
            return VK_PRIMITIVE_TOPOLOGY_PATCH_LIST;
            break;
        default:
            return VK_PRIMITIVE_TOPOLOGY_TRIANGLE_LIST;
            break;
        }
    }

    VkCompareOp translateCompareOp(CompareOp op)
    {
        switch (op)
        {
        case e_compareOpType::never:
            return VK_COMPARE_OP_NEVER;
            break;
        case e_compareOpType::less:
            return VK_COMPARE_OP_LESS;
            break;
        case e_compareOpType::equal:
            return VK_COMPARE_OP_EQUAL;
            break;
        case e_compareOpType::lessOrEqual:
            return VK_COMPARE_OP_LESS_OR_EQUAL;
            break;
        case e_compareOpType::greater:
            return VK_COMPARE_OP_GREATER;
            break;
        case e_compareOpType::notEqual:
            return VK_COMPARE_OP_NOT_EQUAL;
            break;
        case e_compareOpType::greatorOrEqual:
            return VK_COMPARE_OP_GREATER_OR_EQUAL;
            break;
        case e_compareOpType::always:
            return VK_COMPARE_OP_ALWAYS;
            break;
        default:
            return VK_COMPARE_OP_NEVER;
            break;
        }
    }

    VkStencilOp translateStencilOp(StencilOp op)
    {
        switch (op)
        {
        case e_stencilOpType::keep:
            return VK_STENCIL_OP_KEEP;
            break;
        case e_stencilOpType::zero:
            return VK_STENCIL_OP_ZERO;
            break;
        case e_stencilOpType::replace:
            return VK_STENCIL_OP_REPLACE;
            break;
        case e_stencilOpType::incrementClamp:
            return VK_STENCIL_OP_INCREMENT_AND_CLAMP;
            break;
        case e_stencilOpType::decrementClamp:
            return VK_STENCIL_OP_DECREMENT_AND_CLAMP;
            break;
        case e_stencilOpType::invert:
            return VK_STENCIL_OP_INVERT;
            break;
        case e_stencilOpType::incrementWrap:
            return VK_STENCIL_OP_INCREMENT_AND_WRAP;
            break;
        case e_stencilOpType::decrementWrap:
            return VK_STENCIL_OP_DECREMENT_AND_WRAP;
            break;
        default:
            return VK_STENCIL_OP_KEEP;
            break;
        }
    }

    VkPolygonMode translatePolygonMode(PolygonMode mode)
    {
        switch (mode)
        {
        case e_polygonModeType::fill:
            return VK_POLYGON_MODE_FILL;
            break;
        case e_polygonModeType::line:
            return VK_POLYGON_MODE_LINE;
            break;
        case e_polygonModeType::point:
            return VK_POLYGON_MODE_POINT;
            break;
        default:
            return VK_POLYGON_MODE_FILL;
            break;
        }
    }

    VkCullModeFlags translateCullMode(CullMode mode)
    {
        VkCullModeFlags flags = 0;
        if (mode & e_cullModeFlags::front)
            flags |= VK_CULL_MODE_FRONT_BIT;
        if (mode & e_cullModeFlags::back)
            flags |= VK_CULL_MODE_BACK_BIT;
        return flags;
    }

    VkFrontFace translateFrontFace(FrontFace face)
    {
        switch (face)
        {
        case e_frontFaceType::counterClockwise:
            return VK_FRONT_FACE_COUNTER_CLOCKWISE;
            break;
        case e_frontFaceType::clockwise:
            return VK_FRONT_FACE_CLOCKWISE;
            break;
        default:
            return VK_FRONT_FACE_COUNTER_CLOCKWISE;
            break;
        }
    }

    VkLogicOp translateLogicOp(LogicOp op)
    {
        switch (op)
        {
        case e_logicOpType::clear:
            return VK_LOGIC_OP_CLEAR;
            break;
        case e_logicOpType::and:
            return VK_LOGIC_OP_AND;
            break;
        case e_logicOpType::andReverse:
            return VK_LOGIC_OP_AND_REVERSE;
            break;
        case e_logicOpType::copy:
            return VK_LOGIC_OP_COPY;
            break;
        case e_logicOpType::andInverted:
            return VK_LOGIC_OP_AND_INVERTED;
            break;
        case e_logicOpType::noOp:
            return VK_LOGIC_OP_NO_OP;
            break;
        case e_logicOpType::xor:
            return VK_LOGIC_OP_XOR;
            break;
        case e_logicOpType:: or :
            return VK_LOGIC_OP_OR;
            break;
        case e_logicOpType::nor:
            return VK_LOGIC_OP_NOR;
            break;
        case e_logicOpType::equivalent:
            return VK_LOGIC_OP_EQUIVALENT;
            break;
        case e_logicOpType::invert:
            return VK_LOGIC_OP_INVERT;
            break;
        case e_logicOpType::orReverse:
            return VK_LOGIC_OP_OR_REVERSE;
            break;
        case e_logicOpType::copyInverted:
            return VK_LOGIC_OP_COPY_INVERTED;
            break;
        case e_logicOpType::orInverted:
            return VK_LOGIC_OP_OR_INVERTED;
            break;
        case e_logicOpType::nand:
            return VK_LOGIC_OP_NAND;
            break;
        case e_logicOpType::set:
            return VK_LOGIC_OP_SET;
            break;
        default:
            return VK_LOGIC_OP_NO_OP;
            break;
        }
    }

    VkDescriptorType translateDescriptorType(DescriptorType type)
    {
        switch (type)
        {
        case e_descriptorType::combinedImageSampler:
            return VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
            break;
        case e_descriptorType::sampler:
            return VK_DESCRIPTOR_TYPE_SAMPLER;
            break;
        case e_descriptorType::sampledImage:
            return VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE;
            break;
        case e_descriptorType::storageImage:
            return VK_DESCRIPTOR_TYPE_STORAGE_IMAGE;
            break;
        case e_descriptorType::uniformBuffer:
            return VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
            break;
        case e_descriptorType::storageBuffer:
            return VK_DESCRIPTOR_TYPE_STORAGE_BUFFER;
            break;
        case e_descriptorType::uniformBufferDynamic:
            return VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC;
            break;
        case e_descriptorType::storageBufferDynamic:
            return VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC;
            break;
        case e_descriptorType::inputAttachment:
            return VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT;
            break;
        default:
            return VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
            break;
        }
    }

    VkShaderStageFlagBits translateShaderStage(ShaderStageFlags stages)
    {
        VkShaderStageFlagBits flags;
        if (stages & e_shaderStageFlags::vertex)
            flags = VK_SHADER_STAGE_VERTEX_BIT;
        if (stages & e_shaderStageFlags::fragment)
            flags = VK_SHADER_STAGE_FRAGMENT_BIT;
        if (stages & e_shaderStageFlags::compute)
            flags = VK_SHADER_STAGE_COMPUTE_BIT;
        if (stages & e_shaderStageFlags::tessellationControl)
            flags = VK_SHADER_STAGE_TESSELLATION_CONTROL_BIT;
        if (stages & e_shaderStageFlags::tessellationEvaluation)
            flags = VK_SHADER_STAGE_TESSELLATION_EVALUATION_BIT;
        if (stages & e_shaderStageFlags::geometry)
            flags = VK_SHADER_STAGE_GEOMETRY_BIT;
        return flags;
    }

    VkVertexInputRate translateInputRate(InputRate rate)
    {
        switch (rate)
        {
        case e_inputRateType::vertex:
            return VK_VERTEX_INPUT_RATE_VERTEX;
            break;
        case e_inputRateType::instance:
            return VK_VERTEX_INPUT_RATE_INSTANCE;
            break;
        default:
            return VK_VERTEX_INPUT_RATE_VERTEX;
            break;
        }
    }

    VkFilter translateFilter(Filter filter)
    {
        switch (filter)
        {
        case e_filterType::linear:
            return VK_FILTER_LINEAR;
            break;
        case e_filterType::nearest:
            return VK_FILTER_NEAREST;
            break;
        default:
            return VK_FILTER_LINEAR;
            break;
        }
    }

    VkSamplerMipmapMode translateMipmapMode(MipmapMode mipmap)
    {
        switch (mipmap)
        {
        case e_mipmapMode::linear:
            return VK_SAMPLER_MIPMAP_MODE_LINEAR;
            break;
        case e_mipmapMode::nearest:
            return VK_SAMPLER_MIPMAP_MODE_NEAREST;
            break;
        default:
            return VK_SAMPLER_MIPMAP_MODE_LINEAR;
            break;
        }
    }

    VkSamplerAddressMode translateAddressMode(AddressMode mode)
    {
        switch (mode)
        {
        case e_addressMode::repeat:
            return VK_SAMPLER_ADDRESS_MODE_REPEAT;
            break;
        case e_addressMode::mirroredRepeat:
            return VK_SAMPLER_ADDRESS_MODE_MIRRORED_REPEAT;
            break;
        case e_addressMode::clampToEdge:
            return VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_EDGE;
            break;
        case e_addressMode::clampToBorder:
            return VK_SAMPLER_ADDRESS_MODE_CLAMP_TO_BORDER;
            break;
        case e_addressMode::mirrorClampToEdge:
            return VK_SAMPLER_ADDRESS_MODE_MIRROR_CLAMP_TO_EDGE;
            break;
        default:
            return VK_SAMPLER_ADDRESS_MODE_REPEAT;
            break;
        }
    }

    VkBorderColor translateBorderColor(BorderColor border)
    {
        switch (border)
        {
        case e_borderColor::transparentBlack:
            return VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
            break;
        case e_borderColor::opaqueBlack:
            return VK_BORDER_COLOR_FLOAT_OPAQUE_BLACK;
            break;
        case e_borderColor::opaqueWhite:
            return VK_BORDER_COLOR_FLOAT_OPAQUE_WHITE;
            break;
        default:
            return VK_BORDER_COLOR_FLOAT_TRANSPARENT_BLACK;
            break;
        }
    }

    VkImageAspectFlags translateImageAspect(FormatChannelFlags flags)
    {
        VkImageAspectFlags aspect = 0;
        if (flags & e_formatChannelFlags::r)
            aspect |= VK_IMAGE_ASPECT_COLOR_BIT;
        else if (flags & e_formatChannelFlags::d)
            aspect |= VK_IMAGE_ASPECT_DEPTH_BIT;
        if (flags == e_formatChannelFlags::ds)
            aspect |= VK_IMAGE_ASPECT_STENCIL_BIT;
        return aspect;
    }

    namespace e_queueAlias
    {
        enum queueAlias
        {
            none = -1,
            graphics = 0,
            transfer = 1,
            compute = 2,
            present = 3,
            queueCount = present + 1
        };
    }
    typedef e_queueAlias::queueAlias QueueAlias;

    namespace e_attachmentType
    {
        enum attachmentType : u32
        {
            input,
            colour,
            resolve,
            depth,
            COUNT
        };
    }
    typedef e_attachmentType::attachmentType AttachmentType;

    struct VulkanQueueFamily
    {
        u32     m_family = (u32)-1;
        u32     m_count = 0;
        bool    m_suboptimal = false;
    };

    struct VulkanDescriptor
    {
        VkDescriptorType    m_type;
        u32                 m_count;
    };

    u32 s_numBufferedResources = 1;
    u32 s_bb = 0;
    u32 s_fb = 0;

    void incrementWrapBufferedResources()
    {
        s_bb = (s_bb + 1) % s_numBufferedResources;
        s_fb = (s_fb + 1) % s_numBufferedResources;
    }

    struct VulkanTextureInternal
    {
        VkImage*            m_image = nullptr;
        VkImageView*        m_view = nullptr;
        VkImageView*        m_cubemapViews = nullptr;
        VkDeviceMemory*     m_memory = nullptr;
        VkImageLayout*      m_currentLayout = nullptr;
        VkImageLayout       m_desiredLayout = VK_IMAGE_LAYOUT_GENERAL;
        bool                m_active = false;
        bool                m_buffered = false;
        bool                m_resolved = false;
        size_t              m_count = 1;
        u32                 m_layers = 1;
        VkImageAspectFlags  m_aspect = VK_IMAGE_ASPECT_COLOR_BIT;

        void destroy(VkDevice device, VkAllocationCallbacks* callbacks)
        {
            if (m_cubemapViews)
            {
                for (u32 i = 0; i < 6; ++i)
                    vkDestroyImageView(device, m_cubemapViews[i], callbacks);
                memDeleteArray(m_cubemapViews);
            }

            if (m_memory)
            {
                for (u32 i = 0; i < m_count; ++i)
                {
                    vkFreeMemory(device, m_memory[i], callbacks);
                    vkDestroyImageView(device, m_view[i], callbacks);
                    vkDestroyImage(device, m_image[i], callbacks);
                }
                memDeleteArray(m_image);
                memDeleteArray(m_view);
                memDeleteArray(m_memory);
            }
        }

        VkImage         frontImage() { return m_buffered ? m_image[s_fb] : m_image[0]; }
        VkImage         backImage() { return m_buffered ? m_image[s_bb] : m_image[0]; }
        VkImageView     frontView() { return m_buffered ? m_view[s_fb] : m_view[0]; }
        VkImageView     backView() { return m_buffered ? m_view[s_bb] : m_view[0]; }
        VkDeviceMemory  frontMemory() { return m_buffered ? m_memory[s_fb] : m_memory[0]; }
        VkDeviceMemory  backMemory() { return m_buffered ? m_memory[s_bb] : m_memory[0]; }
        VkImageLayout   frontLayout() { return m_buffered ? m_currentLayout[s_fb] : m_currentLayout[0]; }
        VkImageLayout   backLayout() { return m_buffered ? m_currentLayout[s_bb] : m_currentLayout[0]; }
        void            setFrontLayout(VkImageLayout layout) { if (m_buffered) m_currentLayout[s_fb] = layout; else m_currentLayout[0] = layout; }
        void            setBackLayout(VkImageLayout layout) { if (m_buffered) m_currentLayout[s_bb] = layout; else m_currentLayout[0] = layout; }
    };

    struct VulkanTexture : public VulkanDescriptor
    {
        TextureDesc             m_desc = {};
        VulkanTextureInternal*  m_texture = nullptr;
        VulkanTextureInternal* m_resolveTexture = nullptr;

        VulkanTextureInternal* getBindingTexture()
        {
            if (m_texture->m_resolved)
                return m_resolveTexture;
            return m_texture;
        }
    };

    struct VulkanBuffer : public VulkanDescriptor
    {
        VkBuffer* m_buffer = nullptr;
        VkDeviceMemory* m_memory = nullptr;
        size_t* m_size = nullptr;
        VkBufferUsageFlags m_usage = 0;
        VkMemoryPropertyFlags m_access = 0;
        Mem m_data;
        bool m_buffered = false;
        bool m_active = false;

        VkBuffer frontBuffer() { return m_buffered ? m_buffer[s_bb] : m_buffer[0]; }
        VkBuffer backBuffer() { return m_buffered ? m_buffer[s_bb] : m_buffer[0]; }
        VkDeviceMemory frontMemory() { return m_buffered ? m_memory[s_bb] : m_memory[0]; }
        VkDeviceMemory backMemory() { return m_buffered ? m_memory[s_bb] : m_memory[0]; }        
        size_t frontSize() { return m_buffered ? m_size[s_bb] : m_size[0]; }
        size_t backSize() { return m_buffered ? m_size[s_bb] : m_size[0]; }
    };

    struct VulkanSampler : public VulkanDescriptor
    {
        VkSampler m_sampler = VK_NULL_HANDLE;
    };

    struct VulkanShader
    {
        VkShaderModule                  m_module = VK_NULL_HANDLE;
        u32                             m_numDescriptors = 0;
        VkDescriptorSetLayoutBinding*   m_descriptors = nullptr;
        VkDescriptorBindingFlags*       m_descriptorBindingFlags = nullptr;
        u32*                            m_descriptorMappings = nullptr;
        u32                             m_numPushConstants = 0;
        VkPushConstantRange*            m_pushConstants = nullptr;
        std::string                     m_entry = "";
        VkShaderStageFlagBits           m_stage = VK_SHADER_STAGE_ALL;
    };

    struct VulkanFramebuffer
    {
        std::vector<VulkanTextureInternal*> m_textures;
        std::vector<VkImageView>            m_views;
        VkFramebuffer                       m_framebuffer = VK_NULL_HANDLE;
        VkExtent2D                          m_extent = {};
        u32                                 m_layers = 1;
        VkSemaphore                         m_semaphore = VK_NULL_HANDLE;
    };

    struct VulkanRenderPass
    {
        bool            m_resolves = false;
        u32             m_samples = 1;
        VkRenderPass    m_renderPass;
    };

    struct VulkanDescriptorSet
    {
        VkDevice                        m_device = VK_NULL_HANDLE;
        VkDescriptorSetAllocateInfo     m_info = {};
        VkDescriptorSetLayout           m_layout = VK_NULL_HANDLE;
        std::vector<VkDescriptorSet>    m_sets;
        int                             m_write = 0;
        int                             m_read = -1;
        bool                            m_dirty = true;

        void destroy()
        {
            if (m_sets.empty())
                return;
            VkResult result;
            CHECK_RESULT((result = vkFreeDescriptorSets(m_device, m_info.descriptorPool, (u32)m_sets.size(), m_sets.data())) == VK_SUCCESS);
            m_sets.clear();
        }

        void reset()
        {
            m_write = 0;
            m_read = -1;
            m_dirty = true;
        }

        VkDescriptorSet get(bool updating)
        {
            // descriptor set has been used, need a new set
            if (m_dirty && (updating || m_read < 0))
            {
                if (m_write >= m_sets.size())
                {
                    // create new descriptor sets
                    size_t prevSize = m_sets.size();
                    size_t newSize = (prevSize > 0) ? prevSize + 1: 1;
                    m_sets.resize(newSize);
                    VkResult result;
                    m_info.descriptorSetCount = (u32)(newSize - prevSize);
                    std::vector<VkDescriptorSetLayout> layouts;
                    for (size_t i = 0; i < m_info.descriptorSetCount; ++i)
                    {
                        layouts.push_back(m_layout);
                    }
                    m_info.pSetLayouts = layouts.data();
                    CHECK_RESULT((result = vkAllocateDescriptorSets(m_device, &m_info, &m_sets[prevSize])) == VK_SUCCESS);
                }
                ++m_write;
                ++m_read;
                m_dirty = false;
            }
            return m_sets[m_read];
        }
    };

    struct VulkanPipeline
    {
        VkDescriptorPool        m_descriptorPool = VK_NULL_HANDLE;
        VkDescriptorSetLayout   m_descriptorSetLayout = VK_NULL_HANDLE;
        VulkanDescriptorSet*    m_descriptorSet = nullptr;
        u32                     m_descriptorImageMappings[MAX_DESCRIPTOR_WRITES];
        u32                     m_descriptorBufferMappings[MAX_DESCRIPTOR_WRITES];
        VkPipelineLayout        m_pipelineLayout = VK_NULL_HANDLE;
        VkPipeline              m_pipeline = VK_NULL_HANDLE;
    };

    struct VulkanRenderPipeline : public VulkanPipeline
    {
        RenderPipelineDesc  m_desc;
    };

    struct VulkanComputePipeline : public VulkanPipeline
    {

    };

    struct VulkanSync
    {
        VkSemaphore     m_semaphore = VK_NULL_HANDLE;
        VkFence         m_fence = VK_NULL_HANDLE;
        bool            m_used = false;
        bool            m_consumed = false;
    };
    
    struct VulkanCommandBuffer
    {
        VulkanSync*     m_sync = nullptr;;
        VkCommandBuffer m_commandBuffer = VK_NULL_HANDLE;
        bool            m_used = false;
    };

    struct VulkanCommandPool
    {
        size_t                  m_numCommandBuffers = 0;
        VulkanCommandBuffer*    m_commandBuffers = nullptr;
        VkCommandPool           m_commandPool = VK_NULL_HANDLE;

        void grow(VkDevice device, VkAllocationCallbacks* callbacks)
        {
            VkResult result = VK_SUCCESS;

            VulkanCommandBuffer* prev = m_commandBuffers;
            size_t prevNum = m_numCommandBuffers;

            m_numCommandBuffers = prevNum + INITIAL_NUM_COMMAND_BUFFERS;
            m_commandBuffers = memNewArray<VulkanCommandBuffer>(m_numCommandBuffers);
            memcpy(m_commandBuffers, prev, sizeof(VulkanCommandBuffer) * prevNum);

            VkCommandBufferAllocateInfo alloc = {};
            alloc.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
            alloc.commandBufferCount = INITIAL_NUM_COMMAND_BUFFERS;
            alloc.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
            alloc.commandPool = m_commandPool;

            VkSemaphoreCreateInfo semaphore = {};
            semaphore.sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO;

            VkFenceCreateInfo fence = {};
            fence.sType = VK_STRUCTURE_TYPE_FENCE_CREATE_INFO;

            VkCommandBuffer buffers[INITIAL_NUM_COMMAND_BUFFERS];
            alloc.commandPool = m_commandPool;
            CHECK_RESULT((result = vkAllocateCommandBuffers(device, &alloc, buffers)) == VK_SUCCESS);
            size_t i = 0;
            for (size_t k = prevNum; k < m_numCommandBuffers; ++k)
            {
                m_commandBuffers[k].m_commandBuffer = buffers[i++];
                m_commandBuffers[k].m_used = false;
                m_commandBuffers[k].m_sync = memNew<VulkanSync>();
                m_commandBuffers[k].m_sync->m_consumed = false;
                CHECK_RESULT((result = vkCreateSemaphore(device, &semaphore, callbacks, &m_commandBuffers[k].m_sync->m_semaphore)) == VK_SUCCESS);
                CHECK_RESULT((result = vkCreateFence(device, &fence, callbacks, &m_commandBuffers[k].m_sync->m_fence)) == VK_SUCCESS);
            }
        }
    };

    struct RenderPassState
    {
        VkRenderPass                m_pass = VK_NULL_HANDLE;
        VkFramebuffer               m_framebuffer = VK_NULL_HANDLE;
        VkExtent2D                  m_extent = {};
        std::vector<VkClearValue>   m_clears;
    };

    struct VulkanObjectDestroy
    {
        u32 m_age = 0;

        virtual void destroy(VkDevice device, VkAllocationCallbacks* callbacks) = 0;
    };

    struct VulkanTextureDestroy : public VulkanObjectDestroy
    {
        VulkanTexture* m_texture = nullptr;

        void destroy(VkDevice device, VkAllocationCallbacks* callbacks) override
        {
            m_texture->m_texture->destroy(device, callbacks);
            memDelete(m_texture->m_texture);
            if (m_texture->m_resolveTexture)
            {
                m_texture->m_resolveTexture->destroy(device, callbacks);
                memDelete(m_texture->m_resolveTexture);
            }
            memDelete(m_texture);
        }
    };

    struct VulkanBufferDestroy : public VulkanObjectDestroy
    {
        VulkanBuffer* m_buffer = nullptr;

        void destroy(VkDevice device, VkAllocationCallbacks* callbacks) override
        {
            int bufferCount = m_buffer->m_buffered ? s_numBufferedResources : 1;
            for (int i = 0; i < bufferCount; ++i)
            {
                vkFreeMemory(device, m_buffer->m_memory[i], callbacks);
                vkDestroyBuffer(device, m_buffer->m_buffer[i], callbacks);
            }
            delete[] m_buffer->m_buffer;
            delete[] m_buffer->m_memory;
            delete[] m_buffer->m_size;
            memDelete(m_buffer);
        }
    };

    struct VulkanRenderPassDestroy : public VulkanObjectDestroy
    {
        VulkanRenderPass* m_pass = nullptr;

        void destroy(VkDevice device, VkAllocationCallbacks* callbacks) override
        {
            vkDestroyRenderPass(device, m_pass->m_renderPass, callbacks);
            memDelete(m_pass);
        }
    };

    struct VulkanRenderPipelineDestroy : public VulkanObjectDestroy
    {
        VulkanRenderPipeline*   m_pipeline = nullptr;
        size_t                  m_numSwapChainImages = 0;

        void destroy(VkDevice device, VkAllocationCallbacks* callbacks) override
        {
            for (size_t i = 0; i < m_numSwapChainImages; ++i)
            {
                m_pipeline->m_descriptorSet[i].destroy();
            }
            delete[] m_pipeline->m_descriptorSet;
            vkDestroyDescriptorSetLayout(device, m_pipeline->m_descriptorSetLayout, callbacks);
            vkDestroyPipelineLayout(device, m_pipeline->m_pipelineLayout, callbacks);
            vkDestroyPipeline(device, m_pipeline->m_pipeline, callbacks);
            memDelete(m_pipeline);
        }
    };

    struct VulkanComputePipelineDestroy : public VulkanObjectDestroy
    {
        VulkanComputePipeline*  m_pipeline = nullptr;
        size_t                  m_numSwapChainImages = 0;

        void destroy(VkDevice device, VkAllocationCallbacks* callbacks) override
        {
            for (size_t i = 0; i < m_numSwapChainImages; ++i)
            {
                m_pipeline->m_descriptorSet[i].destroy();
            }
            delete[] m_pipeline->m_descriptorSet;
            vkDestroyDescriptorSetLayout(device, m_pipeline->m_descriptorSetLayout, callbacks);
            vkDestroyPipelineLayout(device, m_pipeline->m_pipelineLayout, callbacks);
            vkDestroyPipeline(device, m_pipeline->m_pipeline, callbacks);
            memDelete(m_pipeline);
        }
    };

    struct VulkanImageTransition
    {
        VulkanTextureInternal*  m_texture = nullptr;
        VkImageAspectFlags      m_aspect = VK_IMAGE_ASPECT_COLOR_BIT;
        VkImageLayout           m_target = VK_IMAGE_LAYOUT_GENERAL;
        u32                     m_mips = 1;
    };

    class VulkanContext : public Context
    {
    protected:

    public:
        void setup(void* platform, FeatureFlags featureFlags);
        void createSwapChain(u32 width, u32 height);
        void createCommandPools(u32 count);
        void transitionSwapChainImages();
        void setupAsync(VulkanContext* main, const std::thread::id& id);
        void shutdown();
        void shutdownAsync();

        Shader                  createShader(const ShaderDesc& desc) override;
        Buffer                  createBuffer(const BufferDesc& desc) override;
        Texture                 createTexture(const TextureDesc& desc) override;
        VulkanTextureInternal*  createTextureInternal(const TextureDesc& desc);
        Sampler                 createSampler(const SamplerDesc& desc) override;
        RenderPass              createRenderPass(const RenderPassDesc& desc) override;
        RenderPipeline          createRenderPipeline(const RenderPipelineDesc& desc) override;
        ComputePipeline         createComputePipeline(const ComputePipelineDesc& desc) override;
        Framebuffer             obtainFramebuffer(const FramebufferDesc& desc) override;

        void destroyShader(Shader shader) override;
        void destroyBuffer(Buffer buffer) override;
        void destroyTexture(Texture texture) override;
        void destroySampler(Sampler sampler) override;
        void destroyRenderPass(RenderPass pass) override;
        void destroyRenderPipeline(RenderPipeline pipeline) override;
        void destroyComputePipeline(ComputePipeline pipeline) override;
        void eraseDescriptorSet(VulkanDescriptorSet* set);

        void                        begin() override;
        void                        setViewport(const Viewport& vp) override;
        void                        setScissorRect(const Rect& sr) override;
        void                        beginComputeEncoding() override;
        void                        beginRenderPass(RenderPass pass, Framebuffer framebuffer, Clear* clears = nullptr, size_t numClears = 0) override;
        void                        beginRenderPassInternal();
        void                        beginNextSubpass() override;
        void                        bindRenderPipeline(RenderPipeline pipeline) override;
        void                        bindComputePipeline(ComputePipeline pipeline) override;
        void                        bindBuffer(Buffer buf, u32 slot, BindFlags flags, size_t offset = 0) override;
        void                        bindBufferInternal(Buffer buf, u32 slot, BindFlags flags, size_t offset, IndexType indexType);
        void                        bindTexture(Texture tex, Sampler samp, u32 slot, BindFlags flags) override;
        void                        updateBuffer(Buffer buf, const Mem& data, size_t offset = 0) override;
        void                        updateTexture(Texture tex, const TextureUpdateDesc& data) override;
        void                        updateConstants(ShaderStageFlags stage, size_t size, size_t offset, void* data) override;
        void                        insertBarrier(PipelineStage srcStage, PipelineStage dstStage, BarrierFlags barrierType) override;
        void                        copyTexture(Texture src, Texture dst, u32 srcLayer = 0, u32 dstLayer = 0) override;
        void                        generateMips(Texture tex) override;

        void                        draw(u32 vertexCount, u32 startVertex = 0) override;
        void                        drawIndexed(u32 indexCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset = 0, u32 vertexOffset = 0) override;
        void                        drawIndexedInstanced(u32 indexCount, u32 instanceCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset = 0, u32 vertexOffset = 0) override;
        void                        drawIndirect(u32 drawCount, Buffer drawBuffer, u32 drawBufferOffset = 0) override;
        void                        drawIndexedIndirect(u32 drawCount, Buffer indexBuffer, IndexType indexType, Buffer drawBuffer, u32 indexBufferOffset = 0, u32 drawBufferOffset = 0) override;
        void                        dispatchCompute(u32 grid[3], u32 numThreads[3]) override;
        void                        dispatchComputeIndirect(Buffer drawBuffer, u32 offset = 0) override;
        void                        bindDescriptorSet();
        void                        clearDescriptorWrites();

        void                        endRenderEncoding() override;
        void                        endComputeEncoding() override;
        void                        enqueue() override;
        Semaphore                   commit(Semaphore* wait = nullptr, size_t count = 0, PipelineStage waitStage = e_pipelineStageType::top) override;
        void                        present() override;
        void                        flushCommandBuffers() override;
        void                        doEndRecordingImageTransitions();

        bool                        isSignalled(Semaphore semaphore) override;
        TextureDesc                 getTextureDesc(Texture tex) override;
        TextureDesc                 getBackbufferTextureDesc() override;

        VkFormat                    findSupportedFormat(const VkFormat* candidates, int numCandidates, VkImageTiling tiling, VkFormatFeatureFlags features);
        u32                         findMemoryType(u32 type, VkMemoryPropertyFlags props, VkDeviceSize size);

        VkCommandBuffer             beginSingleTimeCommands(VkCommandPool commandPool);
        void                        endSingleTimeCommands(VkQueue queue, VkCommandPool commandPool, VkCommandBuffer commandBuffer);
        void                        transitionImageLayout(QueueAlias queue, VkCommandBuffer command, VkImage image, VkImageAspectFlags aspect, u32 layers, u32 levels, VkImageLayout src, VkImageLayout dst);
        VulkanCommandBuffer*        obtainCommandBuffer(QueueAlias usage);

        VkInstance                  m_instance = VK_NULL_HANDLE;
        VkDebugReportCallbackEXT    m_debugCallback = VK_NULL_HANDLE;
        VkSurfaceKHR                m_surface = VK_NULL_HANDLE;
        VkPhysicalDevice            m_gpu = VK_NULL_HANDLE;
        VkDevice                    m_device = VK_NULL_HANDLE;
        VkSwapchainKHR              m_swapChain = VK_NULL_HANDLE;
        VkExtent2D                  m_swapChainExtent = {};
        u32                         m_numSwapChainImages = 0;
        u32                         m_currentSwapChainImage = 0;
        VulkanTexture               m_swapChainTextures;
        VulkanQueueFamily           m_queueFamilies[e_queueAlias::queueCount];     // graphics, transfer, compute, present
        VkQueue                     m_queues[e_queueAlias::queueCount];            // graphics, transfer, compute, present
        std::thread::id             m_threadID = std::thread::id();
        VulkanCommandPool*          m_commandPools[2] = { nullptr, nullptr };
        VkCommandPool               m_transferCommandPool = VK_NULL_HANDLE;
        VkDescriptorPool            m_descriptorPool = VK_NULL_HANDLE;             // no clue how many of these there's supposed to be
        VkSemaphore*                m_swapChainImageSemaphores = nullptr;   
        std::vector<VulkanFramebuffer*>* m_framebuffers;                            // share list of framebuffers between all contexts
        std::vector<VulkanDescriptorSet*> m_descriptorSets;

        VulkanCommandBuffer*        m_boundCommandBuffer = VK_NULL_HANDLE;
        VulkanSync*                 m_queueSemaphore = nullptr;
        VulkanPipeline*             m_boundPipeline = nullptr;
        QueueAlias                  m_boundQueueType = e_queueAlias::none;
        VkCommandBuffer             m_queuedCommandBuffers[MAX_QUEUED_COMMAND_BUFFERS];
        VkSemaphore                 m_boundFramebufferSemaphore = VK_NULL_HANDLE;
        IndexType                   m_boundIndexType = e_indexType::uint16;
        VkWriteDescriptorSet        m_descriptorWrites[MAX_DESCRIPTOR_WRITES];
        u32                         m_numDescriptorWrites = 0;
        u32                         m_numChangedDescriptorWrites = 0;

        RenderPassState             m_renderPassState;
        bool                        m_renderPassStarted = false;
        bool                        m_backBufferUsed = false;

        VkAllocationCallbacks*      m_allocationCallbacks = nullptr;

        std::vector<VulkanObjectDestroy*>   m_destroys;
        std::vector<VulkanImageTransition>  m_imageTransitions;
    };

    void VulkanContext::setup(void* platform, FeatureFlags featureFlags)
    {
        VkResult result;

        // init extensions
        const char* requiredExtensions[] = {
            VK_EXT_DEBUG_REPORT_EXTENSION_NAME,
            VK_KHR_SURFACE_EXTENSION_NAME,
            VK_KHR_WIN32_SURFACE_EXTENSION_NAME
        };
        u32 numRequiredExts = sizeof(requiredExtensions) / sizeof(requiredExtensions[0]);
        u32 numExts = 0;
        vkEnumerateInstanceExtensionProperties(nullptr, &numExts, nullptr);
        VkExtensionProperties* extensions = new VkExtensionProperties[numExts];
        vkEnumerateInstanceExtensionProperties(nullptr, &numExts, extensions);
        bool found = false;
        for (u32 i = 0; i < numRequiredExts; ++i)
        {
            found = false;
            for (u32 k = 0; k < numExts; ++k)
            {
                if (std::strcmp(requiredExtensions[i], extensions[k].extensionName) == 0)
                {
                    found = true;
                    break;
                }
            }
            ASSERT(found); // required extension not found
        }
        delete[] extensions;

        // init validation layers
#ifdef DEBUG
        const char* requiredLayers[] = {
            "VK_LAYER_KHRONOS_validation"
        };
        u32 numRequiredLayers = sizeof(requiredLayers) / sizeof(requiredLayers[0]);
#else
        const char** requiredLayers = nullptr;
        u32 numRequiredLayers = 0;
#endif
        u32 numLayers = 0;
        vkEnumerateInstanceLayerProperties(&numLayers, nullptr);
        VkLayerProperties* layers = new VkLayerProperties[numLayers];
        vkEnumerateInstanceLayerProperties(&numLayers, layers);
        for (u32 i = 0; i < numRequiredLayers; ++i)
        {
            found = false;
            for (u32 k = 0; k < numLayers; ++k)
            {
                if (strcmp(requiredLayers[i], layers[k].layerName) == 0)
                {
                    found = true;
                    break;
                }
            }
            ASSERT(found); // required validation layer not found
        }
        delete[] layers;

        // init instance
        VkApplicationInfo appInfo = {};
        appInfo.sType = VK_STRUCTURE_TYPE_APPLICATION_INFO;
        appInfo.pApplicationName = "";
        appInfo.applicationVersion = VK_MAKE_VERSION(0, 0, 1);
        appInfo.pEngineName = "";
        appInfo.engineVersion = VK_MAKE_VERSION(0, 0, 1);
        appInfo.apiVersion = VK_API_VERSION_1_2;

        VkInstanceCreateInfo instanceInfo = {};
        instanceInfo.sType = VK_STRUCTURE_TYPE_INSTANCE_CREATE_INFO;
        instanceInfo.pApplicationInfo = &appInfo;
        instanceInfo.ppEnabledExtensionNames = requiredExtensions;
        instanceInfo.enabledExtensionCount = numRequiredExts;
        instanceInfo.ppEnabledLayerNames = requiredLayers;
        instanceInfo.enabledLayerCount = numRequiredLayers;
        CHECK_RESULT((result = vkCreateInstance(&instanceInfo, m_allocationCallbacks, &m_instance)) == VK_SUCCESS);

        // setup callback
        VkDebugReportCallbackCreateInfoEXT callbackInfo = {};
        callbackInfo.sType = VK_STRUCTURE_TYPE_DEBUG_REPORT_CALLBACK_CREATE_INFO_EXT;
        callbackInfo.flags = VK_DEBUG_REPORT_ERROR_BIT_EXT | VK_DEBUG_REPORT_WARNING_BIT_EXT;
        callbackInfo.pfnCallback = debugCallback;
        auto func = (PFN_vkCreateDebugReportCallbackEXT)vkGetInstanceProcAddr(m_instance, "vkCreateDebugReportCallbackEXT");
        ASSERT(func != nullptr); // failed to create debug callback
        func(m_instance, &callbackInfo, nullptr, &m_debugCallback);
        
        // create surface
        VkWin32SurfaceCreateInfoKHR surfaceInfo = {};
        surfaceInfo.sType = VK_STRUCTURE_TYPE_WIN32_SURFACE_CREATE_INFO_KHR;
        surfaceInfo.hwnd = ((PlatformGFX*)platform)->m_hwnd;
        surfaceInfo.hinstance = ((PlatformGFX*)platform)->m_hinstance;
        CHECK_RESULT((result = vkCreateWin32SurfaceKHR(m_instance, &surfaceInfo, m_allocationCallbacks, &m_surface)) == VK_SUCCESS);

        // init physical device
        VkPhysicalDeviceFeatures requiredFeatures = {};
        requiredFeatures.samplerAnisotropy = VK_TRUE;
        requiredFeatures.independentBlend = VK_TRUE;
        requiredFeatures.shaderStorageImageExtendedFormats = VK_TRUE;
        requiredFeatures.geometryShader = VK_TRUE;
        requiredFeatures.shaderTessellationAndGeometryPointSize = VK_TRUE;
        requiredFeatures.sampleRateShading = VK_TRUE;
        requiredFeatures.shaderFloat64 = (featureFlags & e_featureFlags::shaderFloat64);

        const char* requiredDeviceExtensions[] = {
            VK_KHR_SWAPCHAIN_EXTENSION_NAME
        };
        u32 numRequiredDeviceExtensions = sizeof(requiredDeviceExtensions) / sizeof(requiredDeviceExtensions[0]);

        u32 numDevices = 0;
        vkEnumeratePhysicalDevices(m_instance, &numDevices, nullptr);
        ASSERT(numDevices > 0);
        VkPhysicalDevice* devices = new VkPhysicalDevice[numDevices];
        vkEnumeratePhysicalDevices(m_instance, &numDevices, devices);
        for (u32 i = 0; i < numDevices; ++i)
        {
            VkPhysicalDevice device = devices[i];

            for (int j = 0; j < e_queueAlias::queueCount; ++j)
                m_queueFamilies[j].m_family = (u32)-1;

            VkPhysicalDeviceProperties props = {};
            vkGetPhysicalDeviceProperties(device, &props);

            // check queue support
            u32 numFamilies = 0;
            vkGetPhysicalDeviceQueueFamilyProperties(device, &numFamilies, nullptr);
            VkQueueFamilyProperties* qFamilyProps = new VkQueueFamilyProperties[numFamilies];
            vkGetPhysicalDeviceQueueFamilyProperties(device, &numFamilies, qFamilyProps);
            int index = 0;
            bool queuesComplete;
            for (u32 j = 0; j < numFamilies; ++j)
            {
                VkQueueFamilyProperties qProps = qFamilyProps[j];
                if (qProps.queueCount > 0)
                {
                    if (qProps.queueFlags & VK_QUEUE_GRAPHICS_BIT && m_queueFamilies[QueueAlias::graphics].m_family == (u32)-1)
                    {
                        m_queueFamilies[QueueAlias::graphics].m_family = j;
                        m_queueFamilies[QueueAlias::graphics].m_count = qProps.queueCount;
                    }
                    if ((qProps.queueFlags & VK_QUEUE_TRANSFER_BIT) && (m_queueFamilies[QueueAlias::transfer].m_family == (u32)-1 || m_queueFamilies[QueueAlias::transfer].m_suboptimal))
                    {
                        // allow fallback to non-optimal transfer queue if no other suitable ones exist
                        m_queueFamilies[QueueAlias::transfer].m_family = j;
                        m_queueFamilies[QueueAlias::transfer].m_count = qProps.queueCount;
                        m_queueFamilies[QueueAlias::transfer].m_suboptimal = (qProps.queueFlags & VK_QUEUE_GRAPHICS_BIT);
                    }
                    if (qProps.queueFlags & VK_QUEUE_COMPUTE_BIT && m_queueFamilies[QueueAlias::compute].m_family == (u32)-1)
                    {
                        m_queueFamilies[QueueAlias::compute].m_family = j;
                        m_queueFamilies[QueueAlias::compute].m_count = qProps.queueCount;
                    }
                }
                if (m_queueFamilies[QueueAlias::present].m_family == (u32)-1)
                {
                    VkBool32 present;
                    vkGetPhysicalDeviceSurfaceSupportKHR(device, index, m_surface, &present);
                    if (present)
                        m_queueFamilies[QueueAlias::present].m_family = j;
                }
                queuesComplete = true;
                for (int k = 0; k < e_queueAlias::queueCount; ++k)
                {
                    if (m_queueFamilies[k].m_family == (u32)-1)
                    {
                        queuesComplete = false;
                    }
                    else if (m_queueFamilies[k].m_suboptimal && ((j < numFamilies - 1) || (i < numDevices - 1)))
                    {
                        queuesComplete = false;
                    }
                }
                if (queuesComplete)
                    break;
                ++index;
            }
            delete[] qFamilyProps;
            if (!queuesComplete)
                continue;

            // check extension support
            u32 numDeviceExtensions = 0;
            vkEnumerateDeviceExtensionProperties(device, nullptr, &numDeviceExtensions, nullptr);
            VkExtensionProperties* deviceExtensions = new VkExtensionProperties[numDeviceExtensions];
            vkEnumerateDeviceExtensionProperties(device, nullptr, &numDeviceExtensions, deviceExtensions);
            bool foundAll = true;
            for (u32 j = 0; j < numRequiredDeviceExtensions; ++j)
            {
                bool found = false;
                for (u32 k = 0; k < numDeviceExtensions; ++k)
                {
                    if (strcmp(deviceExtensions[k].extensionName, requiredDeviceExtensions[j]) == 0)
                    {
                        found = true;
                        break;
                    }
                }
                if (!found)
                {
                    foundAll = false;
                    break;
                }
            }
            delete[] deviceExtensions;
            if (!foundAll)
                continue;

            // check device supports required features
            VkPhysicalDeviceFeatures features;
            vkGetPhysicalDeviceFeatures(device, &features);
            bool featuresSupported = true;
            for (int j = 0; j < sizeof(VkPhysicalDeviceFeatures) / sizeof(VkBool32); ++j)
            {
                // A pointer VkPhysicalDeviceFeatures is equivalent to a pointer to an array of VkBool32
                VkBool32 featureRequired = ((VkBool32*)&requiredFeatures)[j];
                if (featureRequired && !((VkBool32*)&features)[j])
                {
                    featuresSupported = false;
                    break;
                }
            }

            // check descriptor indexing features
            VkPhysicalDeviceVulkan12Features vk12Features = {};
            vk12Features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES;

            VkPhysicalDeviceFeatures2 features2 = {};
            features2.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_FEATURES_2;
            features2.pNext = &vk12Features;
            vkGetPhysicalDeviceFeatures2(device, &features2);
            
            if (!vk12Features.descriptorBindingPartiallyBound || !vk12Features.runtimeDescriptorArray)
            {
                //featuresSupported = false;
            }
            
            if (!featuresSupported)
                continue;

            // check device supports required surface features
            u32 formatCount = 0;
            vkGetPhysicalDeviceSurfaceFormatsKHR(device, m_surface, &formatCount, nullptr);
            if (!formatCount)
                continue;

            u32 presentCount = 0;
            vkGetPhysicalDeviceSurfacePresentModesKHR(device, m_surface, &presentCount, nullptr);
            if (!presentCount)
                continue;

            m_gpu = device;
            break;
        }
        ASSERT(m_gpu != VK_NULL_HANDLE);
        delete[] devices;

        // setup queues
        std::vector<VkDeviceQueueCreateInfo> queueInfos;
        float queuePriorities[16];
        for (int i = 0; i < 16; ++i)
        {
            if (i == 0)
                queuePriorities[i] = 1.0f;
            else
                queuePriorities[i] = 0.9f;
        }

        for (int i = 0; i < e_queueAlias::queueCount; ++i)
        {
            bool found = false;
            for (auto& info : queueInfos)
            {
                if (info.queueFamilyIndex == m_queueFamilies[i].m_family)
                {
                    found = true;
                    break;
                }
            }
            if (found)
                continue;
            VkDeviceQueueCreateInfo info = {};
            info.sType = VK_STRUCTURE_TYPE_DEVICE_QUEUE_CREATE_INFO;
            info.pNext = nullptr;
            info.queueFamilyIndex = m_queueFamilies[i].m_family;
            info.queueCount = min(16, m_queueFamilies[i].m_count);
            info.pQueuePriorities = queuePriorities;
            queueInfos.push_back(info);
        }

        // set the dynamic descriptor indexing creation info
        VkPhysicalDeviceVulkan12Features vk12Features = {};
        vk12Features.sType = VK_STRUCTURE_TYPE_PHYSICAL_DEVICE_VULKAN_1_2_FEATURES;
        vk12Features.descriptorBindingPartiallyBound = VK_TRUE;
        vk12Features.runtimeDescriptorArray = VK_TRUE;

        // create logical device
        VkDeviceCreateInfo deviceCreateInfo = {};
        deviceCreateInfo.sType = VK_STRUCTURE_TYPE_DEVICE_CREATE_INFO;
        deviceCreateInfo.pNext = &vk12Features;
        deviceCreateInfo.pQueueCreateInfos = queueInfos.data();
        deviceCreateInfo.queueCreateInfoCount = (u32)queueInfos.size();
        deviceCreateInfo.pEnabledFeatures = &requiredFeatures;
        deviceCreateInfo.ppEnabledExtensionNames = requiredDeviceExtensions;
        deviceCreateInfo.enabledExtensionCount = numRequiredDeviceExtensions;
        deviceCreateInfo.ppEnabledLayerNames = requiredLayers;
        deviceCreateInfo.enabledLayerCount = numRequiredLayers;
        CHECK_RESULT((result = vkCreateDevice(m_gpu, &deviceCreateInfo, m_allocationCallbacks, &m_device)) == VK_SUCCESS);

        // create descriptor pool for main context
        VkDescriptorPoolSize poolSizes[] = {
            { VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER, 4096 },
            { VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE, 4096 },
            { VK_DESCRIPTOR_TYPE_STORAGE_IMAGE, 4096 },
            { VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER, 4096 },
            { VK_DESCRIPTOR_TYPE_STORAGE_BUFFER, 4096 },
            { VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER_DYNAMIC, 4096 },
            { VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC, 4096 },
            { VK_DESCRIPTOR_TYPE_INPUT_ATTACHMENT, 4096},
        };
        VkDescriptorPoolCreateInfo descriptorPoolInfo = {};
        descriptorPoolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
        descriptorPoolInfo.flags = VK_DESCRIPTOR_POOL_CREATE_FREE_DESCRIPTOR_SET_BIT;
        descriptorPoolInfo.maxSets = 8096;
        descriptorPoolInfo.poolSizeCount = sizeof(poolSizes) / sizeof(poolSizes[0]);
        descriptorPoolInfo.pPoolSizes = poolSizes;
        CHECK_RESULT((result = vkCreateDescriptorPool(m_device, &descriptorPoolInfo, m_allocationCallbacks, &m_descriptorPool)) == VK_SUCCESS);

        // get queues for main context
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::graphics].m_family, 0, &m_queues[QueueAlias::graphics]);
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::transfer].m_family, 0, &m_queues[QueueAlias::transfer]);
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::compute].m_family, 0, &m_queues[QueueAlias::compute]);
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::present].m_family, 0, &m_queues[QueueAlias::present]);

        // create swap chain
        u32 width = utils::getWindowWidth();
        u32 height = utils::getWindowHeight();
        createSwapChain(width, height);        

        // create command pools, buffers and synchronisation objects for main context
        createCommandPools(m_numSwapChainImages);

        // create the transfer command pool
        VkCommandPoolCreateInfo poolInfo = {};
        poolInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
        poolInfo.flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
        poolInfo.queueFamilyIndex = m_queueFamilies[QueueAlias::transfer].m_family;
        CHECK_RESULT((result = vkCreateCommandPool(m_device, &poolInfo, m_allocationCallbacks, &m_transferCommandPool)) == VK_SUCCESS);

        transitionSwapChainImages();

        // create swap chain image synchronisation objects
        VkSemaphoreCreateInfo semaphore = {};
        semaphore.sType = VK_STRUCTURE_TYPE_SEMAPHORE_CREATE_INFO;

        m_swapChainImageSemaphores = new VkSemaphore[m_numSwapChainImages + 1];
        for (u32 i = 0; i < m_numSwapChainImages + 1; ++i)
            CHECK_RESULT((result = vkCreateSemaphore(m_device, &semaphore, m_allocationCallbacks, &m_swapChainImageSemaphores[i])) == VK_SUCCESS);

        m_framebuffers = new std::vector<VulkanFramebuffer*>();
        memset(m_queuedCommandBuffers, 0, sizeof(VkCommandBuffer) * MAX_QUEUED_COMMAND_BUFFERS);
    }

    void VulkanContext::createCommandPools(u32 count)
    {
        VkResult result = VK_SUCCESS;

        VkCommandPoolCreateInfo poolInfo = {};
        poolInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
        poolInfo.flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;

        QueueAlias aliases[2] = { e_queueAlias::graphics, e_queueAlias::compute };

        for (int i = 0; i < 2; ++i)
        {
            m_commandPools[i] = memNewArray<VulkanCommandPool>(count);
            poolInfo.queueFamilyIndex = m_queueFamilies[aliases[i]].m_family;
            for (u32 j = 0; j < count; ++j)
            {
                VulkanCommandPool& pool = m_commandPools[i][j];

                CHECK_RESULT((result = vkCreateCommandPool(m_device, &poolInfo, m_allocationCallbacks, &pool.m_commandPool)) == VK_SUCCESS);

                // create command buffers and synchronisation objects for this pool
                pool.grow(m_device, m_allocationCallbacks);

                
            }
        }
    }

    void VulkanContext::createSwapChain(u32 width, u32 height)
    {
        HRESULT result = {};

        bool recreating = (m_swapChain != VK_NULL_HANDLE);

        m_swapChainTextures.m_desc = {};
        m_swapChainTextures.m_desc.m_format.m_channels = e_formatChannelFlags::rgba;
        m_swapChainTextures.m_desc.m_format.m_bitDepth = 8;
        m_swapChainTextures.m_desc.m_format.m_type = e_formatType::unsignedNorm;
        m_swapChainTextures.m_desc.m_format.m_standard = true;
        m_swapChainTextures.m_desc.m_format.m_reverse = true;
        m_swapChainTextures.m_desc.m_layers = 1;

        VkSwapchainKHR oldSwapChain = m_swapChain;
        VkSurfaceCapabilitiesKHR surfaceCapabilities;
        vkGetPhysicalDeviceSurfaceCapabilitiesKHR(m_gpu, m_surface, &surfaceCapabilities);
        u32 numFormats, numPresents;
        vkGetPhysicalDeviceSurfaceFormatsKHR(m_gpu, m_surface, &numFormats, nullptr);
        VkSurfaceFormatKHR* formats = new VkSurfaceFormatKHR[numFormats];
        vkGetPhysicalDeviceSurfaceFormatsKHR(m_gpu, m_surface, &numFormats, formats);
        vkGetPhysicalDeviceSurfacePresentModesKHR(m_gpu, m_surface, &numPresents, nullptr);
        VkPresentModeKHR* presents = new VkPresentModeKHR[numPresents];
        vkGetPhysicalDeviceSurfacePresentModesKHR(m_gpu, m_surface, &numPresents, presents);
        VkSurfaceFormatKHR format = {};
        if (numFormats == 1 && formats[0].format == VK_FORMAT_UNDEFINED)
        {
            format = { VK_FORMAT_B8G8R8A8_UNORM, VK_COLOR_SPACE_SRGB_NONLINEAR_KHR };
        }
        else
        {
            for (u32 i = 0; i < numFormats; ++i)
            {
                if (formats[i].colorSpace != VK_COLORSPACE_SRGB_NONLINEAR_KHR)
                    continue;

                if (formats[i].format == VK_FORMAT_B8G8R8A8_SRGB)
                    format = formats[i];

                // prefer srgb output
                if (formats[i].format == VK_FORMAT_B8G8R8A8_UNORM && format.format != VK_FORMAT_B8G8R8A8_SRGB)
                    format = formats[i];
            }
        }
        delete[] formats;
        VkPresentModeKHR present = VK_PRESENT_MODE_IMMEDIATE_KHR;
        for (u32 i = 0; i < numPresents; ++i)
        {
            if (presents[i] == VK_PRESENT_MODE_MAILBOX_KHR)
            {
                present = presents[i];
                break;
            }
            else
            {
                present = presents[i];
            }
        }
        delete[] presents;
        m_swapChainExtent = surfaceCapabilities.currentExtent;
        m_swapChainTextures.m_desc.m_width = m_swapChainExtent.width;
        m_swapChainTextures.m_desc.m_height = m_swapChainExtent.height;
        u32 numImages = surfaceCapabilities.minImageCount + 1;
        if (surfaceCapabilities.maxImageCount > 0 && numImages > surfaceCapabilities.maxImageCount)
            numImages = surfaceCapabilities.maxImageCount;
        VkSwapchainCreateInfoKHR swapChainInfo = {};
        swapChainInfo.sType = VK_STRUCTURE_TYPE_SWAPCHAIN_CREATE_INFO_KHR;
        swapChainInfo.surface = m_surface;
        swapChainInfo.minImageCount = numImages;
        swapChainInfo.imageFormat = format.format;
        swapChainInfo.imageColorSpace = format.colorSpace;
        swapChainInfo.imageExtent = m_swapChainExtent;
        swapChainInfo.imageArrayLayers = 1;
        swapChainInfo.imageUsage = VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT;
        swapChainInfo.imageSharingMode = VK_SHARING_MODE_EXCLUSIVE;
        swapChainInfo.queueFamilyIndexCount = 0;
        swapChainInfo.pQueueFamilyIndices = nullptr;
        swapChainInfo.preTransform = surfaceCapabilities.currentTransform;
        swapChainInfo.compositeAlpha = VK_COMPOSITE_ALPHA_OPAQUE_BIT_KHR;
        swapChainInfo.presentMode = present;
        swapChainInfo.clipped = VK_TRUE;
        swapChainInfo.oldSwapchain = oldSwapChain;
        CHECK_RESULT((result = vkCreateSwapchainKHR(m_device, &swapChainInfo, m_allocationCallbacks, &m_swapChain)) == VK_SUCCESS);
        if (recreating)
        {
            vkDestroySwapchainKHR(m_device, oldSwapChain, m_allocationCallbacks);
        }
        vkGetSwapchainImagesKHR(m_device, m_swapChain, &m_numSwapChainImages, nullptr);

        if (!recreating)
        {
            m_swapChainTextures.m_texture = memNew<VulkanTextureInternal>();
            m_swapChainTextures.m_texture->m_image = new VkImage[m_numSwapChainImages];
            m_swapChainTextures.m_texture->m_view = new VkImageView[m_numSwapChainImages];
            m_swapChainTextures.m_texture->m_currentLayout = new VkImageLayout[m_numSwapChainImages];
        }

        vkGetSwapchainImagesKHR(m_device, m_swapChain, &m_numSwapChainImages, m_swapChainTextures.m_texture->m_image);
        m_swapChainTextures.m_desc.m_buffered = true;
        m_swapChainTextures.m_desc.m_format.m_channels = e_formatChannelFlags::rgba;
        s_numBufferedResources = m_numSwapChainImages;
        s_bb = s_numBufferedResources > 1 ? 1 : 0;

        if(recreating)
        {
            for (u32 i = 0; i < m_numSwapChainImages; ++i)
            {
                vkDestroyImageView(m_device, m_swapChainTextures.m_texture->m_view[i], m_allocationCallbacks);
            }
        }

        VkImageViewCreateInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
        info.viewType = VK_IMAGE_VIEW_TYPE_2D;
        info.format = format.format;
        info.subresourceRange.aspectMask = VK_IMAGE_ASPECT_COLOR_BIT;
        info.subresourceRange.baseMipLevel = 0;
        info.subresourceRange.levelCount = 1;
        info.subresourceRange.baseArrayLayer = 0;
        info.subresourceRange.layerCount = 1;
        for (u32 i = 0; i < m_numSwapChainImages; ++i)
        {
            info.image = m_swapChainTextures.m_texture->m_image[i];
            CHECK_RESULT((result = vkCreateImageView(m_device, &info, m_allocationCallbacks, &m_swapChainTextures.m_texture->m_view[i])) == VK_SUCCESS);

            // assign to layout tracked swap chain textures
            m_swapChainTextures.m_texture->m_currentLayout[i] = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
        }
        m_swapChainTextures.m_texture->m_desiredLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
        m_swapChainTextures.m_texture->m_buffered = true;
        m_swapChainTextures.m_texture->m_aspect = VK_IMAGE_ASPECT_COLOR_BIT;
        m_swapChainTextures.m_texture->m_layers = 1;
    }

    void VulkanContext::transitionSwapChainImages()
    {
        VkCommandBuffer commandBuffer = beginSingleTimeCommands(m_transferCommandPool);
        for (u32 i = 0; i < m_numSwapChainImages; ++i)
            transitionImageLayout(e_queueAlias::transfer, commandBuffer, m_swapChainTextures.m_texture->m_image[i], VK_IMAGE_ASPECT_COLOR_BIT, 1, 1, VK_IMAGE_LAYOUT_UNDEFINED, VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL);
        endSingleTimeCommands(m_queues[e_queueAlias::transfer], m_transferCommandPool, commandBuffer);
    }

    void VulkanContext::setupAsync(VulkanContext* mainContext, const std::thread::id& id)
    {
        VkResult result;

        m_threadID = id;
        m_gpu = mainContext->m_gpu;
        m_device = mainContext->m_device;
        m_framebuffers = mainContext->m_framebuffers;
        m_numSwapChainImages = 1;   // async contexts should not be messing with the swapchain
        m_currentSwapChainImage = 0;
        memcpy(m_queueFamilies, mainContext->m_queueFamilies, sizeof(m_queueFamilies[0]) * e_queueAlias::queueCount);
        
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::graphics].m_family, 1, &m_queues[QueueAlias::graphics]);
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::transfer].m_family, 1, &m_queues[QueueAlias::transfer]);
        vkGetDeviceQueue(m_device, m_queueFamilies[QueueAlias::compute].m_family, 1, &m_queues[QueueAlias::compute]);

        // create command pools for async context
        createCommandPools(1);

        // create the transfer command pool
        VkCommandPoolCreateInfo poolInfo = {};
        poolInfo.sType = VK_STRUCTURE_TYPE_COMMAND_POOL_CREATE_INFO;
        poolInfo.flags = VK_COMMAND_POOL_CREATE_RESET_COMMAND_BUFFER_BIT;
        poolInfo.queueFamilyIndex = m_queueFamilies[QueueAlias::transfer].m_family;
        CHECK_RESULT((result = vkCreateCommandPool(m_device, &poolInfo, m_allocationCallbacks, &m_transferCommandPool)) == VK_SUCCESS);

        memset(m_queuedCommandBuffers, 0, sizeof(VkCommandBuffer) * MAX_QUEUED_COMMAND_BUFFERS);
    }

    void VulkanContext::shutdown()
    {

    }

    void VulkanContext::shutdownAsync()
    {

    }
    
    VkFormat VulkanContext::findSupportedFormat(const VkFormat* candidates, int numCandidates, VkImageTiling tiling, VkFormatFeatureFlags features)
    {
        for (int i = 0; i < numCandidates; ++i)
        {
            VkFormatProperties props;
            vkGetPhysicalDeviceFormatProperties(m_gpu, candidates[i], &props);
            if ((tiling == VK_IMAGE_TILING_LINEAR && (props.linearTilingFeatures & features) == features)
                || (tiling == VK_IMAGE_TILING_OPTIMAL && (props.optimalTilingFeatures & features) == features))
                return candidates[i];
        }
        ASSERT(false);
        return VK_FORMAT_R8G8B8A8_UNORM;
    }

    u32 VulkanContext::findMemoryType(u32 type, VkMemoryPropertyFlags props, VkDeviceSize size)
    {
        VkPhysicalDeviceMemoryProperties memProps;
        vkGetPhysicalDeviceMemoryProperties(m_gpu, &memProps);
        for (u32 i = 0; i < memProps.memoryTypeCount; ++i)
            if ((type & (1 << i)) && ((memProps.memoryTypes[i].propertyFlags & props) == props) && memProps.memoryHeaps[memProps.memoryTypes[i].heapIndex].size >= size)
                return i;
        ASSERT(false);
        return 0;
    }

    void VulkanContext::transitionImageLayout(QueueAlias queue, VkCommandBuffer command, VkImage image, VkImageAspectFlags aspect, u32 layers, u32 levels, VkImageLayout src, VkImageLayout dst)
    {
        if (src == dst)
            return;

        VkImageMemoryBarrier barrier = {};
        barrier.sType = VK_STRUCTURE_TYPE_IMAGE_MEMORY_BARRIER;
        barrier.oldLayout = src;
        barrier.newLayout = dst;
        barrier.srcQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
        barrier.dstQueueFamilyIndex = VK_QUEUE_FAMILY_IGNORED;
        barrier.image = image;
        barrier.subresourceRange.aspectMask = aspect;
        barrier.subresourceRange.baseMipLevel = 0;
        barrier.subresourceRange.levelCount = levels;
        barrier.subresourceRange.baseArrayLayer = 0;
        barrier.subresourceRange.layerCount = layers;
        
        VkPipelineStageFlags stage[] = { 0, 0 };
        VkAccessFlags* flag[] = { &barrier.srcAccessMask, &barrier.dstAccessMask };
        VkImageLayout layout[] = { src, dst };

        for (int i = 0; i < 2; ++i)
        {
            switch (layout[i])
            {
            case VK_IMAGE_LAYOUT_UNDEFINED:
                *flag[i] = 0;
                stage[i] = VK_PIPELINE_STAGE_TOP_OF_PIPE_BIT;
                break;
            case VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL:
                *flag[i] = VK_ACCESS_TRANSFER_WRITE_BIT;
                stage[i] = VK_PIPELINE_STAGE_TRANSFER_BIT;
                break;
            case VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL:
                *flag[i] = VK_ACCESS_TRANSFER_READ_BIT;
                stage[i] = VK_PIPELINE_STAGE_TRANSFER_BIT;
                break;
            case VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL:
                *flag[i] = VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_READ_BIT | VK_ACCESS_DEPTH_STENCIL_ATTACHMENT_WRITE_BIT;
                stage[i] = VK_PIPELINE_STAGE_EARLY_FRAGMENT_TESTS_BIT;
                break;
            case VK_IMAGE_LAYOUT_PRESENT_SRC_KHR:
                *flag[i] = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT;
                stage[i] = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
                break;
            case VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL:
                *flag[i] = VK_ACCESS_COLOR_ATTACHMENT_WRITE_BIT | VK_ACCESS_COLOR_ATTACHMENT_READ_BIT;
                stage[i] = VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT;
                break;
            case VK_IMAGE_LAYOUT_GENERAL:
                *flag[i] = VK_ACCESS_SHADER_READ_BIT | VK_ACCESS_SHADER_WRITE_BIT;
                stage[i] = VK_PIPELINE_STAGE_COMPUTE_SHADER_BIT;
                break;
            case VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL:
                *flag[i] = VK_ACCESS_SHADER_READ_BIT;
                stage[i] = VK_PIPELINE_STAGE_VERTEX_SHADER_BIT;
                break;
            default:
                // Unsupported layout transition!
                ASSERT(false);
                break;
            }
        }

        if (queue == e_queueAlias::transfer)
        {
            stage[0] = VK_PIPELINE_STAGE_TRANSFER_BIT;
            stage[1] = VK_PIPELINE_STAGE_TRANSFER_BIT;
            *flag[0] = VK_ACCESS_TRANSFER_READ_BIT | VK_ACCESS_TRANSFER_WRITE_BIT;
            *flag[1] = VK_ACCESS_TRANSFER_READ_BIT | VK_ACCESS_TRANSFER_WRITE_BIT;
        }

        vkCmdPipelineBarrier(
            command,
            stage[0], stage[1],
            0,
            0, nullptr,
            0, nullptr,
            1, &barrier
        );
    }

    VulkanTextureInternal* VulkanContext::createTextureInternal(const TextureDesc& desc)
    {
        VulkanTextureInternal* tex = memNew<VulkanTextureInternal>();
        tex->m_buffered = desc.m_buffered;
        tex->m_aspect = translateImageAspect(desc.m_format.m_channels);
        tex->m_layers = desc.m_layers;

        // determine image and view type from params
        VkImageType imageType = VK_IMAGE_TYPE_1D;
        VkImageViewType viewType = VK_IMAGE_VIEW_TYPE_1D;
        if (desc.m_height > 1)
        {
            imageType = VK_IMAGE_TYPE_2D;
            viewType = VK_IMAGE_VIEW_TYPE_2D;
        }
        if (desc.m_depth > 1)
        {
            imageType = VK_IMAGE_TYPE_3D;
            viewType = VK_IMAGE_VIEW_TYPE_3D;
        }
        else if (desc.m_width == desc.m_height && desc.m_layers % 6 == 0)
        {
            imageType = VK_IMAGE_TYPE_2D;
            viewType = VK_IMAGE_VIEW_TYPE_CUBE;
        }
        if (desc.m_layers > 1 && viewType != VK_IMAGE_VIEW_TYPE_CUBE)
            viewType = (VkImageViewType)(viewType + VK_IMAGE_VIEW_TYPE_1D_ARRAY);

        VkFormat imageFormat = translateFormat(desc.m_format);
        VkImageUsageFlags imageUsage = translateImageUsage(desc.m_usage);
        imageUsage |= VK_IMAGE_USAGE_TRANSFER_SRC_BIT;
        imageUsage |= VK_IMAGE_USAGE_TRANSFER_DST_BIT;

        tex->m_desiredLayout = VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        if (!desc.m_data.m_data)
        {
            if ((desc.m_format.m_channels & e_formatChannelFlags::d) == e_formatChannelFlags::d)
                tex->m_desiredLayout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
            else if (imageUsage & VK_IMAGE_USAGE_STORAGE_BIT && !(imageUsage & VK_IMAGE_USAGE_COLOR_ATTACHMENT_BIT))
                tex->m_desiredLayout = VK_IMAGE_LAYOUT_GENERAL;
            else
                tex->m_desiredLayout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
        }

        VkResult result;
        VkImageCreateInfo imageInfo = {};
        imageInfo.sType = VK_STRUCTURE_TYPE_IMAGE_CREATE_INFO;
        imageInfo.imageType = imageType;
        imageInfo.extent.width = desc.m_width;
        imageInfo.extent.height = desc.m_height;
        imageInfo.extent.depth = desc.m_depth;
        imageInfo.mipLevels = desc.m_mips;
        imageInfo.arrayLayers = desc.m_layers;
        imageInfo.format = imageFormat;
        imageInfo.tiling = VK_IMAGE_TILING_OPTIMAL;
        imageInfo.usage = imageUsage;
        imageInfo.samples = (VkSampleCountFlagBits)desc.m_samples;
        imageInfo.initialLayout = VK_IMAGE_LAYOUT_UNDEFINED;
        imageInfo.flags = (viewType == VK_IMAGE_VIEW_TYPE_CUBE) ? VK_IMAGE_CREATE_CUBE_COMPATIBLE_BIT : 0;
        u32 count = (desc.m_buffered) ? s_numBufferedResources : 1;
        tex->m_image = new VkImage[count];
        tex->m_view = new VkImageView[count];
        tex->m_memory = new VkDeviceMemory[count];
        tex->m_currentLayout = new VkImageLayout[count];
        for (u32 i = 0; i < count; ++i)
        {
            tex->m_currentLayout[i] = VK_IMAGE_LAYOUT_UNDEFINED;
            CHECK_RESULT((result = vkCreateImage(m_device, &imageInfo, m_allocationCallbacks, &tex->m_image[i])) == VK_SUCCESS);
        }

        // determine memory properties from access flags
        VkMemoryPropertyFlags memProps = VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
        if (desc.m_access & e_accessFlags::write || desc.m_access & e_accessFlags::read)
            memProps = VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT;

        for (u32 i = 0; i < count; ++i)
        {
            VkMemoryRequirements requirements;
            vkGetImageMemoryRequirements(m_device, tex->m_image[i], &requirements);
            VkMemoryAllocateInfo alloc = {};
            alloc.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            alloc.allocationSize = requirements.size;
            alloc.memoryTypeIndex = findMemoryType(requirements.memoryTypeBits, memProps, requirements.size);
            CHECK_RESULT((result = vkAllocateMemory(m_device, &alloc, m_allocationCallbacks, &tex->m_memory[i])) == VK_SUCCESS);
            vkBindImageMemory(m_device, tex->m_image[i], tex->m_memory[i], 0);
        }

        VkImageAspectFlags aspect = translateImageAspect(desc.m_format.m_channels);
        VkImageViewCreateInfo viewInfo = {};
        viewInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
        viewInfo.viewType = viewType;
        viewInfo.format = imageFormat;
        viewInfo.subresourceRange.aspectMask = aspect;
        viewInfo.subresourceRange.baseMipLevel = 0;
        viewInfo.subresourceRange.levelCount = desc.m_mips;
        viewInfo.subresourceRange.baseArrayLayer = 0;
        viewInfo.subresourceRange.layerCount = desc.m_layers;
        for (u32 i = 0; i < count; ++i)
        {
            viewInfo.image = tex->m_image[i];
            viewInfo.subresourceRange.aspectMask &= ~VK_IMAGE_ASPECT_STENCIL_BIT;
            CHECK_RESULT((result = vkCreateImageView(m_device, &viewInfo, m_allocationCallbacks, &tex->m_view[i])) == VK_SUCCESS);
        }

        // create image views for individual cubemap layers
        if (viewType == VK_IMAGE_VIEW_TYPE_CUBE)
        {
            tex->m_cubemapViews = memNewArray<VkImageView>(6);

            VkImageViewCreateInfo cubeViewInfo = {};
            cubeViewInfo.sType = VK_STRUCTURE_TYPE_IMAGE_VIEW_CREATE_INFO;
            cubeViewInfo.viewType = VK_IMAGE_VIEW_TYPE_2D;
            cubeViewInfo.image = tex->m_image[0];
            cubeViewInfo.format = imageFormat;
            cubeViewInfo.subresourceRange.aspectMask = aspect;
            cubeViewInfo.subresourceRange.baseMipLevel = 0;
            cubeViewInfo.subresourceRange.levelCount = desc.m_mips;
            cubeViewInfo.subresourceRange.layerCount = 1;
            for (u32 i = 0; i < 6; ++i)
            {
                cubeViewInfo.subresourceRange.baseArrayLayer = i;
                CHECK_RESULT((result = vkCreateImageView(m_device, &cubeViewInfo, m_allocationCallbacks, &tex->m_cubemapViews[i])) == VK_SUCCESS);
            }
        }

        // if no data was provided transition to the desired layout now
        if (!desc.m_data.m_data)
        {
            for (u32 i = 0; i < count; ++i)
            {
                VkCommandBuffer commandBuffer = beginSingleTimeCommands(m_transferCommandPool);
                transitionImageLayout(e_queueAlias::transfer, commandBuffer, tex->m_image[i], aspect, desc.m_layers, desc.m_mips, tex->m_currentLayout[i], tex->m_desiredLayout);
                endSingleTimeCommands(m_queues[QueueAlias::transfer], m_transferCommandPool, commandBuffer);
                tex->m_currentLayout[i] = tex->m_desiredLayout;
            }

        }

        return tex;
    }

    Texture VulkanContext::createTexture(const TextureDesc& desc)
    {
        VulkanTexture* tex = memNew<VulkanTexture>();
        tex->m_count = 1;
        tex->m_desc = desc;

        tex->m_texture = createTextureInternal(desc);

        // if initial data has been provided copy to image
        if (desc.m_data.m_data)
        {
            TextureUpdateDesc textureDesc;
            textureDesc.x = 0;
            textureDesc.y = 0;
            textureDesc.z = 0;
            textureDesc.m_width = desc.m_width;
            textureDesc.m_height = desc.m_height;
            textureDesc.m_depth = desc.m_depth;
            textureDesc.m_data = desc.m_data;
            updateTexture(tex, textureDesc);
        }
        
        return (Texture)tex;
    }
    
    Shader VulkanContext::createShader(const ShaderDesc& desc)
    {
        VulkanShader* shader = memNew<VulkanShader>();

        VkResult result;
        VkShaderModuleCreateInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_SHADER_MODULE_CREATE_INFO;
        info.codeSize = desc.m_data.m_size;
        info.pCode = (u32*)desc.m_data.m_data;
        CHECK_RESULT((result = vkCreateShaderModule(m_device, &info, m_allocationCallbacks, &shader->m_module)) == VK_SUCCESS);
        
        shader->m_stage = translateShaderStage(desc.m_stage);
        shader->m_entry = desc.m_entry;
        shader->m_numDescriptors = (u32)desc.m_descriptors.size();
        shader->m_descriptors = new VkDescriptorSetLayoutBinding[shader->m_numDescriptors];
        shader->m_descriptorBindingFlags = new VkDescriptorBindingFlags[shader->m_numDescriptors];
        shader->m_descriptorMappings = new u32[shader->m_numDescriptors];
        for (u32 i = 0; i < shader->m_numDescriptors; ++i)
        {
            shader->m_descriptors[i].binding = desc.m_descriptors[i].m_binding;
            shader->m_descriptorMappings[i] = (desc.m_descriptors[i].m_mapping == (u32)-1) ? desc.m_descriptors[i].m_binding : desc.m_descriptors[i].m_mapping;
            shader->m_descriptors[i].stageFlags = shader->m_stage;
            shader->m_descriptors[i].descriptorCount = desc.m_descriptors[i].m_count;
            shader->m_descriptors[i].descriptorType = translateDescriptorType(desc.m_descriptors[i].m_type);
            shader->m_descriptors[i].pImmutableSamplers = nullptr;
            shader->m_descriptorBindingFlags[i] = VK_DESCRIPTOR_BINDING_PARTIALLY_BOUND_BIT;
        }

        shader->m_numPushConstants = (u32)desc.m_pushConstants.size();
        shader->m_pushConstants = new VkPushConstantRange[shader->m_numPushConstants];
        for (u32 i = 0; i < shader->m_numPushConstants; ++i)
        {
            shader->m_pushConstants[i].stageFlags = shader->m_stage;
            shader->m_pushConstants[i].offset = desc.m_pushConstants[i].m_offset;
            shader->m_pushConstants[i].size = desc.m_pushConstants[i].m_size;
        }

        return (Shader)shader;
    }

    Buffer VulkanContext::createBuffer(const BufferDesc& desc)
    {
        VulkanBuffer* buffer = memNew<VulkanBuffer>();
        
        VkBufferUsageFlags usage = translateBufferUsage(desc.m_usage);
        if (!(usage & VK_BUFFER_USAGE_TRANSFER_SRC_BIT))
            usage |= VK_BUFFER_USAGE_TRANSFER_DST_BIT;

        VkResult result;
        VkBufferCreateInfo bufferInfo = {};
        bufferInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
        bufferInfo.size = desc.m_data.m_size;
        bufferInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE; // should implement non exclusive buffers
        bufferInfo.usage = usage;
        
        int bufferCount = 1;
        if ((usage & VK_BUFFER_USAGE_STORAGE_BUFFER_BIT) || (usage & VK_BUFFER_USAGE_TRANSFER_SRC_BIT))
            bufferCount = 1;
        buffer->m_buffered = bufferCount > 1;
        buffer->m_usage = usage;
        buffer->m_buffer = new VkBuffer[bufferCount];
        buffer->m_memory = new VkDeviceMemory[bufferCount];
        buffer->m_size = new size_t[bufferCount];
        buffer->m_count = 1;
        for (int i = 0; i < bufferCount; ++i)
        {
            CHECK_RESULT((result = vkCreateBuffer(m_device, &bufferInfo, m_allocationCallbacks, &buffer->m_buffer[i])) == VK_SUCCESS);

            // determine memory properties from access flags
            buffer->m_access = VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT;
            if (desc.m_access & e_accessFlags::write || desc.m_access & e_accessFlags::read)
                buffer->m_access = VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT | VK_MEMORY_PROPERTY_HOST_COHERENT_BIT;

            VkMemoryRequirements memRequirements;
            vkGetBufferMemoryRequirements(m_device, buffer->m_buffer[i], &memRequirements);

            VkMemoryAllocateInfo allocInfo = {};
            allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
            allocInfo.allocationSize = memRequirements.size;
            allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, buffer->m_access, desc.m_data.m_size);
            CHECK_RESULT((result = vkAllocateMemory(m_device, &allocInfo, m_allocationCallbacks, &buffer->m_memory[i])) == VK_SUCCESS);

            vkBindBufferMemory(m_device, buffer->m_buffer[i], buffer->m_memory[i], 0);
            buffer->m_size[i] = desc.m_data.m_size;

        }

        if (desc.m_data.m_data)
            updateBuffer(buffer, desc.m_data);

        return (Buffer)buffer;
    }

    Sampler VulkanContext::createSampler(const SamplerDesc& desc)
    {
        VulkanSampler* sampler = memNew<VulkanSampler>();

        VkResult result;
        VkSamplerCreateInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_SAMPLER_CREATE_INFO;
        info.magFilter = translateFilter(desc.m_magFilter);
        info.minFilter = translateFilter(desc.m_minFilter);
        info.mipmapMode = translateMipmapMode(desc.m_mipmapMode);
        info.addressModeU = translateAddressMode(desc.m_addressModeU);
        info.addressModeV = translateAddressMode(desc.m_addressModeV);
        info.addressModeW = translateAddressMode(desc.m_addressModeW);
        info.mipLodBias = desc.m_mipLodBias;
        info.anisotropyEnable = desc.m_anisotropyEnable;
        info.maxAnisotropy = desc.m_maxAnisotropy;
        info.compareEnable = desc.m_compareEnable;
        info.compareOp = translateCompareOp(desc.m_compareOp);
        info.minLod = desc.m_minLod;
        info.maxLod = desc.m_maxLod;
        info.borderColor = translateBorderColor(desc.m_borderColor);
        info.unnormalizedCoordinates = desc.m_unnormalizedCoordinates;
        CHECK_RESULT((result = vkCreateSampler(m_device, &info, m_allocationCallbacks, &sampler->m_sampler)) == VK_SUCCESS);

        return (Sampler)sampler;
    }
    
    RenderPass VulkanContext::createRenderPass(const RenderPassDesc& desc)
    {
        VulkanRenderPass* pass = memNew<VulkanRenderPass>();

        VkResult result;
        VkRenderPassCreateInfo passInfo = {};
        passInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_CREATE_INFO;
        
        std::vector<VkAttachmentDescription> attachments;
        std::vector<TextureDesc*> attachmentTextures;
        std::vector<VkSubpassDescription> subpassDescs;
        std::vector<VkSubpassDependency> subpassDeps;
        std::vector<VkAttachmentReference> subpassRefs[e_attachmentType::COUNT];
        for (int i = 0; i < desc.m_subpasses.size(); ++i)
        {
            const SubpassDesc& subpass = desc.m_subpasses[i];
            VkSubpassDescription subpassDesc = {};
            VkSubpassDependency subpassDep = {};
            std::vector<AttachmentDesc> subpassAttachmentDescs[] = { subpass.m_inputAttachments, subpass.m_colorAttachments, subpass.m_resolveAttachments, subpass.m_depthAttachments };
            for (int j = 0; j < e_attachmentType::COUNT; ++j)
            {
                std::vector<VkAttachmentReference> refs;
                for (auto& attachment : subpassAttachmentDescs[j])
                {
                    if (j == e_attachmentType::resolve)
                        pass->m_resolves = true;
                    pass->m_samples = max(pass->m_samples, attachment.m_texture->m_samples);

                    VkImageLayout layout = VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL;
                    if (attachment.m_texture->m_format.m_channels & e_formatChannelFlags::d)
                        layout = VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;

                    // find in existing attachments
                    int index = -1;
                    for (int k = 0; k < attachmentTextures.size(); ++k)
                    {
                        if (attachmentTextures[k] == attachment.m_texture)
                        {
                            index = k;
                            break;
                        }
                    }
                    if (index >= 0)
                    {
                        // existing attachment
                        VkAttachmentReference attachmentRef = {};
                        attachmentRef.attachment = index;
                        attachmentRef.layout = layout;
                        refs.push_back(attachmentRef);
                    }
                    else
                    {
                        // new attachment
                        VkAttachmentDescription attachmentDesc = {};
                        attachmentDesc.format = translateFormat(attachment.m_texture->m_format);
                        attachmentDesc.samples = (VkSampleCountFlagBits)attachment.m_texture->m_samples;
                        attachmentDesc.loadOp = translateLoadOp(attachment.m_loadOp);
                        attachmentDesc.storeOp = translateStoreOp(attachment.m_storeOp);
                        attachmentDesc.stencilLoadOp = translateLoadOp(attachment.m_stencilLoadOp);
                        attachmentDesc.stencilStoreOp = translateStoreOp(attachment.m_stencilStoreOp);
                        attachmentDesc.initialLayout = layout;
                        attachmentDesc.finalLayout = layout;
                        attachments.push_back(attachmentDesc);
                        attachmentTextures.push_back(attachment.m_texture);
                        
                        VkAttachmentReference attachmentRef = {};
                        attachmentRef.attachment = (u32)attachments.size() - 1;
                        attachmentRef.layout = layout;
                        refs.push_back(attachmentRef);
                    }
                }
                subpassRefs[j] = refs;
            }

            subpassDesc.inputAttachmentCount = (u32)(subpassRefs[e_attachmentType::input].size());
            subpassDesc.pInputAttachments = (subpassDesc.inputAttachmentCount > 0) ? subpassRefs[e_attachmentType::input].data() : nullptr;
            subpassDesc.colorAttachmentCount = (u32)max(subpassRefs[e_attachmentType::resolve].size(), subpassRefs[e_attachmentType::colour].size());
            subpassDesc.pColorAttachments = (subpassRefs[e_attachmentType::colour].size() > 0) ? subpassRefs[e_attachmentType::colour].data() : nullptr;
            subpassDesc.pDepthStencilAttachment = (subpassRefs[e_attachmentType::depth].size() > 0) ? subpassRefs[e_attachmentType::depth].data() : nullptr;
            subpassDesc.pResolveAttachments = (subpassRefs[e_attachmentType::resolve].size() > 0) ? subpassRefs[e_attachmentType::resolve].data() : nullptr;
            subpassDesc.pipelineBindPoint = (subpass.m_computeSubpass) ? VK_PIPELINE_BIND_POINT_COMPUTE : VK_PIPELINE_BIND_POINT_GRAPHICS;
            subpassDescs.push_back(subpassDesc);
            
            subpassDep.srcSubpass = (i == 0) ? VK_SUBPASS_EXTERNAL : i - 1;
            subpassDep.srcStageMask = translatePipelineStage(subpass.m_prevSubpassStageDependency);
            subpassDep.srcAccessMask = translateMemoryAccess(subpass.m_prevSubpassAccessFlags);
            subpassDep.dstSubpass = i;
            subpassDep.dstStageMask = translatePipelineStage(subpass.m_subpassWaitStage);
            subpassDep.dstAccessMask = translateMemoryAccess(subpass.m_subpassAccessFlags);
            subpassDeps.push_back(subpassDep);

        }
        passInfo.attachmentCount = (u32)attachments.size();
        passInfo.pAttachments = attachments.data();
        passInfo.subpassCount = (u32)subpassDescs.size();
        passInfo.pSubpasses = subpassDescs.data();
        passInfo.dependencyCount = (u32)subpassDeps.size();
        passInfo.pDependencies = subpassDeps.data();
        CHECK_RESULT((result = vkCreateRenderPass(m_device, &passInfo, m_allocationCallbacks, &pass->m_renderPass)) == VK_SUCCESS);

        return (RenderPass)pass;
    }

    Framebuffer VulkanContext::obtainFramebuffer(const FramebufferDesc& desc)
    {
        VkResult result;

        VulkanRenderPass* pass = (VulkanRenderPass*)desc.m_pass;

        bool backBuffer = false;
        std::vector<VulkanTextureInternal*> textures;
        std::vector<VkImageView> views;
        VkExtent2D extent = { UINT_MAX, UINT_MAX };
        u32 layers = UINT_MAX;
        for (auto tex : desc.m_textures)
        {
            if (!tex)
            {
                views.push_back(m_swapChainTextures.m_texture->m_view[m_currentSwapChainImage]);
                textures.push_back(m_swapChainTextures.m_texture);
                extent.width = min(extent.width, m_swapChainExtent.width);
                extent.height = min(extent.height, m_swapChainExtent.height);
                layers = 1;
                backBuffer = true;
            }
            else
            {
                VulkanTexture* vk = (VulkanTexture*)tex;
                extent.width = min(extent.width, vk->m_desc.m_width);
                extent.height = min(extent.height, vk->m_desc.m_height);
                layers = min(layers, vk->m_desc.m_layers);

                VulkanTextureInternal* in = vk->m_texture;
                textures.push_back(in);
                views.push_back(in->backView());
                vk->m_texture->m_resolved = false;

                // depth textures cannot be resolved
                if (pass->m_resolves && !(vk->m_texture->m_aspect & VK_IMAGE_ASPECT_DEPTH_BIT))
                {
                    if (!vk->m_resolveTexture)
                    {
                        TextureDesc td = vk->m_desc;
                        td.m_samples = 1;
                        vk->m_resolveTexture = createTextureInternal(td);
                    }
                    VulkanTextureInternal* rin = vk->m_resolveTexture;
                    vk->m_texture->m_resolved = true;
                    views.push_back(rin->backView());
                    textures.push_back(rin);
                }
            }
        }

        // check for existing framebuffers that match
        for (auto fb : *m_framebuffers)
        {
            if (fb->m_views.size() != views.size())
                continue;
            bool match = true;
            for (int i = 0; i < views.size(); ++i)
            {
                if (views[i] != fb->m_views[i])
                    match = false;
            }
            if (match)
            {
                // need to update to the current swap chain semaphore
                if (fb->m_semaphore != VK_NULL_HANDLE)
                    fb->m_semaphore = m_swapChainImageSemaphores[m_currentSwapChainImage];
                
                // flag all used textures as resolved/unresolved as appropriate
                for (auto& t : fb->m_textures)
                    t->m_resolved = pass->m_resolves;
                return fb;
            }
        }

        // create a new framebuffer
        VulkanFramebuffer* framebuffer = memNew<VulkanFramebuffer>();
        VkFramebufferCreateInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_FRAMEBUFFER_CREATE_INFO;
        framebuffer->m_textures = textures;
        framebuffer->m_views = views;
        framebuffer->m_extent = extent;
        framebuffer->m_layers = layers;

        if (backBuffer)
            framebuffer->m_semaphore = m_swapChainImageSemaphores[m_currentSwapChainImage];

        info.renderPass = pass->m_renderPass;
        info.attachmentCount = (u32)views.size();
        info.pAttachments = views.data();
        info.width = framebuffer->m_extent.width;
        info.height = framebuffer->m_extent.height;
        info.layers = framebuffer->m_layers;
        CHECK_RESULT((result = vkCreateFramebuffer(m_device, &info, m_allocationCallbacks, &framebuffer->m_framebuffer)) == VK_SUCCESS);
        m_framebuffers->push_back(framebuffer);
        return (Framebuffer)framebuffer;
    }

    RenderPipeline VulkanContext::createRenderPipeline(const RenderPipelineDesc& desc)
    {
        VulkanRenderPipeline* pipeline = memNew<VulkanRenderPipeline>();

        // init descriptor mappings to -1 so they can be skipped
        for (u32 i = 0; i < MAX_DESCRIPTOR_WRITES; ++i)
        {
            pipeline->m_descriptorBufferMappings[i] = -1;
            pipeline->m_descriptorImageMappings[i] = -1;
        }

        VkResult result;
        VkGraphicsPipelineCreateInfo info = {};

        std::vector<VkPipelineShaderStageCreateInfo> shaders;
        std::vector<VkDescriptorSetLayoutBinding> descriptors;
        std::vector<VkDescriptorBindingFlags> descriptorBindingFlags;
        std::vector<VkPushConstantRange> pushConstants = {};
        for (int i = 0; i < desc.m_shaders.size(); ++i)
        {
            VulkanShader* vkShader = (VulkanShader*)desc.m_shaders[i];
            VkPipelineShaderStageCreateInfo info = {};
            info.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
            info.stage = vkShader->m_stage;
            info.module = vkShader->m_module;
            info.pName = vkShader->m_entry.c_str();
            shaders.push_back(info);

            for (u32 j = 0; j < vkShader->m_numDescriptors; ++j)
            {
                VkDescriptorSetLayoutBinding descriptor = vkShader->m_descriptors[j];

                // check if this is a duplicate descriptor
                bool found = false;
                for (auto& d : descriptors)
                {
                    if (d.binding == descriptor.binding && d.descriptorType == descriptor.descriptorType)
                    {
                        found = true;
                        d.stageFlags |= descriptor.stageFlags;
                    }
                }

                if (found)
                    continue;

                descriptors.push_back(descriptor);
                descriptorBindingFlags.push_back(vkShader->m_descriptorBindingFlags[j]);
                if (descriptor.descriptorType >= VK_DESCRIPTOR_TYPE_SAMPLER && descriptor.descriptorType <= VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER)
                {
                    pipeline->m_descriptorImageMappings[vkShader->m_descriptorMappings[j]] = descriptor.binding;
                }
                else if (descriptor.descriptorType >= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER && descriptor.descriptorType <= VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC)
                {
                    pipeline->m_descriptorBufferMappings[vkShader->m_descriptorMappings[j]] = descriptor.binding;
                }
            }

            for (u32 j = 0; j < vkShader->m_numPushConstants; ++j)
                pushConstants.push_back(vkShader->m_pushConstants[j]);
        }
        /*
        VkDescriptorPoolCreateInfo descriptorPoolInfo = {};
        descriptorPoolInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_POOL_CREATE_INFO;
        descriptorPoolInfo.flags = VK_DESCRIPTOR_POOL_CREATE_FREE_DESCRIPTOR_SET_BIT;
        descriptorPoolInfo.maxSets = 4096;
        descriptorPoolInfo.poolSizeCount = sizeof(poolSizes) / sizeof(poolSizes[0]);
        descriptorPoolInfo.pPoolSizes = poolSizes;
        CHECK_RESULT((result = vkCreateDescriptorPool(m_device, &descriptorPoolInfo, m_allocationCallbacks, &m_descriptorPool)) == VK_SUCCESS);
        */

        VkDescriptorSetLayoutCreateInfo descriptorLayout = {};
        descriptorLayout.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;

        // only create with binding flags if there are any non-zero binding flags
        VkDescriptorSetLayoutBindingFlagsCreateInfo extendedInfo = {};
        extendedInfo.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_BINDING_FLAGS_CREATE_INFO;
        for (size_t i = 0; i < descriptorBindingFlags.size(); ++i)
        {
            if (descriptorBindingFlags[i] != 0)
            {
                extendedInfo.pNext = nullptr;
                extendedInfo.bindingCount = (u32)descriptorBindingFlags.size();
                extendedInfo.pBindingFlags = descriptorBindingFlags.data();
                descriptorLayout.pNext = &extendedInfo;
            }
        }

        descriptorLayout.bindingCount = (u32)descriptors.size();
        descriptorLayout.pBindings = descriptors.data();
        CHECK_RESULT((result = vkCreateDescriptorSetLayout(m_device, &descriptorLayout, m_allocationCallbacks, &pipeline->m_descriptorSetLayout)) == VK_SUCCESS);

        VkDescriptorSetAllocateInfo descriptorSet = {};
        descriptorSet.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
        descriptorSet.descriptorPool = m_descriptorPool;
        descriptorSet.descriptorSetCount = 1;
        descriptorSet.pSetLayouts = &pipeline->m_descriptorSetLayout;
        pipeline->m_descriptorSet = new VulkanDescriptorSet[m_numSwapChainImages];
        for (u32 i = 0; i < m_numSwapChainImages; ++i)
        {
            VulkanDescriptorSet& set = pipeline->m_descriptorSet[i];
            set.m_device = m_device;
            set.m_info = descriptorSet;
            set.m_layout = pipeline->m_descriptorSetLayout;
            m_descriptorSets.push_back(&pipeline->m_descriptorSet[i]);
        }

        VkPipelineLayoutCreateInfo layout = {};
        layout.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
        layout.setLayoutCount = 1;
        layout.pSetLayouts = &pipeline->m_descriptorSetLayout;
        layout.pushConstantRangeCount = (u32)pushConstants.size();
        layout.pPushConstantRanges = pushConstants.data();
        CHECK_RESULT((result = vkCreatePipelineLayout(m_device, &layout, m_allocationCallbacks, &pipeline->m_pipelineLayout)) == VK_SUCCESS);
        
        VkPipelineTessellationStateCreateInfo tessellation = {};
        tessellation.sType = VK_STRUCTURE_TYPE_PIPELINE_TESSELLATION_STATE_CREATE_INFO;
        tessellation.patchControlPoints = desc.m_tessellation.m_controlPoints;

        VkPipelineVertexInputStateCreateInfo vertexInput = {};
        vertexInput.sType = VK_STRUCTURE_TYPE_PIPELINE_VERTEX_INPUT_STATE_CREATE_INFO;
        std::vector<VkVertexInputBindingDescription> bindings;
        std::vector<VkVertexInputAttributeDescription> attributes;
        for (auto& binding : desc.m_vertexInput.m_bindings)
        {
            VkVertexInputBindingDescription bindingDesc = {};
            bindingDesc.binding = binding.m_binding;
            bindingDesc.stride = (u32)binding.m_stride;
            bindingDesc.inputRate = translateInputRate(binding.m_rate);
            bindings.push_back(bindingDesc);
        }
        for (auto& attribute : desc.m_vertexInput.m_attributes)
        {
            VkVertexInputAttributeDescription attribDesc = {};
            attribDesc.location = attribute.m_location;
            attribDesc.binding = attribute.m_binding;
            attribDesc.format = translateFormat(attribute.m_format);
            attribDesc.offset = attribute.m_offset;
            attributes.push_back(attribDesc);
        }
        vertexInput.vertexBindingDescriptionCount = (u32)bindings.size();
        vertexInput.pVertexBindingDescriptions = bindings.data();
        vertexInput.vertexAttributeDescriptionCount = (u32)attributes.size();
        vertexInput.pVertexAttributeDescriptions = attributes.data();

        VkPipelineInputAssemblyStateCreateInfo inputAssembly = {};
        inputAssembly.sType = VK_STRUCTURE_TYPE_PIPELINE_INPUT_ASSEMBLY_STATE_CREATE_INFO;
        inputAssembly.topology = translatePrimitiveTopology(desc.m_inputAssembly.m_topology);
        inputAssembly.primitiveRestartEnable = desc.m_inputAssembly.m_primitiveRestartEnable;
        
        VkPipelineViewportStateCreateInfo viewport = {};
        viewport.sType = VK_STRUCTURE_TYPE_PIPELINE_VIEWPORT_STATE_CREATE_INFO;
        std::vector<VkViewport> viewports;
        std::vector<VkRect2D> scissors;
        for (auto& v : desc.m_viewports)
            viewports.push_back({ v.x, v.y, v.m_width, v.m_height, v.m_minDepth, v.m_maxDepth });
        for (auto& s : desc.m_scissors)
            scissors.push_back({ s.x, s.y, s.w, s.h });
        viewport.viewportCount = (u32)viewports.size();
        viewport.pViewports = viewports.data();
        viewport.scissorCount = (u32)scissors.size();
        viewport.pScissors = scissors.data();

        VkPipelineRasterizationStateCreateInfo rasterization = {};
        rasterization.sType = VK_STRUCTURE_TYPE_PIPELINE_RASTERIZATION_STATE_CREATE_INFO;
        rasterization.depthClampEnable = desc.m_rasterizer.m_depthClamp;
        rasterization.rasterizerDiscardEnable = desc.m_rasterizer.m_rasterizerDiscard;
        rasterization.polygonMode = translatePolygonMode(desc.m_rasterizer.m_polygonMode);
        rasterization.cullMode = translateCullMode(desc.m_rasterizer.m_cullMode);
        rasterization.frontFace = translateFrontFace(desc.m_rasterizer.m_frontFace);
        rasterization.depthBiasEnable = desc.m_rasterizer.m_depthBias;
        rasterization.depthBiasConstantFactor = desc.m_rasterizer.m_depthBiasConstant;
        rasterization.depthBiasClamp = desc.m_rasterizer.m_depthBiasClamp;
        rasterization.depthBiasSlopeFactor = desc.m_rasterizer.m_depthBiasSlope;
        rasterization.lineWidth = desc.m_rasterizer.m_lineWidth;

        //  desc can override the multisample count
        u32 samples = ((VulkanRenderPass*)desc.m_renderPass)->m_samples;
        if (desc.m_multisample.m_samples > 1)
            samples = desc.m_multisample.m_samples;

        VkPipelineMultisampleStateCreateInfo multisample = {};
        multisample.sType = VK_STRUCTURE_TYPE_PIPELINE_MULTISAMPLE_STATE_CREATE_INFO;
        multisample.rasterizationSamples = (VkSampleCountFlagBits)samples;
        multisample.sampleShadingEnable = desc.m_multisample.m_sampleShading;
        multisample.minSampleShading = desc.m_multisample.m_minSampleShading;
        multisample.alphaToCoverageEnable = desc.m_multisample.m_alphaToCoverage;
        multisample.alphaToOneEnable = desc.m_multisample.m_alphaToOne;

        VkPipelineDepthStencilStateCreateInfo depthStencil = {};
        depthStencil.sType = VK_STRUCTURE_TYPE_PIPELINE_DEPTH_STENCIL_STATE_CREATE_INFO;
        depthStencil.depthTestEnable = desc.m_depthStencil.m_depthTest;
        depthStencil.depthWriteEnable = desc.m_depthStencil.m_depthWrite;
        depthStencil.depthCompareOp = translateCompareOp(desc.m_depthStencil.m_depthCompareOp);
        depthStencil.depthBoundsTestEnable = desc.m_depthStencil.m_depthBounds;
        depthStencil.stencilTestEnable = desc.m_depthStencil.m_stencilTest;
        depthStencil.front.failOp = translateStencilOp(desc.m_depthStencil.m_front.m_failOp);
        depthStencil.front.passOp = translateStencilOp(desc.m_depthStencil.m_front.m_passOp);
        depthStencil.front.depthFailOp = translateStencilOp(desc.m_depthStencil.m_front.m_depthFailOp);
        depthStencil.front.compareOp = translateCompareOp(desc.m_depthStencil.m_front.m_compareOP);
        depthStencil.front.compareMask = desc.m_depthStencil.m_front.m_compareMask;
        depthStencil.front.writeMask = desc.m_depthStencil.m_front.m_writeMask;
        depthStencil.front.reference = desc.m_depthStencil.m_front.m_reference;
        depthStencil.back.failOp = translateStencilOp(desc.m_depthStencil.m_back.m_failOp);
        depthStencil.back.passOp = translateStencilOp(desc.m_depthStencil.m_back.m_passOp);
        depthStencil.back.depthFailOp = translateStencilOp(desc.m_depthStencil.m_back.m_depthFailOp);
        depthStencil.back.compareOp = translateCompareOp(desc.m_depthStencil.m_back.m_compareOP);
        depthStencil.back.compareMask = desc.m_depthStencil.m_back.m_compareMask;
        depthStencil.back.writeMask = desc.m_depthStencil.m_back.m_writeMask;
        depthStencil.back.reference = desc.m_depthStencil.m_back.m_reference;
        depthStencil.minDepthBounds = desc.m_depthStencil.m_minDepthBounds;
        depthStencil.maxDepthBounds = desc.m_depthStencil.m_maxDepthBounds;

        VkPipelineColorBlendStateCreateInfo blend = {};
        blend.sType = VK_STRUCTURE_TYPE_PIPELINE_COLOR_BLEND_STATE_CREATE_INFO;
        std::vector<VkPipelineColorBlendAttachmentState> attachments;
        for (auto& attachment : desc.m_blendDesc.m_blends)
        {
            VkPipelineColorBlendAttachmentState state = {};
            state.blendEnable = attachment.m_blendEnable;
            state.srcColorBlendFactor = translateBlendFactor(attachment.m_srcColorBlendFactor);
            state.dstColorBlendFactor = translateBlendFactor(attachment.m_dstColorBlendFactor);
            state.colorBlendOp = translateBlendOp(attachment.m_colorBlendOp);
            state.srcAlphaBlendFactor = translateBlendFactor(attachment.m_srcAlphaBlendFactor);
            state.dstAlphaBlendFactor = translateBlendFactor(attachment.m_dstAlphaBlendFactor);
            state.alphaBlendOp = translateBlendOp(attachment.m_alphaBlendOp);
            state.colorWriteMask = VK_COLOR_COMPONENT_R_BIT | VK_COLOR_COMPONENT_G_BIT | VK_COLOR_COMPONENT_B_BIT | VK_COLOR_COMPONENT_A_BIT;
            attachments.push_back(state);
        }
        blend.attachmentCount = (u32)attachments.size();
        blend.pAttachments = attachments.data();
        blend.logicOpEnable = desc.m_blendDesc.m_logicOpEnable;
        blend.logicOp = translateLogicOp(desc.m_blendDesc.m_logicOp);
        memcpy(blend.blendConstants, desc.m_blendDesc.m_blendConstants, sizeof(blend.blendConstants[0]) * 4);

        VkPipelineDynamicStateCreateInfo dynamic = {};
        std::vector<VkDynamicState> dynamicStates = translateDynamicState(desc.m_dynamic.m_dynamicState);
        dynamic.sType = VK_STRUCTURE_TYPE_PIPELINE_DYNAMIC_STATE_CREATE_INFO;
        dynamic.dynamicStateCount = (u32)dynamicStates.size();
        dynamic.pDynamicStates = dynamicStates.data();

        info.sType = VK_STRUCTURE_TYPE_GRAPHICS_PIPELINE_CREATE_INFO;
        info.layout = pipeline->m_pipelineLayout;
        info.stageCount = (u32)shaders.size();
        info.pStages = shaders.data();
        info.pVertexInputState = &vertexInput;
        info.pInputAssemblyState = &inputAssembly;
        info.pTessellationState = &tessellation;
        info.pViewportState = &viewport;
        info.pRasterizationState = &rasterization;
        info.pMultisampleState = &multisample;
        info.pDepthStencilState = &depthStencil;
        info.pColorBlendState = &blend;
        info.pDynamicState = &dynamic;
        info.renderPass = ((VulkanRenderPass*)desc.m_renderPass)->m_renderPass;
        info.subpass = desc.m_subpass;
        CHECK_RESULT((result = vkCreateGraphicsPipelines(m_device, VK_NULL_HANDLE, 1, &info, m_allocationCallbacks, &pipeline->m_pipeline)) == VK_SUCCESS);

        return (RenderPipeline)pipeline;
    }
        
    ComputePipeline VulkanContext::createComputePipeline(const ComputePipelineDesc& desc)
    {
        VulkanPipeline* pipeline = memNew<VulkanPipeline>();

        // init descriptor mappings to -1 so they can be skipped
        for (u32 i = 0; i < MAX_DESCRIPTOR_WRITES; ++i)
        {
            pipeline->m_descriptorBufferMappings[i] = -1;
            pipeline->m_descriptorImageMappings[i] = -1;
        }

        VkResult result;
        VkComputePipelineCreateInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_COMPUTE_PIPELINE_CREATE_INFO;
        std::vector<VkDescriptorSetLayoutBinding> descriptors;
        std::vector<VkPushConstantRange> pushConstants = {};
        VulkanShader* vkShader = (VulkanShader*)desc.m_shader;
        VkPipelineShaderStageCreateInfo shader = {};
        shader.sType = VK_STRUCTURE_TYPE_PIPELINE_SHADER_STAGE_CREATE_INFO;
        shader.stage = vkShader->m_stage;
        shader.module = vkShader->m_module;
        shader.pName = vkShader->m_entry.c_str();

        for (u32 j = 0; j < vkShader->m_numDescriptors; ++j)
        {
            VkDescriptorSetLayoutBinding descriptor = vkShader->m_descriptors[j];
            if (descriptor.descriptorType >= VK_DESCRIPTOR_TYPE_SAMPLER && descriptor.descriptorType <= VK_DESCRIPTOR_TYPE_STORAGE_TEXEL_BUFFER)
            {
                pipeline->m_descriptorImageMappings[vkShader->m_descriptorMappings[j]] = descriptor.binding;
            }
            else if (descriptor.descriptorType >= VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER && descriptor.descriptorType <= VK_DESCRIPTOR_TYPE_STORAGE_BUFFER_DYNAMIC)
            {
                pipeline->m_descriptorBufferMappings[vkShader->m_descriptorMappings[j]] = descriptor.binding;
            }
            descriptors.push_back(descriptor);
        }

        for (u32 j = 0; j < vkShader->m_numPushConstants; ++j)
        {
            pushConstants.push_back(vkShader->m_pushConstants[j]);
        }

        VkDescriptorSetLayoutCreateInfo descriptorLayout = {};
        descriptorLayout.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_LAYOUT_CREATE_INFO;
        descriptorLayout.bindingCount = (u32)descriptors.size();
        descriptorLayout.pBindings = descriptors.data();
        CHECK_RESULT((result = vkCreateDescriptorSetLayout(m_device, &descriptorLayout, m_allocationCallbacks, &pipeline->m_descriptorSetLayout)) == VK_SUCCESS);

        VkDescriptorSetAllocateInfo descriptorSet = {};
        descriptorSet.sType = VK_STRUCTURE_TYPE_DESCRIPTOR_SET_ALLOCATE_INFO;
        descriptorSet.descriptorPool = m_descriptorPool;
        descriptorSet.descriptorSetCount = 1;
        descriptorSet.pSetLayouts = &pipeline->m_descriptorSetLayout;
        pipeline->m_descriptorSet = new VulkanDescriptorSet[m_numSwapChainImages];
        for (u32 i = 0; i < m_numSwapChainImages; ++i)
        {
            VulkanDescriptorSet& set = pipeline->m_descriptorSet[i];
            set.m_device = m_device;
            set.m_info = descriptorSet;
            set.m_layout = pipeline->m_descriptorSetLayout;
            m_descriptorSets.push_back(&pipeline->m_descriptorSet[i]);
        }

        VkPipelineLayoutCreateInfo layout = {};
        layout.sType = VK_STRUCTURE_TYPE_PIPELINE_LAYOUT_CREATE_INFO;
        layout.setLayoutCount = 1;
        layout.pSetLayouts = &pipeline->m_descriptorSetLayout;
        layout.pushConstantRangeCount = (u32)pushConstants.size();
        layout.pPushConstantRanges = pushConstants.data();
        CHECK_RESULT((result = vkCreatePipelineLayout(m_device, &layout, m_allocationCallbacks, &pipeline->m_pipelineLayout)) == VK_SUCCESS);
        
        info.stage = shader;
        info.layout = pipeline->m_pipelineLayout;
        CHECK_RESULT((result = vkCreateComputePipelines(m_device, VK_NULL_HANDLE, 1, &info, m_allocationCallbacks, &pipeline->m_pipeline)) == VK_SUCCESS);
                
        return (RenderPipeline)pipeline;
    }

    void VulkanContext::destroyShader(Shader shader)
    {
        VulkanShader* vkShader = (VulkanShader*)shader;
        vkDestroyShaderModule(m_device, vkShader->m_module, m_allocationCallbacks);

        delete[] vkShader->m_descriptors;
        delete[] vkShader->m_descriptorBindingFlags;
        delete[] vkShader->m_descriptorMappings;
        memDelete(vkShader);
    }

    void VulkanContext::destroyBuffer(Buffer buffer)
    {
        VulkanBuffer* vkBuffer = (VulkanBuffer*)buffer;
        VulkanBufferDestroy* d = memNew<VulkanBufferDestroy>();
        d->m_buffer = vkBuffer;
        m_destroys.push_back(d);
    }

    void VulkanContext::destroyTexture(Texture texture)
    {
        VulkanTexture* vkTex = (VulkanTexture*)texture;
        VulkanTextureDestroy* d = memNew<VulkanTextureDestroy>();
        d->m_texture = vkTex;
        m_destroys.push_back(d);
    }

    void VulkanContext::destroySampler(Sampler sampler)
    {
        VulkanSampler* vkSampler = (VulkanSampler*)sampler;
        vkDestroySampler(m_device, vkSampler->m_sampler, m_allocationCallbacks);
        memDelete(vkSampler);
    }

    void VulkanContext::destroyRenderPass(RenderPass pass)
    {
        VulkanRenderPass* vkPass = (VulkanRenderPass*)pass;
        VulkanRenderPassDestroy* d = memNew<VulkanRenderPassDestroy>();
        d->m_pass = vkPass;
        m_destroys.push_back(d);
    }

    void VulkanContext::eraseDescriptorSet(VulkanDescriptorSet* set)
    {
        for (size_t j = 0; j < m_descriptorSets.size(); ++j)
        {
            if (set == m_descriptorSets[j])
            {
                m_descriptorSets.erase(m_descriptorSets.begin() + j);
                break;
            }
        }
    }

    void VulkanContext::destroyRenderPipeline(RenderPipeline pipeline)
    {
        VulkanRenderPipeline* vkPipeline = (VulkanRenderPipeline*)pipeline;
        for (u32 i = 0; i < m_numSwapChainImages; ++i)
            eraseDescriptorSet(&vkPipeline->m_descriptorSet[i]);
        VulkanRenderPipelineDestroy* d = memNew<VulkanRenderPipelineDestroy>();
        d->m_pipeline = vkPipeline;
        d->m_numSwapChainImages = m_numSwapChainImages;
    }

    void VulkanContext::destroyComputePipeline(ComputePipeline pipeline)
    {
        VulkanComputePipeline* vkPipeline = (VulkanComputePipeline*)pipeline;
        for (u32 i = 0; i < m_numSwapChainImages; ++i)
            eraseDescriptorSet(&vkPipeline->m_descriptorSet[i]);
        VulkanComputePipelineDestroy* d = memNew<VulkanComputePipelineDestroy>();
        d->m_pipeline = vkPipeline;
        d->m_numSwapChainImages = m_numSwapChainImages;
    }

    VkCommandBuffer VulkanContext::beginSingleTimeCommands(VkCommandPool pool)
    {
        VkResult result;
        VkCommandBuffer buffer;
        VkCommandBufferAllocateInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_ALLOCATE_INFO;
        info.commandBufferCount = 1;
        info.commandPool = pool;
        info.level = VK_COMMAND_BUFFER_LEVEL_PRIMARY;
        CHECK_RESULT((result = vkAllocateCommandBuffers(m_device, &info, &buffer)) == VK_SUCCESS);
       
        VkCommandBufferBeginInfo begin = {};
        begin.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
        begin.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
        CHECK_RESULT((result = vkBeginCommandBuffer(buffer, &begin)) == VK_SUCCESS);

        return buffer;
    }

    void VulkanContext::endSingleTimeCommands(VkQueue queue, VkCommandPool pool, VkCommandBuffer buffer)
    {
        VkResult result;

        CHECK_RESULT((result = vkEndCommandBuffer(buffer)) == VK_SUCCESS);

        VkSubmitInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
        info.commandBufferCount = 1;
        info.pCommandBuffers = &buffer;
        CHECK_RESULT((result = vkQueueSubmit(queue, 1, &info, VK_NULL_HANDLE)) == VK_SUCCESS);
        vkQueueWaitIdle(queue);
        vkResetCommandBuffer(buffer, VK_COMMAND_BUFFER_RESET_RELEASE_RESOURCES_BIT);
        vkFreeCommandBuffers(m_device, pool, 1, &buffer);
    }

    void VulkanContext::setViewport(const Viewport& vp)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
        VkViewport view = {};
        view.x = vp.x;
        view.y = vp.y;
        view.width = vp.m_width;
        view.height = vp.m_height;
        view.minDepth = vp.m_minDepth;
        view.maxDepth = vp.m_maxDepth;
        vkCmdSetViewport(m_boundCommandBuffer->m_commandBuffer, 0, 1, &view);
    }

    void VulkanContext::setScissorRect(const Rect& sr)
    {    
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
        VkRect2D scissor = {};
        scissor.offset.x = sr.x;
        scissor.offset.y = sr.y;
        scissor.extent.width = sr.w;
        scissor.extent.height = sr.h;
        vkCmdSetScissor(m_boundCommandBuffer->m_commandBuffer, 0, 1, &scissor);
    }

    void VulkanContext::begin()
    {
        VkResult result;

        // get the next swap chain frame
        CHECK_RESULT((result = vkAcquireNextImageKHR(m_device, m_swapChain, UINT64_MAX, m_swapChainImageSemaphores[m_numSwapChainImages], VK_NULL_HANDLE, &m_currentSwapChainImage)) == VK_SUCCESS);
        VkSemaphore temp = m_swapChainImageSemaphores[m_currentSwapChainImage];
        m_swapChainImageSemaphores[m_currentSwapChainImage] = m_swapChainImageSemaphores[m_numSwapChainImages];
        m_swapChainImageSemaphores[m_numSwapChainImages] = temp;

        // wait for command buffers that will be used in the upcoming frame to finish processing their previous frame
        std::vector<VkFence> fences;
        QueueAlias aliases[2] = { e_queueAlias::graphics, e_queueAlias::compute };
        for (int i = 0; i < 2; ++i)
        {
            VulkanCommandPool& pool = m_commandPools[i][m_currentSwapChainImage];
            for (int j = 0; j < pool.m_numCommandBuffers; ++j)
            {
                if (pool.m_commandBuffers[j].m_used)
                {
                    fences.push_back(pool.m_commandBuffers[j].m_sync->m_fence);
                    pool.m_commandBuffers[j].m_sync->m_used = false;
                    pool.m_commandBuffers[j].m_sync->m_consumed = false;
                    pool.m_commandBuffers[j].m_used = false;
                }
            }
        }

        if (fences.size() > 0)
        {
            CHECK_RESULT((result = vkWaitForFences(m_device, (u32)fences.size(), fences.data(), VK_TRUE, UINT64_MAX)) == VK_SUCCESS);
            CHECK_RESULT((result = vkResetFences(m_device, (u32)fences.size(), fences.data())) == VK_SUCCESS);
        }

        // reset the command buffers used for next frame
        CHECK_RESULT((result = vkResetCommandPool(m_device, m_commandPools[0][m_currentSwapChainImage].m_commandPool, VK_COMMAND_POOL_RESET_RELEASE_RESOURCES_BIT)) == VK_SUCCESS);
        CHECK_RESULT((result = vkResetCommandPool(m_device, m_commandPools[1][m_currentSwapChainImage].m_commandPool, VK_COMMAND_POOL_RESET_RELEASE_RESOURCES_BIT)) == VK_SUCCESS);

        m_backBufferUsed = false;
    }

    VulkanCommandBuffer* VulkanContext::obtainCommandBuffer(QueueAlias usage)
    {
        VulkanCommandBuffer* commandBuffer = nullptr;
        VulkanCommandPool& pool = m_commandPools[usage == e_queueAlias::graphics ? 0 : 1][m_currentSwapChainImage];
        for (size_t i = 0; i < pool.m_numCommandBuffers; ++i)
        {
            if (!pool.m_commandBuffers[i].m_used)
            {
                commandBuffer = &pool.m_commandBuffers[i];
                pool.m_commandBuffers[i].m_used = true;
                break;
            }
        }

        if (m_threadID != std::thread::id() && commandBuffer == nullptr)
        {
            // if async context and no valid command buffers try flushing used command buffers
            flushCommandBuffers();
            for (size_t i = 0; i < pool.m_numCommandBuffers; ++i)
            {
                if (!pool.m_commandBuffers[i].m_used)
                {
                    commandBuffer = &pool.m_commandBuffers[i];
                    pool.m_commandBuffers[i].m_used = true;
                    break;
                }
            }
        }

        // if there are no free command buffers then allocate a new set of command buffers for the required pool
        if (commandBuffer == nullptr)
        {
            // allocate command buffers for all swap chain images as the number of used command buffers is unlikely to change much between frames
            size_t prevNum = pool.m_numCommandBuffers;
            for (size_t i = 0; i < m_numSwapChainImages; ++i)
            {
                VulkanCommandPool& p = m_commandPools[usage == e_queueAlias::graphics ? 0 : 1][i];
                p.grow(m_device, m_allocationCallbacks);
            }
            // grab the first new command buffer
            commandBuffer = &pool.m_commandBuffers[prevNum];
            commandBuffer->m_used = true;
        }

        return commandBuffer;
    }

    void VulkanContext::beginComputeEncoding()
    {
        ASSERT(m_boundCommandBuffer == VK_NULL_HANDLE);
        m_boundPipeline = nullptr;
        VkResult result;

        // find an unused command buffer
        m_boundCommandBuffer = obtainCommandBuffer(e_queueAlias::compute);
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);

        VkCommandBufferBeginInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
        info.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
        CHECK_RESULT((result = vkBeginCommandBuffer(m_boundCommandBuffer->m_commandBuffer, &info)) == VK_SUCCESS);
        m_boundQueueType = e_queueAlias::compute;
    }

    void VulkanContext::beginRenderPass(RenderPass pass, Framebuffer framebuffer, Clear* clears, size_t numClears)
    {
        ASSERT(m_boundCommandBuffer == nullptr);
        m_boundPipeline = nullptr;
        VulkanRenderPass* vkPass = (VulkanRenderPass*)pass;
        VulkanFramebuffer* fb = (VulkanFramebuffer*)framebuffer;
        VkResult result;

        // find an unused command buffer
        m_boundCommandBuffer = obtainCommandBuffer(e_queueAlias::graphics);
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);

        VkCommandBufferBeginInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
        info.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
        CHECK_RESULT((result = vkBeginCommandBuffer(m_boundCommandBuffer->m_commandBuffer, &info)) == VK_SUCCESS);

        // transition images in the framebuffer to color attachment optimal layout
        for (auto& texture : fb->m_textures)
        {
            VkImageAspectFlags aspect = texture->m_aspect;
            VkImageLayout layout = (aspect & VK_IMAGE_ASPECT_COLOR_BIT) ? VK_IMAGE_LAYOUT_COLOR_ATTACHMENT_OPTIMAL : VK_IMAGE_LAYOUT_DEPTH_STENCIL_ATTACHMENT_OPTIMAL;
            transitionImageLayout(e_queueAlias::graphics, m_boundCommandBuffer->m_commandBuffer, texture->backImage(), aspect, texture->m_layers, 1, texture->backLayout(), layout);
            texture->setBackLayout(layout);
        }

        // shader read image layouts can't be transitioned during a render pass
        // so defer starting the render pass until the draw occurs
        m_renderPassStarted = false;
        m_renderPassState.m_clears.clear();
        for (size_t i = 0; i < numClears; ++i)
        {
            VkClearValue clear;
            clear.color = { clears[i].r, clears[i].g, clears[i].b, clears[i].a };
            clear.depthStencil = { clears[i].depth, clears[i].stencil };
            m_renderPassState.m_clears.push_back(clear);
        }
        m_renderPassState.m_pass = vkPass->m_renderPass;
        m_renderPassState.m_framebuffer = fb->m_framebuffer;
        m_renderPassState.m_extent = fb->m_extent;
        
        // first render pass to use a framebuffer should wait until it is available
        if (fb->m_semaphore != VK_NULL_HANDLE)
            m_boundFramebufferSemaphore = fb->m_semaphore;
        m_boundQueueType = e_queueAlias::graphics;
    }

    void VulkanContext::beginRenderPassInternal()
    {
        VkRenderPassBeginInfo beginInfo = {};
        beginInfo.sType = VK_STRUCTURE_TYPE_RENDER_PASS_BEGIN_INFO;
        beginInfo.renderPass = m_renderPassState.m_pass;
        beginInfo.clearValueCount = (u32)(m_renderPassState.m_clears.size());
        beginInfo.pClearValues = m_renderPassState.m_clears.data();
        beginInfo.renderArea.offset = { 0, 0 };
        beginInfo.renderArea.extent = m_renderPassState.m_extent;
        beginInfo.framebuffer = m_renderPassState.m_framebuffer;
        vkCmdBeginRenderPass(m_boundCommandBuffer->m_commandBuffer, &beginInfo, VK_SUBPASS_CONTENTS_INLINE);
        m_renderPassStarted = true;
    }

    void VulkanContext::beginNextSubpass()
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
        if (!m_renderPassStarted)
            beginRenderPassInternal();
        vkCmdNextSubpass(m_boundCommandBuffer->m_commandBuffer, VK_SUBPASS_CONTENTS_INLINE);
    }

    void VulkanContext::bindRenderPipeline(RenderPipeline pipeline)
    {
        m_boundPipeline = (VulkanPipeline*)pipeline;
        if (m_boundCommandBuffer != VK_NULL_HANDLE)
        {
            vkCmdBindPipeline(m_boundCommandBuffer->m_commandBuffer, VK_PIPELINE_BIND_POINT_GRAPHICS, m_boundPipeline->m_pipeline);
        }
    }

    void VulkanContext::bindComputePipeline(ComputePipeline pipeline)
    {
        m_boundPipeline = (VulkanPipeline*)pipeline;
        if (m_boundCommandBuffer != VK_NULL_HANDLE)
        {
            vkCmdBindPipeline(m_boundCommandBuffer->m_commandBuffer, VK_PIPELINE_BIND_POINT_COMPUTE, m_boundPipeline->m_pipeline);
        }
    }

    void VulkanContext::bindBuffer(Buffer buf, u32 slot, BindFlags flags, size_t offset)
    {
        bindBufferInternal(buf, slot, flags, offset, m_boundIndexType);
    }

    void VulkanContext::bindBufferInternal(Buffer buf, u32 slot, BindFlags flags, size_t offset, IndexType indexType)
    {
        ASSERT(m_boundPipeline || !(flags & (e_bindFlags::vertexBuffer | e_bindFlags::indexBuffer)));

        VulkanBuffer* buffer = (VulkanBuffer*)buf;
        buffer->m_active = true;

        VkBuffer b = buffer->frontBuffer();
        if (flags & e_bindFlags::vertexBuffer)
        {
            ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
            vkCmdBindVertexBuffers(m_boundCommandBuffer->m_commandBuffer, slot, 1, &b, &offset);
        }
        else if (flags & e_bindFlags::indexBuffer)
        {
            ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
            vkCmdBindIndexBuffer(m_boundCommandBuffer->m_commandBuffer, buffer->frontBuffer(), offset, (indexType == e_indexType::uint16) ? VK_INDEX_TYPE_UINT16 : VK_INDEX_TYPE_UINT32);
        }
        else
        {
            VkDescriptorBufferInfo* info = memNew<VkDescriptorBufferInfo>();
            info->buffer = b;
            info->offset = offset;
            info->range = buffer->frontSize() - offset;

            VkWriteDescriptorSet write = {};
            write.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
            write.descriptorType = (flags & e_bindFlags::storage) ? VK_DESCRIPTOR_TYPE_STORAGE_BUFFER : VK_DESCRIPTOR_TYPE_UNIFORM_BUFFER;
            write.descriptorCount = buffer->m_count;
            write.dstBinding = slot;
            write.pBufferInfo = info;
            write.pImageInfo = nullptr;

            // find existing matching descriptor write to update, otherwise add new
            u32 writeIndex = -1;
            for (u32 i = 0; i < m_numDescriptorWrites; ++i)
            {
                if (m_descriptorWrites[i].dstBinding == slot && m_descriptorWrites[i].pBufferInfo)
                {
                    writeIndex = i;
                    // clear existing write
                    memDelete(m_descriptorWrites[i].pBufferInfo);
                    break;
                }
            }

            if (writeIndex == (u32)-1)
                writeIndex = m_numDescriptorWrites++;

            m_descriptorWrites[writeIndex] = write;
            m_numChangedDescriptorWrites++;
        }
    }

    void VulkanContext::bindTexture(Texture tex, Sampler samp, u32 slot, BindFlags flags)
    {
        ASSERT(m_boundCommandBuffer);

        VulkanTexture* texture = (tex) ? (VulkanTexture*)tex : nullptr;
        VulkanSampler* sampler = (samp) ? (VulkanSampler*)samp : nullptr;

        // can't have a sampled storage image
        ASSERT(!(sampler && texture && flags & e_bindFlags::storage));

        VkDescriptorImageInfo* info = memNew<VkDescriptorImageInfo>();
        if (texture)
        {
            VulkanTextureInternal* t = texture->getBindingTexture();
            info->imageView = t->frontView();
            info->imageLayout = (flags & e_bindFlags::storage) ? VK_IMAGE_LAYOUT_GENERAL : VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
        }

        if (sampler)
            info->sampler = sampler->m_sampler;
        
        VkWriteDescriptorSet write = {};
        write.sType = VK_STRUCTURE_TYPE_WRITE_DESCRIPTOR_SET;
        if (sampler && texture)
            write.descriptorType = VK_DESCRIPTOR_TYPE_COMBINED_IMAGE_SAMPLER;
        else if (texture)
            write.descriptorType = (flags & e_bindFlags::storage) ? VK_DESCRIPTOR_TYPE_STORAGE_IMAGE : VK_DESCRIPTOR_TYPE_SAMPLED_IMAGE;
        else
            write.descriptorType = VK_DESCRIPTOR_TYPE_SAMPLER;
        write.descriptorCount = texture->m_count;
        write.dstBinding = slot;
        write.pImageInfo = info;
        write.pBufferInfo = nullptr;

        // find existing matching descriptor write to update, otherwise add new
        u32 writeIndex = -1;
        for (u32 i = 0; i < m_numDescriptorWrites; ++i)
        {
            if (m_descriptorWrites[i].dstBinding == slot && m_descriptorWrites[i].pImageInfo)
            {
                writeIndex = i;
                // clear existing write
                memDelete(m_descriptorWrites[i].pImageInfo);
                break;
            }
        }

        if (writeIndex == (u32)-1)
            writeIndex = m_numDescriptorWrites++;

        m_descriptorWrites[writeIndex] = write;
        m_numChangedDescriptorWrites++;

        if (texture)
        {
            VulkanTextureInternal* t = texture->getBindingTexture();

            VulkanImageTransition transition = {};
            transition.m_texture = t;
            transition.m_aspect = translateImageAspect(texture->m_desc.m_format.m_channels);
            transition.m_mips = texture->m_desc.m_mips;
            transition.m_target = (flags & e_bindFlags::storage) ? VK_IMAGE_LAYOUT_GENERAL : VK_IMAGE_LAYOUT_SHADER_READ_ONLY_OPTIMAL;
            m_imageTransitions.push_back(transition);
        }
    }

    void VulkanContext::updateBuffer(Buffer buf, const Mem& data, size_t offset)
    {
        VulkanBuffer* vkBuf = (VulkanBuffer*)buf;
        
        VkResult result;
        int bufferCount = 1;
        int currentBuffer = 0;
        if (vkBuf->m_active)
        {
            currentBuffer = vkBuf->m_buffered ? s_bb : 0;
        }
        else
        {
            bufferCount = vkBuf->m_buffered ? s_numBufferedResources : 1;
            currentBuffer = 0;
        }

        VkCommandBuffer commandBuffer = VK_NULL_HANDLE;
        VulkanBuffer* staging = nullptr;
        if (vkBuf->m_access & VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT)
        {
            BufferDesc desc;
            desc.m_data = data;
            desc.m_access = e_accessFlags::write;
            desc.m_usage = e_bufferUsageFlags::transferSrc;
            staging = (VulkanBuffer*)createBuffer(desc);
            commandBuffer = beginSingleTimeCommands(m_transferCommandPool);
        }

        for (int i = currentBuffer; i < currentBuffer + bufferCount; ++i)
        {
            // resize the buffer if necessary
            if (data.m_size > vkBuf->m_size[i])
            {
                vkDestroyBuffer(m_device, vkBuf->m_buffer[i], m_allocationCallbacks);
                vkFreeMemory(m_device, vkBuf->m_memory[i], m_allocationCallbacks);
                VkBufferCreateInfo bufferInfo = {};
                bufferInfo.sType = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO;
                bufferInfo.size = data.m_size;
                bufferInfo.sharingMode = VK_SHARING_MODE_EXCLUSIVE;
                bufferInfo.usage = vkBuf->m_usage;

                CHECK_RESULT((result = vkCreateBuffer(m_device, &bufferInfo, m_allocationCallbacks, &vkBuf->m_buffer[i])) == VK_SUCCESS);

                VkMemoryRequirements memRequirements;
                vkGetBufferMemoryRequirements(m_device, vkBuf->m_buffer[i], &memRequirements);

                VkMemoryAllocateInfo allocInfo = {};
                allocInfo.sType = VK_STRUCTURE_TYPE_MEMORY_ALLOCATE_INFO;
                allocInfo.allocationSize = memRequirements.size;
                allocInfo.memoryTypeIndex = findMemoryType(memRequirements.memoryTypeBits, 0, data.m_size);
                CHECK_RESULT((result = vkAllocateMemory(m_device, &allocInfo, m_allocationCallbacks, &vkBuf->m_memory[i])) == VK_SUCCESS);

                vkBindBufferMemory(m_device, vkBuf->m_buffer[i], vkBuf->m_memory[i], 0);
                vkBuf->m_size[i] = data.m_size;
            }

            if (vkBuf->m_access & VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT)
            {
                // buffer is device local so has to be updated from a staging buffer
                VkBufferCopy copy = {};
                copy.size = data.m_size;
                copy.srcOffset = offset;
                copy.dstOffset = offset;
                vkCmdCopyBuffer(commandBuffer, staging->m_buffer[0], vkBuf->m_buffer[i], 1, &copy);
            }
            else
            {
                // buffer is host visible so can be updated using cpu mapping
                void* mapped;
                CHECK_RESULT((result = vkMapMemory(m_device, vkBuf->m_memory[i], offset, data.m_size, 0, &mapped)) == VK_SUCCESS);
                memcpy(mapped, data.m_data, data.m_size);
                vkUnmapMemory(m_device, vkBuf->m_memory[i]);
            }
        }

        if (vkBuf->m_access & VK_MEMORY_PROPERTY_DEVICE_LOCAL_BIT)
        {
            endSingleTimeCommands(m_queues[QueueAlias::transfer], m_transferCommandPool, commandBuffer);
            destroyBuffer(staging);
        }
    }

    void VulkanContext::updateTexture(Texture tex, const TextureUpdateDesc& desc)
    {
        VulkanTexture* vkTex = (VulkanTexture*)tex;

        BufferDesc bufferDesc;
        bufferDesc.m_data = desc.m_data;
        bufferDesc.m_access = e_accessFlags::write;
        bufferDesc.m_usage = e_bufferUsageFlags::transferSrc;
        VulkanBuffer* buffer = (VulkanBuffer*)createBuffer(bufferDesc);

        u32 count = (vkTex->m_desc.m_buffered && !vkTex->m_texture->m_active) ? s_numBufferedResources : 1;
        u32 first = (vkTex->m_desc.m_buffered && vkTex->m_texture->m_active) ? s_bb : 0;

        VkCommandBuffer commandBuffer = beginSingleTimeCommands(m_transferCommandPool);
        for (u32 i = first; i < first + count; ++i)
        {
            VkImageAspectFlags aspect = translateImageAspect(vkTex->m_desc.m_format.m_channels);
            transitionImageLayout(e_queueAlias::transfer, commandBuffer, vkTex->m_texture->m_image[i], aspect, vkTex->m_desc.m_layers, vkTex->m_desc.m_mips, vkTex->m_texture->m_currentLayout[i], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL);
            VkBufferImageCopy bufferCopy = {};
            bufferCopy.bufferOffset = 0;
            bufferCopy.bufferRowLength = 0;
            bufferCopy.bufferImageHeight = 0;
            bufferCopy.imageSubresource.aspectMask = aspect;
            bufferCopy.imageSubresource.mipLevel = desc.m_mip;
            bufferCopy.imageSubresource.baseArrayLayer = 0;
            bufferCopy.imageSubresource.layerCount = vkTex->m_desc.m_layers;
            bufferCopy.imageOffset = { (int)desc.x, (int)desc.y, (int)desc.z };
            bufferCopy.imageExtent = { desc.m_width, desc.m_height, desc.m_depth };
            vkCmdCopyBufferToImage(commandBuffer, buffer->frontBuffer(), vkTex->m_texture->m_image[i], VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, 1, &bufferCopy);
            transitionImageLayout(e_queueAlias::transfer, commandBuffer, vkTex->m_texture->m_image[i], aspect, vkTex->m_desc.m_layers, vkTex->m_desc.m_mips, VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL, vkTex->m_texture->m_desiredLayout);
            vkTex->m_texture->m_currentLayout[i] = vkTex->m_texture->m_desiredLayout;
        }
        endSingleTimeCommands(m_queues[QueueAlias::transfer], m_transferCommandPool, commandBuffer);

        destroyBuffer(buffer);

        vkTex->m_texture->m_resolved = false;
    }

    void VulkanContext::updateConstants(ShaderStageFlags stage, size_t size, size_t offset, void* data)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);

        VkShaderStageFlags s = translateShaderStage(stage);
        vkCmdPushConstants(m_boundCommandBuffer->m_commandBuffer, m_boundPipeline->m_pipelineLayout, s, (u32)offset, (u32)size, data);
    }

    void VulkanContext::insertBarrier(PipelineStage srcStage, PipelineStage dstStage, BarrierFlags barrierType)
    {
        VkMemoryBarrier barrier = {};
        barrier.sType = VK_STRUCTURE_TYPE_MEMORY_BARRIER;
        if (barrierType == BarrierFlags::global)
        {
            barrier.srcAccessMask = VK_ACCESS_MEMORY_WRITE_BIT;
            barrier.dstAccessMask = VK_ACCESS_MEMORY_READ_BIT;
        }
        else if (barrierType & BarrierFlags::texture)
        {
            barrier.srcAccessMask = VK_ACCESS_SHADER_WRITE_BIT;
            barrier.dstAccessMask = VK_ACCESS_SHADER_READ_BIT;
        }
        else if (barrierType & BarrierFlags::buffer)
        {
            barrier.srcAccessMask = VK_ACCESS_SHADER_WRITE_BIT;
            barrier.dstAccessMask = VK_ACCESS_UNIFORM_READ_BIT | VK_ACCESS_SHADER_READ_BIT;
        }
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
        vkCmdPipelineBarrier(
            m_boundCommandBuffer->m_commandBuffer,
            translatePipelineStage(srcStage),
            translatePipelineStage(dstStage),
            0,
            1, &barrier,
            0, nullptr,
            0, nullptr            
        );
    }

    void VulkanContext::copyTexture(Texture src, Texture dst, u32 srcLayer, u32 dstLayer)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE);
        VulkanTexture* srcTex = (VulkanTexture*)src;
        VulkanTexture* dstTex = (VulkanTexture*)dst;
        VulkanTextureInternal* srcInternal = srcTex->getBindingTexture();
        VulkanTextureInternal* dstInternal = dstTex->m_texture;

        {
            VkImageAspectFlags aspect = srcInternal->m_aspect;
            VkImageLayout target = VK_IMAGE_LAYOUT_TRANSFER_SRC_OPTIMAL;
            transitionImageLayout(e_queueAlias::graphics, m_boundCommandBuffer->m_commandBuffer, srcInternal->frontImage(), aspect, srcInternal->m_layers, srcTex->m_desc.m_mips, srcInternal->frontLayout(), target);
            srcInternal->setFrontLayout(target);
        }

        if (dstInternal->backLayout() != VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL && dstInternal->backLayout() != VK_IMAGE_LAYOUT_GENERAL)
        {
            VkImageAspectFlags aspect = dstInternal->m_aspect;
            VkImageLayout target = VK_IMAGE_LAYOUT_TRANSFER_DST_OPTIMAL;
            transitionImageLayout(e_queueAlias::graphics, m_boundCommandBuffer->m_commandBuffer, dstInternal->backImage(), aspect, dstInternal->m_layers, dstTex->m_desc.m_mips, dstInternal->backLayout(), target);
            dstInternal->setBackLayout(target);
        }

        VkImageBlit blit = {};
        blit.srcOffsets[0] = { 0, 0, 0 };
        blit.srcOffsets[1] = { (int)srcTex->m_desc.m_width, (int)srcTex->m_desc.m_height, (int)srcTex->m_desc.m_depth };
        blit.srcSubresource.baseArrayLayer = srcLayer;
        blit.srcSubresource.layerCount = 1;
        blit.srcSubresource.mipLevel = 0;
        blit.srcSubresource.aspectMask = srcInternal->m_aspect;
        blit.dstOffsets[0] = { 0, 0, 0 };
        blit.dstOffsets[1] = { (int)dstTex->m_desc.m_width, (int)dstTex->m_desc.m_height, (int)dstTex->m_desc.m_depth };
        blit.dstSubresource.baseArrayLayer = dstLayer;
        blit.dstSubresource.layerCount = 1;
        blit.dstSubresource.mipLevel = 0;
        blit.dstSubresource.aspectMask = dstInternal->m_aspect;

        vkCmdBlitImage(m_boundCommandBuffer->m_commandBuffer, srcInternal->frontImage(), srcInternal->frontLayout(), dstInternal->backImage(), dstInternal->backLayout(), 1, &blit, VK_FILTER_NEAREST);
    
        dstInternal->m_resolved = false;
    }

    void VulkanContext::generateMips(Texture texture)
    {
        VulkanTexture* tex = (VulkanTexture*)texture;
        VulkanTextureInternal* internal = tex->getBindingTexture();

        VkImageBlit blit = {};
        blit.srcOffsets[0] = { 0, 0, 0 };
        blit.srcSubresource.baseArrayLayer = 0;
        blit.srcSubresource.layerCount = 1;
        blit.srcSubresource.aspectMask = tex->m_texture->m_aspect;
        blit.dstOffsets[0] = { 0, 0, 0 };
        blit.dstSubresource.baseArrayLayer = 0;
        blit.dstSubresource.layerCount = 1;
        blit.dstSubresource.aspectMask = tex->m_texture->m_aspect;

        u32 level = 0;
        int width = tex->m_desc.m_width;
        int height = tex->m_desc.m_height;
        int depth = tex->m_desc.m_depth;

        VulkanCommandPool& pool = m_commandPools[0][m_currentSwapChainImage];
        VkCommandBuffer cmd = beginSingleTimeCommands(pool.m_commandPool);

        // transition the image to the general layout to facilitate blitting
        VkImage img = internal->frontImage();
        VkImageLayout current = internal->frontLayout();
        transitionImageLayout(e_queueAlias::transfer, cmd, img, tex->m_texture->m_aspect, 1, tex->m_desc.m_mips, current, VK_IMAGE_LAYOUT_GENERAL);
        internal->setFrontLayout(VK_IMAGE_LAYOUT_GENERAL);
        for (u32 i = 0; i < tex->m_desc.m_mips - 1; ++i)
        {
            blit.srcSubresource.mipLevel = level;
            blit.srcOffsets[1] = { width, height, depth };

            width = max(width / 2, 1);
            height = max(height / 2, 1);
            depth = max(depth / 2, 1);
            blit.dstSubresource.mipLevel = ++level;
            blit.dstOffsets[1] = { width, height, depth };
            vkCmdBlitImage(cmd, img, VK_IMAGE_LAYOUT_GENERAL, img, VK_IMAGE_LAYOUT_GENERAL, 1, &blit, VK_FILTER_LINEAR);
        }
        endSingleTimeCommands(m_queues[e_queueAlias::graphics], pool.m_commandPool, cmd);
    }

    void VulkanContext::clearDescriptorWrites()
    {
        for (u32 i = 0; i < m_numDescriptorWrites; ++i)
        {
            if (m_descriptorWrites[i].pBufferInfo)
            {
                memDelete(m_descriptorWrites[i].pBufferInfo);
            }

            if (m_descriptorWrites[i].pImageInfo)
            {
                memDelete(m_descriptorWrites[i].pImageInfo);
            }
        }
        m_numDescriptorWrites = 0;
        m_numChangedDescriptorWrites = 0;
    }

    void VulkanContext::bindDescriptorSet()
    {
        VkDescriptorSet set = m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].get(m_numChangedDescriptorWrites > 0);

        // update the descriptor set with all the required resources
        if (m_numChangedDescriptorWrites > 0)
        {
            VkWriteDescriptorSet mappedWrites[MAX_DESCRIPTOR_WRITES];
            u32 actualNumWrites = 0;
            for (u32 i = 0; i < m_numDescriptorWrites; ++i)
            {
                mappedWrites[actualNumWrites] = m_descriptorWrites[i];
                mappedWrites[actualNumWrites].dstSet = set;
                u32 dstBinding = -1;
                if (mappedWrites[actualNumWrites].pImageInfo)
                    dstBinding = m_boundPipeline->m_descriptorImageMappings[mappedWrites[actualNumWrites].dstBinding];
                else if (mappedWrites[actualNumWrites].pBufferInfo)
                    dstBinding = m_boundPipeline->m_descriptorBufferMappings[mappedWrites[actualNumWrites].dstBinding];
                if (dstBinding != (u32)-1)
                {
                    mappedWrites[actualNumWrites].dstBinding = dstBinding;
                    actualNumWrites++;
                }
            }
            if (actualNumWrites > 0)
            {
                vkUpdateDescriptorSets(m_device, actualNumWrites, mappedWrites, 0, nullptr);
            }
            m_numChangedDescriptorWrites = 0;
        }

        // bind the descriptor set
        VkPipelineBindPoint bind = (m_boundQueueType == e_queueAlias::graphics) ? VK_PIPELINE_BIND_POINT_GRAPHICS : VK_PIPELINE_BIND_POINT_COMPUTE;
        vkCmdBindDescriptorSets(m_boundCommandBuffer->m_commandBuffer, bind, m_boundPipeline->m_pipelineLayout, 0, 1, &set, 0, nullptr);
    }

    void VulkanContext::draw(u32 vertexCount, u32 startVertex)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);
        if (!m_renderPassStarted)
            beginRenderPassInternal();
        bindDescriptorSet();
        vkCmdDraw(m_boundCommandBuffer->m_commandBuffer, vertexCount, 1, startVertex, 0);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }

    void VulkanContext::drawIndexed(u32 indexCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset, u32 vertexOffset)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);
        if (!m_renderPassStarted)
            beginRenderPassInternal();
        bindDescriptorSet();
        m_boundIndexType = indexType;
        bindBuffer(indexBuffer, 0, e_bindFlags::indexBuffer, indexBufferOffset);
        vkCmdDrawIndexed(m_boundCommandBuffer->m_commandBuffer, indexCount, 1, 0, vertexOffset, 0);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }

    void VulkanContext::drawIndexedInstanced(u32 indexCount, u32 instanceCount, Buffer indexBuffer, IndexType indexType, u32 indexBufferOffset, u32 vertexOffset)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);
        if (!m_renderPassStarted)
            beginRenderPassInternal();
        bindDescriptorSet();
        m_boundIndexType = indexType;
        bindBuffer(indexBuffer, 0, e_bindFlags::indexBuffer, indexBufferOffset);
        vkCmdDrawIndexed(m_boundCommandBuffer->m_commandBuffer, indexCount, instanceCount, 0, vertexOffset, 0);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }
    
    void VulkanContext::drawIndirect(u32 drawCount, Buffer drawBuffer, u32 drawBufferOffset)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);
        if (!m_renderPassStarted)
            beginRenderPassInternal();
        bindDescriptorSet();
        vkCmdDrawIndirect(m_boundCommandBuffer->m_commandBuffer, ((VulkanBuffer*)drawBuffer)->frontBuffer(), drawBufferOffset, drawCount, 0);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }
    
    void VulkanContext::drawIndexedIndirect(u32 drawCount, Buffer indexBuffer, IndexType indexType, Buffer drawBuffer, u32 indexBufferOffset , u32 drawBufferOffset)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);
        if (!m_renderPassStarted)
            beginRenderPassInternal();
        bindDescriptorSet();
        m_boundIndexType = indexType;
        bindBuffer(indexBuffer, 0, e_bindFlags::indexBuffer, indexBufferOffset);
        vkCmdDrawIndexedIndirect(m_boundCommandBuffer->m_commandBuffer, ((VulkanBuffer*)drawBuffer)->frontBuffer(), drawBufferOffset, drawCount, 0);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }
    
    void VulkanContext::dispatchCompute(u32 grid[3], u32 numThreads[3])
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);

        // transition all sampled textures
        doEndRecordingImageTransitions();

        bindDescriptorSet();
        vkCmdDispatch(m_boundCommandBuffer->m_commandBuffer, grid[0], grid[1], grid[2]);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }

    void VulkanContext::dispatchComputeIndirect(Buffer drawBuffer, u32 offset)
    {
        ASSERT(m_boundCommandBuffer != VK_NULL_HANDLE && m_boundPipeline);

        // transition all sampled textures
        doEndRecordingImageTransitions();

        bindDescriptorSet();
        vkCmdDispatchIndirect(m_boundCommandBuffer->m_commandBuffer, ((VulkanBuffer*)drawBuffer)->frontBuffer(), offset);
        m_boundPipeline->m_descriptorSet[m_currentSwapChainImage].m_dirty = true;
    }

    void VulkanContext::endRenderEncoding()
    {
        ASSERT(m_boundQueueType == e_queueAlias::none || m_boundQueueType == e_queueAlias::graphics);
        
        // start render pass here if nothing else started the render pass
        if (!m_renderPassStarted)
            beginRenderPassInternal();

        vkCmdEndRenderPass(m_boundCommandBuffer->m_commandBuffer);

        // transition all sampled textures
        doEndRecordingImageTransitions();

        vkEndCommandBuffer(m_boundCommandBuffer->m_commandBuffer);
        clearDescriptorWrites();
        m_boundPipeline = nullptr;
        m_renderPassStarted = false;
    }

    void VulkanContext::doEndRecordingImageTransitions()
    {
        for (auto& t : m_imageTransitions)
        {
            VulkanTextureInternal* texture = t.m_texture;
            VkImageAspectFlags aspect = t.m_aspect;
            transitionImageLayout(e_queueAlias::graphics, m_boundCommandBuffer->m_commandBuffer, texture->frontImage(), aspect, texture->m_layers, t.m_mips, texture->frontLayout(), t.m_target);
            texture->setFrontLayout(t.m_target);
        }
        m_imageTransitions.clear();
    }

    void VulkanContext::endComputeEncoding()
    {
        ASSERT(m_boundQueueType == e_queueAlias::none || m_boundQueueType == e_queueAlias::compute);


        vkEndCommandBuffer(m_boundCommandBuffer->m_commandBuffer);
        m_boundPipeline = nullptr;
    }

    void VulkanContext::enqueue()
    {        
        int i = 0;
        for (i = 0; i < MAX_QUEUED_COMMAND_BUFFERS; ++i)
        {
            if (m_queuedCommandBuffers[i] == VK_NULL_HANDLE)
            {
                m_queuedCommandBuffers[i] = m_boundCommandBuffer->m_commandBuffer;
                m_queueSemaphore = m_boundCommandBuffer->m_sync;
                break;
            }
        }
        ASSERT(i < MAX_QUEUED_COMMAND_BUFFERS);
        m_boundCommandBuffer = nullptr;
    }

    Semaphore VulkanContext::commit(Semaphore* waits, size_t semCount, PipelineStage stage)
    {
        VkSubmitInfo info = {};
        info.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
        info.pCommandBuffers = m_queuedCommandBuffers;
        u32& count = info.commandBufferCount;
        for (count = 0; count < MAX_QUEUED_COMMAND_BUFFERS; ++count)
        {
            if (m_queuedCommandBuffers[count] == VK_NULL_HANDLE)
                break;
        }
        ASSERT(info.commandBufferCount > 0);

        std::vector<VkSemaphore> waitSemaphores;
        std::vector<VkPipelineStageFlags> waitStages;
        if (waits)
        {
            for(size_t i = 0; i < semCount; ++i)
            {   
                ((VulkanSync*)waits[i])->m_consumed = true;
                waitSemaphores.push_back(((VulkanSync*)waits[i])->m_semaphore);
                waitStages.push_back(translatePipelineStage(stage));
            }
        }
        if (m_boundFramebufferSemaphore != VK_NULL_HANDLE && !m_backBufferUsed)
        {
            m_backBufferUsed = true;
            waitSemaphores.push_back(m_boundFramebufferSemaphore);
            waitStages.push_back(VK_PIPELINE_STAGE_COLOR_ATTACHMENT_OUTPUT_BIT);
        }
        info.waitSemaphoreCount = (u32)waitSemaphores.size();
        info.pWaitSemaphores = waitSemaphores.size() > 0 ? waitSemaphores.data() : nullptr;
        info.pWaitDstStageMask = waitStages.size() > 0 ? waitStages.data() : nullptr;

        Semaphore signal = m_queueSemaphore;
        VkFence fence = m_queueSemaphore->m_fence;
        m_queueSemaphore->m_used = true;
        info.pSignalSemaphores = &m_queueSemaphore->m_semaphore;
        info.signalSemaphoreCount = 1;
        
        vkQueueSubmit(m_queues[m_boundQueueType], 1, &info, fence);
        m_boundQueueType = e_queueAlias::none;
        memset(m_queuedCommandBuffers, 0, sizeof(VkCommandBuffer) * MAX_QUEUED_COMMAND_BUFFERS);
        m_boundFramebufferSemaphore = VK_NULL_HANDLE;
        return signal;
    }

    void VulkanContext::present()
    {
        VkResult result;

        // swap chain presentation should wait on completion of all rendering for the frame
        std::vector<VkSemaphore> semaphores;
        for (int i = 0; i < 2; ++i)
        {
            VulkanCommandPool& pool = m_commandPools[i][m_currentSwapChainImage];
            for (size_t j = 0; j < pool.m_numCommandBuffers; ++j)
            {
                if (pool.m_commandBuffers[j].m_sync->m_used && !pool.m_commandBuffers[j].m_sync->m_consumed)
                {
                    semaphores.push_back(pool.m_commandBuffers[j].m_sync->m_semaphore);
                }
            }
        }

        // transition swap chain image to presentation layout
        if (m_swapChainTextures.m_texture->m_currentLayout[m_currentSwapChainImage] != VK_IMAGE_LAYOUT_PRESENT_SRC_KHR)
        {
            VulkanCommandPool& pool = m_commandPools[0][m_currentSwapChainImage];

            VkCommandBuffer command = VK_NULL_HANDLE;
            VkSemaphore semaphore = VK_NULL_HANDLE;
            VkFence fence = VK_NULL_HANDLE;
            for (size_t i = 0; i < pool.m_numCommandBuffers; ++i)
            {
                if (!pool.m_commandBuffers[i].m_used)
                {
                    command = pool.m_commandBuffers[i].m_commandBuffer;
                    semaphore = pool.m_commandBuffers[i].m_sync->m_semaphore;
                    fence = pool.m_commandBuffers[i].m_sync->m_fence;
                    pool.m_commandBuffers[i].m_sync->m_used = true;
                    pool.m_commandBuffers[i].m_used = true;
                    break;
                }
            }

            // grow command pool if there are no other free command buffers
            if (command == VK_NULL_HANDLE)
            {
                size_t prev = pool.m_numCommandBuffers;
                pool.grow(m_device, m_allocationCallbacks);
                command = pool.m_commandBuffers[prev].m_commandBuffer;
                semaphore = pool.m_commandBuffers[prev].m_sync->m_semaphore;
                fence = pool.m_commandBuffers[prev].m_sync->m_fence;
                pool.m_commandBuffers[prev].m_sync->m_used = true;
                pool.m_commandBuffers[prev].m_used = true;
            }

            VkCommandBufferBeginInfo cmd = {};
            cmd.sType = VK_STRUCTURE_TYPE_COMMAND_BUFFER_BEGIN_INFO;
            cmd.flags = VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT;
            CHECK_RESULT((result = vkBeginCommandBuffer(command, &cmd)) == VK_SUCCESS);
            transitionImageLayout(e_queueAlias::graphics, command, 
                                    m_swapChainTextures.m_texture->m_image[m_currentSwapChainImage],
                                    VK_IMAGE_ASPECT_COLOR_BIT, 1, 1,
                                    m_swapChainTextures.m_texture->m_currentLayout[m_currentSwapChainImage],
                                    VK_IMAGE_LAYOUT_PRESENT_SRC_KHR);
            CHECK_RESULT((result = vkEndCommandBuffer(command)) == VK_SUCCESS);

            VkSubmitInfo submit = {};
            submit.sType = VK_STRUCTURE_TYPE_SUBMIT_INFO;
            submit.pCommandBuffers = &command;
            submit.commandBufferCount = 1;
            submit.waitSemaphoreCount = 0;
            submit.pWaitSemaphores = nullptr;
            submit.pWaitDstStageMask = nullptr;
            submit.signalSemaphoreCount = 1;
            submit.pSignalSemaphores = &semaphore;
            ASSERT(submit.pSignalSemaphores);
            vkQueueSubmit(m_queues[e_queueAlias::graphics], 1, &submit, fence);
            semaphores.push_back(semaphore);
            m_swapChainTextures.m_texture->m_currentLayout[m_currentSwapChainImage] = VK_IMAGE_LAYOUT_PRESENT_SRC_KHR;
        }

        // present the swap chain
        VkPresentInfoKHR info = {};
        info.sType = VK_STRUCTURE_TYPE_PRESENT_INFO_KHR;
        info.waitSemaphoreCount = (u32)semaphores.size();
        info.pWaitSemaphores = (semaphores.size() > 0) ? semaphores.data() : nullptr;
        info.swapchainCount = 1;
        info.pSwapchains = &m_swapChain;
        info.pImageIndices = &m_currentSwapChainImage;
        info.pResults = nullptr;

        // present the swap chain image to the screen
        CHECK_RESULT((result = vkQueuePresentKHR(m_queues[QueueAlias::present], &info)) == VK_SUCCESS);

        // reset all the used descriptor sets
        for (auto& s : m_descriptorSets)
        {
            s->reset();
        }

        // destroy any resources that are no longer in use
        for (intptr_t i = m_destroys.size() - 1; i >= 0; --i)
        {
            if (m_destroys[i]->m_age >= (m_numSwapChainImages))
            {
                m_destroys[i]->destroy(m_device, m_allocationCallbacks);
                memDelete(m_destroys[i]);
                m_destroys.erase(m_destroys.begin() + i);
            }
            else
            {
                m_destroys[i]->m_age++;
            }
        }

        incrementWrapBufferedResources();
    }

    // async context clean up used command buffers without calling begin()
    void VulkanContext::flushCommandBuffers()
    {
        // reset all the used descriptor sets
        for (auto& s : m_descriptorSets)
        {
            s->reset();
        }

        VkResult result;
        for (int i = 0; i < 2; ++i)
        {
            VulkanCommandPool& pool = m_commandPools[i][m_currentSwapChainImage];
            for (size_t j = 0; j < pool.m_numCommandBuffers; ++j)
            {
                VulkanSync* sync = pool.m_commandBuffers[j].m_sync;
                if (!sync->m_used)
                    continue;


                if (isSignalled(sync))
                {
                    CHECK_RESULT((result = vkResetFences(m_device, 1, &sync->m_fence)) == VK_SUCCESS);
                    sync->m_consumed = false;
                    sync->m_used = false;

                    // recycle command buffer
                    pool.m_commandBuffers[j].m_used = false;
                    CHECK_RESULT((result = vkResetCommandBuffer(pool.m_commandBuffers[j].m_commandBuffer, VK_COMMAND_BUFFER_RESET_RELEASE_RESOURCES_BIT)) == VK_SUCCESS);
                }
            }
        }
    }

    bool VulkanContext::isSignalled(Semaphore semaphore)
    {
        VulkanSync* sync = (VulkanSync*)semaphore;
        return vkGetFenceStatus(m_device, sync->m_fence) == VK_SUCCESS;
    }

    TextureDesc VulkanContext::getTextureDesc(Texture tex)
    {
        VulkanTexture* vk = (VulkanTexture*)tex;
        return vk->m_desc;
    }

    TextureDesc VulkanContext::getBackbufferTextureDesc()
    {
        return m_swapChainTextures.m_desc;
    }

    static gfx::VulkanContext* s_mainContext;
    static gfx::VulkanContext* s_asyncContexts[MAX_ASYNC_CONTEXTS];
    void createMainContext(void* platform, FeatureFlags featureFlags)
    {
        // function also gets called when window is resized
        if (!s_mainContext)
        {
            s_mainContext = memNew<VulkanContext>();
            s_mainContext->setup(platform, featureFlags);

            // make sure all the async contexts are null before creating any
            memset(s_asyncContexts, 0, sizeof(s_asyncContexts[0]) * MAX_ASYNC_CONTEXTS);
        }
        else
        {
            u32 width = utils::getWindowWidth();
            u32 height = utils::getWindowHeight();

            s_mainContext->createSwapChain(width, height);
            s_mainContext->transitionSwapChainImages();
        }
    }

    Context* getMainContext()
    {
        return s_mainContext;
    }

    Context* getAsyncContext()
    {
        // search for existing async context for this thread
        std::thread::id threadID = std::this_thread::get_id();
        int first = INT_MAX;
        for (int i = 0; i < MAX_ASYNC_CONTEXTS; ++i)
        {
            if (s_asyncContexts[i] && s_asyncContexts[i]->m_threadID == threadID)
                return s_asyncContexts[i];

            if (!s_asyncContexts[i] && i < first)
                first = i;
        }
                
        ASSERT(first < INT_MAX); // no free async contexts remaining!!!

        // create a new async context for this thread id
        s_asyncContexts[first] = memNew<VulkanContext>();
        s_asyncContexts[first]->setupAsync(s_mainContext, threadID);
        return s_asyncContexts[first];
    }

    void destroyContext(Context* context)
    {
        if (context == s_mainContext)
        {
            // shutdown main context
            ((VulkanContext*)context)->shutdown();
            memDelete(context);
        }
        else
        {
            // shutdown async context
            for (int i = 0; i < MAX_ASYNC_CONTEXTS; ++i)
            {
                if (((VulkanContext*)context)->m_threadID == s_asyncContexts[i]->m_threadID)
                {
                    s_asyncContexts[i]->shutdownAsync();
                    memDelete(s_asyncContexts[i]);
                    s_asyncContexts[i] = nullptr;
                    break;
                }
            }
        }
    }

    std::string getAPI()
    {
        return "vulkan";
    }
}
