#include "audio.h"

#include "platform.h"

#include <atomic>
#include <mutex>
#include <dsound.h>

#define MAX_CONCURRENT_STREAMING_SOUNDS 128
#define STREAMING_TARGET_RATE 16

namespace utils
{ 
	extern HWND getHWND();
}

namespace
{
	audio::AudioDevice	s_devices[16];
	std::atomic<int>	s_numDevices = { 0 };

	BOOL CALLBACK DSEnumCallback(LPGUID pGUID, LPCTSTR pDesc, LPCTSTR pName, LPVOID pDevice)
	{
		LPDIRECTSOUND* device = (LPDIRECTSOUND*)pDevice;

		// use the primary device
		audio::AudioDevice d;
		if (!pGUID)
		{
			HRESULT hr = DirectSoundCreate(pGUID, device, NULL);
			ASSERT(hr == S_OK);

			// set cooperative level
			HWND hwnd = utils::getHWND();
			(*device)->SetCooperativeLevel(hwnd, DSSCL_PRIORITY);

			d.m_name = "Default Device";
			d.m_device = nullptr;
		}
		else
		{
			d.m_device = memNew<GUID>();
			*((GUID*)d.m_device) = *pGUID;
			d.m_name = pDesc;
		}

		s_devices[s_numDevices++] = d;
		return TRUE;
	}

	size_t copyAudioData(void* dst, void* src, size_t loopStartOffset, size_t loopEndOffset, size_t size, size_t offset, bool loop)
	{
		size_t returnOffset = offset + size;

		size_t size0 = size;
		size_t size1 = 0;
		if(offset + size0 > loopEndOffset)
		{
			// resize copy to only remaining data
			size0 = loopEndOffset - offset;
			if (loop)
			{
				// copy a second chunk from the beginning of the data
				size1 = size - size0;
				returnOffset = loopStartOffset + size1;
			}
			else
			{
				// memset everything after the data to 0
				memset((void*)((intptr_t)dst + size0), 0, size - size0);
				returnOffset = loopEndOffset;
			}
		}

		memcpy(dst, (void*)((intptr_t)src + offset), size0);
		if (size1 > 0)
		{
			// copy second chunk of data from start
			memcpy((void*)((intptr_t)dst + size0), (void*)((intptr_t)src + loopStartOffset), size1);
		}

		// return the new offset into the source data
		return returnOffset;
	}
}

namespace audio
{
	struct SoundInternal : public SoundDesc
	{
		AudioStreamingType      m_streamingType = e_audioStreamingType::nonStreaming;
		LPDIRECTSOUNDBUFFER     m_buffer = nullptr;
		WAVEFORMATEX		    m_format = {};
		float				    m_startTime = 0.0f;
		Mem					    m_streamingData;
		size_t				    m_streamingSrcOffset = 0;
		size_t				    m_streamingDstOffset = 0;
		size_t				    m_bufferSize = 0;
        std::recursive_mutex    m_streamingMutex;
		bool				    m_streamingPlayheadLooped = false;
        size_t                  m_loopStartOffset = 0;
        size_t                  m_loopEndOffset = 0;
	};

	SoundInternal*		s_streamingSounds[MAX_CONCURRENT_STREAMING_SOUNDS];
	Thread				s_streamingThread = nullptr;
	std::atomic_bool	s_closeStreamingThread = false;

    void updateStreamingSoundData(SoundInternal* s, size_t cpysize)
    {
        // copy next chunk of data to sound buffer
        void* dst0 = nullptr;
        void* dst1 = nullptr;
        DWORD size0 = 0;
        DWORD size1 = 0;

        HRESULT hr = s->m_buffer->Lock((DWORD)s->m_streamingDstOffset, (DWORD)cpysize, &dst0, &size0, &dst1, &size1, 0);
        ASSERT(hr == S_OK);

        s->m_streamingSrcOffset = copyAudioData(dst0, s->m_streamingData.m_data, s->m_loopStartOffset, s->m_loopEndOffset, size0, s->m_streamingSrcOffset, s->m_loopType == e_audioLoopType::looping);
        if (dst1)
        {
            // need to copy to wrapped around section
            s->m_streamingSrcOffset = copyAudioData(dst1, s->m_streamingData.m_data, s->m_loopStartOffset, s->m_loopEndOffset, size1, s->m_streamingSrcOffset, s->m_loopType == e_audioLoopType::looping);
        }

        hr = s->m_buffer->Unlock(dst0, size0, dst1, size1);
        ASSERT(hr == S_OK);
    }

	void* streamingUpdateThread(void* data)
	{
		while (!s_closeStreamingThread.load())
		{
			u64 timeBefore = utils::getSystemTimeMs();
			for (int i = 0; i < MAX_CONCURRENT_STREAMING_SOUNDS; ++i)
			{
				if (s_streamingSounds[i])
				{
                    {
                        auto s = s_streamingSounds[i];
                        std::lock_guard<std::recursive_mutex>lg(s->m_streamingMutex);

                        DWORD play = 0;
                        DWORD write = 0;
                        s->m_buffer->GetCurrentPosition(&play, &write);

                        size_t cpysize = 0;
                        if (play < s->m_streamingDstOffset)
                        {
                            // playhead has looped back to the start of the buffer
                            size_t toEnd = s->m_bufferSize - s->m_streamingDstOffset;
                            cpysize = toEnd + play;
                        }
                        else if (play > s->m_streamingDstOffset)
                        {
                            cpysize = (size_t)play - s->m_streamingDstOffset;
                        }

                        if (cpysize > 0)
                        {
                            if (s->m_loopType != e_audioLoopType::looping && s->m_streamingSrcOffset >= s->m_loopEndOffset)
                            {
                                if (!s->m_streamingPlayheadLooped && play < s->m_streamingDstOffset)
                                {
                                    // track that the playhead has looped for the last time
                                    s->m_streamingPlayheadLooped = true;

                                    // lock and zero everything in the buffer from the playhead onwards
                                    size_t zeroSize = s->m_bufferSize - s->m_streamingDstOffset;
                                    void* dst = nullptr;
                                    DWORD size = 0;
                                    s->m_buffer->Lock((DWORD)s->m_streamingDstOffset, (DWORD)zeroSize, &dst, &size, nullptr, nullptr, 0);
                                    memset(dst, 0, (size_t)size);
                                    s->m_buffer->Unlock(dst, size, nullptr, 0);
                                }
                                else if (s->m_streamingPlayheadLooped && play >= s->m_streamingDstOffset)
                                {
                                    // stop playing if we've reached the end of audio data
                                    s->m_buffer->Stop();
                                }
                                continue;
                            }

                            updateStreamingSoundData(s, cpysize);
							s->m_streamingDstOffset = play;
                        }
                    }
				}
			}
			u64 processingDuration = utils::getSystemTimeMs() - timeBefore;

			utils::sleepMs(1);
		}
		return nullptr;
	}


	class DirectSoundAudioManager : public AudioManager
	{
	public:
		LPDIRECTSOUND				m_device = nullptr;
		
		void setup() override
		{
			DirectSoundEnumerate(DSEnumCallback, &m_device);

			s_streamingThread = utils::createThread(streamingUpdateThread, nullptr, utils::e_threadFlags::detached);
		}

		void shutdown() override
		{
			for (int i = 0; i < s_numDevices.load(); ++i)
			{
				if (s_devices[i].m_device)
				{
					memDelete(s_devices[i].m_device);
				}
				s_devices[i].m_name = "";
			}
			m_device->Release();
			m_device = nullptr;

			s_closeStreamingThread = true;
		}

		void update(float dt) override
		{
		}

		void getAudioDevices(AudioDevice*& devices, int& numDevices) override
		{
			devices = s_devices;
			numDevices = s_numDevices.load();
		}

		void setAudioDevice(const AudioDevice& device) override
		{
			// stop any playing sounds

			// shut down any existing device
			if (m_device)
			{
				m_device->Release();
				m_device = nullptr;
			}

			// set new sound device
			LPGUID pGUID = (LPGUID)device.m_device;
			HRESULT hr = DirectSoundCreate(pGUID, &m_device, NULL);
			ASSERT(hr == S_OK);

			HWND hwnd = utils::getHWND();
			m_device->SetCooperativeLevel(hwnd, DSSCL_PRIORITY);

			// restart any sounds that were playing			
		}

		Sound loadSound(const SoundDesc& desc) override
		{
			HRESULT hr = {};

			SoundInternal* sound = memNew<SoundInternal>();
			*(SoundDesc*)sound = desc;
			sound->m_startTime = desc.m_startTime;
			sound->m_streamingType = desc.m_streamingType;
			sound->m_loopType = desc.m_loopType;

			WAVEFORMATEX wfx = {};
			wfx.wFormatTag = WAVE_FORMAT_PCM;
			wfx.nChannels = desc.m_channels;
			wfx.nSamplesPerSec = desc.m_samplesPerSecond;
			wfx.wBitsPerSample = desc.m_bitsPerSample;
			wfx.nBlockAlign = ((wfx.wBitsPerSample / 8) * wfx.nChannels);
			wfx.nAvgBytesPerSec = wfx.nBlockAlign * wfx.nSamplesPerSec;

			DSBUFFERDESC dsbd = {};
			dsbd.dwSize = (DWORD)sizeof(DSBUFFERDESC);
			dsbd.dwFlags = DSBCAPS_CTRLPAN | DSBCAPS_CTRLVOLUME | DSBCAPS_CTRLFREQUENCY | DSBCAPS_GLOBALFOCUS;
			dsbd.lpwfxFormat = &wfx;

			// non streaming buffers will be sized to hold full data
			if (desc.m_streamingType == e_audioStreamingType::nonStreaming)
				dsbd.dwBufferBytes = (DWORD)desc.m_data.m_size;
			else if (desc.m_streamingType == e_audioStreamingType::streaming)
				dsbd.dwBufferBytes = 3 * wfx.nAvgBytesPerSec;
			sound->m_bufferSize = (size_t)dsbd.dwBufferBytes;

            sound->m_loopStartOffset = (size_t)(desc.m_loopStartTime * wfx.nAvgBytesPerSec);
            if(desc.m_loopEndTime < 0.0f)
                sound->m_loopEndOffset = desc.m_data.m_size;
            else
                sound->m_loopEndOffset = (size_t)(desc.m_loopEndTime * wfx.nAvgBytesPerSec);
            
			// create sound buffer
			LPDIRECTSOUNDBUFFER buffer = {};
			hr = m_device->CreateSoundBuffer(&dsbd, &buffer, NULL);
			ASSERT(hr == S_OK);

			// load initial data into buffer
			DWORD lflags = DSBLOCK_ENTIREBUFFER;
			DWORD lsize = 0;
			size_t cpysize = desc.m_data.m_size;
			if (desc.m_streamingType == e_audioStreamingType::streaming)
			{
				// fill the buffer with the first 3 seconds of audio data
				lflags = 0;
				lsize = (DWORD)sound->m_bufferSize;
				cpysize = lsize;
			}

			LPVOID data = nullptr;
			DWORD size = 0;
			hr = buffer->Lock(0, lsize, &data, &size, nullptr, nullptr, lflags);
			ASSERT(hr == S_OK);

			memcpy(data, desc.m_data.m_data, cpysize);

			hr = buffer->Unlock(data, size, nullptr, 0);
			ASSERT(hr == S_OK);

			sound->m_buffer = buffer;
			sound->m_format = wfx;

			if (desc.m_streamingType == e_audioStreamingType::streaming)
			{
				sound->m_streamingSrcOffset = cpysize;

				// take a copy of the audio data to continually stream
				sound->m_streamingData.m_size = desc.m_data.m_size;
				sound->m_streamingData.m_data = memAlloc(desc.m_data.m_size);
				memcpy(sound->m_streamingData.m_data, desc.m_data.m_data, desc.m_data.m_size);

				// track streaming sounds to continuously new data
				for (int i = 0; i < MAX_CONCURRENT_STREAMING_SOUNDS; ++i)
				{
					if (!s_streamingSounds[i])
					{
						s_streamingSounds[i] = sound;
						break;
					}
				}
			}

			return sound;
		}

		void unloadSound(Sound sound) override
		{
			// make sure sound is stopped
			stopSound(sound);

			// destroy sound buffer
			SoundInternal* si = (SoundInternal*)sound;

			// untrack streaming sound and delete stored data
			if (si->m_streamingType == e_audioStreamingType::streaming)
			{
				for (int i = 0; i < MAX_CONCURRENT_STREAMING_SOUNDS; ++i)
				{
					if (s_streamingSounds[i] == si)
					{
						s_streamingSounds[i] = nullptr;
						break;
					}
				}

				{
					std::lock_guard<std::recursive_mutex>lg(si->m_streamingMutex);
					memFree(si->m_streamingData.m_data);
				}
			}

			si->m_buffer->Release();
			memDelete(sound);
		}

		void playSound(Sound sound) override
		{
			HRESULT hr = {};

			SoundInternal* si = (SoundInternal*)sound;

            // seek to the start time before playing
            seekSound(si, si->m_startTime);

			// play the sound
			DWORD flags = (si->m_loopType == e_audioLoopType::looping) ? DSBPLAY_LOOPING : 0;

			// streaming buffers always use the looping flag
			if (si->m_streamingType == e_audioStreamingType::streaming)
				flags = DSBPLAY_LOOPING;
			hr = si->m_buffer->Play(0, 0, flags);
			ASSERT(hr == S_OK);
		}

		void setSoundPaused(Sound sound, bool paused) const override
		{
			SoundInternal* si = (SoundInternal*)sound;
			if (paused)
			{
				si->m_buffer->Stop();
			}
			else
			{
                // play the sound
                DWORD flags = (si->m_loopType == e_audioLoopType::looping) ? DSBPLAY_LOOPING : 0;
                if (si->m_streamingType == e_audioStreamingType::streaming)
                    flags = DSBPLAY_LOOPING;
				si->m_buffer->Play(0, 0, flags);
			}
		}

		void seekSound(Sound sound, float t) const override
		{
			SoundInternal* si = (SoundInternal*)sound;
			u64 offset = (u64)(si->m_format.nAvgBytesPerSec * t);
			
            if(si->m_streamingType == e_audioStreamingType::streaming)
            {
                {
                    std::lock_guard<std::recursive_mutex>lg(si->m_streamingMutex);
                    
                    // skip the seek if trying to seek to the current position
                    if(si->m_streamingSrcOffset == offset)
                        return;
                    
                    // stop the buffer for safety
                    si->m_buffer->Stop();
                    
                    // seek within streaming data buffer
                    si->m_streamingSrcOffset = offset;
                    
                    // get the write position for the buffer to start overwriting the data from
                    DWORD play = 0;
                    DWORD write = 0;
                    si->m_buffer->GetCurrentPosition(&play, &write);
                    si->m_streamingDstOffset = write;
                    
                    // update the streaming buffer memory from the write cursor wrapping round to the play cursor
                    size_t cpysize = (si->m_bufferSize - write) + play;
                    
                    updateStreamingSoundData(si, cpysize);
					si->m_streamingDstOffset = play;

                    // play the buffer again from the written position
                    si->m_buffer->SetCurrentPosition(write);
                    si->m_buffer->Play(0, 0, DSBPLAY_LOOPING);
                }
            }
            else
            {
                // non-streaming sounds can simply seekt to the correct position in the buffer
                si->m_buffer->SetCurrentPosition(offset);
            }
		}

		void stopSound(Sound sound) override
		{
			SoundInternal* si = (SoundInternal*)sound;

			// stop sound playing
			si->m_buffer->Stop();
		}
        
        float getSoundCurrentTime(Sound sound) override
        {
            SoundInternal* si = (SoundInternal*)sound;
            
            // get current play position
            DWORD play = 0;
            DWORD write = 0;
            si->m_buffer->GetCurrentPosition(&play, &write);
            
            if(si->m_streamingType == e_audioStreamingType::streaming)
            {
                // subtract the difference the play head and the last written end from the source offset
				size_t playDiff = 0;
				if (play > si->m_streamingDstOffset)
					playDiff = si->m_bufferSize - ((size_t)play - si->m_streamingDstOffset);
				else
					playDiff = si->m_streamingDstOffset - (size_t)play;

				if (playDiff > si->m_streamingSrcOffset)
				{
					return (si->m_streamingData.m_size - (playDiff - si->m_streamingSrcOffset)) / (float)si->m_format.nAvgBytesPerSec;
				}
				else
				{
					return (si->m_streamingSrcOffset - playDiff) / (float)si->m_format.nAvgBytesPerSec;
				}
            }
            else
            {
                return play / (float)si->m_format.nAvgBytesPerSec;
            }
        }
	};

	AudioManager* createAudioManager()
	{
		return memNew<DirectSoundAudioManager>();
	}

	void destroyAudioManager(AudioManager* manager)
	{
		memDelete(manager);
	}

	AudioManager* AudioManager::get()
	{
		static AudioManager* s_self = nullptr;
		if (!s_self)
			s_self = createAudioManager();
		return s_self;
	}
}
