#include "platform.h"

#include <windows.h>
#include <windowsx.h>
#include <fstream>
#include <thread>
#include <unordered_map>

#include "gfx.h"
#include "gfx_win32.h"
#include "app.h"
#include "defs.h"

using namespace utils;

namespace
{
    std::unordered_map<u8, Key> s_keyMappings = 
    {
        { VK_ESCAPE,    e_key::esc, },
        { VK_RETURN,    e_key::ret, },
        { VK_TAB,       e_key::tab, },
        { VK_SHIFT,     e_key::shift, },
        { VK_CONTROL,   e_key::ctrl, },
        { VK_DELETE,    e_key::del, },
        { VK_SPACE,     e_key::space, },
        { VK_UP,        e_key::up, },
        { VK_DOWN,      e_key::down, },
        { VK_LEFT,      e_key::left, },
        { VK_RIGHT,     e_key::right, },
    };

    struct Mouse
    {
        int x, y;
        bool m_buttonState[MouseButton::count][MouseButtonState::count];
        bool m_prevButtonState[MouseButton::count][MouseButtonState::count];
        u64  m_pressedTime[MouseButton::count];
    };

    struct Keyboard
    {
        bool m_keyState[Key::count][KeyState::count];
        bool m_prevKeyState[Key::count][KeyState::count];
    };

    struct Context
    {
        INT                     m_nCmdShow = 0;
        HINSTANCE               m_hinstance = NULL;
        HWND                    m_hwnd = NULL;
        RECT                    m_dims = {};
        App*                    m_app = nullptr;
        bool                    m_recreatedHwnd = false;
        size_t                  m_frame = 0;
        double                  m_ticksToMs = 0.0;
        double                  m_ticksToUs = 0.0;
        double                  m_ticksToNs = 0.0;
        std::vector<InputEvent> m_inputEvents;
        Mouse                   m_mouse = {};
        Keyboard                m_keyboard = {};
        bool                    m_changeFullscreenState = false;
        bool                    m_fullscreen = false;
        bool                    m_minimised = false;
        RECT                    m_prefullscreenRect = {};
    };
    Context s_context;


    struct WinThread
    {
        std::thread m_thread;
        ThreadFlags m_flags = e_threadFlags::none;
    };
}


namespace utils
{
    HWND getHWND()
    {
        return s_context.m_hwnd;
    }

    u32 getWindowWidth()
    {
        return s_context.m_dims.right - s_context.m_dims.left;
    }

    u32 getWindowHeight()
    {
        return s_context.m_dims.bottom - s_context.m_dims.top;
    }

    size_t getFrame()
    {
        return s_context.m_frame;
    }

    void setCursorVisible(bool visible)
    {
        ShowCursor(visible);
    }

    void setWindowFullscreen(bool fullscreen)
    {
        // defer state change until state of next update
        s_context.m_changeFullscreenState = true;
        s_context.m_fullscreen = fullscreen;
    }

    void getMousePos(int& x, int& y)
    {
        x = s_context.m_mouse.x;
        y = s_context.m_mouse.y;
    }

    bool isMouseButtonPressed(MouseButton button)
    {
        return s_context.m_mouse.m_buttonState[button][MouseButtonState::pressed];
    }

    bool isMouseButtonDown(MouseButton button)
    {
        return s_context.m_mouse.m_buttonState[button][MouseButtonState::down];
    }

    bool isKeyPressed(Key key)
    {
        return s_context.m_keyboard.m_keyState[key][KeyState::pressed];
    }

    bool isKeyDown(Key key)
    {
        return s_context.m_keyboard.m_keyState[key][KeyState::down];
    }

    bool fileExists(const char* path)
    {
        std::ifstream file(path, std::ios::ate | std::ios::binary);
        if(!file.is_open())
            return false;
        file.close();
        return true;
    }

    void readFileToBuffer(const char* path, void*& buf, size_t& len)
    {
        std::ifstream file(path, std::ios::ate | std::ios::binary);
        if (!file.is_open())
            return;
        len = (size_t)file.tellg();
        char* c = new char[len];
        file.seekg(0);
        file.read(c, len);
        file.close();
        buf = (void*)c;
    }

    void writeFileFromBuffer(const char* path, void* buf, size_t len)
    {
        std::ofstream file(path, std::ios::out);
        if (!file.is_open())
            return;
        char* c = (char*)buf;
        file.write(c, len);
        file.close();
    }

    void freeFileBuffer(void* buf)
    {
        delete[] buf;
    }

    void createDirectory(const char* path)
    {
        CreateDirectory(path, nullptr);
    }

    int runCommand(const char* command)
    {
        return system(command);
    }

    Thread createThread(ThreadFunc func, void* data, ThreadFlags flags)
    {
       WinThread* thread = memNew<WinThread>();
       thread->m_thread = std::thread(func, data);
       thread->m_flags = flags;
        if (flags & e_threadFlags::detached)
            thread->m_thread.detach();
        return (Thread)thread;
    }

    void destroyThread(Thread thread)
    {
        WinThread* t = (WinThread*)thread;
        if (!(t->m_flags & e_threadFlags::detached))
            t->m_thread.join();
        memDelete(t);
    }

    u64 getSystemTimeMs()
    {
        LARGE_INTEGER t = {};
        QueryPerformanceCounter(&t);
        return (u64)((double)t.QuadPart * s_context.m_ticksToMs);
    }

    void sleepMs(u64 ms)
    {
        Sleep((DWORD)ms);
    }

    void onWindowResize(HWND hwnd, HINSTANCE hinstance)
    {
        GetClientRect(s_context.m_hwnd, &s_context.m_dims);

        if (getWindowWidth() == 0 || getWindowHeight() == 0)
        {
            s_context.m_minimised = true;
            return;
        }
        s_context.m_minimised = false;

        gfx::PlatformGFX platform = {
            hwnd,
            hinstance
        };

        gfx::createMainContext(&platform, s_context.m_app->getGFXFeatureFlags());

        if (s_context.m_app)
        {
            s_context.m_app->resize(getWindowWidth(), getWindowHeight());
        }
    }

    LRESULT CALLBACK WindowProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg)
        {
        case WM_KEYDOWN:
            if (s_keyMappings.find((Key)wParam) != s_keyMappings.end())
            {
                s_context.m_keyboard.m_keyState[s_keyMappings[wParam]][KeyState::down] = true;
                s_context.m_keyboard.m_keyState[s_keyMappings[wParam]][KeyState::up] = false;
            }
            break;
        case WM_KEYUP:
            if (s_keyMappings.find((Key)wParam) != s_keyMappings.end())
            {
                s_context.m_keyboard.m_keyState[s_keyMappings[wParam]][KeyState::down] = false;
                s_context.m_keyboard.m_keyState[s_keyMappings[wParam]][KeyState::up] = true;
            }
            break;
        case WM_LBUTTONDOWN:
            s_context.m_mouse.m_buttonState[MouseButton::left][MouseButtonState::down] = true;
            s_context.m_mouse.m_buttonState[MouseButton::left][MouseButtonState::up] = false;
            break;
        case WM_LBUTTONUP:
        {
            InputEvent up = {};
            up.m_type = e_inputEventType::mouse;
            up.m_mouse.m_type = e_mouseEventType::button;
            up.m_mouse.m_button.m_button = e_mouseButton::left;
            up.m_mouse.m_button.m_buttonState = e_mouseButtonState::up;
            s_context.m_inputEvents.push_back(up);

            // detect clicked
            if (s_context.m_mouse.m_pressedTime[MouseButton::left] <= 100)
            {
                s_context.m_mouse.m_buttonState[MouseButton::left][MouseButtonState::clicked] = true;

                InputEvent clicked = {};
                clicked.m_type = e_inputEventType::mouse;
                clicked.m_mouse.m_type = e_mouseEventType::button;
                clicked.m_mouse.m_button.m_button = e_mouseButton::left;
                clicked.m_mouse.m_button.m_buttonState = e_mouseButtonState::clicked;
                s_context.m_inputEvents.push_back(clicked);
            }

            s_context.m_mouse.m_buttonState[MouseButton::left][MouseButtonState::down] = false;
            s_context.m_mouse.m_buttonState[MouseButton::left][MouseButtonState::up] = true;
        }
            break;
        case WM_MOUSEMOVE:
        {
            s_context.m_mouse.x = GET_X_LPARAM(lParam);
            s_context.m_mouse.y = GET_Y_LPARAM(lParam);
        }
            break;
        case WM_DESTROY:
            PostQuitMessage(0);
            break;
        case WM_SIZE:
            onWindowResize(s_context.m_hwnd, s_context.m_hinstance);
            break;
        default:
            return DefWindowProc(hwnd, uMsg, wParam, lParam);
            break;
        }
        return S_OK;
    }

    HWND createWindow()
    {
        HWND prevHwnd = s_context.m_hwnd;
        WNDCLASS wc = {};
        wc.lpfnWndProc = WindowProc;
        wc.hInstance = GetModuleHandle(NULL);
        wc.lpszClassName = "Window";

        RegisterClass(&wc);

        s_context.m_hwnd = CreateWindow("Window", "Window", WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL, s_context.m_hinstance, NULL);
        ShowWindow(s_context.m_hwnd, s_context.m_nCmdShow);

        if (prevHwnd)
        {
            DestroyWindow(prevHwnd);
            s_context.m_recreatedHwnd = true;
        }

        return s_context.m_hwnd;
    }
}


INT WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, PSTR pCmdLine, INT nCmdShow)
{
    AppDesc desc = {};
    s_context.m_app = createApp(desc);

    s_context.m_hinstance = hInstance;
    s_context.m_nCmdShow = nCmdShow;
    createWindow();

    // setup tick to system time conversions
    LARGE_INTEGER freq;
    QueryPerformanceFrequency(&freq);
    s_context.m_ticksToMs = (double)(1.0 / (freq.QuadPart / 1000.0));
    s_context.m_ticksToUs = s_context.m_ticksToMs * 1000.0;
    s_context.m_ticksToNs = s_context.m_ticksToUs * 1000.0;

    s_context.m_mouse.x = 0;
    s_context.m_mouse.y = 0;
    memset(s_context.m_mouse.m_buttonState, 0, sizeof(bool) * MouseButton::count * MouseButtonState::count);

    // setup some key mappings
    for (u8 i = 0x30; i < 0x3A; ++i)
        s_keyMappings[i] = (Key)(Key::zero + (i - 0x30));
    for (u8 i = 0x41; i < 0x5B; ++i)
        s_keyMappings[i] = (Key)(Key::a + (i - 0x41));
    memset(s_context.m_keyboard.m_keyState, 0, sizeof(bool) * Key::count * KeyState::count);

    gfx::Context* ctx = gfx::getMainContext();

    s_context.m_app->setup();

    MSG msg = {};

    u64 prevTime = getSystemTimeMs();
    while (true)
    {
        // resize window if fullscreen state change has been requested
        if (s_context.m_changeFullscreenState)
        {
            if (s_context.m_fullscreen)
            {
                GetWindowRect(s_context.m_hwnd, &s_context.m_prefullscreenRect);
                SetWindowLongPtr(s_context.m_hwnd, GWL_STYLE, WS_POPUP | WS_VISIBLE);
                ShowWindow(s_context.m_hwnd, SW_MAXIMIZE);
            }
            else
            {
                int x = s_context.m_prefullscreenRect.left;
                int y = s_context.m_prefullscreenRect.top;
                int w = s_context.m_prefullscreenRect.right - s_context.m_prefullscreenRect.left;
                int h = s_context.m_prefullscreenRect.bottom - s_context.m_prefullscreenRect.top;
                SetWindowLongPtr(s_context.m_hwnd, GWL_STYLE, WS_OVERLAPPEDWINDOW | WS_VISIBLE);
                ShowWindow(s_context.m_hwnd, s_context.m_nCmdShow);
                MoveWindow(s_context.m_hwnd, x, y, w, h, TRUE);
            }
            s_context.m_changeFullscreenState = false;
        }

        u64 currentTime = getSystemTimeMs();
        u64 dt = currentTime - prevTime;
        prevTime = currentTime;

        s_context.m_inputEvents.clear();

        // copy prev mouse button and key state
        memcpy(s_context.m_mouse.m_prevButtonState, s_context.m_mouse.m_buttonState, sizeof(bool) * MouseButton::count * MouseButtonState::count);
        memcpy(s_context.m_keyboard.m_prevKeyState, s_context.m_keyboard.m_keyState, sizeof(bool) * Key::count * KeyState::count);

        // clear mouse button clicked
        for (int i = 0; i < MouseButton::count; ++i)
        {
            s_context.m_mouse.m_buttonState[i][e_mouseButtonState::clicked] = false;
        }

        // always send a mouse position update event first
        InputEvent mouse = {};
        mouse.m_type = e_inputEventType::mouse;
        mouse.m_mouse.m_type = e_mouseEventType::move;
        mouse.m_mouse.m_pos.m_prevPosX = s_context.m_mouse.x;
        mouse.m_mouse.m_pos.m_prevPosY = s_context.m_mouse.y;
        s_context.m_inputEvents.push_back(mouse);

        if (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
        {
            if (msg.message == WM_QUIT)
            {
                // eat the quit message if the window has been recreated
                if (s_context.m_recreatedHwnd)
                    s_context.m_recreatedHwnd = false;
                else
                    break;
            }
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }

        // skip main update if minimised
        // TODO: handle minimised behaviour better eg. go through custom app pausing flow
        if (s_context.m_minimised)
            continue;

        s_context.m_inputEvents[0].m_mouse.m_pos.m_posX = s_context.m_mouse.x;
        s_context.m_inputEvents[0].m_mouse.m_pos.m_posY = s_context.m_mouse.y;

        // update pressed/clicked states
        for (int i = 0; i < Key::count; ++i)
        {
            //s_context.m_keyboard.m_keyState[i][e_keyState::pressed] = s_context.m_keyboard.m_keyState[i][e_keyState::down] && s_context.m_keyboard.m_prevKeyState[i][e_keyState::down];
        }

        for (int i = 0; i < MouseButton::count; ++i)
        {
            s_context.m_mouse.m_buttonState[i][e_mouseButtonState::pressed] = s_context.m_mouse.m_buttonState[i][e_mouseButtonState::down] && s_context.m_mouse.m_prevButtonState[i][e_keyState::down];
            if (s_context.m_mouse.m_buttonState[i][e_mouseButtonState::pressed])
                s_context.m_mouse.m_pressedTime[i] += dt;
            else
                s_context.m_mouse.m_pressedTime[i] = 0;
        }

        // send input events for changes in key state
        for (int i = 0; i < Key::count; ++i)
        {
            for (int j = 0; j < e_keyState::count; ++j)
            {
                if (s_context.m_keyboard.m_keyState[i][j] && !s_context.m_keyboard.m_prevKeyState[i][j])
                {
                    InputEvent e = {};
                    e.m_type = e_inputEventType::key;
                    e.m_key.m_key = (Key)i;
                    e.m_key.m_state = (KeyState)j;
                    s_context.m_inputEvents.push_back(e);
                }
            }
        }

        ctx->begin();
        s_context.m_app->processInput(s_context.m_inputEvents);
        s_context.m_app->update();
        s_context.m_app->render();
        ctx->present();

        s_context.m_frame++;
    }
    
    s_context.m_app->shutdown();
    memDelete(s_context.m_app);
    gfx::destroyContext(ctx);
    return 0;
}