#define TINYOBJLOADER_IMPLEMENTATION
#include <tiny_obj_loader.h>
#define GLM_ENABLE_EXPERIMENTAL
#include <glm\glm.hpp>
#include <glm/gtx/hash.hpp>
#include <glm/gtx/transform.hpp>
#include <glm/gtx/norm.hpp>
#include <pystring.h>
#include <unordered_map>
#include <algorithm>
#include <iostream>

struct AABB
{
    glm::vec4   m_min;
    glm::vec4   m_max;
    float       m_extentLength;
};

struct Vertex
{
    glm::vec4 m_pos;
    glm::vec3 m_normal;
    glm::vec2 m_tex;
    glm::vec4 m_colour;

    bool operator==(const Vertex& other) const
    {
        return m_pos == other.m_pos && m_normal == other.m_normal && m_tex == other.m_tex;
    }
};

struct Material
{

};

struct MeshInstance
{
    glm::mat4   m_matrix;
    uint32_t    m_materialIndex;
};

struct Mesh
{
    std::vector<uint32_t> m_indices;
    std::vector<Vertex> m_vertices;
    std::vector<glm::mat4> m_instanceMatrices;
    std::vector<std::vector<uint32_t>> m_instanceMaterials;
    AABB m_aabb;
};

namespace std
{
    template<> struct hash<Vertex>
    {
        size_t operator()(const Vertex& vertex) const
        {
            return (std::hash<glm::vec4>()(vertex.m_pos) ^ (std::hash<glm::vec3>()(vertex.m_normal) << 1) ^ (std::hash<glm::vec2>()(vertex.m_tex) << 2));
        }
    };
}
std::vector<Mesh> s_meshes;
std::vector<Material> s_materials;

int main(int argc, char *argv[])
{
    std::map<std::string, std::string> args;
    for (int i = 1; i < argc; i += 2)
    {
        std::string arg1 = std::string(argv[i]);
        std::string arg2 = std::string(argv[i + 1]);
        args[arg1] = arg2;
    }

    if (args.size() < 2)
    {
        printf("Not enough arguments!\n");
        return -1;
    }

    if (args.find("-m") == args.end())
        if (pystring::os::path::dirname(args["-i"]) == args["-i"])
            args["-m"] = "";
        else
            args["-m"] = pystring::os::path::dirname(args["-i"]) + "/";
    
    tinyobj::attrib_t attrib;
    std::vector<tinyobj::shape_t> shapes;
    std::vector<tinyobj::material_t> materials;
    std::string err, warn;
    if (!tinyobj::LoadObj(&attrib, &shapes, &materials, &warn, &err, args["-i"].c_str(), args["-m"].c_str()))
    {
        printf("Failed to load obj: %s\n", args["-i"].c_str());
        return -1;
    }

    auto& verts = attrib.vertices;
    auto& norms = attrib.normals;
    auto& texs = attrib.texcoords;
    auto& colors = attrib.colors;
    printf("Input mesh count %zi\n", shapes.size());
    for (auto& shape : shapes)
    {
        std::vector<uint32_t> indices;
        std::unordered_map<Vertex, uint32_t> vertices;
        glm::vec4 min = glm::vec4(FLT_MAX, FLT_MAX, FLT_MAX, 1.0f);
        glm::vec4 max = glm::vec4(-FLT_MAX, -FLT_MAX, -FLT_MAX, 1.0f);
        auto& inds = shape.mesh.indices;
        for (int i = 0, n = inds.size(); i < n; ++i)
        {
            Vertex vertex;
            vertex.m_pos.x = verts[3 * inds[i].vertex_index + 0];
            vertex.m_pos.y = verts[3 * inds[i].vertex_index + 1];
            vertex.m_pos.z = verts[3 * inds[i].vertex_index + 2];
            vertex.m_pos.w = 1.0f;
            vertex.m_normal.x = norms[3 * inds[i].normal_index + 0];
            vertex.m_normal.y = norms[3 * inds[i].normal_index + 1];
            vertex.m_normal.z = norms[3 * inds[i].normal_index + 2];
            vertex.m_tex.x = texs[2 * inds[i].texcoord_index + 0];
            vertex.m_tex.y = texs[2 * inds[i].texcoord_index + 1];

            uint32_t index = 0;
            std::unordered_map<Vertex, uint32_t>::const_iterator got = vertices.find(vertex);
            if (got == vertices.end())
            {
                index = (uint32_t)vertices.size();
                vertices[vertex] = index;
                min.x = (vertex.m_pos.x < min.x) ? vertex.m_pos.x : min.x;
                min.y = (vertex.m_pos.y < min.y) ? vertex.m_pos.y : min.y;
                min.z = (vertex.m_pos.z < min.z) ? vertex.m_pos.z : min.z;
                max.x = (vertex.m_pos.x > max.x) ? vertex.m_pos.x : max.x;
                max.y = (vertex.m_pos.y > max.y) ? vertex.m_pos.y : max.y;
                max.z = (vertex.m_pos.z > max.z) ? vertex.m_pos.z : max.z;
            }
            else
            {
                index = got->second;
            }

            indices.push_back(index);
        }

        Mesh m;
        m.m_vertices.resize(vertices.size());
        for (auto& vert : vertices)
            m.m_vertices[vert.second] = vert.first;

        // find the centre of the rough bounds
        glm::vec4 centre = min + (max - min) * 0.5f;
        centre.w = 1.0f;
        
        // use the centre of the rough bounds to find the tight non aligned bounding box
        float maxDist = -FLT_MAX;
        glm::vec4 f;
        for (auto& v : m.m_vertices)
        {
            float dist = glm::length(v.m_pos - centre);
            if(dist > maxDist)
            {
                f = v.m_pos;
                maxDist = dist;
            }
        }
        glm::vec4 invF = centre + (centre - f);
        invF.w = 1.0f;
        min = (glm::length2(f) < glm::length2(invF)) ? f : invF;
        max = (glm::length2(f) > glm::length2(invF)) ? f : invF;

        // create an aabb from the min+max verts
        glm::vec4 extent = max - min;
        float extentLength = glm::length(extent);
        glm::vec4 scale = glm::vec4(extentLength);
        scale.w = 1.0f;

        // move the mesh vertices to be centred around zero and scale down
        for (auto& v : m.m_vertices)
        {
            v.m_pos -= centre;
            v.m_pos /= scale;
            v.m_pos.w = 1.0f;
        }

        size_t existingIndex = -1;
        glm::mat4 rotation;
        for (int i = 0, n = s_meshes.size(); i < n; ++i)
        {
            Mesh& mesh = s_meshes[i];
            if (mesh.m_indices.size() != indices.size() || mesh.m_vertices.size() != m.m_vertices.size())
                continue;

            float extentScale = mesh.m_aabb.m_extentLength / extentLength;
            bool match = true;
            for (uint32_t k = 0, ns = indices.size(); k < ns; ++k)
            {
                Vertex sv0 = m.m_vertices[indices[k]];
                sv0.m_pos *= glm::vec4(glm::vec3(extentScale), 1.0f);
                uint32_t index = -1;
                for (uint32_t j = 0, nd = mesh.m_indices.size(); j < nd; ++j)
                {
                    Vertex dv0 = m.m_vertices[mesh.m_indices[j]];
                    
                    if (sv0.m_tex.x != dv0.m_tex.x || sv0.m_tex.y != dv0.m_tex.y)
                        continue;

                    if (k == 0)
                    {
                        // find the rotation matrix for first index
                        glm::vec3 a = glm::normalize(glm::vec3(sv0.m_pos.x, sv0.m_pos.y, sv0.m_pos.z));
                        glm::vec3 b = glm::normalize(glm::vec3(dv0.m_pos.x, dv0.m_pos.y, dv0.m_pos.z));
                        if (a == b)
                        {
                            index = j;
                            rotation = glm::mat4(1.0f);
                            break;
                        }
                        glm::vec3 v = glm::cross(b, a);
                        float angle = acos(glm::dot(b, a) / (glm::length((b)* glm::length(a))));
                        glm::mat4 rot = glm::rotate(angle, v);
                        glm::vec4 v0 = rot * sv0.m_pos;

                        bool equal = signbit(v0.x) == signbit(dv0.m_pos.x) && trunc(1000. * v0.x) == trunc(1000. * dv0.m_pos.x);
                        equal |= signbit(v0.y) == signbit(dv0.m_pos.y) && trunc(1000. * v0.y) == trunc(1000. * dv0.m_pos.y);
                        equal |= signbit(v0.z) == signbit(dv0.m_pos.z) && trunc(1000. * v0.z) == trunc(1000. * dv0.m_pos.z);
                        if (equal)
                        {
                            rotation = rot;
                            index = j;
                            break;
                        }
                    }
                    else
                    {
                        // test for matches using existing rotation matrix
                        glm::vec4 v0 = rotation * sv0.m_pos;
                        bool equal = signbit(v0.x) == signbit(dv0.m_pos.x) && trunc(1000. * v0.x) == trunc(1000. * dv0.m_pos.x);
                        equal |= signbit(v0.y) == signbit(dv0.m_pos.y) && trunc(1000. * v0.y) == trunc(1000. * dv0.m_pos.y);
                        equal |= signbit(v0.z) == signbit(dv0.m_pos.z) && trunc(1000. * v0.z) == trunc(1000. * dv0.m_pos.z);
                        if (equal)
                        {
                            index = j;
                            break;
                        }
                    }
                }
                if (index == -1)
                {
                    match = false;
                    break;
                }
            }
            if (!match)
                continue;
            existingIndex = i;
        }

        if(existingIndex != -1)
        {
            glm::mat4 w = glm::translate(glm::vec3(centre)) * rotation * glm::scale(glm::vec3(scale));
            s_meshes[existingIndex].m_instanceMatrices.push_back(w);
        }
        else
        {
            m.m_indices = indices;
            m.m_aabb.m_min = min;
            m.m_aabb.m_max = max;
            m.m_aabb.m_extentLength = extentLength;            
            glm::mat4 w = glm::translate(glm::vec3(centre)) * glm::scale(glm::vec3(scale));
            m.m_instanceMatrices.push_back(w);
            s_meshes.push_back(m);
        }
    }
    printf("Output mesh count %zi\n", s_meshes.size());
    std::cin.ignore();
    return 0;
}