#include <regex>
#include <iostream>
#define RAPIDJSON_HAS_STDSTRING 1
#include "rapidjson\document.h"
#include "rapidjson\writer.h"
#include "rapidjson\prettywriter.h"
#include "rapidjson\stringbuffer.h"
#include "pystring.h"
#include <fstream>
#include <map>
#include <windows.h>
#include <filesystem>
#include <set>
#include <unordered_map>

namespace e_layoutType
{
    enum layoutType
    {
        in,
        out,
        uniform
    };
}
typedef e_layoutType::layoutType LayoutType;

namespace e_shaderStage
{
    enum shaderStage
    {
        vert,
        tesc,
        tese,
        geom,
        frag,
        comp,
        COUNT
    };
}
typedef e_shaderStage::shaderStage ShaderStage;

namespace e_primitiveType
{
    enum primitiveType
    {
        points,
        lines,
        triangles,
        COUNT
    };
}
typedef e_primitiveType::primitiveType PrimitiveType;

namespace e_bindingType
{
    enum bindingType
    {
        image,
        buffer
    };
}
typedef e_bindingType::bindingType BindingType;

std::string k_fileEndings[] =
{
    "vert",
    "tesc",
    "tese",
    "geom",
    "frag",
    "comp"
};

ShaderStage shaderStageNameToEnum(const std::string& name)
{
    for(int i = 0; i < e_shaderStage::COUNT; ++i)
    {
        if(name == k_fileEndings[i])
        {
            return (ShaderStage)i;
        }
    }
    return e_shaderStage::vert;
}

std::string k_primitiveTypes[] =
{
    "points",
    "lines",
    "triangles"
};

PrimitiveType primitiveTypeNameToEnum(const std::string& name)
{
    for (int i = 0; i < i < e_primitiveType::COUNT; ++i)

    {
        if (name == k_primitiveTypes[i])
        {
            return (PrimitiveType)i;
        }
    }
    return e_primitiveType::triangles;
}

struct BindingInfo
{
    std::string name = "";
    BindingType type = e_bindingType::image;
};

struct ShaderVariationJob
{
    std::vector<std::string>                variationNames[e_shaderStage::COUNT];
    bool                                    usesStage[e_shaderStage::COUNT] = { false };
    std::vector<std::string>                filesToDo;
    std::unordered_map<int, BindingInfo>    usedBindings;
    std::string                             outputPath = "";
    bool                                    isDefaultVariation = true;
    bool                                    builtVariations = false;
};

std::map<std::string, rapidjson::Value> structs;
std::map<std::string, int> defines;

void processStruct(const std::string& buf, rapidjson::Value& out, bool hasArrays, rapidjson::Document::AllocatorType& alloc)
{
    std::regex memberRegex("\\S+\\s\\S+;\r");
    std::sregex_iterator memberBegin = std::sregex_iterator(buf.begin(), buf.end(), memberRegex);
    std::sregex_iterator memberEnd = std::sregex_iterator();
    for (std::sregex_iterator j = memberBegin; j != memberEnd; ++j) {
        std::string member = j->str();
        rapidjson::Value m;
        m.SetObject();
        bool isArray = false;
        if (hasArrays)
            isArray = (member.find("[]") != std::string::npos);

        if (member.find("mat") != std::string::npos)
        {
            // matrices are always floats according to the internet so if this turns out to be wrong then blame the internet not me
            m.AddMember("Type", "float", alloc);
            std::string typeString = member.substr(0, member.find_first_of(' '));
            if (typeString[typeString.size() - 2] == 'x')
            {
                int columns = typeString[typeString.size() - 3] - '0';
                int rows = typeString[typeString.size() - 1] - '0';
                m.AddMember("Count", columns*rows, alloc);
            }
            else
            {
                int dims = typeString[typeString.size() - 1] - '0';
                m.AddMember("Count", dims*dims, alloc);
            }
        }
        else if (member.find("vec") != std::string::npos)
        {
            std::string typeString = member.substr(0, member.find_first_of(' '));
            switch (member[0])
            {
            case 'u':
                m.AddMember("Type", "uint", alloc);
                break;
            case 'i':
                m.AddMember("Type", "int", alloc);
                break;
            default:
                m.AddMember("Type", "float", alloc);
                break;
            }
            int num = typeString.back() - '0';
            m.AddMember("Count", num, alloc);
        }
        else if (member.find("float") != std::string::npos)
        {
            m.AddMember("Type", "float", alloc);
            m.AddMember("Count", 1, alloc);
        }
        else if (member.find("uint") != std::string::npos)
        {
            m.AddMember("Type", "uint", alloc);
            m.AddMember("Count", 1, alloc);
        }
        else
        {
            // look for the type in the defined structs
            std::string typeString = member.substr(0, member.find_first_of(' '));
            if (structs.find(typeString) != structs.end())
            {
                rapidjson::Value& structType = structs[typeString];
                m.AddMember("Members", structType, alloc);
            }
        }
        size_t start = member.find_last_of(' ') + 1;
        size_t len = 0;
        if (isArray)
            len = member.find_last_of('[') - start;
        else
            len = member.find_last_of(';') - start;
        std::string nameString = member.substr(start, len);
        m.AddMember("Name", "", alloc);
        m["Name"].SetString(nameString.c_str(), alloc);
        m.AddMember("IsArray", isArray, alloc);
        out.PushBack(m, alloc);
    }
}

void processIOStruct(const std::string& buf, bool instanced, int& location, rapidjson::Value& out, rapidjson::Document::AllocatorType& alloc)
{
    std::regex memberRegex("\\S+\\s\\S+\\s:\\s\\S+;\r");
    std::sregex_iterator memberBegin = std::sregex_iterator(buf.begin(), buf.end(), memberRegex);
    std::sregex_iterator memberEnd = std::sregex_iterator();
    for (std::sregex_iterator j = memberBegin; j != memberEnd; ++j) {
        std::string member = j->str();
        rapidjson::Value m;
        m.SetObject();

        std::vector<std::string> strings;
        pystring::split(member, strings);
        for(auto& s : strings)
            s = pystring::strip(s);
        
        std::string typeName = strings[0];
        std::string typeDims = "1";
        for(size_t i = 0; i < typeName.size(); ++i)
        {
            // find first numeric character
            if(typeName[i] >= '0'  && typeName[i] <= '9')
            {
                typeDims = typeName.substr(i, typeName.size());
                typeName = typeName.substr(0, i);
            }
        }
        
        if(structs.find(typeName) != structs.end())
        {
            rapidjson::Value& structType = structs[typeName];
            m.AddMember("Members", structType, alloc);
        }
        else
        {
            m.AddMember("Type", "", alloc);
            m["Type"].SetString(typeName.c_str(), alloc);
        }
        
        // extract dimensions of input and total count
        std::vector<std::string> dims;
        pystring::split(typeDims, dims, "x");
        int count = 1;
        m.AddMember("Dims", "", alloc);
        m["Dims"].SetArray();
        for(auto& d : dims)
        {
            int dim = std::stoi(d);
            m["Dims"].PushBack(dim, alloc);
            count = count * dim;
        }
        
        m.AddMember("Count", count, alloc);
        
        std::string varName = strings[1];
        int arrayCount = 0;
        int arrBegin = pystring::find(varName, "[");
        int arrEnd = pystring::find(varName, "]");
        if(arrBegin != -1 &&  arrEnd != -1)
        {
            arrayCount = stoi(varName.substr(arrBegin + 1, arrEnd - 1));
            varName = varName.substr(0, arrBegin);
        }
        std::string semName = strings[3];
        int semIndex = 0;
        for(size_t i = 0; i < semName.size(); ++i)
        {
            if(semName[i] >= '0' && semName[i] <= '9')
            {
                semIndex = std::stoi(semName.substr(i, semName.size() - 1));
                semName = semName.substr(0, i);
            }
        }
        m.AddMember("Semantic Name", "", alloc);
        m["Semantic Name"].SetString(semName, alloc);
        m.AddMember("Semantic Index", semIndex, alloc);
        
        m.AddMember("Name", "", alloc);
        m["Name"].SetString(varName.c_str(), alloc);
        m.AddMember("Instanced", instanced, alloc);
        m.AddMember("Location", location++, alloc);
        m.AddMember("IsArray", arrayCount > 0, alloc);
        out.PushBack(m, alloc);
    }
}

std::string replaceIOStruct(const std::string& source, const std::string& ioName, const std::string& definition, ShaderStage stage, LayoutType layoutType, PrimitiveType primitiveType, int maxVertices, const rapidjson::Value& strct, const rapidjson::Value& replacement)
{
    std::string out = source;
    
    std::string layoutDef = "";
    std::string perVertexDef = "";

    // add geometry shader input/output primitive definitions
    if (stage == e_shaderStage::geom)
    {
        if (layoutType == e_layoutType::in)
        {
            switch (primitiveType)
            {
            case e_primitiveType::points:
                layoutDef += "layout(points) in;\n";
                break;
            case e_primitiveType::lines:
                layoutDef += "layout(lines) in;\n";
                break;
            case e_primitiveType::triangles:
                layoutDef += "layout(triangles) in;\n";
                break;
            }
        }
        else if (layoutType == e_layoutType::out)
        {
            switch (primitiveType)
            {
            case e_primitiveType::points:
                layoutDef += "layout(points, max_vertices=";
                break;
            case e_primitiveType::lines:
                layoutDef += "layout(line_strip, max_vertices=";
                break;
            case e_primitiveType::triangles:
                layoutDef += "layout(triangle_strip, max_vertices=";
                break;
            }

            layoutDef += std::to_string(maxVertices);
            layoutDef += ") out;\n";
        }
    }

    for(auto& m : replacement.GetArray())
    {
        std::string memberName = m["Name"].GetString();
        
        // create type definition string
        std::string typeDef = "";
        int numDims = m["Dims"].Size();
        if(numDims == 2)
        {
            typeDef = "mat";
            typeDef += std::to_string(m["Dims"][0].GetInt());
            typeDef += "x";
            typeDef += std::to_string(m["Dims"][1].GetInt());
        }
        else
        {
            std::string typeName = m["Type"].GetString();
            
            int dim = m["Dims"][0].GetInt();
            if(dim > 1)
            {
                typeDef = "vec";
                typeDef += std::to_string(dim);
                
                if(typeName == "uint")
                    typeDef = "u" + typeDef;
                else if (typeName == "int")
                    typeDef = "i" + typeDef;
            }
            else
            {
                typeDef = typeName;
            }
        }
        
        std::string semanticName = m["Semantic Name"].GetString();
        int semanticIndex = m["Semantic Index"].GetInt();
        
        std::string def = "layout(location = ";
        def += std::to_string(m["Location"].GetInt());
        def += ") ";
        if (layoutType == e_layoutType::in)
            def += "in ";
        else if (layoutType == e_layoutType::out)
            def += "out ";

        def += typeDef + " ";
        def += ioName + "_" + memberName;
        if (stage == e_shaderStage::geom && layoutType == e_layoutType::in)
        {
            // in geometry shaders most vertices are input in groups
            switch (primitiveType)
            {
            case e_primitiveType::points:
                def += "[1]";
                break;
            case e_primitiveType::lines:
                def += "[2]";
                break;
            case e_primitiveType::triangles:
                def += "[3]";
                break;
            }
        }
        def += ";\n";

        // handle position 0 semantic separately
        if (semanticName == "POSITION" && semanticIndex == 0)
        {
            if (layoutType == e_layoutType::out && (stage == e_shaderStage::vert || stage == e_shaderStage::geom))
            {
                // on output specify per vertex layout
                // on input we don't need to specify the semantic
                if (perVertexDef.empty())
                {
                    perVertexDef = "out gl_PerVertex{};";
                }

                std::string vdef = typeDef + " gl_Position;";
                size_t insertPos = perVertexDef.find_last_of('}');
                perVertexDef.insert(insertPos, vdef);

                def = "";
            }
            else if (layoutType == e_layoutType::in && stage != e_shaderStage::vert)
            {
                // on input to fragment shader hide this semantic
                def = "";
            }
        }
        layoutDef += def;
    }
    
    // replace old definition format with new layout
    std::string completeDef = perVertexDef + "\n" + layoutDef;
    out = pystring::replace(out, definition, completeDef);
    
    // replace references to io members with new layout
    for(auto& m : strct.GetArray())
    {
        // get name to search for
        std::string memberName = m["Name"].GetString();
        std::string searchName = ioName + "." + memberName;
        
        // find matching semantic in the replacement struct
        std::string replacementName = ioName + "_" + memberName;
        std::string semanticName = m["Semantic Name"].GetString();
        int semanticIndex = m["Semantic Index"].GetInt();
        for (auto& r : replacement.GetArray())
        {
            std::string rs = r["Semantic Name"].GetString();
            int rsi = r["Semantic Index"].GetInt();
            if (rs == semanticName && rsi == semanticIndex)
            {
                replacementName = ioName + "_" + r["Name"].GetString();
                break;
            }
        }

        // custom replacement for built in shader variables
        if(semanticName == "POSITION" && semanticIndex == 0)
        {
            if ((stage == e_shaderStage::vert || stage == e_shaderStage::geom) && layoutType == e_layoutType::out)
                replacementName = "gl_Position";
            else if (stage == e_shaderStage::frag && layoutType == e_layoutType::in)
                replacementName = "gl_FragCoord";
        }
        out = pystring::replace(out, searchName, replacementName);
    }
    
    return out;
}

std::string findDimension(std::string source, std::string typeName, std::string post)
{
    size_t loc = source.find(typeName);
    size_t end = post.empty() ? source.size() : source.find(post);
    loc += typeName.size();
    return source.substr(loc, end - loc);
}

void extractNameAndCount(const std::string& source, char preNameChar, std::string& name, int& count)
{
    name = source.substr(source.find_last_of(preNameChar) + 1);
    if (name.find_last_of('[') == std::string::npos)
    {
        name = name.substr(0, name.length() - 1);
        count = 1;
    }
    else
    {
        size_t countStart = name.find_last_of('[') + 1;
        size_t countLen = name.find_last_of(']') - countStart;
        count = countLen > 0 ? std::stoi(name.substr(countStart, countLen)) : 0;
        name = name.substr(0, name.find_last_of('['));
    }
}

int getBindingIndex(const std::string& source, std::string bindingName)
{
    bindingName = bindingName + " = ";
    size_t indexBegin = source.find(bindingName);
    if (indexBegin == std::string::npos)
        return -1;
    indexBegin += bindingName.size();
    size_t indexLen = source.find_first_of(')') - indexBegin;
    std::string binding = source.substr(indexBegin, indexLen);
    if (binding[0] >= '0' && binding[0] <= '9')
    {
        return std::stoi(source.substr(indexBegin, indexLen));
    }
    else
    {
        if (defines.find(binding) != defines.end())
        {
            return defines[binding];
        }
        else
        {
            return -1;
        }
    }
}

void pushJobs(ShaderVariationJob j, std::vector<std::string> names, ShaderStage stage, std::vector<ShaderVariationJob>& out)
{
    if (names.size() > 0)
    {
        std::string current = names.front();
        names.erase(names.begin());

        ShaderVariationJob with = j;
        with.variationNames[stage].push_back(current);
        pushJobs(with, names, stage, out);
        j.isDefaultVariation = false;
        pushJobs(j, names, stage, out);
    }
    else
    {
        out.push_back(j);
    }
}

bool mergeObjects(rapidjson::Value& dst, const rapidjson::Value& src, rapidjson::Document::AllocatorType& allocator)
{
    for (auto srcIt = src.MemberBegin(); srcIt != src.MemberEnd(); ++srcIt)
    {
        auto dstIt = dst.FindMember(srcIt->name);
        if (dstIt == dst.MemberEnd())
        {
            rapidjson::Value dstName;
            dstName.CopyFrom(srcIt->name, allocator);
            rapidjson::Value dstValue;
            dstValue.CopyFrom(srcIt->value, allocator);

            dst.AddMember(dstName, dstValue, allocator);
        }
        else
        {
            auto srcType = srcIt->value.GetType();
            auto dstType = dstIt->value.GetType();
            if (srcType != dstType)
                return false;

            if (srcIt->value.IsArray())
            {
                for (auto arrayIt = srcIt->value.Begin(); arrayIt != srcIt->value.End(); ++arrayIt)
                {
                    rapidjson::Value dstValue;
                    dstValue.CopyFrom(*arrayIt, allocator);
                    dstIt->value.PushBack(dstValue, allocator);
                }
            }
            else if (srcIt->value.IsObject())
            {
                if (!mergeObjects(dstIt->value, srcIt->value, allocator))
                    return false;
            }
            else
            {
                dstIt->value.CopyFrom(srcIt->value, allocator);
            }
        }
    }
    return true;
}

int main(int argc, char *argv[])
{
    std::map<std::string, std::string> args;
    for (int i = 1; i < argc; i+=2)
    {
        std::string arg1 = std::string(argv[i]);
        std::string arg2 = (i + 1 < argc) ? std::string(argv[i + 1]) : "";
        args[arg1] = arg2;
    }

    // if input/output isn't provided just look in the working dir
    if (args.find("-o") == args.end())
        args["-o"] = "";

    std::string glslang = "glslangvalidator";
    if (args.find("-glslang") != args.end())
        glslang = args["-glslang"];

    std::vector<std::string> inputPaths;
    pystring::split(args["-i"], inputPaths, ",");
    std::string outputPath = args["-o"];
    std::vector<ShaderVariationJob> jobs;
    std::vector<ShaderVariationJob> finals;

    size_t numFileEndings = sizeof(k_fileEndings) / sizeof(k_fileEndings[0]);
    if (std::filesystem::is_directory(inputPaths[0]))
    {
        // process all the files in a directory into compiled shaders
        std::string workingPath = std::filesystem::current_path().string();
        std::string absoluteInputPath = pystring::os::path::join(workingPath, inputPaths[0]);
        for (auto& p : std::filesystem::directory_iterator(absoluteInputPath))
        {
            std::string ext, root;
            pystring::os::path::splitext(root, ext, p.path().string());
            ext = pystring::strip(ext, ".");
            
            for (size_t i = 0; i < numFileEndings; ++i)
            {
                if (ext == k_fileEndings[i])
                {
                    std::string baseInputName = pystring::os::path::basename(p.path().string());

                    // skip temp files
                    if (pystring::startswith(baseInputName, "temp_"))
                        continue;
                    

                    ShaderVariationJob job;
                    job.filesToDo.push_back(p.path().string());
                    baseInputName = pystring::replace(baseInputName, ".", "_") + ".shader";
                    job.outputPath = pystring::os::path::join(outputPath, baseInputName);
                    jobs.push_back(job);
                    break;
                }
            }
        }
    }
    else
    {
        // process multiple files (.vert, .frag etc.) into a single compiled shader
        ShaderVariationJob job;
        for(auto& p : inputPaths)
        {
            job.filesToDo.push_back(p);
        }
        job.outputPath = outputPath;
        jobs.push_back(job);
    }

    // pre-emptively erase any jobs that don't need rebuilding
    for (intptr_t i = jobs.size() - 1; i >= 0; --i)
    {
        const ShaderVariationJob& j = jobs[i];

        // -f flag will for a rebuild, otherwise check stored output modified time against input file modified times
        if (args.find("-f") == args.end())
        {
            // load the output file if it exists and check
            std::ifstream outputFile(j.outputPath, std::ios::ate | std::ios::binary);
            if (outputFile.is_open())
            {
                size_t len = (size_t)outputFile.tellg();
                std::string buf;
                buf.resize(len);
                outputFile.seekg(0);
                outputFile.read((char*)buf.c_str(), len);
                outputFile.close();

                rapidjson::Document d;
                bool result = d.Parse<rapidjson::kParseStopWhenDoneFlag>(buf.c_str()).HasParseError();
                if (!result)
                {
                    if (d.HasMember("Last Modified"))
                    {
                        FILETIME stored;
                        stored.dwHighDateTime = d["Last Modified"]["High"].GetUint();
                        stored.dwLowDateTime = d["Last Modified"]["Low"].GetUint();

                        // check if any files are newer than the last rebuild
                        bool rebuild = false;
                        for (auto& in : j.filesToDo)
                        {
                            // check last modified time on the input file
                            HANDLE inFile = CreateFile(in.c_str(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL);
                            if (inFile == 0)
                            {
                                std::cout << "Failed to open shader " << in << " for read!\n";
                                continue;
                            }
                            FILETIME ftCreate, ftAccess, ftWrite;
                            bool gotFileTime = GetFileTime(inFile, &ftCreate, &ftAccess, &ftWrite);
                            CloseHandle(inFile);

                            if (!gotFileTime || CompareFileTime(&stored, &ftWrite) < 0)
                            {
                                if (!gotFileTime)
                                    std::cout << "Failed to get file modified time for " << in << ", forcing a rebuild\n";
                                rebuild = true;
                            }
                        }

                        // skip rebuild for this input if no new files
                        if (!rebuild)
                        {
                            std::cout << j.outputPath << " does not need a rebuild\n";
                            jobs.erase(jobs.begin() + i);
                        }
                    }
                }
                else
                {
                    auto error = d.GetParseError();
                    std::cout << "Error\n";
                }
            }
        }
    }

    rapidjson::Document tracker;
    tracker.SetObject();

    while(!jobs.empty())
    {
        rapidjson::Document::AllocatorType& alloc = tracker.GetAllocator();

        ShaderVariationJob j = jobs[0];
        jobs.erase(jobs.begin());

        std::unordered_map<int, BindingInfo>& usedBindings = j.usedBindings;

        std::string in = j.filesToDo[0];
        std::cout << "Processing file: \"" << in << "\"\n";

        std::ifstream shader(in.c_str(), std::ios::ate | std::ios::binary);
        if (!shader.is_open())
        {
            std::cout << "Failed to open shader " << in << " for read!\n";
            continue;
        }
        size_t len = (size_t)shader.tellg();
        std::string buf;
        buf.resize(len);
        shader.seekg(0);
        shader.read((char*)buf.c_str(), len);
        shader.close();

        // store shader type
        std::string root, ext;
        pystring::os::path::splitext(root, ext, in);
        ext = pystring::strip(ext, ".");
        ShaderStage stage = shaderStageNameToEnum(ext);
        j.usesStage[stage] = true;

        // get json data on any shader variations
        rapidjson::Value variationStateJson;
        variationStateJson.SetObject();
        if (pystring::find(buf, "\"variations\":") != -1)
        {
            std::vector<std::string> lines;
            pystring::splitlines(buf, lines, true);
            int currentIndent = 0;
            bool processingVariations = false;
            std::string variationsJsonString = "";
            size_t currentLine = 0;
            for (currentLine = 0; currentLine < lines.size(); ++currentLine)
            {
                std::string& l = lines[currentLine];
                if (pystring::find(l, "\"variations\":") != -1)
                {
                    processingVariations = true;
                    continue;
                }

                variationsJsonString += l;
                if (pystring::find(l, "{") != -1)
                {
                    currentIndent++;
                }
                else if (pystring::find(l, "}") != -1)
                {
                    currentIndent--;
                    if (currentIndent == 0)
                    {
                        ++currentLine;
                        break;
                    }
                }
            }

            // strip variations json from the top of the file
            std::string withoutVariations = "";
            for (size_t cl = currentLine; cl < lines.size(); ++cl)
            {
                withoutVariations += lines[cl];
            }
            buf = withoutVariations;

            rapidjson::Document variationsJson;
            variationsJson.Parse<rapidjson::kParseStopWhenDoneFlag>(variationsJsonString);
            if (variationsJson.HasParseError())
            {
                std::cout << "Error parsing variations JSON: " << variationsJson.GetParseError();
            }

            if (!j.builtVariations)
            {
                j.builtVariations = true;

                std::vector<std::string> subvariationNames;
                for (auto& vit = variationsJson.MemberBegin(); vit != variationsJson.MemberEnd(); ++vit)
                {
                    if (!vit->value.HasMember("sub_variation") || vit->value["sub_variation"].GetBool() == false)
                        continue;
                    
                    subvariationNames.push_back(vit->name.GetString());
                }

                bool setCurrent = true;
                ShaderVariationJob current = {};
                for (auto& vit = variationsJson.MemberBegin(); vit != variationsJson.MemberEnd(); ++vit)
                {
                    if (vit->value.HasMember("sub_variation") && vit->value["sub_variation"].GetBool())
                        continue;

                    ShaderVariationJob next = j;
                    next.variationNames[stage].push_back(vit->name.GetString());
                    next.isDefaultVariation = next.isDefaultVariation && (vit->value.HasMember("default") && vit->value["default"].GetBool());

                    // create jobs for each combination of sub variations
                    std::vector<ShaderVariationJob> subvariationJobs;
                    pushJobs(next, subvariationNames, stage, subvariationJobs);

                    for (auto& sub : subvariationJobs)
                    {
                        if (setCurrent)
                        {
                            setCurrent = false;
                            current = sub;
                        }
                        else
                        {
                            jobs.push_back(sub);
                        }
                    }
                }
                j = current;
            }


            std::string combinedVariationNames = "";
            for (int i = 0; i < e_shaderStage::COUNT; ++i)
            {
                if (!j.usesStage[i])
                    continue;

                for (auto& s : j.variationNames[i])
                {
                    if (combinedVariationNames.empty())
                        combinedVariationNames = s;
                    else
                        combinedVariationNames += "/" + s;
                }
            }
            std::cout << "Variation: \"" << combinedVariationNames << "\"\n";

            // take a copy of the variation json for tracking
            for (auto& s : j.variationNames[stage])
            {
                mergeObjects(variationStateJson, variationsJson[s], alloc);
            }

            // strip code for other variations and remove wrappers around code for this variation
            pystring::splitlines(buf, lines, true);
            std::string replaced = "";
            int currentVariationIndentation = -1;
            bool keepCode = false;
            for (size_t i = 0; i < lines.size(); ++i)
            {
                std::string& l = lines[i];

                // variation blocks always start with "variation_name:"
                std::string noWs = pystring::strip(l);
                if (pystring::endswith(noWs, ":"))
                {
                    std::string nameString = pystring::rstrip(noWs, ":");
                    std::vector<std::string> names;
                    pystring::split(nameString, names, "/");
                    for (auto& n : names)
                    {
                        if (std::find(j.variationNames[stage].begin(), j.variationNames[stage].end(), n) != j.variationNames[stage].end())
                        {
                            keepCode = true;
                            break;
                        }
                    }
                    currentVariationIndentation = 0;
                    continue;
                }

                if (currentVariationIndentation >= 0)
                {
                    if (pystring::find(l, "{") != -1)
                    {
                        currentVariationIndentation++;
                        if (currentVariationIndentation == 1)
                            continue;
                    }
                    else if (pystring::find(l, "}") != -1)
                    {
                        currentVariationIndentation--;
                        if (currentVariationIndentation == 0)
                        {
                            // check if there is an else block after this
                            bool hasElse = false;
                            if (i < lines.size() - 1)
                            {
                                std::string& nl = lines[i + 1];
                                if (pystring::startswith(pystring::strip(nl), "else"))
                                {
                                    hasElse = true;
                                }
                            }

                            if (hasElse)
                            {
                                // move onto the else block and invert whether we're keeping the code
                                ++i;
                                keepCode = !keepCode;
                            }
                            else
                            {
                                currentVariationIndentation = -1;
                                keepCode = false;
                            }
                            continue;
                        }
                    }

                    if (keepCode)
                    {
                        // strip the extra indentation from the variation wrapping
                        if (pystring::startswith(l, "\t"))
                            l = l.substr(1, l.size() - 1);
                        replaced += l;
                    }
                }
                else
                {
                    replaced += l;
                }
            }
            buf = replaced;
        }

        if (!tracker.HasMember(j.outputPath))
        {
            rapidjson::Value name;
            name.SetString(j.outputPath, alloc);
            rapidjson::Value value;
            value.SetObject();
            tracker.AddMember(name, value, alloc);
        }

        // pass output down through the json layers
        bool passOutput = false;
        rapidjson::Value prevOutputName, prevOutputValue;
        prevOutputName.SetString("prev_output");
        prevOutputValue.SetObject();

        rapidjson::Value* valIt = &tracker[j.outputPath];
        for (int i = 0; i < e_shaderStage::COUNT; ++i)
        {
            if (!j.usesStage[i])
                continue;

            std::string combined = "";
            for (auto& s : j.variationNames[i])
            {
                if (!combined.empty())
                    combined += "/";
                combined += s;
            }

            if (!combined.empty())
            {
                if (valIt->HasMember(combined))
                {
                    valIt = &(*valIt)[combined];
                    
                }
                else
                {
                    rapidjson::Value name;
                    name.SetString(combined, alloc);
                    rapidjson::Value value;
                    value.SetObject();
                    valIt->AddMember(name, value, alloc);
                    valIt = &(*valIt)[combined];
                }
            }
            if (valIt->HasMember("output"))
            {
                passOutput = true;
                prevOutputValue.CopyFrom(valIt->FindMember("output")->value, alloc);
            }
        }

        if (passOutput && !valIt->HasMember("prev_output"))
        {
            valIt->AddMember(prevOutputName, prevOutputValue, alloc);
        }

        rapidjson::Value& variationJson = *valIt;

        // copy the variation json data
        if (!variationStateJson.ObjectEmpty())
        {
            if (variationStateJson.HasMember("skip") && variationStateJson["skip"].GetBool())
            {
                // pass through the output struct
                if (variationJson.HasMember("prev_output"))
                {
                    rapidjson::Value outputName, outputValue;
                    outputName.SetString("output", alloc);
                    outputValue.SetObject();
                    outputValue.CopyFrom(variationJson["prev_output"], alloc);
                    variationJson.AddMember(outputName, outputValue, alloc);
                }

                if (j.filesToDo.size() > 1)
                {
                    ShaderVariationJob nextFileJob = j;
                    nextFileJob.builtVariations = false;
                    nextFileJob.filesToDo.erase(nextFileJob.filesToDo.begin());
                    jobs.insert(jobs.begin(), nextFileJob);
                }
                continue;
            }
            variationJson.AddMember("json", variationStateJson, alloc);
        }

        rapidjson::Value val(ext.c_str(), alloc);
        variationJson.AddMember(val, "", alloc);
        variationJson[ext.c_str()].SetObject();
        rapidjson::Value& sj = variationJson[ext.c_str()];

        // bring in any includes to the shader buffer
        std::regex includeRegex("#include \"\\w+\.\\w\"");
        std::sregex_iterator includeBegin = std::sregex_iterator(buf.begin(), buf.end(), includeRegex);
        std::sregex_iterator includeEnd = std::sregex_iterator();
        std::string includes = "";
        int includesCount = 0;
        std::string workingPath = pystring::os::path::dirname(in);
        for (std::sregex_iterator i = includeBegin; i != includeEnd; ++i) 
        {
            std::string match = i->str();

            size_t start = match.find_first_of('\"') + 1;
            size_t len = match.find_last_of('\"') - start;
            std::string filename = pystring::os::path::join(workingPath, match.substr(start, len));
            std::ifstream include(filename.c_str(), std::ios::ate | std::ios::binary);
            if (!include.is_open())
            {
                std::cout << "Failed to find include shader " << filename << "!\n";
                continue;
            }
            len = (size_t)include.tellg();
            std::string incBuf;
            incBuf.resize(len);
            include.seekg(0);
            include.read((char*)incBuf.c_str(), len);
            include.close();
            includes += "\n\n" + incBuf;
            includesCount++;
        }

        // cut the #includes from the top of the file
        if (includesCount > 0)
        {
            for (size_t i = 0; i < buf.size(); ++i)
            {
                if (buf[i] == '\n')
                    includesCount--;
                if (includesCount <= 0)
                {
                    buf = buf.substr(i + 1, buf.size() - (i + 1));
                    break;
                }
            }
        }

        // version identifier and extensions should come before included code
        std::regex extensionRegex("#extension.+");
        std::sregex_iterator extensionBegin = std::sregex_iterator(buf.begin(), buf.end(), extensionRegex);
        std::sregex_iterator extensionEnd = std::sregex_iterator();
        std::sregex_iterator lastExtension = std::sregex_iterator();
        std::string header = "";
        for (std::sregex_iterator i = extensionBegin; i != extensionEnd; ++i)
        {
            lastExtension = i;
            header += lastExtension->prefix().str();
            header += lastExtension->str();
        }
        std::string body = lastExtension->suffix().str();
        buf = header + includes + body;
        std::string finalShader = buf;

        sj.AddMember("Uniforms", "", alloc);
        sj["Uniforms"].SetArray();
        sj.AddMember("Push Constants", "", alloc);
        sj["Push Constants"].SetArray();
        sj.AddMember("Inputs", "", alloc);
        sj["Inputs"].SetArray();
        sj.AddMember("Outputs", "", alloc);
        sj["Outputs"].SetArray();
        sj.AddMember("Shader Files", "", alloc);
        sj["Shader Files"].SetObject();

        // map all of the shader defines
        std::regex defineRegex("#define \\w+ \\d+");
        std::sregex_iterator defineBegin = std::sregex_iterator(buf.begin(), buf.end(), defineRegex);
        std::sregex_iterator defineEnd = std::sregex_iterator();
        for (std::sregex_iterator i = defineBegin; i != defineEnd; ++i)
        {
            std::string match = i->str();
            size_t first = match.find_first_of(' ') + 1;
            size_t last = match.find_last_of(' ');
            std::string name = match.substr(first, last - first);
            int define = std::stoi(match.substr(last));
            defines[name] = define;
        }

        // find defined structs for if they're needed later
        int inLocation = 0;
        int outLocation = 0;
        std::regex structRegex("struct .+[\n\r]+\\{[^\\}]+\\};");
        std::sregex_iterator structBegin = std::sregex_iterator(buf.begin(), buf.end(), structRegex);
        std::sregex_iterator structEnd = std::sregex_iterator();
        for (std::sregex_iterator i = structBegin; i != structEnd; ++i)
        {
            std::string match = i->str();

            size_t start = match.find_first_of(' ') + 1;
            size_t len = match.find_first_of('\r') - start;
            std::string name = match.substr(start, len);

            // parse and replace io structs and references to members
            if (pystring::startswith(name, "shader_input_") || pystring::startswith(name, "shader_output_"))
            {
                rapidjson::Value strct;
                strct.SetArray();

                int index = stoi(name.substr(pystring::rfind(name, "_") + 1, name.size()));

                LayoutType layoutType;
                if (pystring::find(name, "_input_") != -1)
                {
                    layoutType = e_layoutType::in;
                }
                else if (pystring::find(name, "_output_") != -1)
                {
                    layoutType = e_layoutType::out;
                }
                    
                bool instanced = pystring::find(name, "instanced") != -1;
                    
                processIOStruct(match, instanced, (layoutType == e_layoutType::in) ? inLocation : outLocation, strct, alloc);

                // replace input/output struct with its gl layout equivalent
                bool replaceInput = variationJson.HasMember("prev_output") && layoutType == e_layoutType::in;
                PrimitiveType primitiveType = e_primitiveType::triangles;
                int maxVertices = 3;
                if (stage == e_shaderStage::geom)
                {
                    if (layoutType == e_layoutType::in && variationJson["json"].HasMember("input_primitive"))
                    {
                        std::string inPrimitive = variationJson["json"]["input_primitive"].GetString();
                        primitiveType = primitiveTypeNameToEnum(inPrimitive);
                    }
                    else if (layoutType == e_layoutType::out)
                    {
                        if (variationJson["json"].HasMember("output_primitive"))
                        {
                            std::string outPrimitive = variationJson["json"]["output_primitive"].GetString();
                            primitiveType = primitiveTypeNameToEnum(outPrimitive);
                        }
                        if (variationJson["json"].HasMember("max_vertices"))
                        {
                            maxVertices = variationJson["json"]["max_vertices"].GetInt();
                        }
                    }
                }

                finalShader = replaceIOStruct(finalShader, name, match, stage, layoutType, primitiveType, maxVertices, strct, replaceInput ? variationJson["prev_output"] : strct);

                rapidjson::Value strctcpy;
                if (replaceInput)
                    strctcpy.CopyFrom(variationJson["prev_output"], alloc);
                else
                    strctcpy.CopyFrom(strct, alloc);

                for (auto& entry : strctcpy.GetArray())
                {
                    if (layoutType == e_layoutType::in)
                    {
                        sj["Inputs"].PushBack(entry, alloc);
                    }
                    else if (layoutType == e_layoutType::out)
                    {
                        sj["Outputs"].PushBack(entry, alloc);
                    }
                }

                if (layoutType == e_layoutType::out)
                {
                    rapidjson::Value name;
                    name.SetString("output", alloc);
                    rapidjson::Value value;
                    value.SetObject();
                    value.CopyFrom(strct, alloc);
                    variationJson.AddMember(name, value, alloc);
                }
            }
            else
            {
                rapidjson::Value strct;
                strct.SetArray();
                processStruct(match, strct, false, alloc);
                structs[name] = strct;
            }
        }

        // find compute shader workgroup definition
            
        size_t lastBinding = 0;

        // find the uniform structs
        std::regex uniformRegex("layout\\(binding = \\d+\\).+[\n\r]+\\{[^\\}]+\\}.+;");
        std::sregex_iterator uniformBegin = std::sregex_iterator(buf.begin(), buf.end(), uniformRegex);
        std::sregex_iterator uniformEnd = std::sregex_iterator();
        for (std::sregex_iterator i = uniformBegin; i != uniformEnd; ++i)
        {
            std::string match = i->str();
            rapidjson::Value uniform;
            uniform.SetObject();
            // check if this struct is a storage buffer
            bool isStorage = (match.find(" buffer ") != std::string::npos);
            std::string nameString;
            int count;
            extractNameAndCount(match, ' ', nameString, count);
            uniform.AddMember("Members", "", alloc);
            uniform["Members"].SetArray();
            processStruct(match, uniform["Members"], isStorage, alloc);

            int requestedBinding = getBindingIndex(match, "binding");
            if (requestedBinding < 0)
                continue;

            int binding = requestedBinding;

            auto bit = usedBindings.find(requestedBinding);
            int searchBinding = 0;
            while (bit != usedBindings.end())
            {
                // check if this is a matching binding from a previous shader
                if (bit->second.type == e_bindingType::buffer && bit->second.name == nameString)
                {
                    break;
                }
                else
                {
                    // try the next binding slot
                    binding = searchBinding++;
                    bit = usedBindings.find(binding);
                }
            }

            usedBindings[binding] = { nameString, e_bindingType::buffer };
            lastBinding = max(lastBinding, binding);
            uniform.AddMember("Binding", binding, alloc);
            uniform.AddMember("Mapping", requestedBinding, alloc);

            uniform.AddMember("Count", count, alloc);
            uniform.AddMember("Name", "", alloc);
            uniform["Name"].SetString(nameString.c_str(), alloc);
            std::string typeString = isStorage ? "Storage Buffer" : "Uniform Buffer";
            uniform.AddMember("Type", "", alloc);
            uniform["Type"].SetString(typeString.c_str(), alloc);

            std::vector<std::string> split;
            pystring::split(match, split, "\r");
            std::string oldDefine = split[0];
            std::string oldBinding = "binding = " + std::to_string(requestedBinding);
            std::string newBinding = "binding = " + std::to_string(binding);
            std::string newDefine = pystring::replace(oldDefine, oldBinding, newBinding);
            finalShader = pystring::replace(finalShader, oldDefine, newDefine);

            sj["Uniforms"].PushBack(uniform, alloc);
        }
        
        // find uniform textures
        std::regex textureRegex("layout\\(binding = .+;");
        std::sregex_iterator textureBegin = std::sregex_iterator(buf.begin(), buf.end(), textureRegex);
        std::sregex_iterator textureEnd = std::sregex_iterator();
        for (std::sregex_iterator i = textureBegin; i != textureEnd; ++i)
        {
            std::string match = i->str();
            rapidjson::Value tex;
            tex.SetObject();
            
            // find the name and number of uniforms
            std::string nameString;
            int count = 1;
            extractNameAndCount(match, ' ', nameString, count);
            tex.AddMember("Count", count, alloc);
            tex.AddMember("Name", "", alloc);
            tex["Name"].SetString(nameString.c_str(), alloc);

            int requestedBinding = getBindingIndex(match, "binding");
            if (requestedBinding < 0)
                continue;

            int binding = requestedBinding;

            auto bit = usedBindings.find(requestedBinding);
            int searchBinding = 0;
            while (bit != usedBindings.end())
            {
                // check if this is a matching binding from a previous shader
                if (bit->second.type == e_bindingType::image && bit->second.name == nameString)
                {
                    break;
                }
                else
                {
                    // try the next binding slot
                    binding = searchBinding++;
                    bit = usedBindings.find(binding);
                }
            }

            usedBindings[binding] = { nameString, e_bindingType::image };
            lastBinding = max(lastBinding, binding);
            tex.AddMember("Binding", binding, alloc);
            tex.AddMember("Mapping", requestedBinding, alloc);

            // get the type of texture uniform
            std::string uniform = "";

            // check for image sampler
            std::regex samplerRegex("sampler ");
            std::sregex_iterator samplerBegin = std::sregex_iterator(match.begin(), match.end(), samplerRegex);
            std::sregex_iterator samplerEnd = std::sregex_iterator();
            if (samplerBegin != samplerEnd)
            {
                uniform = samplerBegin->str();
                tex.AddMember("Type", "Sampler", alloc);
            }

            // check for combined image sampler
            std::regex combinedRegex("\\S*sampler\\S+");
            std::sregex_iterator combinedBegin = std::sregex_iterator(match.begin(), match.end(), combinedRegex);
            std::sregex_iterator combinedEnd = std::sregex_iterator();
            if (combinedBegin != combinedEnd)
            {
                uniform = combinedBegin->str();
                bool arr = (uniform.find("DArray") != std::string::npos);
                tex.AddMember("Type", "Combined Sampler", alloc);
                std::string dims = findDimension(uniform, "sampler", arr ? "Array" : "");
                tex.AddMember("Dimensions", "", alloc);
                tex["Dimensions"].SetString(dims.c_str(), alloc);
                tex.AddMember("Array", arr, alloc);
            }

            // check for texture
            std::regex texRegex("\\S*texture\\d\\S+");
            std::sregex_iterator texBegin = std::sregex_iterator(match.begin(), match.end(), texRegex);
            std::sregex_iterator texEnd = std::sregex_iterator();
            if (texBegin != texEnd)
            {
                uniform = texBegin->str();
                bool arr = (uniform.find("DArray") != std::string::npos);
                tex.AddMember("Type", "Texture", alloc);
                std::string dims = findDimension(uniform, "texture", arr ? "Array" : "");
                tex.AddMember("Dimensions", "", alloc);
                tex["Dimensions"].SetString(dims.c_str(), alloc);
                tex.AddMember("Array", arr, alloc);
            }

            // check for storage image
            std::regex storageRegex("\\S*image\\d\\S+");
            std::sregex_iterator storageBegin = std::sregex_iterator(match.begin(), match.end(), storageRegex);
            std::sregex_iterator storageEnd = std::sregex_iterator();
            std::string formatLayoutStr = "";
            if (storageBegin != storageEnd)
            {
                uniform = storageBegin->str();
                bool arr = (uniform.find("DArray") != std::string::npos);
                tex.AddMember("Type", "Storage Image", alloc);
                std::string dims = findDimension(uniform, "image", arr ? "Array" : "");
                tex.AddMember("Dimensions", "", alloc);
                tex["Dimensions"].SetString(dims.c_str(), alloc);
                tex.AddMember("Array", arr, alloc);

                // get layout format specifier
                std::regex formatRegex(", .+\\)");
                std::sregex_iterator formatBegin = std::sregex_iterator(match.begin(), match.end(), formatRegex);
                std::sregex_iterator formatEnd = std::sregex_iterator();
                if (formatBegin != formatEnd)
                {
                    formatLayoutStr = pystring::strip(formatBegin->str(), ")");
                }
            }

            std::string finalDefinition = "layout(binding = " + std::to_string(binding) + formatLayoutStr + ") uniform ";
            finalDefinition += uniform;
            finalDefinition += " ";
            finalDefinition += nameString;
            if (count != 1)
            {
                finalDefinition += "[";
                if (count > 0)
                    finalDefinition += std::to_string(count);
                finalDefinition += "]";
            }
            finalDefinition += ";";

            finalShader = pystring::replace(finalShader, match, finalDefinition);
            sj["Uniforms"].PushBack(tex, alloc);
        }

        // find push constant description
        std::string pushConstantOriginal = "";
        size_t pushConstantBinding = 0;
        std::regex pushConstantRegex("layout\\(push_constant\\).+[\n\r]+\\{[^\\}]+\\}.+;");
        std::sregex_iterator pushConstantBegin = std::sregex_iterator(buf.begin(), buf.end(), pushConstantRegex);
        std::sregex_iterator pushConstantEnd = std::sregex_iterator();
        if(pushConstantBegin != pushConstantEnd)
        {
            std::string match = pushConstantBegin->str();
            rapidjson::Value pushConstant;
            pushConstant.SetObject();
            pushConstant.AddMember("Members", "", alloc);
            pushConstant["Members"].SetArray();
            processStruct(match, pushConstant["Members"], false, alloc);

            pushConstantOriginal = match;
            pushConstantBinding = ++lastBinding;

            std::string nameString;
            int count;
            extractNameAndCount(match, ' ', nameString, count);
            pushConstant.AddMember("Name", "", alloc);
            pushConstant["Name"].SetString(nameString.c_str(), alloc);
            pushConstant.AddMember("Binding", pushConstantBinding, alloc);

            sj["Push Constants"].PushBack(pushConstant, alloc);
        }

        // strip excess carriage return characters
        finalShader = pystring::replace(finalShader, "\r\n", "\n");

        // write out the temp buffer with the includes
        std::string tempName = "temp_" + pystring::os::path::basename(in);
        std::string failedName = pystring::replace(tempName, "temp_", "failed_");
        std::ofstream temp(tempName.c_str(), std::ios::out);
        if (!temp.is_open())
        {
            std::cout << "Failed to open file " << tempName << " for write!\n";
            continue;
        }
        temp.write(finalShader.c_str(), finalShader.size());
        temp.close();

        std::ifstream tt(tempName.c_str(), std::ios::ate | std::ios::binary);
        size_t l = (size_t)tt.tellg();
        std::string b;
        b.resize(l);
        tt.seekg(0);
        tt.read((char*)b.c_str(), l);
        tt.close();

        std::string cmd = glslang;
        cmd += " -V ";
        cmd += tempName;
        cmd += " -o temp.spv";
        int res = system(cmd.c_str());
            
        std::string cBuf = " ";
        std::ifstream compiled("temp.spv", std::ios::ate | std::ios::binary);
        if (!compiled.is_open())
        {
            std::cout << "Failed to open compiled " << ext << " shader for read!\n";
            std::string failedName = pystring::replace(tempName, "temp_", "failed_");
            std::filesystem::copy_file(tempName, failedName);
            continue;
        }
        else
        {
            size_t cLen = (size_t)compiled.tellg();
            cBuf.resize(cLen);
            compiled.seekg(0);
            compiled.read((char*)cBuf.c_str(), cLen);
            compiled.close();
                
            // store compiled vulkan shader
            sj["Shader Files"].AddMember("vulkan", "", alloc);
            sj["Shader Files"]["vulkan"].SetString(cBuf.c_str(), cBuf.length(), alloc);
            cBuf.clear();

            // create openGL shader
            std::string glShader = finalShader;
            if (!pushConstantOriginal.empty())
            {
                std::string newBinding = "binding = " + std::to_string(pushConstantBinding);
                glShader = pystring::replace(glShader, pushConstantOriginal, pystring::replace(pushConstantOriginal, "push_constant", newBinding));
            }
                
            // add vertical flipping uniform into vertex shaders
            if (stage == e_shaderStage::vert)
            {
                size_t endOfShader = glShader.find_last_of('}') - 1;
                glShader.insert(endOfShader, "\ngl_Position.y *= vFlip;\n");

                const std::string mainDecl = "void main()";
                glShader = pystring::replace(glShader, mainDecl, "uniform float vFlip;\n" + mainDecl);
            }

            // store opengl shader
            glShader = pystring::replace(glShader, "\r", "");
            sj["Shader Files"].AddMember("opengl", "", alloc);
            sj["Shader Files"]["opengl"].SetString(glShader.c_str(), glShader.length(), alloc);
        }

        std::remove("temp.spv");
        std::remove(tempName.c_str());
        std::remove(failedName.c_str());

        // if there are any files left to process put the job to process them first
        if (j.filesToDo.size() > 1)
        {
            ShaderVariationJob nextFileJob = j;
            nextFileJob.builtVariations = false;
            nextFileJob.filesToDo.erase(nextFileJob.filesToDo.begin());
            jobs.insert(jobs.begin(), nextFileJob);
        }
        else 
        {
            if (j.isDefaultVariation)
            {
                std::string defaultVariationString = "";
                for (const std::vector<std::string>& dns : j.variationNames)
                {
                    if (dns.empty())
                        continue;

                    for (const std::string& s : dns)
                    {
                        if (!defaultVariationString.empty())
                            defaultVariationString += "/";
                        defaultVariationString += s;
                    }
                }
                tracker[j.outputPath].AddMember("default_variation", defaultVariationString, alloc);
            }
            finals.push_back(j);
        }
    }

    for (auto dit = tracker.MemberBegin(); dit != tracker.MemberEnd(); ++dit)
    {
        std::string outPath = dit->name.GetString();
        rapidjson::Value& variations = dit->value;
        rapidjson::Value d;
        d.SetObject();

        auto alloc = tracker.GetAllocator();

        d.AddMember("variations", "", alloc);
        d["variations"].SetArray();
        for (auto& j : finals)
        {
            if (j.outputPath != outPath)
                continue;

            rapidjson::Value variationValue;
            variationValue.SetObject();

            variationValue.AddMember("default_variation", j.isDefaultVariation, alloc);
            variationValue.AddMember("variation", "", alloc);
            variationValue["variation"].SetArray();


            rapidjson::Value* valIt = &variations;
            std::string lastName = "";
            for (int i = 0; i < e_shaderStage::COUNT; ++i)
            {
                if (!j.usesStage[i])
                    continue;

                std::string combined = "";
                for (auto& s : j.variationNames[i])
                {
                    if (!combined.empty())
                        combined += "/";
                    combined += s;

                    if (s != "default" && s != lastName)
                    {
                        rapidjson::Value inValue;
                        inValue.SetString(s, alloc);
                        variationValue["variation"].PushBack(inValue, alloc);
                    }
                }
                lastName = combined;

                if (!combined.empty())
                {
                    valIt = &(*valIt)[combined];
                }

                const rapidjson::Value& value = *valIt;
                if (value.HasMember("json"))
                {
                    for (auto jit = value["json"].MemberBegin(); jit != value["json"].MemberEnd(); ++jit)
                    {
                        rapidjson::Value name, value;
                        name.CopyFrom(jit->name, alloc);
                        value.CopyFrom(jit->value, alloc);
                        variationValue.AddMember(name, value, alloc);
                    }
                }

                if (value.HasMember(k_fileEndings[i]))
                {
                    rapidjson::Value newName;
                    newName.SetString(k_fileEndings[i], alloc);
                    rapidjson::Value newValue;
                    newValue.CopyFrom(value[k_fileEndings[i]], alloc);
                    variationValue.AddMember(newName, newValue, alloc);
                }
            }
            d["variations"].PushBack(variationValue, alloc);
        }

        // store the last time this shader was rebuilt
        SYSTEMTIME currentST;
        GetSystemTime(&currentST);
        FILETIME currentFT;
        SystemTimeToFileTime(&currentST, &currentFT);

        d.AddMember("Last Modified", "", alloc);
        d["Last Modified"].SetObject();
        d["Last Modified"].AddMember("Low", "", alloc);
        d["Last Modified"]["Low"].SetUint(currentFT.dwLowDateTime);
        d["Last Modified"].AddMember("High", "", alloc);
        d["Last Modified"]["High"].SetUint(currentFT.dwHighDateTime);

        rapidjson::StringBuffer buffer;
        rapidjson::PrettyWriter<rapidjson::StringBuffer> writer(buffer);
        d.Accept(writer);

        std::ofstream json(outPath, std::ios::out);
        if (!json.is_open())
        {
            std::cout << "Failed to open file " << outPath << " for write!\n";
            continue;
        }

        json.write(buffer.GetString(), buffer.GetSize());
        json.close();
    }

    return 0;
}
